/* -*- Mode: C++ -*-
 * $Id: Actuators.cc,v 1.7 2003/09/02 17:55:01 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Actuators.cc
 * Unitname: BasicSystem
 * $Revision: 1.7 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2003/09/02 17:55:01 $
 *
 *
 * HISTORY
 *
 * $Log: Actuators.cc,v $
 * Revision 1.7  2003/09/02 17:55:01  frehe
 * Added the semantics of the command pointers. RoboSoc will now take control of them.
 *
 * Revision 1.6  2003/07/21 08:54:26  s02davro
 * Acutators now handles the commands with dynamic_cast
 *
 * Revision 1.5  2001/11/15 16:48:47  frehe
 * Added support for a coach.
 *
 * Revision 1.4  2001/08/19 23:32:22  frehe
 * Updated 2001-08-19
 *
 * Revision 1.3  2000/09/09 11:50:32  frehe
 * Modified the code based on the new level in the command hieararchy, namely ImmediateCommand and PeriodicCommand.
 *
 * Revision 1.2  2000/09/05 21:15:52  frehe
 * Made minor changes.
 *
 * Revision 1.1  2000/09/03 20:41:12  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file Actuators.cc
 * Implementation file for the class Actuators.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.7 $
 * $Author: frehe $
 * $Date: 2003/09/02 17:55:01 $
 */


#include "Actuators.h"
#include "ServerCommand.h"
#include "ControlCommand.h"
#include "CoachCommand.h"
#include "TrainerCommand.h"
#include "CoachOrTrainerCommand.h"
#include "ImmediateCommand.h"
#include "PeriodicCommand.h"
#include "rs_error.h"
#include "rs_log.h"


RS_BEGIN_NAMESPACE


Actuators::Actuators(ActuatorInterface* actuator_interface)
  : theController(NULL),
    theActuatorInterface(actuator_interface),
    periodicActionQueue(),
    immediateActionQueue(),
    controlCommandQueue()
{
}

void Actuators::init(Controller* controller)
{
  theController = controller;
  theActuatorInterface->init(this);
}


void Actuators::executeCommand(Command* cmd)
{
  if ( cmd == NULL )
    RS_ERROR(theController->getCurrentTime() <<
	     ": Actuators::executeCommand got a NULL command!");

  if ( ServerCommand* s_cmd = dynamic_cast<ServerCommand *>(cmd) ) {
    if ( CoachCommand* c_cmd = dynamic_cast<CoachCommand *>(s_cmd) ) {
      // Check if player is coach.
      RS_LOG( LA_ACTUATORS, theController->getCurrentTime() <<
	      ": Send a coach command " << c_cmd->serverString() );
      theController->sendCoachCommand(c_cmd->serverString());
      delete cmd;
    } else if ( TrainerCommand* t_cmd = dynamic_cast<TrainerCommand *>(s_cmd) ) {
      // Check if player is trainer.
      RS_LOG( LA_ACTUATORS, theController->getCurrentTime() <<
	      ": Send a trainer command " << t_cmd->serverString() );
      theController->sendTrainerCommand(t_cmd->serverString());
      delete cmd;
    } else if ( CoachOrTrainerCommand* ct_cmd = dynamic_cast<CoachOrTrainerCommand *>(s_cmd) ) {
      // Check if player is trainer of coach.
      RS_LOG( LA_ACTUATORS, theController->getCurrentTime() <<
	      ": Send a trainer/coach command " << ct_cmd->serverString() );
      theController->sendCoachOrTrainerCommand(ct_cmd->serverString());
      delete cmd;
    } else if ( ImmediateCommand* i_cmd = dynamic_cast<ImmediateCommand *>(s_cmd) ) {
      RS_LOG( LA_ACTUATORS, theController->getCurrentTime() <<
	      ": Pushed a " << i_cmd->serverString()
	      << " on the immediateActionQueue." );
      immediateActionQueue.push(i_cmd);
    } else if ( PeriodicCommand* p_cmd = dynamic_cast<PeriodicCommand *>(s_cmd) ) {
      RS_LOG( LA_ACTUATORS, theController->getCurrentTime() <<
	      ": Pushed a " << p_cmd->serverString()
	      << " on the periodicActionQueue." );
      periodicActionQueue.push(p_cmd);
    }
    else {
      RS_ERROR(theController->getCurrentTime() <<
	       ": Unknown server command sent to the actuators!");
      delete cmd;
    }
  } else if ( ControlCommand* c_cmd = dynamic_cast<ControlCommand *>(cmd) ) {
    controlCommandQueue.push(c_cmd);
  } else {
    RS_ERROR(theController->getCurrentTime() <<
	     ": Neither server nor control command sent to the actuators!");
    delete cmd;
  }
}

UInt16 Actuators::removeAllPeriodicActions()
{
  UInt16 no_elt = periodicActionQueue.size();
  Command* cmd = NULL;
  while ( !periodicActionQueue.empty() ) {
    cmd = periodicActionQueue.front();
    periodicActionQueue.pop();
    delete cmd;
  }
  return no_elt;
}

UInt16 Actuators::removeAllImmediateActions()
{
  UInt16 no_elt = immediateActionQueue.size();
  Command* cmd = NULL;
  while ( !immediateActionQueue.empty() ) {
    cmd = immediateActionQueue.front();
    immediateActionQueue.pop();
    delete cmd;
  }
  return no_elt;
}

UInt16 Actuators::removeAllControlCommands()
{
  UInt16 no_elt = controlCommandQueue.size();
  Command* cmd = NULL;
  while ( !controlCommandQueue.empty() ) {
    cmd = controlCommandQueue.front();
    controlCommandQueue.pop();
    delete cmd;
  }
  return no_elt;
}

void Actuators::executeNextPeriodicAction()
{
  if ( periodicActionQueue.empty() ) {
    // This is *NOT* good!
    RS_WARNING(theController->getCurrentTime() <<
	       ": Nothing to send to the server!");
  }
  else {
    theController->sendCommand(periodicActionQueue.front());
    periodicActionQueue.pop();
  }
}

void Actuators::executeNextImmediateAction()
{
  if ( !immediateActionQueue.empty() ) {
    theController->sendCommand(immediateActionQueue.front());
    immediateActionQueue.pop();
  }
}

void Actuators::executeNextControlCommand()
{
  if ( !controlCommandQueue.empty() ) {
    theController->sendCommand(controlCommandQueue.front());
    controlCommandQueue.pop();
  }
}


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "Actuators.icc"
#endif
