/* -*- Mode: C++ -*-
 * $Id: Controller.cc,v 1.25 2004/08/31 17:12:47 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Controller.cc
 * Unitname: BasicSystem
 * $Revision: 1.25 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2004/08/31 17:12:47 $
 *
 *
 * HISTORY
 *
 * $Log: Controller.cc,v $
 * Revision 1.25  2004/08/31 17:12:47  frehe
 * Added a stop method.
 *
 * Revision 1.24  2004/08/29 13:36:19  frehe
 * Updated the handling of time for the coach and added a callback for coach visual data.
 *
 * Revision 1.23  2003/08/20 12:28:04  s02davro
 * New count counter, attentiontos etc.
 *
 * Revision 1.22  2003/07/22 14:44:48  s02davro
 * Added count for signals
 *
 * Revision 1.21  2003/07/21 08:51:26  s02davro
 * Fixed starvation problem, updated signal handlers, added support for new message types. etc.
 *
 * Revision 1.20  2002/12/07 17:29:04  s02davro
 * Hack in Sensor to fix trainer added CP_trainer to WorldFacts
 *
 * Revision 1.19  2002/11/29 12:44:20  frehe
 * Added a check to see whether to use time.h or sys/time.h
 *
 * Revision 1.18  2002/09/21 19:08:04  s02davro
 * The Controller now handles serverAlive in a correct way and terminates agent when the server is killed
 *
 * Revision 1.17  2002/09/04 13:25:09  s02davro
 * Removed deprecated methods in Controller and Decision
 *
 * Revision 1.16  2002/08/31 15:04:34  s02davro
 * Updated the signal handlers, new main loop - run()
 *
 * Revision 1.15  2002/08/02 21:49:24  s02davro
 * Small changes, added somm big todos.
 *
 * Revision 1.14  2002/06/18 20:06:29  frehe
 * Stop the command timer when the controller is stopped.
 *
 * Revision 1.13  2002/01/15 17:19:23  frehe
 * Changed the error message when receiving unsuported messages.
 *
 * Revision 1.12  2001/11/15 16:48:47  frehe
 * Added support for a coach.
 *
 * Revision 1.11  2001/10/14 21:21:42  frehe
 * Added better support for messages from self and from other players.
 *
 * Revision 1.10  2001/09/03 19:39:36  frehe
 * Changed code to use the RS_BEGIN_TRY, RS_END_TRY and RS_CATCH macros.
 *
 * Revision 1.9  2001/08/30 21:42:10  frehe
 * Added more exception handling.
 *
 * Revision 1.8  2001/08/19 23:32:22  frehe
 * Updated 2001-08-19
 *
 * Revision 1.7  2001/08/17 07:15:15  frehe
 * Made changes to compile RoboSoc with RCCParser-0.10.
 *
 * Revision 1.6  2001/08/13 16:43:15  frehe
 * Made updates to compile with new autotools and RCCParser.
 *
 * Revision 1.5  2000/10/24 15:15:17  frehe
 * Fixed the generation of the onNewCycle event.
 *
 * Revision 1.4  2000/10/23 12:05:29  frehe
 * Removed all explicit output to std::cout.
 *
 * Revision 1.3  2000/09/05 21:15:52  frehe
 * Made minor changes.
 *
 * Revision 1.2  2000/09/04 15:23:46  frehe
 * Fixed a naming miss.
 *
 * Revision 1.1  2000/09/03 20:41:13  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file Controller.cc
 * Implementation file for the class Controller.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.25 $
 * $Author: frehe $
 * $Date: 2004/08/31 17:12:47 $
 */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <cstdlib>
#ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
#else 
#  include <time.h>
#endif 

#include "Controller.h"
#include "Actuators.h"
#include "Sensors.h"
#include "ServerInterface.h"
#include "ControlInterface.h"
#include "ServerCommand.h"
#include "ControlCommand.h"
#include "SetCyclesSinceIOBeforeShutdown.h"
#include "SetMinTicksBeforeSendingCommand.h"
#include "SetMaxTicksBeforeSendingCommand.h"
#include "SetTicksBeforeCommandWarning.h"
#include "SetTicksBetweenCycles.h"
#include "SetTicksBetweenSenses.h"
#include "SetTicksBetweenVisuals.h"

#include "rs_error.h"
#include "rs_log.h"
#include "rs_try_catch.h"


RS_BEGIN_NAMESPACE


Controller* Controller::currentControllerObject = NULL;

/** Signal handler for sigalrm. It is declared as an external
  C function since the Unix signal handler requires this. */
extern "C" RETSIGTYPE externSigAlarmHandler(int sig);
/** Signal handler for sigio. It is declared as an external
  C function since the Unix signal handler requires this. */
extern "C" RETSIGTYPE externSigIOHandler(int sig);


Controller::Controller(ServerInterface* const server_interface,
		       Sensors* const sensors,
		       Actuators* const actuators,
		       ControlInterface* const control_interface,
		       const UInt16 cycle_length,
		       const UInt16 tick_length,
		       const UInt16 sense_step,
		       const UInt16 visual_step,
		       const UInt16 max_sigalrms,
		       const UInt16 min_ticks,
		       const UInt16 ticks_before_cmd_warning,
		       const UInt16 max_ticks, 
		       const bool _coach,
		       const bool _trainer)
  throw()
  : theServerInterface(server_interface),
    theSensors(sensors),
    theActuators(actuators),
    theControlInterface(control_interface),
#ifdef RS_ENABLE_COACH_DEBUGGING
    theCoachInterface(NULL),
#endif
    maxCyclesSinceIOBeforeShutdown(max_sigalrms),
    minTicksBeforeSendingCommand(min_ticks),
    ticksBeforeCommandWarning(ticks_before_cmd_warning),
    maxTicksBeforeSendingCommand(max_ticks),
    cycleLength(cycle_length),
    tickLength(tick_length),
    ticksBetweenCycles(cycle_length / tick_length),
    ticksBetweenSenses(sense_step / tick_length),
    ticksBetweenVisuals(visual_step / tick_length),
    commandTimer(),
    startClock(false),
    coach(_coach),
    trainer(_trainer),
    ticks(0),
    senseTicks(0),
    visualTicks(0),
    lastCycleTime(Time(0,0)),
    lastAuralTime(0),
    lastSenseTime(0),
    lastVisualTime(0),
    lastKnownServerTime(0),
    currentTime(Time::currentTime),
    estimatedTime(0),
    stoppedCycles(0),
    sensorData(NULL),
    sigalarmsSinceSigio(0),
    lastActTime(Time(0,0)),
    active(false),
    IOFlag(false),
    AlarmFlag(false),
    ticksSinceLastHandle(0),
    ioFlagsSinceHandle(0) {
  sensorData = new char[theServerInterface->getServerBufferSize()];
  
  currentControllerObject = this;

  // The clock is initially stopped
  Time::stopClock();
  
  if ( coach )
    RS_LOG( LA_CONTROLLER, "Starting the controller as a coach. " << this);
  else if (trainer)
    RS_LOG( LA_CONTROLLER, "Starting the controller as a trainer. " << this);
  else 
    RS_LOG( LA_CONTROLLER, "Starting the controller as a player. " << this);
}


Controller::~Controller()
  throw()
{
  delete[] sensorData;
#ifdef RS_ENABLE_COACH_DEBUGGING
  if ( theCoachInterface != NULL )
    delete theCoachInterface;
#endif
}


bool Controller::init()
  throw(RSException)
{
  RS_LOG (LA_CONTROLLER, "Controller: Init sensors, actuators "
	  << "and signal handlers. ");
  // Init the sensors and actuators.
  theSensors->init(this);
  theActuators->init(this);

  // Init the signal handlers.
  allSigsMask = initHandlers();

  // Open the connection to the server and return the result.
  return theServerInterface->connect();
}


bool Controller::start()
  throw(RSException)
{
  // Set the active flag
  active = true;
  
  // Set the start flag, the internal timer will be started
  // after the first sense_body message.
  startClock = true;

  sigprocmask(SIG_UNBLOCK, &sigiomask, NULL);
  sigprocmask(SIG_UNBLOCK, &sigalarmask, NULL);

  // Initialize the decision maker,
  // and send an init message to the server.
  ServerCommand* cmd = theControlInterface->initAgent();
  if ( cmd == NULL )
    RS_ERROR("[Start]: no init command generated by the decision!");
  else {
    RS_LOG( LA_CONTROLLER, "Sent init message: " << cmd->serverString() );
    sendCommand(cmd);
  }

  return true;
}


bool Controller::stop()
  throw(RSException)
{
  // Reset the active flag
  active = false;

  return true;
}


bool Controller::sendCoachCommand(const std::string& msg) throw(RSException) {
  if (coach)
    return theServerInterface->sendCommand(msg);
  else 
    RS_WARNING("Tried sending a coach command as a non-coach agent.");
  return false;
}

bool Controller::sendTrainerCommand(const std::string& msg) throw(RSException) {
  if (trainer)
    return theServerInterface->sendCommand(msg);
  else 
    RS_WARNING("Tried sending a trainer command as a non-trainer agent.");
  return false;
}

bool Controller::sendCoachOrTrainerCommand(const std::string& msg) throw(RSException) {
  if (coach || trainer)
    return theServerInterface->sendCommand(msg);
  else 
    RS_WARNING("Tried sending a coach/trainer command as a non-coach/trainer agent.");
  return false;
}

sigset_t Controller::initHandlers() {
  // Ngot gr snett hr!.

  // Adopted from CMUnited'98 by Peter Stone et al.
  sigemptyset(&sigalarmask);
  sigaddset(&sigalarmask, SIGALRM);
  sigemptyset(&sigiomask);
  sigaddset(&sigiomask, SIGIO);
  
  struct sigaction act, oact;
  
  //Signal(SIGALRM, externSigAlarmHandler, &sigiomask);
  act.sa_handler = externSigAlarmHandler;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;
//    act.sa_mask = sigiomask;
#ifdef SA_INTERRUPT
  act.sa_flags |= SA_INTERRUPT;   // SunOS 4.x
#endif
#ifdef SA_RESTART
  act.sa_flags |= SA_RESTART;     // SVR4, 4.4BSD
#endif 
  if (sigaction(SIGALRM, &act, &oact) == -1)
    RS_ERROR("Problems with initiation of signal handlers");
  
  //Signal(SIGIO, externSigIOHandler, &sigalarmask);
  act.sa_handler = externSigIOHandler;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;
  //    act.sa_mask = sigalarmask;
#ifdef SA_RESTART
  act.sa_flags |= SA_RESTART;     // SVR4, 4.4BSD
#endif
#ifdef SA_INTERRUPT
  act.sa_flags |= SA_INTERRUPT;   // SunOS 4.x
#endif
  if (sigaction(SIGIO, &act, &oact) == -1)
    RS_ERROR("Problems with initiation of signal handlers");
  
  sigprocmask(SIG_UNBLOCK, &sigiomask, NULL);
  sigprocmask(SIG_UNBLOCK, &sigalarmask, NULL);

  sigset_t sigsetmask;
  // get's the currently unblocked signals
  sigprocmask(SIG_BLOCK, NULL, &sigsetmask);   
  return sigsetmask;   
}

void Controller::sigAlarmHandler(int) {
  AlarmFlag = true;
  ++ticksSinceLastHandle;
  ++sigalarmsSinceSigio;
}

void Controller::sigIOHandler(int) {
  IOFlag = true;
  ++ioFlagsSinceHandle;
  sigalarmsSinceSigio = 0;
}

int Controller::run() {
  gettimeofday(&timeOfFirstTick, NULL);

  try {
    while ( active && serverAlive()) {
      try {
	if ( pause() == -1 && errno != EINTR )
	  RS_ERROR("Something went wrong in pause!");
	while (IOFlag || AlarmFlag) {
	  if (IOFlag) {
	    IOFlag = false;
	    handleSigIO();
	  }
	  if (Controller::AlarmFlag) {
	    AlarmFlag = false;
	    handleSigAlarm();
	  }	
	}
      } catch (RSException &excpt) {
	std::cerr << "Caught a RoboSoc exception: " 
		  << excpt.What() << std::endl;
	return 1;
      } catch (std::exception &excpt) {
	std::cerr << "Caught a standard exception: " 
		  << excpt.what() << std::endl;
	return 1;
      } catch (...) {
	std::cerr << "Caught an unknown exception: " << std::endl;
	return 1;
      }
    }
    
    std::cerr << std::endl 
	      << "Server is assumed dead. Terminating agent..." << std::endl;
    return true;
  } catch (RSException &excpt) {
    std::cerr << "Caught a RoboSoc exception: " << excpt.What() << std::endl;
    return 1;
  } catch (std::exception &excpt) {
    std::cerr << "Caught a standard exception: " << excpt.what() << std::endl;
    return 1;
  } catch (...) {
    std::cerr << "Caught an unknown exception: " << std::endl;
    return 1;
  }
}


void Controller::handleSigAlarm() {
  RS_LOG( LA_CONTROLLER, "Entering HandleSigAlarm..." );
  
  gettimeofday(&currentTickTime, NULL);
  Int usecSinceFirstTick = currentTickTime.tv_usec - timeOfFirstTick.tv_usec;
  if (usecSinceFirstTick < 0) usecSinceFirstTick += 1000000;
  RS_LOG( LA_CONTROLLER, "Alarmtick: " << ticks << " time (us): "
	  << usecSinceFirstTick << " number of ticksSinceLastHandle " 
	  << ticksSinceLastHandle);  
  
  // Notify the decision maker that a new tick is to be processed.
  theControlInterface->onBeforeTick();
  
  // For each SIGALRM add a tick to each counter.
  ticks += ticksSinceLastHandle;
  visualTicks += ticksSinceLastHandle;
  senseTicks += ticksSinceLastHandle;
  ticksSinceLastHandle = 0;


  // If ticks is greater than ticksBetweenCycles, 
  // then a new cycle should've begun.
  // Therefore increase the estimated time and the current time.
  if (ticks > ticksBetweenCycles) {
    estimatedTime += (ticks / ticksBetweenCycles);
    RS_LOG( LA_CONTROLLER, "ticks (" << ticks <<") > ticksBetweenCycles ("
	    << ticksBetweenCycles << " estimatedTime = " << estimatedTime); 
    
    // Update the current time depending on whether the game is stopped or not.
    if ( currentTime.isClockStopped() )
      currentTime.setS(stoppedCycles + estimatedTime);
    else {
      currentTime.setT(lastKnownServerTime + estimatedTime);
      currentTime.setS(0);
    }

    // Since it is ticksBetweenCycles+1 we should remove 
    // ticksBetweenCycles ticks,
    // which leaves us with 1 tick left.    
    ticks = 1;
    gettimeofday(&timeOfFirstTick, NULL);
  }
  
  // Since we might have lost messages from the server we need to notify
  // the decision maker that there should've been a new cycle now.
  //theControlInterface->OnEstimatedNewCycle(estimatedTime, ticks);
  if ( estimatedTime > 0 ) {
    if ( theControlInterface->onEstimatedNewCycle(estimatedTime, ticks) ) {
      RS_LOG( LA_CONTROLLER, "Controller - alarmticks indicate new cycle"); 
      lastCycleTime = currentTime;
      theControlInterface->onNewCycle(true);
    }
  }
  
  // Check whether the agent sent a future action already.
  if ( currentTime < lastActTime ) {
    RS_WARNING(currentTime 
	       << ": Sent future action(s) already! Last action time = " 
	       << lastActTime);
    // Assume it depends on some timing problem and set lastActTime to 
    // currentTime to prevent losses of action oppertunities.
    lastActTime = currentTime;
  }
  
  // Check whether the agent sent an action last cycle or not.
  if (( ticks == 1 &&
	( currentTime.getT() - lastActTime.getT() > 1 ||
	  ( currentTime.isClockStopped() &&
	    currentTime.getS() - lastActTime.getS() > 1) ) ) &&
      // Give the agent some startuptime.
      ( (currentTime.getT() != 0) || (currentTime.getS() > 5)) &&
      // Coach do not need to act every cycle.
      (!coach && !trainer))      
    RS_WARNING(currentTime << ": Missed action(s)! Last action time = "
	       << lastActTime);

  // If the actuators want to send a periodic action and one haven't been
  // sent this cycle then send it.
  RS_LOG( LA_CONTROLLER, "lastActTime: " << lastActTime
	  << "currentTime: " << currentTime
	  << "waitingAction: " << theActuators->waitingPeriodicAction()
	  << "ticks: " << ticks
	  << "minTicks: " << minTicksBeforeSendingCommand 
	  << "maxTicks: " << maxTicksBeforeSendingCommand);

  if ( lastActTime != currentTime &&
       theActuators->waitingPeriodicAction() &&
       ticks >= minTicksBeforeSendingCommand &&
       ticks <= maxTicksBeforeSendingCommand ) {
    RS_LOG (LA_CONTROLLER, "Execute next periodic action");
    theActuators->executeNextPeriodicAction();
    lastActTime = currentTime;
  } else {
    if ( lastActTime != currentTime &&
	 !theActuators->waitingPeriodicAction() &&
	 ticks >= ticksBeforeCommandWarning &&
	 (!coach && !trainer)) {
      // Warn the decision maker about the fact that no command 
      // has been sent yet!
      RS_LOG (LA_CONTROLLER, "Warning decisionmaker");
      theControlInterface->onCommandWarning();
      // If we got a decision, execute it.
      if (theActuators->waitingPeriodicAction()) {
	RS_LOG (LA_CONTROLLER, "Execute next periodic action");      
	theActuators->executeNextPeriodicAction();
	lastActTime = currentTime;	
      }
    }
  }
  // If the actuators want to send an immediate action then send it.
  while ( theActuators->waitingImmediateAction() ) {
    RS_LOG (LA_CONTROLLER, "Execute next immediate action");      
    theActuators->executeNextImmediateAction();
  }
  
  // If there is a control command waiting execute it.
  while ( theActuators->waitingControlCommand() ) {
    RS_LOG (LA_CONTROLLER, "Execute next control command");      
    theActuators->executeNextControlCommand();  
  }
  
  // Notify the decision maker that a new tick was processed.
  theControlInterface->onAfterTick();
  RS_LOG( LA_CONTROLLER, "Leaving HandleSigAlarm..." );
}

void Controller::handleSigIO() {
  RS_LOG( LA_CONTROLLER, "Entering HandleSigIO..." );
  RS_LOG( LA_CONTROLLER, "numberOfFlags: " << ioFlagsSinceHandle);

  ioFlagsSinceHandle = 0;
  sigalarmsSinceSigio = 0;

  // Notify the decision maker that a new sensors data is to be processed.
  theControlInterface->onBeforeSensorData();
  
  bool first = true;
  int recivedChunks = 0; // To save actuators (alarmhandler) from starvation.
  UInt16 time_of_first_data = 0;

  theSensors->setLastSensorType(SENSOR_UNKNOWN);
  
  while (recivedChunks < 2) {
    if (!receiveInput(sensorData)) {
      IOFlag = false;
      break;
    }
    else {
      IOFlag = true;
    }
    
    ++recivedChunks;
    RS_LOG( LA_CONTROLLER,
	    "NEW INFO! " << theSensors->getCurrentTime() << ": "
	    << sensorData );
    
    theSensors->analyze(sensorData);
    
    if ( first ) {
      time_of_first_data = lastKnownServerTime;
      first = false;
    }
    
    switch ( theSensors->getLastSensorType() ) {
    case SENSOR_SENSE:
      RS_LOG( LA_CONTROLLER, "SENSE_BODY " << lastKnownServerTime );

      if ( currentTime.getT() == lastSenseTime.getT() &&
	   currentTime.getS() == 0 && !currentTime.isClockStopped() ) {
 	RS_WARNING(currentTime << ": Two sensebodies in one cycle!");
	RS_WARNING("Global clock must have been stopped, stopping local clock.");
	Time::stopClock();
      }
      
      // If more than one cycle between senses sense_body messages are lost.
      if ( currentTime.getT() - lastSenseTime.getT() > 1 ||
	   ( currentTime.isClockStopped() &&
	     currentTime.getS() - lastSenseTime.getS() > 1) )
	RS_WARNING(currentTime << ": More than one cycle between sense_body's!"
		   << " lastSenseTime = " << lastSenseTime);
      
      lastSenseTime = currentTime;
      theControlInterface->onPhysicalSensorData();
      break;
    case SENSOR_VISUAL:
      RS_LOG( LA_CONTROLLER, "VISUAL " << lastKnownServerTime );
      lastVisualTime = currentTime;
      theControlInterface->onVisualSensorData();
      break;
    case SENSOR_AURAL_COACH:
      RS_LOG( LA_CONTROLLER, "AURAL_COACH " << lastKnownServerTime );
      lastAuralTime = currentTime;
      theControlInterface->onAuralCoachSensorData();
      theControlInterface->onAuralSensorData();
      break;
    case SENSOR_AURAL_PLAYER:
      RS_LOG( LA_CONTROLLER, "AURAL_PLAYER " << lastKnownServerTime );
      lastAuralTime = currentTime;
      theControlInterface->onAuralPlayerSensorData();
      theControlInterface->onAuralSensorData();
      break;
    case SENSOR_AURAL_REFEREE:
    case SENSOR_AURAL_SCORE:
      RS_LOG( LA_CONTROLLER, "AURAL_REFEREE " << lastKnownServerTime );
      lastAuralTime = currentTime;
      theControlInterface->onAuralRefereeSensorData();
      theControlInterface->onAuralSensorData();
      break;
    case SENSOR_AURAL_SELF:
      RS_LOG( LA_CONTROLLER, "AURAL_SELF " << lastKnownServerTime );
      lastAuralTime = currentTime;
      theControlInterface->onAuralSelfSensorData();
      theControlInterface->onAuralSensorData();
      break;
    case SENSOR_SUBSTITUTION:
      RS_LOG( LA_CONTROLLER, "SENSOR_SUBSTITUTION " << lastKnownServerTime);
      theControlInterface->onSubstitution();
      break;
    case SENSOR_ACTUATORS:
      // Should not be posible!
      RS_WARNING(currentTime << 
		 "SENSOR_ACTUATORS should not be possible to get!");
      break;
    case SENSOR_INIT:
      RS_LOG( LA_CONTROLLER, "INIT " << lastKnownServerTime );
      theControlInterface->onInit();
      break;
    case SENSOR_ERROR:
      RS_LOG( LA_CONTROLLER, "ERROR " << lastKnownServerTime );
      theControlInterface->onSensorError();
      break;
    case SENSOR_COACH_AURAL:
      RS_LOG( LA_CONTROLLER, "SENSOR_COACH_AURAL " << lastKnownServerTime );
      break;
    case SENSOR_COACH_VISUAL:
      RS_LOG( LA_CONTROLLER, "SENSOR_COACH_VISUAL " << lastKnownServerTime );
      if ( coach || trainer ) {
	if ( currentTime.getT() == lastCoachVisualTime.getT() &&
	     currentTime.getS() == 0 && !currentTime.isClockStopped() ) {
	  RS_WARNING(currentTime << ": Two coach visuals in one cycle!");
	  RS_WARNING("Global clock must have been stopped, stopping local clock.");
	  Time::stopClock();
	}
      
	// If more than one cycle between visuals messages are lost.
	if ( currentTime.getT() - lastCoachVisualTime.getT() > 1 ||
	     ( currentTime.isClockStopped() &&
	       currentTime.getS() - lastCoachVisualTime.getS() > 1) ) {
	  RS_WARNING(currentTime
		     << ": More than one cycle between coach visuals!"
		     << " lastCoachVisualTime = " << lastCoachVisualTime);
	}
	
	lastCoachVisualTime = currentTime;
	theControlInterface->onVisualCoachSensorData();
	theControlInterface->onVisualSensorData();
	break;
      } else {
	RS_WARNING(currentTime <<  "SENSOR_COACH is only handled if coach or trainer is true");
	break;
      }
    case SENSOR_PARAM:
      RS_LOG( LA_CONTROLLER, "SENSOR_PARAM " << lastKnownServerTime );
      break;
    case SENSOR_PLAYER_TYPE:
      RS_LOG( LA_CONTROLLER, "SENSOR_PLAYER_TYPE " << lastKnownServerTime );
      break;
    case SENSOR_PLAYER_PARAM:      
      RS_LOG( LA_CONTROLLER, "SENSOR_PLAYER_PARAM " << lastKnownServerTime );
      break;
    case SENSOR_OK:
      RS_LOG ( LA_CONTROLLER, "SENSOR_OK " << lastKnownServerTime );
      break;
    case SENSOR_UNKNOWN:
      RS_LOG ( LA_CONTROLLER, "Unknown sensor input!" );
      RS_ERROR(currentTime << ": unknown sensor input!!!");
      break;
    default:
      RS_ERROR(currentTime << ": Strange sensor input!!!");
    }
  
    // Check if the currentTime differs from the lastKnownServerTime.
    if ( false && currentTime != lastKnownServerTime )
      RS_WARNING("Diff: " << lastKnownServerTime 
		 << " != " << currentTime 
		 << " (" << lastSenseTime << ", " << lastVisualTime
		 << ", " << lastAuralTime << ") (" 
		 << estimatedTime << ", " << ticks << ")");

    // Notify the decision maker that a new sensors data was processed.
    theControlInterface->onAfterSensorData();
    RS_LOG( LA_CONTROLLER, "Leaving HandleSigIO..." );
  }   
}


bool Controller::updateTime(const UInt16 newTime) throw() {
  RS_LOG ( LA_CONTROLLER,  currentTime << ": updateTime(" << newTime << ")");
  Time lastTime = currentTime;

  // Update the current time with the time of the last sensor data.
  // Since that time is assumed to be correct if the game isn't stopped
  // reset the estimated time if the game is running.
  if ( !currentTime.isClockStopped() ) {
    if ( lastKnownServerTime != newTime ) {
      currentTime.setS(0);
      stoppedCycles = 0;
    }
    estimatedTime = 0;
    // If the clock is stopped then the time shouldn't be different...
  } else if ( lastKnownServerTime != newTime ) {
    RS_WARNING("Clock is stopped but the current time is not equal to the new time.");
    // But since it is start the clock again...
    // Small change, seems like we can get a free_kick_fault one cycle and the clock
    // is stopped next cycle. This (hopfully) fix the problem.
    if ((lastKnownServerTime + 1) < newTime)
      currentTime.startClock();
  }

  // Update the last known server time and the current time.
  lastKnownServerTime = newTime;
  currentTime.setT(lastKnownServerTime);

  switch ( theSensors->getLastSensorType() ) {
  case SENSOR_SENSE:
  case SENSOR_COACH_AURAL:
  case SENSOR_COACH_VISUAL:
    // This is used to start the timer after the first sense_body
    if ( startClock ) {
      commandTimer.start(tickLength / 2, tickLength);
      senseTicks = 1;
//        lastSenseTime = currentTime;
      startClock = false;
    } else {
      // If the server and the internal clock are running on time the
      // ticks should be zero, since the estimated time should be updated
      // before a new time arrive.
      if ( false && abs(ticks - ticksBetweenCycles) > 1 )
	RS_WARNING(currentTime << ": ticks = " << ticks);
      // If the game is stopped then the
      if ( currentTime.isClockStopped() &&
	   currentTime.getT() == lastSenseTime.getT() ) {
	if ( false )
	  RS_WARNING(currentTime << ": stoppedCycles = " << stoppedCycles
		     << " estimatedTime = " << estimatedTime
		     << " ticks " << ticks);
	if ( estimatedTime > 0 ) {
	  stoppedCycles+=estimatedTime;
	  estimatedTime = 0;
	  if ( ticks == ticksBetweenCycles ) {
	    ++stoppedCycles;
	    ticks = 0;
	    gettimeofday(&timeOfFirstTick, NULL);
	  }
	} else {
	  ++stoppedCycles;
	  ticks = 0;
	  gettimeofday(&timeOfFirstTick, NULL);
	}
	currentTime.setS(stoppedCycles);
	if ( false )
	  RS_WARNING(currentTime << ": stoppedCycles = " << stoppedCycles
		     << " estimatedTime = " << estimatedTime
		     << " ticks " << ticks);
      } else {
	// Reset the ticks since last cycle.
	estimatedTime = 0;
	ticks = 0;
	gettimeofday(&timeOfFirstTick, NULL);
      }
	
      // If senseTicks isn't equal to ticksBetweenSenses either the
      // server or the internal clock is not running correctly.
      if ( false && abs(senseTicks - ticksBetweenSenses) > 1 )
	RS_WARNING(currentTime << ": senseTicks = " << senseTicks);
      // Reset the number of ticks since last sense.
      senseTicks = 0;
      
    }
    break;
  case SENSOR_VISUAL:
    // If visualTicks isn't equal to ticksBetweenVisuals either the
    // server or the internal clock is not running correctly.
    if ( false && abs(visualTicks - ticksBetweenVisuals) > 1 )
      RS_WARNING(currentTime << ": visualTicks = " << visualTicks);
    // Reset the number of ticks since last visual.

    visualTicks = 0;
    break;
  case SENSOR_AURAL_COACH:
  case SENSOR_AURAL_PLAYER:
  case SENSOR_AURAL_REFEREE:
  case SENSOR_AURAL_SCORE:
  case SENSOR_AURAL_SELF:
    break;
  case SENSOR_ACTUATORS:
    // Should not be posible!
    RS_WARNING(currentTime << "SENSOR_ACTUATORS should not be possible to get!");
    break;
  case SENSOR_INIT:
  case SENSOR_PARAM:
  case SENSOR_PLAYER_TYPE:
  case SENSOR_PLAYER_PARAM:
  case SENSOR_SUBSTITUTION:
  case SENSOR_OK:
    break;
  case SENSOR_ERROR:
    RS_ERROR(currentTime << ": sensor error!!!");
    break;
  case SENSOR_UNKNOWN:
    RS_ERROR(currentTime << ": unknown sensor input!!!");
    break;
  default:
    RS_ERROR(currentTime << ": Strange sensor input!!!");
  }

  if ( currentTime.getT() - lastCycleTime.getT() > 1 ||
       ( currentTime.isClockStopped() &&
	 currentTime.getS() - lastCycleTime.getS() > 1) )
   RS_WARNING(currentTime << ": Missed cycles! lastCycleTime = "
	      << lastCycleTime);

  // RS_LOG ( LA_CONTROLLER, "currentTime = " << currentTime << " lastCycleTime = " << lastCycleTime );
  bool new_cycle = ( currentTime > lastCycleTime );
  
  if ( new_cycle ) {
    lastCycleTime = currentTime;
    theControlInterface->onNewCycle(false);
  }

  if ( lastTime > currentTime ) {
    RS_WARNING(currentTime << ": Incorrectly estimated server time! "
	       << "lastTime = " << lastTime);
  }

  return new_cycle;
  
  //return ( currentTime > lastTime ||
  //   last_estimatedTime > estimatedTime );
}


bool Controller::sendCommand(ServerCommand* cmd)
  throw(RSException)
{
  bool res = false;
  if ( cmd != NULL ) {
    RS_LOG( LA_CONTROLLER, theSensors->getCurrentTime() << " " << cmd->serverString() );
    res = theServerInterface->sendCommand(cmd->serverString());
    if ( !coach && !trainer )
      theSensors->analyze(cmd);
    theControlInterface->onActuatorSensorData();
  } else {
    RS_LOG( LA_CONTROLLER, theSensors->getCurrentTime() << " NULL command" );
  }
  return res;  
}


bool Controller::sendCommand(ControlCommand* cmd) throw() {
  if (SetCyclesSinceIOBeforeShutdown* ccmd = 
      dynamic_cast<SetCyclesSinceIOBeforeShutdown *>(cmd) ) {
    maxCyclesSinceIOBeforeShutdown = ccmd->getCycles();
  } else if (SetMinTicksBeforeSendingCommand* ccmd = 
	     dynamic_cast<SetMinTicksBeforeSendingCommand *>(cmd) ) {
    minTicksBeforeSendingCommand = ccmd->getTicks();
  } else if (SetMaxTicksBeforeSendingCommand* ccmd = 
	     dynamic_cast<SetMaxTicksBeforeSendingCommand *>(cmd) ) {
    maxTicksBeforeSendingCommand = ccmd->getTicks();
  } else if (SetTicksBeforeCommandWarning* ccmd = 
	     dynamic_cast<SetTicksBeforeCommandWarning *>(cmd) ) {
    ticksBeforeCommandWarning = ccmd->getTicks();
  } else if (SetTicksBetweenCycles* ccmd = 
	     dynamic_cast<SetTicksBetweenCycles *>(cmd) ) {
    ticksBetweenCycles = ccmd->getTicks();
  } else if (SetTicksBetweenSenses* ccmd = 
	     dynamic_cast<SetTicksBetweenSenses *>(cmd) ) {
    ticksBetweenSenses = ccmd->getTicks();
  } else if (SetTicksBetweenVisuals* ccmd = 
	     dynamic_cast<SetTicksBetweenVisuals *>(cmd) ) {
    ticksBetweenVisuals = ccmd->getTicks();
  } else 
    RS_ERROR ("Unknown controller command sent to the controller!");
  theSensors->executeControlCommand(cmd);
  return true;
}

bool
Controller::receiveInput(char* const input) throw(RSException) {
  return theServerInterface->receiveInput(input);
}

void
Controller::delayedActions() throw() {
  if ((currentTime.getT() != 0)) {
    RS_WARNING("The server missed action (arrived to late?) " + toString(currentTime));
    theControlInterface->onDelayedActions();
  }
}

void
Controller::missingActions() throw() {
  if ((currentTime.getT() != 0)) {
    RS_WARNING("The server executed to many actions last cycle " + toString(currentTime));
    theControlInterface->onMissingActions();
  }
}

// Ugly fix to avoid anachronism when initializing the signal handlers.
extern "C" RETSIGTYPE externSigAlarmHandler(int sig) {
  Controller::currentControllerObject->sigAlarmHandler(sig);
}

// Ugly fix to avoid anachronism when initializing the signal handlers.
extern "C" RETSIGTYPE externSigIOHandler(int sig) {
  Controller::currentControllerObject->sigIOHandler(sig);
}

RS_END_NAMESPACE

/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "Controller.icc"
#endif
