/* -*- Mode: C++ -*-
 * $Id: Controller.h,v 1.12 2004/08/31 17:12:47 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Controller.h
 * Unitname: BasicSystem
 * $Revision: 1.12 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2004/08/31 17:12:47 $
 *
 *
 * HISTORY
 *
 * $Log: Controller.h,v $
 * Revision 1.12  2004/08/31 17:12:47  frehe
 * Added a stop method.
 *
 * Revision 1.11  2004/08/29 13:36:19  frehe
 * Updated the handling of time for the coach and added a callback for coach visual data.
 *
 * Revision 1.10  2003/08/20 12:28:04  s02davro
 * New count counter, attentiontos etc.
 *
 * Revision 1.9  2003/07/21 08:51:27  s02davro
 * Fixed starvation problem, updated signal handlers, added support for new message types. etc.
 *
 * Revision 1.8  2002/09/04 13:25:09  s02davro
 * Removed deprecated methods in Controller and Decision
 *
 * Revision 1.7  2002/08/31 15:04:34  s02davro
 * Updated the signal handlers, new main loop - run()
 *
 * Revision 1.6  2002/08/02 21:49:24  s02davro
 * Small changes, added somm big todos.
 *
 * Revision 1.5  2001/11/15 16:48:48  frehe
 * Added support for a coach.
 *
 * Revision 1.4  2001/08/19 23:32:20  frehe
 * Updated 2001-08-19
 *
 * Revision 1.3  2001/08/13 16:43:12  frehe
 * Made updates to compile with new autotools and RCCParser.
 *
 * Revision 1.2  2000/09/04 15:22:58  frehe
 * Added an #include statement to make the documentation better.
 *
 * Revision 1.1  2000/09/03 20:41:08  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file Controller.h
 * Include file for the class Controller.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.12 $
 * $Author: frehe $
 * $Date: 2004/08/31 17:12:47 $
 */


#ifndef CONTROLLER_H
#define CONTROLLER_H


#include <signal.h>
#include <string>

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif


#include "basic_system_compiler_directives.h"
#include "Time.h"
#include "Timer.h"
#include "RSException.h"


RS_BEGIN_NAMESPACE


class Actuators;
class Sensors;
class ControlInterface;
class ServerInterface;
class ServerCommand;
class ControlCommand;


/** Used to represent signal handling functions. */
typedef RETSIGTYPE Sigfunc(int);


/** Represent the different results the controller can return. */
enum ControllerResult {
  CONTROLLER_OK,
  CONTROLLER_ERROR
};


/**
 * @class Controller Controller.h
 * This class is the controller of the agent.
 * It will take care of all synchronization between the different
 * parts of the agent, including sending commands to the soccer
 * server and generating events to the decision maker.
 * <!--Controller is a model of the design pattern Singleton. The only
 * instance is accessible through Controller* Controller::instance(); -->
 *
 * @todo Add documentation.
 * @todo Fix the syncronization problems. The commands should be able
 *       to updated many times during the cycle, and sent att the best
 *       time. After see information is an example of a good time.
 * @todo Fix tick problem. If we get visual information the alarm ticks
 *       isn't running as fast as they should and we risk missing 
 *       actions. Futhermore the Controller should take care of the
 *       hole count and avoid generating clashes. This is a rather
 *       _big_ todo. Good information about syncronization whith server
 *       can be found in the UvA Trilearn documentation.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.12 $
 * $Author: frehe $
 * $Date: 2004/08/31 17:12:47 $
 */
class Controller {
public:
  Controller(ServerInterface* const, Sensors* const,
	     Actuators* const, ControlInterface* const,
	     const UInt16 cycle_length, const UInt16 recv_step,
	     const UInt16 sense_step, const UInt16 visual_step,
	     const UInt16 max_sigalrms, const UInt16 min_ticks,
	     const UInt16 ticks_before_cmd_warning,
	     const UInt16 max_ticks, 
	     const bool _coach = false,
	     const bool _trainer = false) throw();
  
  // Not used since the constructor is nice.
  // The only access point to the controller object,
  // since it is a singleton.
  // static Controller* instance();
  
  ~Controller() throw();
  
  bool serverAlive() const throw();

  bool init() throw(RSException);
  bool start() throw(RSException);
  bool stop() throw(RSException);
  int  run();

  bool sendCoachCommand(const std::string& msg) throw(RSException);
  bool sendTrainerCommand(const std::string& msg) throw(RSException);
  bool sendCoachOrTrainerCommand(const std::string& msg) throw(RSException);
  
  bool updateTime(const UInt16 new_time) throw();
  Time getCurrentTime() const throw();
  
  bool sendCommand(ServerCommand*) throw(RSException);
  bool sendCommand(ControlCommand*) throw();
  bool receiveInput(char* const input) throw(RSException);

  void delayedActions() throw();
  void missingActions() throw();

  static Controller* currentControllerObject;
  void sigAlarmHandler(int);
  void sigIOHandler(int);

  void handleSigAlarm();
  void handleSigIO();

private:
  ServerInterface* const theServerInterface;
  Sensors* const theSensors;
  Actuators* const theActuators;
  ControlInterface* const theControlInterface;

#ifdef RS_ENABLE_COACH_DEBUGGING
  ServerInterface* theCoachInterface;
#endif

  UInt16 maxCyclesSinceIOBeforeShutdown;
  UInt16 minTicksBeforeSendingCommand;
  UInt16 ticksBeforeCommandWarning;
  UInt16 maxTicksBeforeSendingCommand;

  UInt16 cycleLength;
  UInt16 tickLength;
  UInt16 ticksBetweenCycles;
  UInt16 ticksBetweenSenses;
  UInt16 ticksBetweenVisuals;

  /** The timer used for the internal clock. */
  Timer commandTimer;

  /** Flag used to start the internal clock. */
  bool startClock;

  /** If true then control a coach. */
  bool coach; 

  /** If true then control a trainer. */
  bool trainer;
 
  /** Count the number of ticks this cycle. */
  UInt16 ticks;
  /** Count the number of ticks since last sensory sensor data. */
  UInt16 senseTicks;
  /** Count the number of ticks since last visual sensor data. */
  UInt16 visualTicks;
  /** The time the last cycle started. */
  Time lastCycleTime;
  /** The time of the last aural sensor data. */
  Time lastAuralTime;
  /** The time of the last sensory sensor data. */
  Time lastSenseTime;
  /** The time of the last coach visual sensor data. */
  Time lastCoachVisualTime;
  /** The time of the last visual sensor data. */
  Time lastVisualTime;
  /** The time last reported from the server. */
  UInt16 lastKnownServerTime;
  /** The current time, according to the agent. Stored in CTime. */
  Time& currentTime;
  /** The number of cycles added to the last known server
    time to get the current time, aCording to the agent. */
  UInt16 estimatedTime;
  /** The number of stopped cycles. */
  UInt16 stoppedCycles;
  
  char* sensorData;
  UInt16 sigalarmsSinceSigio;
  Time lastActTime;

  sigset_t initHandlers();

  sigset_t sigalarmask;
  sigset_t sigiomask;
  sigset_t allSigsMask;

  bool active;
  bool IOFlag;
  bool AlarmFlag;

  UInt ticksSinceLastHandle;
  UInt ioFlagsSinceHandle;

  struct timeval timeOfFirstTick;
  struct timeval currentTickTime;
  //Sigfunc* Signal(int signo, Sigfunc* func, sigset_t* mask = NULL);

  Controller(const Controller&);
  Controller& operator= (const Controller&);
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "Controller.icc"
#endif


#endif // !CONTROLLER_H
