/* -*- Mode: C++ -*-
 * $Id: SensorInterface.h,v 1.13 2003/10/14 18:07:52 s02davro Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: SensorInterface.h
 * Unitname: BasicSystem
 * $Revision: 1.13 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: s02davro $ $Date: 2003/10/14 18:07:52 $
 *
 *
 * HISTORY
 *
 * $Log: SensorInterface.h,v $
 * Revision 1.13  2003/10/14 18:07:52  s02davro
 * Added call for command to sensorinterface
 *
 * Revision 1.12  2003/07/21 08:55:26  s02davro
 * Change the communication interfaces to support v. 9
 *
 * Revision 1.11  2002/12/07 17:29:04  s02davro
 * Hack in Sensor to fix trainer added CP_trainer to WorldFacts
 *
 * Revision 1.10  2002/08/31 15:05:44  s02davro
 * Updated the sensor interface to the new rccparser
 *
 * Revision 1.9  2002/08/13 10:42:22  s02davro
 * Corrected small documentation misstake.
 *
 * Revision 1.8  2002/08/13 10:38:59  s02davro
 * Added support for hearPlayerType, hearPlayerParams and hearServerParams.
 *
 * Revision 1.7  2002/08/02 21:49:49  s02davro
 * Added documentation, converted to abstract baseclass, old SensorInterface is now BasicSensorInterface.
 *
 * Revision 1.6  2002/01/16 09:53:52  frehe
 * Fixed the coach parsing of referee messages.
 *
 * Revision 1.5  2001/10/14 21:21:42  frehe
 * Added better support for messages from self and from other players.
 *
 * Revision 1.4  2001/09/17 21:53:38  frehe
 * Added the relative speed to the seeBall method.
 *
 * Revision 1.3  2001/08/13 16:43:13  frehe
 * Made updates to compile with new autotools and RCCParser.
 *
 * Revision 1.2  2000/09/13 00:02:02  frehe
 * Made minor changes to the forEach methods.
 *
 * Revision 1.1  2000/09/03 20:41:09  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file SensorInterface.h
 * Include file for the class SensorInterface.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.13 $
 * $Author: s02davro $
 * $Date: 2003/10/14 18:07:52 $
 */


#ifndef SENSOR_INTERFACE_H
#define SENSOR_INTERFACE_H


#include <queue>
#include <vector>
#include <string>

#include "library_types.h"
#include "basic_system_types.h"
#include "soccer_types.h"
#include "Sensors.h"
#include "Time.h"


RS_BEGIN_NAMESPACE


/**
 * @class SensorInterface SensorInterface.h
 * This class is the interface between the sensors and the rest of the agent.
 * It's a abstract base class for FrameworkSensorInterface and 
 * BasicSensorInterface.
 * 
 * @todo hearPlayer -> hearOpponent and hearTeammate?
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.13 $
 * $Author: s02davro $
 * $Date: 2003/10/14 18:07:52 $
 */
class SensorInterface {
public:
  /** Constructor */
  SensorInterface();

  /** Virtual destructor since it will be inherited. */
  virtual ~SensorInterface();

  /** Is called when we have a new cycle. */
  virtual void newCycle();

  /** Called when a new cycle is forced. */
  virtual void forceNewCycle();

  /** Called before a new cycle is issued. */
  virtual void updateBeforeNewCycle();

  /** Called from Sensors when we have new visual information. */
  virtual void see(const std::vector<ServerLineObject>& lines,
		   const std::vector<ServerMarkerObject>& markers,
		   const std::vector<ServerPlayerObject>& players,
		   const ServerBallObject& ball) = 0;
  
  /** Called from Sensors when we have new global visual information.(coach) */
  virtual void seeGlobal(const std::vector<CoachPlayerObject>& cplayers,
			 const CoachBallObject& cball) = 0;
  
  /** Called from Sensors when we have new sense information. */
  virtual void sense(const ServerAgentObject& agent) = 0;

  /** Called when we hear the referee. */
  virtual void hearReferee(const PlayMode play_mode) = 0;

  /** Called when we hear a message from our coach. 
   *  Called when hearing trainer (offline coach) or a
   *  non-CLang online coach. */
  virtual void hearCoach(const std::string& message) = 0;

  /** Called when we hear a message from ourself. */
  virtual void hearSelf(const std::string& message) = 0;

  /** Called when we (the online coach) hear another player. */
  virtual void hearPlayerAsCoach(const Team team, 
				 const UIntUnknown nr,
				 const std::string& message) = 0;

  /** Called when we (field player) hears a teammate. */
  virtual void hearTeammate(const UInt unum,
			    const AngleDegUE direction,
			    const std::string& message) = 0;

  /** Called when we (field player) hears another teammate (no message). */
  virtual void hearAnotherTeammate(const UInt unum) = 0;

  /** Called when we (field player) hears a opponent. */
  virtual void hearOpponent(const AngleDegUE direction,
			    const std::string& message) = 0;
  
  /** Called when we (field player) hears another opponent. */
  virtual void hearAnotherOpponent() = 0;

  /** Sets the last command executed by the agent. */
  virtual void setLastSensorType(const SensorType sensor_type) = 0;

  virtual void setLastCommand(ServerCommand* command) = 0;
protected:
  /** Information about if the cycle is forced. */
  bool forcedNewCycle;
  
  /** The forcedCycleTime. */
  Time forcedNewCycleTime;
  
  /** This is a reference to the Time::current_time.
   *  The users can use this value insted of using the Time class directly. */
  Time& currentTime;
  
  /** Reference to Time::clockStopped class variables. 
   *  The users can use this value insted of using the Time class directly. */
  bool& clockStopped;
    
  /** Representation of the last cycles time. */
  Time lastCycleTime;
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "SensorInterface.icc"
#endif


#endif // !SENSOR_INTERFACE_H
