/* -*- Mode: C++ -*-
 * $Id: Sensors.h,v 1.22 2004/08/27 14:57:54 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Sensors.h
 * Unitname: BasicSystem
 * $Revision: 1.22 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2004/08/27 14:57:54 $
 *
 *
 * HISTORY
 *
 * $Log: Sensors.h,v $
 * Revision 1.22  2004/08/27 14:57:54  frehe
 * Added support for (ok team_names...)
 *
 * Revision 1.21  2004/08/23 20:47:20  frehe
 * Added support for move ok.
 *
 * Revision 1.20  2004/08/23 13:23:24  frehe
 * Added support for (ok ear on).
 *
 * Revision 1.19  2003/08/20 12:29:56  s02davro
 * ControlCommand fix, count counter
 *
 * Revision 1.18  2003/07/22 14:45:21  s02davro
 * Logging, arm, tackle, focus + bugfix
 *
 * Revision 1.17  2003/07/21 08:55:59  s02davro
 * Added support for new rccparser and server version 9.
 *
 * Revision 1.16  2002/09/26 15:23:41  s02davro
 * Bugfixes, close, team, coach
 *
 * Revision 1.15  2002/09/23 10:28:42  s02davro
 * Find bug in (close), this information needs to be handled - todo
 *
 * Revision 1.14  2002/09/21 15:38:58  s02davro
 * Update to new (1.0.3) RCCParser
 *
 * Revision 1.13  2002/08/31 15:06:16  s02davro
 * Redesign (again) now handles the coach fullstate information messages
 *
 * Revision 1.12  2002/08/02 21:51:50  s02davro
 * Complete rewrite, now uses rccparser-1.0 and clangparser from the soccerserver.
 *
 * Revision 1.11  2002/01/15 17:19:41  frehe
 * Improved the coach parser.
 *
 * Revision 1.10  2001/11/15 16:48:48  frehe
 * Added support for a coach.
 *
 * Revision 1.9  2001/09/17 21:54:00  frehe
 * Readded support for a debug coach.
 *
 * Revision 1.8  2001/09/17 21:30:09  andbo928
 * Implemented an ugly workaround to get ballspeeds working.
 *
 * Revision 1.7  2001/08/19 23:32:21  frehe
 * Updated 2001-08-19
 *
 * Revision 1.6  2001/08/17 07:15:13  frehe
 * Made changes to compile RoboSoc with RCCParser-0.10.
 *
 * Revision 1.5  2001/08/13 16:43:14  frehe
 * Made updates to compile with new autotools and RCCParser.
 *
 * Revision 1.4  2001/08/13 07:33:16  frehe
 * Updated from laptop 2001-08-13.
 *
 * Revision 1.3  2001/08/02 17:04:22  frehe
 * Updated 2001-08-02
 *
 * Revision 1.2  2000/09/04 08:46:45  frehe
 * Fixed a naming error in the sensors.
 *
 * Revision 1.1  2000/09/03 20:41:10  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file Sensors.h
 * Include file for the class Sensors.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.22 $
 * $Author: frehe $
 * $Date: 2004/08/27 14:57:54 $
 */


#ifndef SENSORS_H
#define SENSORS_H

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <queue>

#include <rccparser/rccparser.h>

#include "basic_system_types.h"
#include "ServerCommand.h"
#include "ControlCommand.h"
#include "WorldFacts.h"

RS_BEGIN_NAMESPACE

/** Used to represent the different types of sensor
  information the agent can receive. */
enum SensorType {
  SENSOR_AURAL_COACH,
  SENSOR_AURAL_PLAYER,
  SENSOR_AURAL_REFEREE,
  SENSOR_AURAL_SCORE,
  SENSOR_AURAL_SELF,
  SENSOR_VISUAL,
  SENSOR_SENSE,
  SENSOR_ACTUATORS,
  SENSOR_INIT,
  SENSOR_ERROR,
  SENSOR_WARNING,
  SENSOR_COACH_AURAL,
  SENSOR_COACH_VISUAL,
  SENSOR_PARAM,
  SENSOR_PLAYER_TYPE,
  SENSOR_PLAYER_PARAM,
  SENSOR_SUBSTITUTION,
  SENSOR_OK,
  SENSOR_UNKNOWN
};

class Sensors;
class Controller;
class SensorInterface;

/**
 * @class Sensors Sensors.h
 * This class is responsible for parsing the server messages
 * and storing them in the sensor interface for further use
 * by the agent.
 *
 * @todo Handle new serverinformation (tackle, point)
 * @todo Syncronization in Controller may force changes here.
 * @todo Handle the information about close objects.
 * @todo Analyze feedback in a correct way.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.22 $
 * $Author: frehe $
 * $Date: 2004/08/27 14:57:54 $
 */

class Sensors : public rcc::Parser {
public:
  /** Constructor, needs a SensorInterface pointer and a coachlanguageparser. 
   *  The rcss::Parser is needed for the rcc:Parser to be able to parse
   *  coach messages. The sensorInterface is called when we see, hear or
   *  sense something new. */
  Sensors(SensorInterface*, 
	  rcss::Parser& coachMessageParser);
  
  /** Init the Sensors
   *  We need a pointer to the Controller to handle holes and clashes 
   *  (Is our commands getting to the soccerserver in time. */
  void init(Controller*);

  /** Analyze new sensor data.
   *  The data is given to the rcc::Parser which calls our doBuildfunctions. */
  void analyze(char* sensor_data);

  /** Analyze feedback from the actuators. */
  void analyze(ServerCommand*);  

  /** Get the current time. 
   *  This meens the last known server time, according to the sensor
   *  information. */
  UInt getCurrentTime() const throw(); 

  /** Get the type of the last sensor information.
   *  This is used by the sensorinterface call the right functions. */
  SensorType getLastSensorType() const throw();

  /** Set the type of the last sensor information. 
   *  This is beeing done in the doBuild functions called by the parser. */
  void setLastSensorType(SensorType sensor_type) throw();

  /** Used to update world facts after a ContolCommand has arrived to controller. */
  void executeControlCommand(ControlCommand* cmd) throw();

protected:

  /** Set the current (last-known-server) time. */
  void setCurrentTime(UInt time) throw();

  /** Update the time in the sensors and the controller */
  /** This according to the time information in the sensor data. */
  void updateTime(int time);

  //@{
  /** @name Builder functions.
   *  This functions are called from the rcc:Parser during the parsing of
   *  sensor data. The functions stores the results in local objects 
   *  (parsed*) and sends the chuncked information to SensorInterface for
   *  more analyzation. */

  //@{
  /** @name Misc parsing */
  void doMsgParsed();
  void doBuildTeamName (const std::string &name);
  void doBuildUNum (int unum);
  void doBuildGoalie();
  void doBuildScore(int time, int our, int opp);
  //@}

  //@{
  /** @name CLang parsing */
  void doParsedCLang( const std::string& msg );
  //@}

  //@{
  /** @name Fullstate information parsing */
  void doBuildFullState( int time );
  void doBuildCounts(int kick, int dash, int turn, int katch,
		     int move, int turn_neck, int change_view, int say);
  void doBuildScore( int our, int opp );
  void doBuildBall( double x, double y, 
		    double delta_x, double delta_y );
  void doBuildPlayer( int unum, int type, double x, double y, 
		      double dx, double dy,
		      double body_dir, double head_dir,
		      int stamina, double effort, double recovery );
  void doBuildPlayer( int unum, int type, double x, double y, 
		      double dx, double dy,
		      double body_dir, double head_dir,
		      double arm_mag, double arm_dir,
		      int stamina, double effort, double recovery );
  void doBuildGoalie( int unum, double x, double y, 
		      double dx, double dy,
		      double body_dir, double head_dir,
		      int stamina, double effort, double recovery );
  void doBuildGoalie( int unum, double x, double y, 
		      double dx, double dy,
		      double body_dir, double head_dir,
		      double arm_mag, double arm_dir,
		      int stamina, double effort, double recovery );
  //@}

  //@{
  /** @name PlayMode parsing */
  void doBuildCatchBallLeftPlayMode();
  void doBuildCatchBallRightPlayMode();
  void doBuildBeforeKickOffPlayMode();
  void doBuildTimeOverPlayMode();
  void doBuildPlayOnPlayMode();
  void doBuildKickOffLeftPlayMode();
  void doBuildKickOffRightPlayMode();
  void doBuildKickInLeftPlayMode();
  void doBuildKickInRightPlayMode();
  void doBuildFreeKickLeftPlayMode();
  void doBuildFreeKickRightPlayMode();
  void doBuildCornerKickLeftPlayMode();
  void doBuildCornerKickRightPlayMode();
  void doBuildGoalKickLeftPlayMode();
  void doBuildGoalKickRightPlayMode();
  void doBuildAfterGoalLeftPlayMode();
  void doBuildAfterGoalRightPlayMode();
  void doBuildDropBallPlayMode();
  void doBuildOffSideLeftPlayMode();
  void doBuildOffsideRightPlayMode();
  void doBuildPenaltyKickLeftPlayMode();
  void doBuildPenaltyKickRightPlayMode();
  void doBuildFirstHalfOverPlayMode();
  void doBuildPausePlayMode();
  void doBuildHumanPlayMode();
  void doBuildFoulLeftPlayMode();
  void doBuildFoulRightPlayMode();
  void doBuildFoulChargeLeftPlayMode();
  void doBuildFoulChargeRightPlayMode();
  void doBuildFoulPushLeftPlayMode();
  void doBuildFoulPushRightPlayMode();
  void doBuildFoulMultipleAttackerLeftPlayMode();
  void doBuildFoulMultipleAttackerRightPlayMode();
  void doBuildFoulBallOutLeftPlayMode();
  void doBuildFoulBallOutRightPlayMode();
  void doBuildBackPassLeftPlayMode();
  void doBuildBackPassRightPlayMode();
  void doBuildFreeKickFaultLeftPlayMode();
  void doBuildFreeKickFaultRightPlayMode();
  void doBuildCatchFaultLeftPlayMode();
  void doBuildCatchFaultRightPlayMode();
  void doBuildIndirectFreeKickLeftPlayMode();
  void doBuildIndirectFreeKickRightPlayMode();
  void doBuildPenaltySetupLeftPlayMode();
  void doBuildPenaltySetupRightPlayMode();
  void doBuildPenaltyReadyLeftPlayMode();
  void doBuildPenaltyReadyRightPlayMode();
  void doBuildPenaltyTakenLeftPlayMode();
  void doBuildPenaltyTakenRightPlayMode();
  void doBuildPenaltyMissLeftPlayMode();
  void doBuildPenaltyMissRightPlayMode();
  void doBuildPenaltyScoreLeftPlayMode();
  void doBuildPenaltyScoreRightPlayMode();
  void doBuildPenaltyOnFieldLeftPlayMode();
  void doBuildPenaltyOnFieldRightPlayMode();
  void doBuildPenaltyFoulLeftPlayMode();
  void doBuildPenaltyFoulRightPlayMode();
  void doBuildPenaltyWinnerLeftPlayMode();
  void doBuildPenaltyWinnerRightPlayMode();
  void doBuildPenaltyDrawPlayMode();
  //@}

  //@{
  /** @name Side parsing */
  void doBuildLeftSide();
  void doBuildRightSide();
  void doBuildOppSide();
  void doBuildOurSide();
  void doBuildNeutralSide();
  //@}

  //@{
  /** @name Sense body parsing */
  void doBuildSenseBody( int time, double stamina, double effort,
			 double speed_mag, double speed_head,
			 double head_angle, int kick_count,
			 int dash_count, int turn_count,
			 int say_count, int turn_neck_count,
			 int catch_count, int move_count,
			 int chg_view_count );
  void doBuildArmState(int moveable_in, int expires_in, 
		       double target_x, double target_y,
		       int count);
  void doBuildFocusState(int count);
  void doBuildFocusState(int unum, int count);
  void doBuildTackleState(int expires_in, int count);
  //@}

  //@{
  /** @name view mode parsing */
  void doBuildViewMode();
  void doBuildViewQualityHigh();
  void doBuildViewQualityLow();
  void doBuildViewWidthNarrow();
  void doBuildViewWidthNormal();
  void doBuildViewWidthWide();
  //@}

  //@{
  /** @name parameter parsing */
  void doBuildParam(const std::string &name, int value);
  void doBuildParam(const std::string &name, double value);
  void doBuildParam(const std::string &name, const std::string &value);
  void doBuildServerParam();
  void doBuildPlayerParam();
  void doBuildPlayerType();
  //@}

  //@{
  /** @name player visual parsing */
  void doBuildVisual( int time );
  void doBuildLine( double dist, double dir );
  void doBuildLine( double dir );
  void doBuildFlag( double dist, double dir,
		    double dist_chg, double dir_chg );
  void doBuildFlag( bool close, double dist, double dir );
  void doBuildFlag( bool close, double dir );
  void doBuildGoal( double dist, double dir,
		    double dist_chg, double dir_chg );
  void doBuildGoal( bool close, double dist, double dir );
  void doBuildGoal( bool close, double dir );
  void doBuildPlayerVisBall( double dist, double dir,
			     double dist_chg, double dir_chg );
  void doBuildBall( bool close, double dist, double dir );
  void doBuildBall( bool close, double dir );
  void doBuildPlayer( double dist, double dir,
		      double dist_chg, double dir_chg,
		      double orientation, double head_orientation,
		      double point_dir, bool tackling);
  void doBuildPlayer( double dist, double dir,
		      double point_dir, bool tackling);
  void doBuildPlayer( double dist, double dir,
		      double dist_chg, double dir_chg,
		      double orientation, double head_orientation,
		      bool tackling);
  void doBuildPlayer( bool close,
		      double dist, double dir,
		      bool tackling);
  void doBuildPlayer( bool close, double dir );
  //@}

  //@{
  /** @name coach visual parsing */
  void doBuildGlobalVisual(int time);
  void doBuildGlobalGoal(double x, double y);
  void doBuildGlobalBall(double x, double y, double dx, double dy);
  void doBuildGlobalPlayer(double x, double y, double dx, double dy, 
			   double bodydir, double neckdir, bool tackle);
  void doBuildGlobalPlayer(double x, double y, double dx, double dy, 
			   double bodydir, double neckdir, 
			   double point_dir, bool tackle);
  //@}

  //@{
  /** @name field location parsing */
  void doBuildFlagOffset( int offset );
  void doBuildTopLocation();
  void doBuildBottomLocation();
  void doBuildLeftLocation();
  void doBuildRightLocation();
  void doBuildCenterLocation();
  void doBuildPenaltyLocation();
  void doBuildGoalLocation();
  //@}

  //@{
  /** @name initialisation parsing */
  void doBuildInit( int unum );
  void doBuildInit();
  void doBuildCoachInit();
  //@}

  //@{
  /** @name audio parsing */
  void doBuildCoachAudio( int time, const std::string& msg );
  void doBuildRefAudio( int time );
  void doBuildGoalRefAudio( int time, int score );
  void doBuildUnknownRefAudio( int time, const std::string& message );
  void doBuildPlayerAudio( int time, const std::string& msg );
  void doBuildPlayerAudio( int time, double dir, 
			   const std::string& msg );
  void doBuildPlayerAudio( int time, double dir, 
			   int unum, const std::string& msg );
  void doBuildPlayerAudio( int time, int unum );
  void doBuildPlayerAudio( int time );
  void doBuildSelfAudio( int time, const std::string& msg );
  //@}

  //@{
  /** @name substitution parsing */
  void doBuildSubstitution( int unum, int type );
  void doBuildSubstitution( int unum );
  //@}

  //@{
  /** @name Clang version parsing */
  void doBuildClangPlayerVersionMsg();
  void doBuildClangPlayerVersion(int min, int max);
  //@}

  //@{
  /** @name error parsing */
  void doBuildCantReconnect();
  void doBuildInitError();
  void doBuildNoMoreTeamOrPlayerOrGoalieError();
  void doBuildNoSuchTeamOrAlreadyHaveCoachError();
  void doBuildTooManyMovesError();
  void doBuildUnknownCommandError();
  void doBuildIllegalCommandError();
  void doBuildSayMessageTooLongError();
  void doBuildIllegalModeError();
  void doBuildIllegalObjectFormError();
  void doBuildSaidTooManyFreeformMessagesError();
  void doBuildCannotSayFreeformWhilePlayonError();
  void doBuildSaidTooManyMetaMessagesError();
  void doBuildSaidTooManyAdviceMessagesError();
  void doBuildSaidTooManyInfoMessagesError();
  void doBuildSaidTooManyDefineMessagesError();
  void doBuildCouldNotParseSayError();
  void doBuildSaidTooManyMessagesError();
  void doBuildUnknownError( const std::string& error );
  //@}

  //@{
  /** @name warning parsing */
  void doBuildNoTeamFoundWarning();
  void doBuildNoSuchPlayerWarning();
  void doBuildCannotSubWhilePlayOnWarning();
  void doBuildNoSubsLeftWarning();
  void doBuildMaxOfThatPlayerTypeOnFieldWarning();
  void doBuildCannotChangeGoalieWarning();
  void doBuildCompressionWarning();
  void doBuildUnknownWarning(const std::string& warning);
  //@}

  //@{
  /** @name OK parsing */
  void doBuildClangVerOK( int min, int max );
  void doBuildEyeOK(bool on);
  void doBuildEarOK(bool on);
  void doBuildMoveOK();
  void doBuildSayOK();
  void doBuildTeamNamesOK();
  void doBuildChangePlayerTypeOK(int unum, int type);
  void doBuildCompressionOK(int level);
  void doBuildUnknownOK(const std::string& msg);
  //@}
  
  /** @name Help fuctions for converting Locations to Lines and Markers. */
  //@{
  /** Convert the locations to a line. */
  SideLineName locationToLine() const;

  /** Convert the locations to a markerer. */
  MarkerName locationToMarker() const;
  //@}

  /** The facts about the world. */
  WorldFacts* facts;

  /** Pointer to the agents controller. */
  Controller* theController;

  /** Pointer to the agents sensorInterface. */
  SensorInterface* theSensorInterface;

  /** The type of the last sensor information. 
   *  Set to SENSOR_UNKNOWN during the parsing of new sensor information.
   *  See basic_system_types.h for more information. */
  SensorType lastSensorType;

  /** The current (last-known-server) time. */
  UInt currentTime;

  /** A queue of commands the agent will execute. */
  std::queue<CommandType> commandQueue;
      
  /** Hidden copy-constructor. */
  Sensors(const Sensors&);

  /** Hidden assignment operator. */
  Sensors& operator= (const Sensors&);

  /** The Side parsed from the current sensor data. */
  Side parsedSide;

  /** The PlayMode parsed from the current sensor data. */
  PlayMode parsedPlayMode;

  /** The ViewQuality parsed from the current sensor data. */
  ViewQuality parsedViewQuality;

  /** The ViewWidth parsed from the current sensor data. */
  ViewWidth parsedViewWidth;

  /** The Locations parsed from the current sensor data. */
  std::vector<FieldLocation> parsedLocations;

  /** The FlagOffset parsed from the current sensor data. */
  UIntUnknown parsedFlagOffset;

  /** The uniform number parsed from the current sensor data. */
  UIntUnknown parsedUniformNumber; 

  /** If the player is a goalie according to the current sensor data. */
  bool parsedGoalie;

  /** The team parsed from the current sensor data. */
  Team parsedTeam;

  /** The parsed arm object. */
  ServerArmState parsedArmState;

  /** The parsed focus state. */
  ServerFocusState parsedFocusState;
  
  /** The parsed tackle state. */
  ServerTackleState parsedTackleState;
  
  /** The markers parsed in the current message. */
  std::vector<ServerMarkerObject> parsedMarkers;

  /** The lines parsed in the current message. */
  std::vector<ServerLineObject> parsedLines;

  /** The players parsed in the current message. */
  std::vector<ServerPlayerObject> parsedPlayers;
  
  /** The ball parsed in the current message. */
  ServerBallObject parsedBall;

  /** The agent parsed in the current (sense) message. */
  ServerAgentObject parsedAgent;

  /** The players observed by the coach parsed in the current message. */
  std::vector<CoachPlayerObject> parsedCoachPlayers;

  /** The ball observed by the coach parsed in the current message. */
  CoachBallObject parsedCoachBall;
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "Sensors.icc"
#endif


#endif // !SENSORS_H

