/* -*- Mode: C++ -*-
 * $Id: WorldFacts.h,v 1.10 2004/09/05 20:35:29 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: WorldFacts.h
 * Unitname: BasicSystem
 * $Revision: 1.10 $
 * Created by: David Rosn 2002-06-27
 * Last modified by $Author: frehe $ $Date: 2004/09/05 20:35:29 $
 *
 *
 * HISTORY
 *
 * $Log: WorldFacts.h,v $
 * Revision 1.10  2004/09/05 20:35:29  frehe
 * Added CP_no_logging.
 *
 * Revision 1.9  2004/08/31 17:12:59  frehe
 * Added CP_random_seed.
 *
 * Revision 1.8  2004/08/29 14:41:55  frehe
 * Added methods to test whether teamnames has been set or not.
 *
 * Revision 1.7  2004/08/23 13:22:48  frehe
 * Added CP_module_dir
 *
 * Revision 1.6  2003/12/02 16:16:23  frehe
 * Added a new parameter CP_cycles_before_ball_unknown to represent the number of cycles the ball hasn't been seen before it's unknown.
 *
 * Revision 1.5  2003/07/22 14:45:59  s02davro
 * Changed type of pen_dist_x
 *
 * Revision 1.4  2003/07/21 08:56:36  s02davro
 * Added support for new server parameters (v.9), player parameters and player types
 *
 * Revision 1.3  2002/12/07 17:29:05  s02davro
 * Hack in Sensor to fix trainer added CP_trainer to WorldFacts
 *
 * Revision 1.2  2002/10/01 16:57:50  s02davro
 * Added version 8 parameters in WorldFacts
 *
 * Revision 1.1  2002/08/31 16:09:06  s02davro
 * Moved WorldFacts from Framework to BasicSystem
 *
 * Revision 1.1  2002/08/02 21:58:29  s02davro
 * Created to replace parameterView, better associated with the agents WorldModel.
 *
 */

/**
 * @file WorldFacts.h
 * Include file for the class WorldFacts.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.10 $
 * $Author: frehe $
 * $Date: 2004/09/05 20:35:29 $
 */


#ifndef WORLDFACTS_H
#define WORLDFACTS_H

#include "basic_system_types.h"
#include "MarkerObject.h"
#include "LineObject.h"
#include "PlayerType.h"
#include <string>
#include <vector>

RS_BEGIN_NAMESPACE

/**
 * @class WorldFacts WorldFacts.h
 * This class contains information that doesn't change about the world.
 * Values are set using default values, values from configfiles or
 * command line argument.
 * This information is used with ObservedWorld to give the agent
 * a good WorldModel.
 * This class replace the parameterView.
 *
 * @todo Inline functions.
 * @todo Decide what should be included in the class.
 * @todo Move client parameters to different location.?
 * @todo Sort the assorted parameters.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.10 $
 * $Author: frehe $
 * $Date: 2004/09/05 20:35:29 $
 */

class WorldFacts {
public:
  /** Static method to get a constant pointer to the WorldFacts. */
  static const WorldFacts* instance();
  
  /** Static method to get a pointer to the WorldFacts. */
  static WorldFacts* nonConstInstance();

  /** Initializer for the class.
   *  Needs command line arguments to read the server and client
   *  parameters from file or command line. */
  static void init(int argc, char** argv, bool verbose = true);
  
  /** The InputTypes for the parameters. */
  enum InputType {
    IT_INT,
    IT_FLOAT,
    IT_BOOL,
    IT_STRING,
    IT_NONE
  };
  

  /** Destructor for the class. */
  virtual ~WorldFacts();

  /** Read and parse the command line options supported by worldFacts.
    They include @c -f @i file to read @i file as if it was a configuration
    file. You can also set individual parameters on the command line by
    @c -parameter @i value to set the parameter @c parameter to @i value.
    The parameter name is the same as used by the parameter view except for
    the prefix CP_ or SP_. */
  void readCommandLine(int argc, char** argv, int verbose = 0);

  void setOption(const std::string &name, std::string value);
  void setOption(const std::string &name, int value);
  void setOption(const std::string &name, double value);

  /** Read and parse a configuration file assigning parameters values. */ 
  void readOptions(const std::string& filename, int verbose = 0);

  /** Read and parse a line buffer containing parameters and their values. */
  bool readOption(const char* linebuf, int verbose = 0);

  /** Read and assign a single parameter a value. */
  bool readOption(const char* option, const char* value);

  /** Get a Marker. */
  const MarkerObject& getMarker(MarkerName marker) const;

  /** Get a Line. */
  const LineObject& getLine(SideLineName line) const;

  /** Set the side of the playing teams. */
  void setSides(Side our, Side their);

  /** Set the agents uniformnumber. */
  void setUniformNumber(UInt16 my_number);

  /** Set the score of the game. */
  void setScore(Side side, UInt score);

  /** Return the side the agent is playing for. */
  Side getOurSide() const;

  /** Return the opponents side. */
  Side getTheirSide() const;

  /** Return the agents uniformnumber. */
  UInt getUniformNumber() const;

  /** Returns the agents teams score. */
  UInt getOurScore() const;

  /** Returns the opponents score. */
  UInt getTheirScore() const;

  /** Return a MarkerObject representing the opponents goal. */
  const MarkerObject& getTheirGoal() const;

  /** Return a MarkerObject representing the agents teams goal. */
  const MarkerObject& getOurGoal() const;

  /** The server parameters has been initialized from the server. */
  void serverParamsInitialized();

  /** The player parameters has been initialized from the server. */
  void playerParamsInitialized();

  /** The heterogenous player types has been initialized from the server. */
  void playerTypeComplete();

  /** Return a vector of the available player types. */
  const std::vector<PlayerType>& getPlayerTypes() const;

  /** Return a player type specificed by id. */
  PlayerType getPlayerType(UInt id) const;

  /** Return a teammates type specificed by number. */
  PlayerType getTeammateType(UInt number) const;

  /** Return current players type. */
  PlayerType getMyType() const;  
  
  /** Sets a teammates type. */
  void setTeammateType(const UInt number, const UInt id);

  /** Returns true iff our teamname is known. */
  bool isOurTeamNameKnown() const;
  
  /** Returns true iff their teamname is known. */
  bool isTheirTeamNameKnown() const;
  
  //@{
  /** @name Server parameters
      Moved here from ParameterView */

  /** Client parameters 
   *  Might be moved to other location. */
  /** If true no logging will be made. */
  bool CP_no_logging;
  
  /** Name of the team. */
  std::string CP_team_name;

  /** Name of the opponent team. */
  std::string CP_their_team_name;

  /** If the agent is a goalie. */
  bool        CP_goalie;

  /** If the agent is a trainer. */
  bool        CP_trainer;

  /** The number of cycles the ball isn't seen before it's unknown. */
  UInt        CP_cycles_before_unknown_ball;
  
  /** Control parameters. */
  UInt        CP_cycles_since_io_before_shutdown;
  UInt        CP_min_ticks_before_sending_command;
  UInt        CP_ticks_before_command_warning;
  UInt        CP_max_ticks_before_sending_command;

  /** CLang version parameters. */
  UInt        CP_clang_version_min;
  UInt        CP_clang_version_max;

  /** Extra quantize parameters. */
  Float       CP_dist_change_qstep;
  Float       CP_dir_change_qstep;
  Float       CP_speed_qstep;

  /** The random seed to be used to initialize random number generators. */
  Int         CP_random_seed;
  
  /** Server parameters */
  //@{
  /** @name Basic field parameters. */
  /** Length of the playing field. */
  Float       SP_pitch_length;

  /** Width of the playing field. */
  Float       SP_pitch_width;

  /** Padding for the field, size of the frame containing the field. */ 
  Float       SP_pitch_margin;

  /** Length of the penalty area. */
  Float       SP_penalty_area_length;

  /** Width of the penalty area. */
  Float       SP_penalty_area_width;

  /** Length of the goal area. */
  Float       SP_goal_area_length;

  /** With of the goal area. */
  Float       SP_goal_area_width;

  /** Distance between the penalty spot and the goal. */
  Float       SP_penalty_spot_dist;

  /** The radius of the corner arc. */
  Float       SP_corner_arc_r;

  /** Number of lines in the field. Usually four. =) */
  Int         SP_num_lines;

  /** Number of markers in the field. A lot, see manual for positions.
   *  The markers as MarkerObjects can be received using this class getMarker
   *  method. */
  Int         SP_num_markers;

  /** The width of the goal. */
  Float       SP_goal_width;

  /** Number of players in the team. */
  Int         SP_team_size;
      
  //@}

  //@{ 
  /** @name Visual Parameters. */
  /** The maximum distance when we always see a players uniformnumber. 
   *  The chance of seeing a number decrease linear between unum_far_length
   *  and unum_too_far_length.  */
  Float       SP_unum_far_length;

  /** The maximum distance when we have a chance to see a uniformnumber. 
   *  The chance of seeing a number decrease linear between unum_far_length
   *  and unum_too_far_length.  */
  Float       SP_unum_too_far_length;

  /** The maximum distance when we always see a players team.
   *  The chance of seeing the team decrease linear between team_far_length
   *  and team_too_far_length. */
  Float       SP_team_far_length;

  /** The maximum distance when we have a chance to see a players team. 
   *  The chance of seeing the team decrease linear between team_far_length
   *  and team_too_far_length. */
  Float       SP_team_too_far_length;
  //@}

  //@{ 
  /** @name Server related parameters. */
  /** The version of server protocol the agent will use. */
  Float       SP_version;

  /** The name of the machine hosting the server. */
  std::string SP_host;

  /** The size of the communication buffer. */
  Int         SP_buffer_size; 

  /** Port number for player connection. */
  Int         SP_port;

  /** Port number for coach connection. */
  Int         SP_coach_port;
  
  /** Port number for old coach connection. */
  Int         SP_olcoach_port;

  /** Flag whether the server should report errors. */
  bool	      SP_verbose;

  /** Cycles for dropping the ball after a free kick. */
  Int         SP_drop_ball_time; 
  //@}

  //@{
  /** @name Automation parameters. */
  /** Time to wait for players to connect. */
  /** Connect wait is the number of cycles to wait for all the players
   * to connect before kicking off
   * -1 indicates wait forever */
  Int         SP_connect_wait;

  /** Cycles to wait when game is over. */
  Int         SP_game_over_wait;

  /** Time to wait before the game is automaticly kicked off. */
  Int         SP_kick_off_wait;

  /** Automatic mode, kick off, end, exit */
  bool        SP_auto_mode; 

  /** Command for staring the left team */
  std::string SP_team_l_start;

  /** Command for starting the right team */
  std::string SP_team_r_start;
 //@}

  //@{
  /** @name Game related parameters. */
  /** Factor to slow down simulator and information sending. */
  Int         SP_slow_down_factor;  

  /** The time in seconds of a game half. */
  Int         SP_half_time;        
  
  /** Distance between opponent and ball when a freekick is taken.
   *  See SP_cornerkick_margin, and SP_offside_kick_margin for values 
   *  in those situations. */
  Float       SP_free_kick_margin;

  /** Distance between opponent and ball when a cornerkick is taken. */
  Float       SP_ckick_margin;

  /** Time in cycles to wait until the game begins again after a goal. */
  Int         SP_after_goal_wait;  

  /** Is back passes allowed. */
  bool        SP_back_passes;

  /** Is freekick faults enabled. */
  bool        SP_free_kick_faults;

  /** Number of extra halfs to play. */
  Int         SP_nr_extra_halfs;

  /** Number of normal halfs to play. */
  Int         SP_nr_normal_halfs;
  //@}

  //@{
  /** @name Player parameters. */
  /** The size (radius) of the players. */
  Float       SP_player_size;	

  /** Player speed decay per cycle. */
  Float       SP_player_decay;    	

  /** Random error in player movement. */
  Float       SP_player_rand;	

  /** The weight of the player (for wind calculations.) */
  Float       SP_player_weight;	

  /** The maximum speed of a player. */
  Float       SP_player_speed_max;

  /** The maximum acceleation of a player. */
  Float       SP_player_accel_max;

  /** The random error in player movement for the left team. */
  Float       SP_player_rand_factor_l;

  /** The random error in player movement for the right team. */
  Float       SP_player_rand_factor_r;

  /** Inertia moment for a player. 
   *  This affects the turning angle depending on speed. */
  Float       SP_inertia_moment;

  /** The maximum distance where the agent can feel the ball or players. 
   *  If other players or the ball is inside this feel circle, the agent 
   *  recives visual information about them although his not looking in
   *  the objects direction. */
  Float       SP_feel_distance;

  /** Is multiplied to the actual dashpower to get the players acceleration.
   *  See the manual for more information about how the player accelerates 
   *  when dashing. */
  Float       SP_dash_power_rate;

  /** The angle of the view cone for a player in standard view mode. */
  Float       SP_visible_angle;

  /** Visible distance?? */
  Float       SP_visible_distance;
  //@}

  //@{
  /** @name Parameters related to auditory perception and aural abilities */
  /** The maximum hearing capacity for a player. 
   *  The player can hear hear_inc messages in hear_decay cycles. */
  Int         SP_hear_max;

  /** The number of messages a player can hear in hear_decay cycles. */
  Int         SP_hear_inc;

  /** Decay rate of player hearing capacity. */
  Int         SP_hear_decay;        

  /** The maximum length in bytes of a (by player) spoken message. */
  Int         SP_say_msg_size;

  /** The maximum distance over which a spoken message can be heard. */
  Float       SP_audio_cut_dist;
  //@}

  //@{
  /** @name Parameters related to the tackling ability. */
  /** Distance for applicable backtackle. */
  Float        SP_tackle_back_dist;

  /** Distance for applicable tackle. */
  Float        SP_tackle_dist;

  /** Tackle exponent, used in propability calculations. */
  Int          SP_tackle_exponent;

  /** Tackle power rate. */
  Float        SP_tackle_power_rate;

  /** Tackle width. */
  Float        SP_tackle_width;

  /** The number of cycles a player is immobile after a tackle. */
  Int         SP_tackle_cycles;
  //@}

  //@{
  /** @name Parameters related to the arm moving, pointing ability. */
  /** The number of cycles before the hand is mobile after a pointto. */
  Int         SP_point_to_ban;

  /** The number of cycles the hand will point to a position after a pointto. */
  Int         SP_point_to_duration;
  //@}

  //@{
  /** @name Stamina, recovery and effort related parameters. */
  /** The minimum stamina for a player. */
  Float       SP_stamina_min;

  /** The maximum stamina for a player. */
  Float       SP_stamina_max;

  /** The maximum stamina increase for a player per cycle. */
  Float       SP_stamina_inc_max;

  /** The percentage of maximum stamina below which the recovery decreases. */
  Float       SP_recover_dec_thr;
  
  /** The recovery decrease per cycle. */
  Float       SP_recover_dec;

  /** The minimum recovery for a player. */
  Float       SP_recover_min;

  /** The maximum recovery for a player. */
  Float       SP_recover_max;

  /** The inital recovery for a player. */
  Float       SP_recover_init;

  /** The percentage of maximum stamina below which the effort decreases. */
  Float       SP_effort_dec_thr;

  /** The effort decrease per cycle. */
  Float       SP_effort_dec;

  /** The percentage of maximum stamina above which the effort increases. */
  Float       SP_effort_inc_thr;

  /** The effort increase per cycle. */
  Float       SP_effort_inc;

  /** The minimum effort for a player. */
  Float       SP_effort_min;

  /** The maximum effort for a player. */
  Float       SP_effort_max;

  /** Effort init. */
  Int          SP_effort_init;
  //@}

  //@{
  /** @name Parameters related to the goalies catching ability. */
  /** The length of the area where the goalies is able to catch the ball. */
  Float       SP_catchable_area_l;

  /** The width of the area where the goalies is able to catch the ball. */
  Float       SP_catchable_area_w;

  /** The propability a catch i successful. */
  Float       SP_catch_probability;

  /** Cycles after a catch in which the goalie cannot catch again. 
   *  Time for goalie to stand up and prepare for a new safe, that is. */
  Int         SP_catch_ban_cycle;

  /** Maximum times the goalie is allowed to move (not dash) after a catch. */
  Int         SP_goalie_max_moves;

  /** Enforce proper goal kicks (out of penalty area). */
  bool        SP_proper_goal_kicks;

  /** Maximum times the goalie may try to do a correct goalkick. */
  Int         SP_max_goal_kicks;
  //@}
      
  //@{
  /** @name Ball related parameters. */
  /** The size (radius) of the ball. */
  Float       SP_ball_size;		

  /** Ball speed decay per cycle. */
  Float       SP_ball_decay;	

  /** Random error in ball movement. */
  Float       SP_ball_rand;		

  /** The weight of the ball (for wind calculations.) */
  Float       SP_ball_weight;

  /** The maximum speed of the ball. */
  Float       SP_ball_speed_max;

  /** The maximum acceleration of the ball. */
  Float       SP_ball_accel_max;

  /** Velocity of a stopped ball. */
  Float        SP_stopped_ball_vel;
  //@}

  //@{
  /** @name Wind related parameters. */
  /** The wind force. */
  Float       SP_wind_force;
  
  /** The wind direction. */
  Float       SP_wind_dir;        

  /** The size of random error in wind force (I think).
   *  Server code says: wind force for force. */
  Float       SP_wind_rand;

  /** The random error in wind direction.
   *  Server code says: wind angle for rand. */
  Float       SP_wind_ang;

  /** If the is no wind (the usual case). */
  bool       SP_wind_none;

  /** If there is a random component in the wind. See parameters above. */
  bool       SP_wind_random;
  //@}

  //@{
  /** @name Kicking related parameters. */
  /** Margin around player in which the ball is kickable. 
   *  The ball is kickable if the distance is less then 
   *  kickable_margin + ball_size + player_size. */
  Float       SP_kickable_margin;

  /** The area in which the ball is kickable. 
   *  kickable_margin + ball_size + player_size. */
  Float       SP_kickable_area;

  /** Is multiplied to the actual kickpower to get the balls acceleration. 
   *  See the manual for more information about how the ball is accelerated 
   *  by kicking */
  Float       SP_kick_power_rate;	

  /** The random error in direction when kicking the ball. */
  Float       SP_kick_rand;

  /** The random error in direction when kicking for the left team. */
  Float       SP_kick_rand_factor_l;

  /** The random error in direction when kicking for the right team. */
  Float       SP_kick_rand_factor_r;

  /** Ball control radius, used to be used when calculating the kick power. 
   *  I think this its deprecated, use kickable_margin instead. */
  Float       SP_control_radius;

  /** This is actually control_radius - player_size. 
   *  I think this its deprecated, use kickable_margin instead. */
  Float       SP_control_radius_width;
  //@}

  //@{
  //@name Quantization parameters. */
  /** Quantization step for moving objects. */
  Float       SP_dist_qstep;
  
  /** Quantization step for stationary objects. */
  Float       SP_land_qstep;        

  /** Quantization step for angles. */
  Float       SP_angle_qstep;

  /** Quantization step for moving objects for the left team. */
  Float       SP_dist_qstep_l;     

  /** Quantization step for moving objects for the right team. */
  Float       SP_dist_qstep_r;        

  /** Quantization step for stationary objects for the left team. */
  Float       SP_land_qstep_l;     

  /** Quantization step for stationary objects for the right team. */
  Float       SP_land_qstep_r;     

  /** Quantization step for angles for the left team. */
  Float       SP_angle_qstep_l;     

  /** Quantization step for angles for the right team. */
  Float       SP_angle_qstep_r;        
  //@}

  //@{
  /** Parameters associated with range for commands. */
  /** Maximum power for kick/dash. */
  Float       SP_maxpower;		

  /** Minimum power for kick/dash. */
  Float       SP_minpower;

  /** Maximum angle for turn/kick. */
  Float       SP_maxmoment;		

  /** Minimum angle for turn/kick. */
  Float       SP_minmoment;

  /** Maximum angle for turnneck. */
  Float       SP_maxneckang;

  /** Minimum angle for turnneck. */
  Float       SP_minneckang;

  /** Maximum neckangle relative to body. */
  Float       SP_maxneckmoment;

  /** Minimum neckangle relative to body. */
  Float       SP_minneckmoment;
  //@}
        
  //@{
  /** @name Coach related parameters. */
  /** Maximum number of coach messages possible. */
  Int         SP_say_coach_cnt_max;

  /** Maximum size of a coach message. */
  Int         SP_say_coach_msg_size;

  /** Time window which controls how many coach messages can be sent. */
  Int         SP_clang_win_size;

  /** Number of define messages by coach per window. */
  Int         SP_clang_define_win;

  /** Number of meta messages by coach per window. */
  Int         SP_clang_meta_win;

  /** Number of advice messages by coach per window. */
  Int         SP_clang_advice_win;

  /** Number of info messages by coach per window. */
  Int         SP_clang_info_win;

  /** Number of del messages from coach per window */
  Int         SP_clang_del_win;

  /** Number of rule messages from coach per window. */
  Int         SP_clang_rule_win;


  /** Delay of coach message, the number of cycles between send and receive. */
  Int         SP_clang_mess_delay;

  /** The number of messages the coach can send per cycle. */
  Int         SP_clang_mess_per_cycle;

  /** Cycles between freeform coach messages. */
  Int         SP_freeform_send_period;

  /** Cycles for ?? */
  Int         SP_freeform_wait_period;

  /** The time in ms between (fullstate) visual information to the coach. */
  Int         SP_send_vi_step;  

  /** Coach mode. */
  bool	      SP_coach;

  /** Coach with referee mode. */
  bool	      SP_coach_w_referee;

  /** Old format for hear command for the coach. */
  bool        SP_old_coach_hear;
  //@}

  //@{
  /** @name Offside related parameters. */
  /** Whether we using offside rules. */
  bool	      SP_use_offside;

  /** Radius around the ball in which a player can be offside. */
  Float       SP_offside_active_area_size;

  /** Distance between opponent and ball when a offsidekick is taken. */
  Float       SP_offside_kick_margin;

  /** Whether it's allowed to kick off from an offside position. */
  bool	      SP_forbid_kick_off_offside;
  //@}


  //@{
  /** @name Handicap parameters. */
  /** Starting score of the left team. */
  Int         SP_start_goal_l;

  /** Starting score of the right team. */
  Int         SP_start_goal_r;

  /** Flag whether to send fullstate information to the left team. */
  Int         SP_fullstate_l;

  /** Flag whether to send fullstate information to the right team. */
  Int         SP_fullstate_r;

  /** Whether the server is using team specific actuator noise. */
  bool        SP_team_actuator_noise;

  /** Slowness factor for left team. */
  Int          SP_slowness_on_top_left;

  /** Slowness factor for right team. */
  Int          SP_slowness_on_top_right;
  //@}

  //@{
  /** @name Penalty parameters. */
  /** Time to wait before penalty setup. */
  Int          SP_pen_before_setup_wait;

  /** Maximum extra penalty kicks. */
  Int          SP_pen_max_extra_kicks;
  
  /** Number of enalty kicks. */
  Int          SP_pen_nr_kicks;

  /** Time to wait when penalty is ready. */
  Int          SP_pen_ready_wait;
  
  /** Time to wait after penalty setup. */
  Int          SP_pen_setup_wait;
  
  /** Time to wait after a penalty has been taken. */
  Int          SP_pen_taken_wait;

  /** Distance from the goal where the ball is placed for a penalty. */
  Float        SP_pen_dist_x;

  /** Maximum allowed distance in front of the goal for the goalie. */
  Int          SP_pen_max_goalie_dist_x;

  /** Is multiple kicks allowed during a penaltyshot. */
  bool         SP_pen_allow_mult_kicks;

  /** Should a random winner be picked when a penaltyround result in a draw.*/
  bool         SP_pen_random_winner;
  
  /** Should a draw game result in a penalty shoot out.*/
  bool         SP_penalty_shoot_outs;
  //@}
 
  //@{
  /** @name Keepaway parameters. */
  /** Keepaway mode. */
  bool        SP_keepaway;

  /** Keepaway region length. */
  Int         SP_keepaway_length;

  /** Keepaway region width. */
  Int         SP_keepaway_width;

  /** Automatic start of keepaway. */
  Int        SP_keepaway_start;

  /** Keepaway logging. */
  bool       SP_keepaway_logging;

  /** Keepaway output. */
  std::string SP_keepaway_log_dir;

  /** Keepaway log name. */
  std::string SP_keepaway_log_fixed_name;

  /** Keepaway fixed log. */
  bool SP_keepaway_log_fixed;

  /** Keepaway date stamp on log. */
  bool SP_keepaway_log_dated; 
 //@}

  //@{
  /** @name Logging parameters. */
  /** Log date format. */
  std::string SP_log_date_format;

  /** Is game logging enabled. */
  bool        SP_game_logging;

  /** Game log version. */
  Int         SP_game_log_version;

  /** Date stamp in game log. */
  bool        SP_game_log_dated;

  /** Fixed game log name. */
  bool        SP_game_log_fixed;

  /** Compression level for game log. */
  Int         SP_game_log_compression;

  /** Game log directory. */
  std::string SP_game_log_dir;
  
  /** Game log fixed name. */
  std::string SP_game_log_fixed_name;
  
  /** Is text logging enabled */
  bool         SP_text_logging;

  /** Date stamp in text log. */
  bool         SP_text_log_dated;
  
  /** Fixed text log name. */
  bool         SP_text_log_fixed;

  /** Text log compression */
  Int          SP_text_log_compression;

  /** Text log directory. */
  std::string SP_text_log_dir;
  
  /** Text log fixed name. */
  std::string SP_text_log_fixed_name;

  /** ?? */
  Int         SP_log_times;

  /** Write out time of different parts to text log. */
  bool         SP_profile;

  /** Record player commands in game log. */
  bool         SP_record_messages;    
  //@}

  //@{
  /** @name Other assorted parameters. */

  /** The time in ms between sense body information. */
  Int         SP_sense_body_step;

  /** The time in ms between simulator step. */
  Int         SP_simulator_step;

  /** The time in ms between simulator udp send step. */
  Int         SP_send_step;

  /** The time in ms between simulator udp receive step. */
  Int         SP_recv_step;

  /** The least common multiplier of all the above steps. */
  Int         SP_lcm_step;

  /** Landmark file. */
  std::string SP_landmark_file;

  /** Module directory. */
  std::string SP_module_dir;
  
  /** Send player commands to monitor. */
  bool        SP_send_comms;

  /** Work in sycronized mode (send (think) and (done)) */
  bool        SP_synch_mode;

  /** Player sends all messages before this offset value a start of cycle. */
  Int         SP_synch_offset;

  /** Micro sleep ?? */
  Int         SP_synch_micro_sleep;
  //@}
  
  //@{
  /** @name Player parameters */
  /** Number of player types. */
  Int         PP_player_types;
  
  /** Number of substitutions allowed during a game. */
  Int         PP_subs_max;
  
  /** Min delta for adjusting player_speed_max. */
  Float       PP_player_speed_max_delta_min;

  /** Max delta for adjusting player_speed_max. */
  Float       PP_player_speed_max_delta_max;

  /** Amount delta is multiplied by for stamina_inc_max. */
  Float       PP_stamina_inc_max_delta_factor;

  /** Min delta for adjusting player_decay. */
  Float       PP_player_decay_delta_min;

  /** Max delta for adjusting player_decay. */
  Float       PP_player_decay_delta_max;

  /** Amount delta is multiplied by for inertia_moment. */
  Float       PP_inertia_moment_delta_factor;

  /** Min delta for adjusting dash_power_rate */
  Float       PP_dash_power_rate_delta_min;

  /** Min delta for adjusting dash_power_rate */
  Float       PP_dash_power_rate_delta_max;
  
  /** Amount delta is multiplied by for player_size. */
  Float       PP_player_size_delta_factor;

  /** Additional min delta for adjusting dash_power_rate  */
  /** Traded with new_stamina_inc_max. */
  Float       PP_new_dash_power_rate_delta_min;

  /** Additional max delta for adjusting dash_power_rate  */
  /** Traded with new_stamina_inc_max. */
  Float       PP_new_dash_power_rate_delta_max;

  /** amount new_dash_power_rate_delta is multiplied by for stamina_inc_max */
  Float       PP_new_stamina_inc_max_delta_factor;

  /** Min delta for adjusting kickable_margin. */
  Float       PP_kickable_margin_delta_min;

  /** Max delta for adjusting kickable_margin. */
  Float       PP_kickable_margin_delta_max;

  /** amount delta is multiplied by for kick_rand */
  Float       PP_kick_rand_delta_factor;

  /** Min delta for adjusting extra_stamina */
  Float       PP_extra_stamina_delta_min;

  /** Max delta for adjusting extra_stamina */
  Float       PP_extra_stamina_delta_max;

  /** amount delta is multiplied by for effort_max */
  Float       PP_effort_max_delta_factor;

  /** amount delta is multiplied by for effort_min */
  Float       PP_effort_min_delta_factor;

private:
  PlayerType playertype;
  Side ourSide;
  Side theirSide;
  UInt uniformNumber;

  UInt ourScore;
  UInt theirScore;
  
  bool serverOptionsInitialized;
  bool playerOptionsInitialized;

  std::vector<MarkerObject> markers;  
  /** Internal type used to parse the parameter options. */
  struct OptionType
  {
    char name[50];
    void *vptr;
    InputType vsize;
  };

  bool readOption(const char* linebuf, int verbose, const OptionType*);
  bool readOption(const char* option, const char* value, const OptionType*);

  /** Set the total number of parameters that can be stored in the
    parameter view. */
  static const int optsize = 300; // Current value is ~204
  OptionType options[optsize];

  static const int poptsize = 50; // Current value is ~26
  OptionType playeroptions[poptsize];

  static const int ptoptsize = 20; // Current value is ~12
  OptionType playertypeoptions[ptoptsize];

  std::vector<PlayerType> playerTypes;

  std::map<UInt, UInt> teammateTypes;
  
  void pM(MarkerName mn, Point p);
  void initMarkers();
  void parameterInit();
  void initOptions();  

  /** The only instance of the class, Singleton. */
  static WorldFacts *onlyInstance;

  /** Hidden constructor. */
  WorldFacts ();

  /** Hidden copy-constructor. */
  WorldFacts (const WorldFacts&);

  /** Hidden assignment operator. */
  WorldFacts& operator= (const WorldFacts&);
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "WorldFacts.icc"
#endif


#endif // !WORLDFACTS_H
