/* -*- Mode: C++ -*-
 * $Id: AttackerDecision.cc,v 1.21 2005/09/19 19:01:27 petan Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: AttackerDecision.cc
 * Unitname: Framework/example/attacker
 * $Revision: 1.21 $
 * Created by: Fredrik Heintz 2004-08-31
 * Last modified by $Author: petan $ $Date: 2005/09/19 19:01:27 $
 *
 *
 * HISTORY
 *
 * $Log: AttackerDecision.cc,v $
 * Revision 1.21  2005/09/19 19:01:27  petan
 * Removed unnecessary couts
 *
 * Revision 1.20  2005/09/19 18:40:13  petan
 * Removed setting scenario to NONE for all types of scenarios.
 *
 * Revision 1.19  2005/09/19 17:41:56  frehe
 * Created a default mode for the attacker to be used in three on three scenarios.
 *
 * Revision 1.18  2005/09/19 16:58:12  frehe
 * Fixed start of scenario.
 *
 * Revision 1.17  2005/09/15 09:12:41  frehe
 * Removed mAct and warnings.
 *
 * Revision 1.16  2005/09/08 14:45:11  frehe
 * Fixed some minor problems with mAct.
 *
 * Revision 1.15  2005/09/08 14:18:46  frehe
 * Handle the new bookkeeping messages from the trainer.
 *
 * Revision 1.14  2005/09/08 13:42:57  frehe
 * Refactored out the different skills.
 *
 * Revision 1.13  2005/09/08 12:42:24  frehe
 * Added a test so that decide isn't called more than once a cycle.
 *
 * Revision 1.12  2005/08/18 20:23:08  frehe
 * Made minor modifications to the score behavior.
 *
 * Revision 1.11  2004/09/20 10:42:50  frehe
 * Improved dribble.
 *
 * Revision 1.10  2004/09/16 11:39:20  frehe
 * Implemented really simple dribbler.
 *
 * Revision 1.9  2004/09/13 19:40:59  frehe
 * Tried to use KickBallView and then changed back to AccelerateBallToVelocity.
 *
 * Revision 1.8  2004/09/13 15:34:57  frehe
 * Implemented simple receiver.
 *
 * Revision 1.7  2004/09/13 14:38:12  frehe
 * Added support for the new coach messages for the scenarios.
 *
 * Revision 1.6  2004/09/12 23:28:45  frehe
 * Added a check to make sure we have new information about the ball.
 *
 * Revision 1.5  2004/09/08 22:29:32  frehe
 * Updated to the new SayCommand.
 *
 * Revision 1.4  2004/09/08 19:00:32  frehe
 * Changed RS_DEBUG into RS_LOG( LA_DEBUG
 *
 * Revision 1.3  2004/09/08 18:33:43  frehe
 * Added support for different scenario modes.
 *
 * Revision 1.2  2004/09/02 21:47:31  frehe
 * Updated logging.
 *
 * Revision 1.1  2004/08/31 23:04:03  frehe
 * Added an example attacker.
 *
 */

/**
 * @file AttackerDecision.cc
 * Implementation of the class AttackerDecision.
 *
 * @par Unit:
 *      Framework/example/attacker
 * $Revision: 1.21 $
 * $Author: petan $
 * $Date: 2005/09/19 19:01:27 $
 */


#include "AttackerDecision.h"
#include "ActuatorInterface.h"
#include "WorldFacts.h"
#include "ViewManager.h"
#include "view_constants.h"
#include "GameView.h"
#include "PlayerView.h"
#include "AgentView.h"
#include "BallView.h"
#include "KickBallView.h"
#include "InitCommand.h"
#include "MoveCommand.h"
#include "AttentionToCommand.h"
#include "DashToPoint.h"
#include "PointToCommand.h"
#include "TackleCommand.h"
#include "TurnBodyToPoint.h"
#include "SearchBall.h"
#include "rs_assert.h"
#include "rs_log.h"
#include "rs_error.h"
#include "string_utils.h"


RS_BEGIN_NAMESPACE


AttackerDecision::AttackerDecision(ActuatorInterface* const actuator_interface)
  : Decision(actuator_interface),
    facts(WorldFacts::instance()),
    mGameView(0),
    mAgentView(0),
    mBallView(0),
    mKickBallView(0),
    mPlayerView(0),
    mRunning(true),
    mScenario(DEFAULT),
    mNewScenario(false),
    mNewScenarioStart(0),
    mScore(),
    mDribble(),
    mPass(),
    mReceivePass(),
    mAlignNeckToBody()
{
  getView(mGameView, GAME_VIEW_ID);
  getView(mAgentView, AGENT_VIEW_ID);
  getView(mBallView, BALL_VIEW_ID);
  getView(mKickBallView, KICKBALL_VIEW_ID);
  getView(mPlayerView, PLAYER_VIEW_ID);
}


AttackerDecision::~AttackerDecision()
{
  ViewManager::instance()->releaseView(GAME_VIEW_ID);
  ViewManager::instance()->releaseView(AGENT_VIEW_ID);
  ViewManager::instance()->releaseView(BALL_VIEW_ID);
  ViewManager::instance()->releaseView(KICKBALL_VIEW_ID);
  ViewManager::instance()->releaseView(PLAYER_VIEW_ID);
}


ServerCommand*
AttackerDecision::initAgent()
{
  InitCommand* init_cmd = new InitCommand(facts->CP_team_name,
					  facts->SP_version,
					  facts->CP_goalie);
  return init_cmd;
}


bool
AttackerDecision::start()
{
  RS_REQUIRE( initialized == true );
  return true;
}


bool
AttackerDecision::decide()
{
  if ( !mRunning ) {
    return false;
  }
  
  static const WorldModelInterface* wmi = WorldModelInterface::instance();
  const WorldModel& wm = wmi->getWorld();
  const AgentObject& agent = wm.getAgent();

  const std::vector<PlayerObject>& opps_wm = wm.getOpponents();
  if ( opps_wm.size() > 0 ) {
    RS_LOG ( LA_DEBUG, "Found " << opps_wm.size() << " opponents" );
  }
  /*
    for ( std::vector<PlayerObject>::const_iterator i = opps_wm.begin();
    i != opps_wm.end(); ++i ) {
    if ( i->isGoalie() ) {
    std::cout << "Found goalie, it's player #" << i->getUniformNumber()
    << std::endl;
    }
    }
  */

  const std::vector<PlayerObject>& opps_view = mPlayerView->getOpponents(0);
  /*
    if ( opps_view.size() > 0 ) {
    std::cout << "Found " << opps_view.size() << " opponents in view."
    << std::endl;
    }
  */

  if ( opps_wm.size() != opps_view.size() ) {
    RS_LOG ( LA_DEBUG, "Something is WRONG opponents in view ("
	     << opps_view.size() << ") and world model (" << opps_wm.size()
	     << ") differ!" );
  }

  const std::vector<PlayerObject>& unknowns_wm = wm.getUnknowns();
  for ( std::vector<PlayerObject>::const_iterator i = unknowns_wm.begin();
	i != unknowns_wm.end(); ++i ) {
    if ( i->isGoalie() ) {
      if ( i->getUniformNumber().isKnown() ) {
	RS_LOG( LA_DEBUG, "Found goalie, it's player #"
		<< i->getUniformNumber() << " on team " << i->getTeam() );
      } else {
	RS_LOG( LA_DEBUG, "Found goalie, it's a player on team "
		<< i->getTeam() );
      }
    }
  }

  /*
    RS_LOG( LA_DEBUG, "Focus team = " << agent.getFocusedTeam()
    << "; focus unum = " << agent.getFocusedUniformNumber() );

    RS_LOG( LA_DEBUG, "PointTo target = " << agent.getPointTarget() );

    RS_LOG( LA_DEBUG, "Arm movable in = "<< agent.getArmMoveableIn()
    << "; expires in = " << agent.getArmExpiresIn() );

    RS_LOG( LA_DEBUG, "Tackle expires in = " << agent.getTackleExpiresIn() );
  */


  // Add a test to make sure that we don't start until we have seen the ball
  if ( mNewScenario ) {
    if ( mGameView->getLastSeeTime().isKnown()
	 && mGameView->getLastSeeTime().getValue() > mNewScenarioStart ) {
      RS_LOG( LA_DEBUG, "saw after the scenario begun: "
	      << mNewScenarioStart << " <= "
	      << mGameView->getLastSeeTime().getValue() << " <= "
	      << Time::currentTime );
      mNewScenario = false;
    } else {
      RS_LOG( LA_DEBUG, "didn't see "
	      << mGameView->getLastSeeTime()
	      << " <= " << mNewScenarioStart );
      return false;
    }
  }

  //static TurnNeckToAbsoluteDirection turn_neck(0, 0);

  /*
    RS_LOG( LA_DEBUG, "AttackerDecision: neckdir = " <<
    agent.getNeckDirection() << " bodydir = " <<
    agent.getBodyDirection() );
  */

  static int count = 0;
  ++count;
  if ( mScenario == SCORE ) {
    //std::cout << "In score mode!" << std::endl;

    // If score is applicable then call it, otherwise nothing
    if ( mScore.applicable() ) {
      actuatorInterface->addToPlan(mScore.generateStep());
      return true;
    } else {
      return false;
    }
  } else if ( mScenario == DRIBBLE ) {
    //std::cout << "In dribble mode!" << std::endl;
    
    // If dribble is applicable then call it, otherwise nothing
    if ( mDribble.applicable() ) {
      actuatorInterface->addToPlan(mDribble.generateStep());
      return true;
    } else {
      return false;
    }
  } else if ( mScenario == PASS ) {
    //std::cout << "In pass mode!" << std::endl;

    if ( mAgentView->getUniformNumber().isUnknown() ) {
      // If I dont know who I am quit
      return false;
    }
    
    if ( mAgentView->getUniformNumber() == 1 ) {
      // I am the passer
      // If pass is applicable then call it, otherwise nothing
      if ( mPass.applicable() ) {
	actuatorInterface->addToPlan(mPass.generateStep());
	return true;
      } else {
	//std::cout << "Pass not applicable" << std::endl;
	return false;
      }
    } else {
      // I am a receiver
      if ( mReceivePass.applicable() ) {
	actuatorInterface->addToPlan(mReceivePass.generateStep());
	return true;
      } else {
	//std::cout << "ReceivePass not applicable" << std::endl;
	return false;
      }
    }

    return false;
  } else if ( mScenario == DEFAULT ) {
    // The agent is a stand alone agent useful in for example three on three

    if ( mScore.applicable() ) {
      actuatorInterface->addToPlan(mScore.generateStep());
      return true;
    }
    
    Point ball_pos;
    if ( mBallView->getPosition(NEXT).isKnown() ) {
      ball_pos = mBallView->getPosition(NEXT).getPoint();
    } else if ( mBallView->getPosition().isKnown() ) {
      ball_pos = mBallView->getPosition().getPoint();
    } else {
      SearchBall search_ball;
      if ( search_ball.applicable() ) {
	//std::cout << "search ball" << std::endl;
	actuatorInterface->addToPlan(search_ball.generateStep());
	return true;
      } else {
	// Something is not right here!!!
	std::cout << "search ball not applicable!" << std::endl;
      }
    }

    Point agent_pos;
    if ( mAgentView->getPosition().isKnown() ) {
      agent_pos = mAgentView->getPosition().getPoint();
    } else {
      // No agent position!
      std::cout << "no agent position" << std::endl;
      return false;
    }
    
    // If we are not turned towards the ball turn towards it
    // otherwise dash towards the ball
    AngleDeg dir_to_ball = agent_pos.getDirectionTo(ball_pos);
    if ( abs(dir_to_ball) > 20 ) {
      TurnBodyToPoint turn(ball_pos);
      if ( turn.applicable() ) {
	//std::cout << "turn to ball dir = " << dir_to_ball << std::endl;
	actuatorInterface->addToPlan(turn.generateStep());
	return true;
      }
    } else {
      DashToPoint dash(ball_pos);
      if ( dash.applicable() ) {
	//std::cout << "dash to ball" << std::endl;
	actuatorInterface->addToPlan(dash.generateStep());
	return true;
      }
    }
    return false;
  } else {
    //std::cout << "No scenario!" << std::endl;
  }

  /*
    if ( mAlignNeckToBody.applicable() ) {
    actuatorInterface->addToPlan(mAlignNeckToBody.generateStep());
    return true;
    }
  */
  /*
    if ( turn_neck.applicable() ) {

    }
  */
  /*
    if ( count == 10 ) {
    count = 0;
    actuatorInterface->addToPlan(new SayCommand("Hello"));
    return true;
    } else if ( count == 1 ) {
    actuatorInterface->addToPlan(new AttentionToCommand(T_OUR_TEAM, 2));
    return true;
    } else if ( count == 2 ) {
    actuatorInterface->addToPlan(new AttentionToCommand(T_THEIR_TEAM, 1));
    return true;
    } else if ( count == 3 ) {
    //actuatorInterface->addToPlan(new AttentionToCommand());
    //return true;
    } else if ( count == 4 ) {
    if ( wm.getBall().getPosition().isKnown()
    && wm.getAgent().getPosition().isKnown() ) {
    actuatorInterface->addToPlan(new PointToCommand(wm.getAgent().getPosition().getPoint().getVectorTo(wm.getBall().getPosition().getPoint())));
    } else {
    actuatorInterface->addToPlan(new PointToCommand(1, 45));
    }
    return true;
    } else if ( count == 5 ) {
    actuatorInterface->addToPlan(new PointToCommand());
    return true;
    } else if ( count == 6 ) {
    actuatorInterface->addToPlan(new TackleCommand(100));
    return true;
    }
  */
  return false;
}


void
AttackerDecision::onAuralCoachSensorData()
{
  std::string message = mAgentView->getCoachMessage();
  //std::cout << "Got coach message '" << message << "'" << std::endl;

  if ( message == "start" ) {
    //std::cout << "starting test cases" << std::endl;
    mRunning = true;
  } else if ( message == "finished" ) {
    //std::cout << "test cases finished" << std::endl;
    mRunning = false;
  } else if ( message == "success" || message == "failure" ) {
    RS_LOG ( LA_DEBUG, "last scenario was a " << message );
    switch ( mScenario ) {
    case SCORE:
      mScore.reset();
      mScenario = NONE;
      break;
    case DRIBBLE:
      mDribble.reset();
      mScenario = NONE;
      break;
    case PASS:
      mPass.reset();
      mReceivePass.reset();
      mScenario = NONE;
      break;
    }
  } else if ( message == "score" ) {
    RS_LOG ( LA_DEBUG, "change to score mode" );
    mScenario = SCORE;
    mNewScenario = true;
  } else if ( message.size() > 7
	      && std::string(message, 0, 7) == "dribble" ) {
    std::string min_dribble_dist_str(message, 7, std::string::npos);
    mMinDribbleDist = from_string<int>(min_dribble_dist_str);
    mScenario = DRIBBLE;
    RS_LOG ( LA_DEBUG, "change to dribble mode, dist = " << mMinDribbleDist );
    mNewScenario = true;
  } else if ( message.size() > 4
	      && std::string(message, 0, 4) == "pass" ) {
    if ( message.size() > 7 ) {
      std::string min_pass_dist_str(message, 5, 2);
      mMinPassDist = from_string<int>(min_pass_dist_str);
      std::string max_pass_dist_str(message, 7, std::string::npos);
      mMaxPassDist = from_string<int>(max_pass_dist_str);
    } else {
      mMinPassDist = 0;
      mMaxPassDist = 0;
    }
    RS_LOG ( LA_DEBUG, "change to pass mode, " << mMinPassDist
	     << " < pass dist < " << mMaxPassDist );
    mScenario = PASS;
    mReceivePass.setMinPassDist(mMinPassDist);
    mReceivePass.setMaxPassDist(mMaxPassDist);
    mNewScenario = true;
  }

  // Save the time when the scenario was started
  if ( mNewScenario ) {
    mNewScenarioStart = Time::currentTime;
    RS_LOG( LA_DEBUG, "new scenario started at " << mNewScenarioStart );
  }
}


void
AttackerDecision::onAuralRefereeSensorData()
{
}


void
AttackerDecision::onAuralPlayerSensorData()
{
}


void
AttackerDecision::onAuralSelfSensorData()
{
}


void
AttackerDecision::onInit()
{
  // Position the attacker in front of the goal
  Point movePoint = Point(-40,0);
  MoveCommand* move_cmd = new MoveCommand(movePoint);
  actuatorInterface->addToPlan(move_cmd);
}


void
AttackerDecision::onNewCycle(const bool forced)
{
}


void
AttackerDecision::onCommandWarning()
{
  decide();
}


RS_END_NAMESPACE
