/* -*- Mode: C++ -*-
 * $Id: Dribble.cc,v 1.1 2005/09/08 13:42:57 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Dribble.cc
 * Unitname: Framework/example/attacker
 * $Revision: 1.1 $
 * Created by: Fredrik Heintz 2005-09-08
 * Last modified by $Author: frehe $ $Date: 2005/09/08 13:42:57 $
 *
 *
 * HISTORY
 *
 * $Log: Dribble.cc,v $
 * Revision 1.1  2005/09/08 13:42:57  frehe
 * Refactored out the different skills.
 *
 */

/**
 * @file Dribble.cc
 * Implementation file for the skill class Dribble.
 *
 * @par Unit:
 *      Framework/example/attacker
 * $Revision: 1.1 $
 * $Author: frehe $
 * $Date: 2005/09/08 13:42:57 $
 */


#include "Dribble.h"
#include "BallView.h"
#include "AgentView.h"
#include "PlayerView.h"
#include "view_constants.h"
#include "sector_helpers.h"
#include "rs_error.h"

#include "AccelerateBallToVelocity.h"


RS_BEGIN_NAMESPACE


// --- CONSTRUCTORS ---

Dribble::Dribble()
  : Skill(),
    worldModelInterface(WorldModelInterface::instance()),
    facts(worldModelInterface->getWorldFacts()),
    ballView(NULL),
    agentView(NULL),
    playerView(NULL)
{
  getView(ballView, BALL_VIEW_ID);
  getView(agentView, AGENT_VIEW_ID);
  getView(playerView, PLAYER_VIEW_ID);
}


// --- DESTRUCTOR ---

Dribble::~Dribble()
{
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(PLAYER_VIEW_ID);
}


// --- METHODS ---

void
Dribble::reset()
{
}


Plan
Dribble::generatePlan()
{
  // Return an empty plan.
  return Plan(generateStep());
}


AgentStep
Dribble::generateStep()
{
  // Create an empty step
  AgentStep step;

  // Kick and run after the ball
  if ( ballView->isBallKickable() ) {
    // if no player straight infront kick straight
    // otherwise kick 30 degrees on either side, minimizing the absolute angle

    AngleDeg kick_dir = 0;
    if ( agentView->getPosition().isKnown() ) {
      Point agent_pos = agentView->getPosition().getPoint();

      // Find the best angle to kick in
      typedef std::vector<PlayerObject> PlayerObjects;
      const PlayerObjects& opps = playerView->getOpponents();
      for ( PlayerObjects::const_iterator p = opps.begin();
	    p != opps.end(); ++p ) {
	if ( p->getPosition().isKnown() ) {
	  AngleDeg player_dir = agent_pos.getDirectionTo(p->getPosition().getPoint());
	  if ( abs(player_dir) < 30 ) {
	    if ( player_dir > 0 ) {
	      kick_dir = player_dir - 30;
	    } else {
	      kick_dir = player_dir + 30;
	    }
	  }
	}
      }
    }

    Vector kick_vector(kick_dir, 0.75);
    AccelerateBallToVelocity kick(kick_vector);
    if ( kick.applicable() ) {
      return kick.generateStep();
    }
  }

  if ( ballView->getRelativeVector().isKnown() ) {
    Vector ball_vec = ballView->getRelativeVector().getVector();

    // Turn towards the ball
    if ( abs(ball_vec.getAngle()) > 15 ) {
      step.setBodyCommand(new TurnCommand(agentView->getAngleForTurn(ball_vec.getAngle())));
      return step;
    } else if ( ballView->getPosition().isKnown() ) {
      Point ball_pos = ballView->getPosition().getPoint();
      if ( ballView->getPosition(NEXT).isKnown() ) {
	ball_pos = ballView->getPosition(NEXT).getPoint();
      }
      step.setBodyCommand(new DashCommand(agentView->getPowerForDash(ball_pos)));
      return step;
    }
  }

  // Return the step
  return step;
}


FuzzyBool
Dribble::persistent()
{
  // Never persistent.
  return FuzzyBool(0.0);
}


FuzzyBool
Dribble::applicable()
{
  bool res = ( ballView->isBallKickable()
	       || ballView->getRelativeVector().isKnown() );

  return res;
}


FuzzyBool
Dribble::succeed()
{
  // The skill has not succeeded yet.
  return FuzzyBool(0.0);
}


FuzzyBool
Dribble::failed()
{
  // The skill has not failed yet.
  return FuzzyBool(0.0);
}


RS_END_NAMESPACE
