/* -*- Mode: C++ -*-
 * $Id: Score.cc,v 1.7 2005/09/19 19:01:27 petan Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Score.cc
 * Unitname: Framework/example/attacker
 * $Revision: 1.7 $
 * Created by: Fredrik Heintz 2004-08-31
 * Last modified by $Author: petan $ $Date: 2005/09/19 19:01:27 $
 *
 *
 * HISTORY
 *
 * $Log: Score.cc,v $
 * Revision 1.7  2005/09/19 19:01:27  petan
 * Removed unnecessary couts
 *
 * Revision 1.6  2005/09/08 13:42:58  frehe
 * Refactored out the different skills.
 *
 * Revision 1.5  2005/09/08 08:58:50  frehe
 * Implemented the reset method.
 *
 * Revision 1.4  2005/08/18 20:23:08  frehe
 * Made minor modifications to the score behavior.
 *
 * Revision 1.3  2004/09/20 10:42:38  frehe
 * Improved score.
 *
 * Revision 1.2  2004/09/01 21:29:09  frehe
 * Use RS_LOG instead of std::cout to debug.
 *
 * Revision 1.1  2004/08/31 23:04:03  frehe
 * Added an example attacker.
 *
 */

/**
 * @file Score.cc
 * Implementation file for the skill class Score.
 *
 * @par Unit:
 *      Framework/example/attacker
 * $Revision: 1.7 $
 * $Author: petan $
 * $Date: 2005/09/19 19:01:27 $
 */


#include "Score.h"
#include "BallView.h"
#include "AgentView.h"
#include "PlayerView.h"
#include "view_constants.h"
#include "sector_helpers.h"
#include "rs_error.h"


RS_BEGIN_NAMESPACE


// --- CONSTRUCTORS ---

Score::Score()
  : Skill(),
    worldModelInterface(WorldModelInterface::instance()),
    facts(worldModelInterface->getWorldFacts()),
    ballView(NULL),
    agentView(NULL),
    playerView(NULL)
{
  getView(ballView, BALL_VIEW_ID);
  getView(agentView, AGENT_VIEW_ID);
  getView(playerView, PLAYER_VIEW_ID);
}


// --- DESTRUCTOR ---

Score::~Score()
{
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(PLAYER_VIEW_ID);
}


// --- METHODS ---

void
Score::reset()
{
}


Plan
Score::generatePlan()
{
  // Return an empty plan.
  return Plan(generateStep());
}


AgentStep
Score::generateStep()
{
  // Cache the position of their goal
  static Point their_goal = facts->getTheirGoal().getPosition();

  // Create an empty step
  AgentStep step;

  // Calculate the direction to the ball
  AngleDegUE ball_dir = ballView->getRelativeVector().getAngle();

  if ( ! ball_dir.isKnown() ) {
    RS_ERROR( "Score: ball direction unknown, violates applicable" );
  }
  
  // If we are not within 5 degrees of the ball turn towards where it will be the next time
  if ( ball_dir.isKnown() && abs(ball_dir.getAngle()) > 5 ) {
    AngleDegUE next_ball_dir = ballView->getRelativeVector(NEXT).getAngle();
    TurnCommand* cmd = new TurnCommand(agentView->getAngleForTurn(next_ball_dir.getAngle()));
    step.setBodyCommand(cmd);
  } else {
    // Collect all the obstacles in front of the agent
    typedef std::vector<Point> Points;
    Points obstacles;
    typedef std::vector<PlayerObject> PlayerObjects;
    const PlayerObjects& opps = playerView->getOpponents();
    for ( PlayerObjects::const_iterator p = opps.begin(); p != opps.end(); ++p ) {
      if ( p->getPosition().isKnown() ) {
	obstacles.push_back(p->getPosition().getPoint());
	//std::cout << "Added obstacle " << p->getPosition().getPoint() << std::endl;
      } else {
	//std::cout << "Position unknown" << std::endl;
      }
    }

    AngleDegUE kick_dir;
    if ( agentView->getPosition().isKnown() ) {
      Point agent_pos = agentView->getPosition().getPoint(); 
      // Calculate the largest sector towards the goal
      Sector goal_sector(agent_pos,
			 facts->getMarker(FLAG_GRT).getPosition(),
			 facts->getMarker(FLAG_GRB).getPosition());
      Sector best_sector = get_largest_sector(agent_pos,
					      obstacles, goal_sector);

      kick_dir = best_sector.center();
    } else {
      // Get the direction towards their goal
      RS_ERROR( "Score: agent position unknown, violates applicable" );
      kick_dir = agentView->getPosition().getDirectionTo(their_goal);
    }

    // If the kick direction is known continue
    if ( kick_dir.isKnown()
	 && agentView->getBodyDirection().isKnown() ) {
      // Kick the ball in the direction with maximum power
      kick_dir -= agentView->getBodyDirection().getAngle();
    
      KickCommand* cmd = new KickCommand(100, kick_dir.getAngle());
      step.setBodyCommand(cmd);
    }
  }

  // Return the step
  return step;
}


FuzzyBool
Score::persistent()
{
  // Never persistent.
  return FuzzyBool(0.0);
}


FuzzyBool
Score::applicable()
{
  bool res = ( ballView->isBallKickable()
	       && ballView->getRelativeVector().isKnown()
	       && agentView->getPosition().isKnown()
	       && agentView->getBodyDirection().isKnown() );

  /*
  if ( !res ) {
    if ( !ballView->isBallKickable() ) {
      std::cout << "Ball is not kickable, distance is "
		<< ballView->getRelativeDistance() << std::endl;
    }

    if ( !agentView->getPosition().isKnown() ) {
      std::cout << "Agent position is unknown" << std::endl;
    }

    if ( !agentView->getBodyDirection().isKnown() ) {
      std::cout << "Agent direction is unknown" << std::endl;
    }
  }
  */

  return res;
}


FuzzyBool
Score::succeed()
{
  // The skill has not succeeded yet.
  return FuzzyBool(0.0);
}


FuzzyBool
Score::failed()
{
  // The skill has not failed yet.
  return FuzzyBool(0.0);
}


RS_END_NAMESPACE
