/* -*- Mode: C++ -*-
 * $Id: DefenderDecision.cc,v 1.8 2005/09/19 19:02:01 petan Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: DefenderDecision.cc
 * Unitname: Framework/test/Team
 * $Revision: 1.8 $
 * Created by: Peter Andersson 2003-08-26
 * Last modified by $Author: petan $ $Date: 2005/09/19 19:02:01 $
 *
 *
 * HISTORY
 *
 * $Log: DefenderDecision.cc,v $
 * Revision 1.8  2005/09/19 19:02:01  petan
 * Using DEFENDER_MODE_6 per default after start unless using normal scenarios
 *
 * Revision 1.7  2005/09/19 16:58:21  frehe
 * Fixed start of scenario.
 *
 * Revision 1.6  2004/09/13 14:56:47  frehe
 * Wait until seen the ball before acting.
 *
 * Revision 1.5  2004/09/10 13:36:13  petan
 * Fixed unknown AngleDeg error
 *
 * Revision 1.4  2004/09/08 18:34:19  frehe
 * Fixed parsing the different defender modes.
 *
 * Revision 1.3  2004/09/08 14:45:57  petan
 * Updated coach messages and set persistent cycle to 1
 *
 * Revision 1.2  2004/09/02 09:33:44  petan
 * Fixed defender mode 3-6
 *
 * Revision 1.1  2004/09/01 08:18:02  petan
 * Added defender
 *
 */

/**
 * @file DefenderDecision.cc
 * Implementation of the class DefenderDecision.
 *
 * @par Unit:
 *      Defender
 * $Revision: 1.8 $
 * $Author: petan $
 * $Date: 2005/09/19 19:02:01 $
 */

#include "DefenderDecision.h"
#include "GameView.h"
#include "BallView.h"
#include "AgentView.h"
#include "PlayerView.h"
#include "ActuatorInterface.h"
#include "ViewManager.h"
#include "view_constants.h"
#include "InitCommand.h"
#include "MoveCommand.h"
#include "SayCommand.h"
#include "DashCommand.h"
#include "KickCommand.h"
#include "TurnCommand.h"
#include "TurnNeckCommand.h"
#include "PersistentSeeStrategy.h"
#include "rs_assert.h"
#include "rs_log.h"
#include "string_utils.h"

RS_BEGIN_NAMESPACE

// At which distance the goalie should rush towards the ball
#define RUSHING_DISTANCE 15

// At which delta angle the goalie should turn when moving about
#define MAX_DASH_ANGLE 20

// How far from a point the goalie considers itself in place
#define MIN_DASH_DISTANCE 1

// How far an opponent need to be from the ball to be considered "having" the ball
#define BALL_POSSESSION_DISTANCE 2

// The default leash length for circular zone defence (mode 5 and 6)
#define DEFAULT_LEASH_LENGTH 10

// The minimum distance to the goal for positioning defenders (mode 4 and 6)
#define MIN_DISTANCE_TO_GOAL 14

// Default starting mode
#define DEFAULT_MODE DEFENDER_MODE_6

DefenderDecision::DefenderDecision(ActuatorInterface* const actuator_interface)
  : Decision(actuator_interface),
    viewManager(ViewManager::instance()),
    worldModelInterface(WorldModelInterface::instance()),
    facts(worldModelInterface->getWorldFacts()),
    mGameView(NULL),
    mBallView(NULL),
    mAgentView(NULL),
    mPlayerView(NULL),
    mRunning(false),
    mNewScenario(false),
    mNewScenarioStart(0),
    mMode(NO_DEFENDER_MODE),
    mState(STATE_NORMAL),
    mZoneLeash(true),
    mZoneLeashLength(DEFAULT_LEASH_LENGTH),
    mAgentZonePosition(Point(-30, 0)),
    mTurnBodyToPoint(Point(0,0)),
    mTurnBackToPoint(Point(0,0)),
    mTurnNeckToBall(NULL),
    mDashToPoint(Point(0,0))
{
  ((PersistentSeeStrategy*)worldModelInterface->getSeeStrategy())->setUncertaintyCycleMargin(1);
  getView(mGameView, GAME_VIEW_ID);
  getView(mBallView, BALL_VIEW_ID);
  getView(mAgentView, AGENT_VIEW_ID);
  getView(mPlayerView, PLAYER_VIEW_ID);
}

DefenderDecision::~DefenderDecision() {
  viewManager->releaseView(GAME_VIEW_ID);
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(PLAYER_VIEW_ID);
}

ServerCommand* DefenderDecision::initAgent() {
  InitCommand* init_cmd = new InitCommand(facts->CP_team_name, 
					  facts->SP_version, 
					  facts->CP_goalie);
  return init_cmd;
}

bool DefenderDecision::start() {
  RS_REQUIRE( initialized == true );
  return true;
}

bool DefenderDecision::decide()
{
  if ( !mRunning ) {
    return false;
  }
  
  // Add a test to make sure that we don't start until we have seen the ball
  if ( mNewScenario ) {
    if ( mGameView->getLastSeeTime().isKnown()
	 && mGameView->getLastSeeTime().getValue() > mNewScenarioStart ) {
      RS_LOG( LA_DEBUG, "saw after the scenario begun: "
	      << mNewScenarioStart << " <= "
	      << mGameView->getLastSeeTime().getValue() << " <= "
	      << Time::currentTime );
      mNewScenario = false;
    } else {
      RS_LOG( LA_DEBUG, "didn't see "
	      << mGameView->getLastSeeTime()
	      << " <= " << mNewScenarioStart );
      return false;
    }
  }

  if ( mMode > NO_DEFENDER_MODE ) {
    actuatorInterface->addToPlan(generateStep());
    return true;
  } else {
    // No scenario
    return false;
  }
}

AgentStep DefenderDecision::generateStep()
{
  AgentStep step;
  static Point goalCenter = facts->getOurGoal().getPosition();


  if (mGameView->getPlayMode() == PM_PLAY_ON) {
    // mode 1 - clear ball if able
    if (mMode >= DEFENDER_MODE_2 &&
	mBallView->isBallKickable() &&
	mAgentView->getPosition().isKnown() &&
	mAgentView->getBodyDirection().isKnown()) {
      Point theirGoal = facts->getTheirGoal().getPosition();
      Point agentPosition = mAgentView->getPosition().getPoint();
      AngleDeg kickdir = (agentPosition.getDirectionTo(theirGoal) -
                          mAgentView->getBodyDirection()).getDeg();
      KickCommand* kick = new KickCommand(facts->SP_maxpower, kickdir);
      RS_LOG ( LA_DECISION, "Defender - clear ball towards enemy goal: " << kickdir);
      step.setBodyCommand(kick);
    }
    // Mode 2 - place closer to goal as attackers close in
    else if (mMode >= DEFENDER_MODE_3) {
      Point preferredPosition = this->preferredPosition();

      PointUE agentPosition = mAgentView->getPosition();
      if (agentPosition.isKnown() &&
	  mAgentView->getBodyDirection().isKnown()) {
	AngleDeg moveDir = (agentPosition.getDirectionTo(preferredPosition) -
			    mAgentView->getBodyDirection()).getDeg();

	// Set the position in the skills
	mTurnBodyToPoint.setPosition(preferredPosition);
	mTurnBackToPoint.setPosition(preferredPosition);
	mDashToPoint.setPosition(preferredPosition);
	
	// Turn body if direction is greater than MAX_DASH_ANGLE forwards
	if (abs(moveDir.getDeg()) > MAX_DASH_ANGLE &&
	    MIN_DASH_DISTANCE < agentPosition.getPoint().getDistanceTo(preferredPosition) &&
	    mTurnBodyToPoint.applicable()) {
	  RS_LOG (LA_DECISION, "Turn to point" << preferredPosition);
	  step = mTurnBodyToPoint.generateStep();
	}
	// Dash towards the point if we aren't there
	else if (MIN_DASH_DISTANCE < agentPosition.getPoint().getDistanceTo(preferredPosition) &&
		 mDashToPoint.applicable()) {
	  RS_LOG (LA_DECISION, "Dash to Point " << preferredPosition);
	  step = mDashToPoint.generateStep();
	}
	// If we're there and turned in some strange way
	else if (mBallView->getPosition(NEXT).isKnown()) {
	  RS_LOG (LA_DECISION, "Turn to ball");
	  mTurnBodyToPoint.setPosition(mBallView->getPosition(NEXT).getPoint());
	  if (mTurnBodyToPoint.applicable())
	    step = mTurnBodyToPoint.generateStep();
	}
	else if (mSearchBall.applicable()) {
	  RS_LOG(LA_DECISION, "Search ball");
	  step = mSearchBall.generateStep();
	}
      }
      // Bodydirection or NeckDirection unknown, do turn to see something.
      else {
	RS_LOG(LA_DECISION, "Turn to see");
	TurnCommand* turncmd = new TurnCommand(45);
	step.setBodyCommand(turncmd);
      }
    }
    else if (mSearchBall.applicable()) {
      RS_LOG(LA_DECISION, "Search ball");
      return mSearchBall.generateStep();
    }
    // Bodydirection or NeckDirection unknown, do turn to see something.
    else if (mBallView->getPosition().isUnknown()) {
      RS_LOG(LA_DECISION, "Turn to see");
      TurnCommand* turncmd = new TurnCommand(45);
      step.setBodyCommand(turncmd);
      return step;
    }
  }
  
  // Always try to look at ball
  if (mTurnNeckToBall.applicable()) {
    RS_LOG(LA_DECISION, "Turn neck to ball");
    mTurnNeckToBall.setCommand(step.getBodyCommand());
    step.merge(mTurnNeckToBall.generateStep());
  }
  else if (mAlignNeckToBody.applicable()) {
    RS_LOG(LA_DECISION, "Align neck to body");
    step.merge(mAlignNeckToBody.generateStep());
  }

  return step;
}

void DefenderDecision::onAuralCoachSensorData() {
  std::string message = mAgentView->getCoachMessage();
  UIntUnknown uniformNumber = mAgentView->getUniformNumber();
  int uniform = 0;
  if (uniformNumber.isKnown())
    uniform = uniformNumber.getValue();

  if ( message == "start" ) {
    mRunning = true;
    mMode = DEFENDER_MODE_6;
  } else if ( message == "finished" ) {
    mRunning = false;
  } else if ( message == "success" || message == "failure" ) {
    mMode = NO_DEFENDER_MODE;
  } else if ( message.size() > 7 ) {
    // Copy the first 6 characters in the message
    std::string mode(message, 0, 6);
    
    // Copy the rest of the characters in the message
    std::string unum_str(message, 6, std::string::npos);
    
    int unum = from_string<int>(unum_str);
    
    if ( unum == uniform ) {
      if ( mode == "DMode1" ) {
	mMode = DEFENDER_MODE_1;
	mNewScenario = true;
      } else if ( mode == "DMode2" ) {
	mMode = DEFENDER_MODE_2;
	mNewScenario = true;
      } else if ( mode == "DMode3" ) {
	mMode = DEFENDER_MODE_3;
	mNewScenario = true;
      } else if ( mode == "DMode4" ) {
	mMode = DEFENDER_MODE_4;
	mNewScenario = true;
      } else if ( mode == "DMode5" ) {
	mMode = DEFENDER_MODE_5;
	mNewScenario = true;
      } else if ( mode == "DMode6" ) {
	mMode = DEFENDER_MODE_6;
	mNewScenario = true;
      }
    }
  }

  // Save the time when the scenario was started
  if ( mNewScenario ) {
    mNewScenarioStart = Time::currentTime;
    RS_LOG( LA_DEBUG, "new scenario started at " << mNewScenarioStart );
  }
}

void DefenderDecision::onAuralRefereeSensorData() {
}

void DefenderDecision::onAuralPlayerSensorData() {
}

void DefenderDecision::onAuralSelfSensorData() {
}

void DefenderDecision::onInit() {
  // Position the Defender in front of the goal
  Point movePoint = Point(-25,0);
  if (mAgentView->getUniformNumber().isKnown()) {
    UInt nr = mAgentView->getUniformNumber().getValue();
    switch (nr) {      
    case 1:
      movePoint = Point(-45, 0);
      mAgentZonePosition = movePoint;
      break;
    case 2:
      movePoint = Point(-30, -10);
      mAgentZonePosition = movePoint;
      break;
    case 3:
      movePoint = Point(-30, 10);
      mAgentZonePosition = movePoint;
      break;
    }
  }
  MoveCommand* move_cmd = new MoveCommand(movePoint);
  actuatorInterface->addToPlan(move_cmd);
}

void DefenderDecision::onCommandWarning() {
  decide();
}

Point DefenderDecision::preferredPosition()
{
  static Point goalCenter = facts->getOurGoal().getPosition();
  PointUE ballPosition = mBallView->getPosition(NEXT);
  PointUE agentPosition = mAgentView->getPosition();

  Point position;
  
  if (mMode == DEFENDER_MODE_3 &&
      ballPosition.isKnown()) {
    // Hunt ball!
    position = ballPosition.getPoint();
  }
  else if (mMode == DEFENDER_MODE_4 &&
	   mPlayerView->getOpponentClosestToAgent().getPosition().isKnown()) {
    // Position ourselves
    Point opponentPosition = mPlayerView->getOpponentClosestToAgent().getPosition().getPoint();
    Vector opponentGoalVector(goalCenter, opponentPosition);

    //place the defender in a straight line right between goal and opponent
    opponentGoalVector /= 2;

    // Minimum distance to goal
    if (opponentGoalVector.getLength() < MIN_DISTANCE_TO_GOAL)
      opponentGoalVector.setLength(MIN_DISTANCE_TO_GOAL);
    position = goalCenter+opponentGoalVector;
  }
  else if (mMode == DEFENDER_MODE_5 &&
	   ballPosition.isKnown()) {
    // Zone defence, hunt ball if it is in our zone
    if (inZone(ballPosition.getPoint()))
      position = ballPosition.getPoint();
    else
      position = getAgentZonePosition();
  }
  else if (mMode == DEFENDER_MODE_6) {
    // Change zone position
    if (mPlayerView->getOpponentClosestToAgent().getPosition().isKnown()) {
      // Position ourselves
      Point opponentPosition = mPlayerView->getOpponentClosestToAgent().getPosition().getPoint();
      Vector opponentGoalVector(goalCenter, opponentPosition);
      
      //place the defender in a straight line right between goal and opponent
      opponentGoalVector /= 2;
      if (opponentGoalVector.getLength() < MIN_DISTANCE_TO_GOAL)
	opponentGoalVector.setLength(MIN_DISTANCE_TO_GOAL);
      setAgentZonePosition(goalCenter+opponentGoalVector);
    }

    // If ball in zone, hunt!
    if (ballPosition.isKnown() && 
	inZone(ballPosition.getPoint())) {
      position = ballPosition.getPoint();
    }    
    else
      // Otherwise keep in center of zone
      position = getAgentZonePosition();
  }
  else if (agentPosition.isKnown())
    // stay put
    position = agentPosition.getPoint();
  else
    position = getAgentZonePosition();

  return position;
}

bool DefenderDecision::inZone(Point point)
{
  if (mZoneLeash)
    return Vector(point, mAgentZonePosition).getLength() < mZoneLeashLength;
  else
    return point.getX() >= mZoneRectangleX &&
      point.getY() >= mZoneRectangleY &&
      point.getX() <= mZoneRectangleX2 &&
      point.getY() <= mZoneRectangleY2;
}

Point DefenderDecision::getAgentZonePosition()
{
  return mAgentZonePosition;
}

void DefenderDecision::setAgentZonePosition(Point point)
{
  mAgentZonePosition = point;
}

void DefenderDecision::setZoneLeash(Float length)
{
  mZoneLeash = true;
  mZoneLeashLength = length;
}

void DefenderDecision::setZoneRectangle(Point upperLeft, Float width, Float height)
{
  mZoneLeash = false;
  mZoneRectangleX = upperLeft.getX();
  mZoneRectangleY = upperLeft.getY();
  mZoneRectangleX2 = upperLeft.getX() + width;
  mZoneRectangleY2 = upperLeft.getY() + height;
}


RS_END_NAMESPACE
