/* -*- Mode: C++ -*-
 * $Id: defender.cc,v 1.5 2004/09/13 14:40:12 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: defender.cc
 * Unitname: Framework/example/defender
 * $Revision: 1.5 $
 * Created by: Peter Andersson 2003-08-26
 * Last modified by $Author: frehe $ $Date: 2004/09/13 14:40:12 $
 *
 *
 * HISTORY
 * $Log: defender.cc,v $
 * Revision 1.5  2004/09/13 14:40:12  frehe
 * Set the persistance to 2 cycles.
 *
 * Revision 1.4  2004/09/05 20:36:27  frehe
 * Added support for CP_no_logging.
 *
 * Revision 1.3  2004/09/02 09:33:44  petan
 * Fixed defender mode 3-6
 *
 * Revision 1.2  2004/09/01 21:25:41  frehe
 * Made RoboSoc the default teamname.
 *
 * Revision 1.1  2004/09/01 09:19:23  frehe
 * Renamed Defender.cc to defender.cc
 *
 * Revision 1.1  2004/09/01 08:18:02  petan
 * Added defender
 *
 */

/**
 * @file defender.cc
 * Implementation file for an example defender agent.
 *
 * @par Unit:
 *      Framework/example/defender
 * $Revision: 1.5 $
 * $Author: frehe $
 * $Date: 2004/09/13 14:40:12 $
 */


#include <fstream>

#include "framework_compiler_directives.h"
#include "view_constants.h"
#include "FrameworkAgent.h"
#include "WorldModelInterface.h"
#include "DefenderDecision.h"
#include "LogStream.h"
#include "PersistentSeeStrategy.h"
#include "BasicHearStrategy.h"
#include "BasicCommandStrategy.h"
#include "BasicSenseStrategy.h"
#include "BasicInitStrategy.h"
#include "BasicCoachMessageStrategy.h"
#include "BasicNewCycleStrategy.h"
#include "BasicCoachInformationStrategy.h"

RS_BEGIN_NAMESPACE

#if defined(RS_USE_ERROR) && !defined(RS_LOG_ERROR)
extern void rs_warning(const std::string& str,
		       const std::string& file,
		       int line)
{
  std::cerr << "WARNING: " << str
	    << " at line " << line
	    << " in file " << file
	    << std::endl;
}

extern void rs_error(const std::string& str,
		     const std::string& file,
		     int line)
{
  std::cerr << "ERROR: " << str
	    << " at line " << line
	    << " in file " << file
	    << std::endl;
}

extern void rs_fatal(const std::string& str,
		     const std::string& file,
		     int line)
{
  std::cerr << "FATAL: " << str
	    << " at line " << line
	    << " in file " << file
	    << std::endl;
  abort();
}
#endif

#if !defined(RS_USE_EXCEPTION) && !defined(RS_LOG_EXCEPTION)
extern void rs_throw(const std::string& exception,
		     const std::string& msg,
		     const std::string& file,
		     int line)
{
  std::cerr << "RAISED: " << exception << "(\"" << msg << "\")"
	    << " at line " << line
	    << " in file " << file
	    << std::endl;
}
#endif


RS_END_NAMESPACE


int main(int argc, char** argv)
{
#ifdef RS_USE_NAMESPACE
  using namespace RS_NAMESPACE;
#endif

  std::cerr << "Initializing the world...\n";
  WorldFacts::nonConstInstance()->CP_team_name = "RoboSoc";
  WorldFacts::init(argc, argv);

  std::ofstream logFile1("logfile.txt");
  LogStream *logStream = LogStream::instance();

  if ( WorldFacts::instance()->CP_no_logging ) {
    logStream->addLogArea(LA_ROBOSOC, LA_TOP, false, &std::cerr);
  } else {
    logStream->addLogArea(LA_ROBOSOC, LA_TOP, false, &std::cerr);
    logStream->addLogArea(LA_WARNING, LA_ROBOSOC, false, &logFile1,
			  false, &std::cerr);
    logStream->addLogArea(LA_ERROR, LA_ROBOSOC, false, &logFile1,
			  true, &std::cerr);
    logStream->addLogArea(LA_FATAL, LA_ROBOSOC, false, &logFile1,
			  true, &std::cerr);

    logStream->addLogArea(LA_ACTUATORS, LA_ROBOSOC, false, &logFile1);
    logStream->addLogArea(LA_CONTROLLER, LA_ROBOSOC, false, &logFile1);
    logStream->addLogArea(LA_SENSORS, LA_ROBOSOC, false, &logFile1);
    logStream->addLogArea(LA_SERVER_COMMANDS, LA_ROBOSOC, false, &logFile1);
    logStream->addLogArea(LA_SENSOR_DATA, LA_ROBOSOC, false, &logFile1);
    logStream->addLogArea(LA_STRATEGY, LA_ROBOSOC, false, &logFile1);
    logStream->addLogArea(LA_SENSOR_INTERFACE, LA_ROBOSOC, false, &logFile1);
    logStream->addLogArea(LA_DECISION, LA_ROBOSOC, false, &logFile1);
    logStream->addLogArea(LA_FRAMEWORK_AGENT, LA_ROBOSOC, false, &logFile1);
    logStream->addLogArea(LA_AGENT, LA_ROBOSOC, false, &logFile1);
    logStream->addLogArea(LA_SKILL, LA_ROBOSOC, false, &logFile1);
    logStream->addLogArea(LA_COACH, LA_ROBOSOC, false, &logFile1);
    logStream->addLogArea(LA_WORLDMODEL, LA_ROBOSOC, false, &logFile1);
    logStream->addLogArea(LA_TIME, LA_ROBOSOC, false, &logFile1);
  }
  
  try {
    std::cerr << "Initializing the worldModelInterface...\n";
    WorldModelInterface *wmi = WorldModelInterface::nonConstInstance();

    std::cerr << "Setting worldModel update strategies...\n";
    wmi->setSeeStrategy(new PersistentSeeStrategy(2));
    wmi->setHearStrategy(new BasicHearStrategy());
    wmi->setCommandStrategy(new BasicCommandStrategy());
    wmi->setInitStrategy(new BasicInitStrategy());
    wmi->setSenseStrategy(new BasicSenseStrategy());
    wmi->setCoachMessageStrategy(new BasicCoachMessageStrategy());
    wmi->setNewCycleStrategy(new BasicNewCycleStrategy());
    wmi->setCoachInformationStrategy(new BasicCoachInformationStrategy());

    std::cerr << "Starting decision...\n";

    Decision* decision = new DefenderDecision();

    std::cerr << "Initializing the FrameworkAgent...\n";
    FrameworkAgent agent(decision);

    std::cerr << "Starting the agent...\n";
    agent.start();

  } catch (RSException &excpt) {
    std::cerr << "Caught a RoboSoc exception: " << excpt.What() << std::endl;
    return 1;
  } catch (std::exception &excpt) {
    std::cerr << "Caught a standard exception: " << excpt.what() << std::endl;
    return 1;
  } catch (...) {
    std::cerr << "Caught an unknown exception: " << std::endl;
    return 1;
  }

  return 0;
}
