/* -*- Mode: C++ -*-
 * $Id: GoalieDecision.cc,v 1.11 2005/09/19 19:02:26 petan Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: GoalieDecision.cc
 * Unitname: Framework/test/Team
 * $Revision: 1.11 $
 * Created by: Peter Andersson 2003-08-26
 * Last modified by $Author: petan $ $Date: 2005/09/19 19:02:26 $
 *
 *
 * HISTORY
 *
 * $Log: GoalieDecision.cc,v $
 * Revision 1.11  2005/09/19 19:02:26  petan
 * Using GOALIE_MODE_5 per default unless using normal scenarios
 *
 * Revision 1.10  2005/09/19 16:58:25  frehe
 * Fixed start of scenario.
 *
 * Revision 1.9  2004/09/13 14:56:40  frehe
 * Wait until seen the ball before acting.
 *
 * Revision 1.8  2004/09/10 13:36:40  petan
 * Fixed goalie goal kick
 *
 * Revision 1.7  2004/09/08 18:34:01  frehe
 * Fixed parsing the different goalie modes.
 *
 * Revision 1.6  2004/09/01 13:13:26  petan
 * Fixed teleportation and goalie modes
 *
 * Revision 1.5  2004/09/01 09:05:54  frehe
 * Change GoalieMode to GMode
 *
 * Revision 1.4  2004/09/01 08:14:03  petan
 * Added first 2 modes of training
 *
 * Revision 1.3  2004/08/30 15:00:57  petan
 * Kick 45 degrees
 *
 * Revision 1.2  2004/08/18 13:06:40  petan
 * Updated not to turn towards goal and then stand there
 *
 * Revision 1.1  2004/08/18 12:05:36  petan
 * Added example goalie
 *
 *
 */

/**
 * @file GoalieDecision.cc
 * Implementation of the class GoalieDecision.
 *
 * @par Unit:
 *      Goalie
 * $Revision: 1.11 $
 * $Author: petan $
 * $Date: 2005/09/19 19:02:26 $
 */

#include "GoalieDecision.h"
#include "GameView.h"
#include "BallView.h"
#include "AgentView.h"
#include "PlayerView.h"
#include "ActuatorInterface.h"
#include "ViewManager.h"
#include "view_constants.h"
#include "InitCommand.h"
#include "MoveCommand.h"
#include "SayCommand.h"
#include "DashCommand.h"
#include "KickCommand.h"
#include "TurnCommand.h"
#include "TurnNeckCommand.h"
#include "PersistentSeeStrategy.h"
#include "rs_assert.h"
#include "rs_log.h"
#include <cstdlib>

RS_BEGIN_NAMESPACE

// At which distance the goalie should rush towards the ball
#define RUSHING_DISTANCE 15

// At which delta angle the goalie should turn when moving about
#define MAX_DASH_ANGLE 20

// How far from a point the goalie considers itself in place
#define MIN_DASH_DISTANCE 1

GoalieDecision::GoalieDecision(ActuatorInterface* const actuator_interface)
  : Decision(actuator_interface),
    viewManager(ViewManager::instance()),
    worldModelInterface(WorldModelInterface::instance()),
    facts(worldModelInterface->getWorldFacts()),
    mGameView(NULL),
    mBallView(NULL),
    mAgentView(NULL),
    mPlayerView(NULL),
    mRunning(false),
    mNewScenario(false),
    mNewScenarioStart(0),
    mState(STATE_NORMAL),
    mMode(NO_GOALIE_MODE),
    mTurnBodyToPoint(Point(0,0)),
    mTurnBackToPoint(Point(0,0)),
    mTurnNeckToBall(NULL),
    mDashToPoint(Point(0,0))
{
  ((PersistentSeeStrategy*)worldModelInterface->getSeeStrategy())->setUncertaintyCycleMargin(1);
  getView(mGameView, GAME_VIEW_ID);
  getView(mBallView, BALL_VIEW_ID);
  getView(mAgentView, AGENT_VIEW_ID);
  getView(mPlayerView, PLAYER_VIEW_ID);
}

GoalieDecision::~GoalieDecision() {
  viewManager->releaseView(GAME_VIEW_ID);
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(PLAYER_VIEW_ID);
}

ServerCommand* GoalieDecision::initAgent() {
  InitCommand* init_cmd = new InitCommand(facts->CP_team_name, 
					  facts->SP_version, 
					  facts->CP_goalie);
  return init_cmd;
}

bool GoalieDecision::start() {
  RS_REQUIRE( initialized == true );
  return true;
}

bool GoalieDecision::decide()
{
  if ( !mRunning ) {
    return false;
  }
  
  // Add a test to make sure that we don't start until we have seen the ball
  if ( mNewScenario ) {
    if ( mGameView->getLastSeeTime().isKnown()
	 && mGameView->getLastSeeTime().getValue() > mNewScenarioStart ) {
      RS_LOG( LA_DEBUG, "saw after the scenario begun: "
	      << mNewScenarioStart << " <= "
	      << mGameView->getLastSeeTime().getValue() << " <= "
	      << Time::currentTime );
      mNewScenario = false;
    } else {
      RS_LOG( LA_DEBUG, "didn't see "
	      << mGameView->getLastSeeTime()
	      << " <= " << mNewScenarioStart );
      return false;
    }
  }

  if ( mMode > NO_GOALIE_MODE ) {
    actuatorInterface->addToPlan(generateStep());
    return true;
  } else {
    // No scenario
    return false;
  }
}

AgentStep GoalieDecision::generateStep()
{
  AgentStep step;

  // Frequently used variables
  PointUE ballPos = mBallView->getPosition();
  VectorUE ballSpeed = mBallView->getSpeedVector();
  PointUE agentPos = mAgentView->getPosition();
  VectorUE relBallVector = mBallView->getRelativeVector();
  Point goalCenter = facts->getOurGoal().getPosition();

  // If it is the goalies turn to kick the ball, do so.
  if (mGameView->getPlayMode() == PM_OUR_GOALIE_GOT_BALL ||
      mGameView->getPlayMode() == PM_OUR_GOAL_KICK ||
      mGameView->getPlayMode() == PM_OUR_FREE_KICK) {
    
    if (mGameView->isGoalieTeleportationAllowed() &&
	mState != STATE_TELEPORTED) {
      AngleDeg turnAngle = 0;
      // Turn goalie towards center and teleport goalie randomly
      if (agentPos.isKnown() && mAgentView->getBodyDirection().isKnown())
	turnAngle = (agentPos.getDirectionTo(facts->getMarker(FLAG_C).getPosition()) -
		     mAgentView->getBodyDirection()).getDeg();
      else
	turnAngle = 45;

      if (abs(turnAngle) > 5) {
	// We're turned strange, turn towards center
	RS_LOG ( LA_DECISION, "Goalie - turn after catch before teleportation: " << turnAngle);
	TurnCommand* turn = new TurnCommand(turnAngle);
	step.setBodyCommand(turn);
      }
      else {
	// Turned correctly, teleport!
	RS_LOG ( LA_DECISION, "Goalie - teleport" );
	int side = random()%2;
	Point point(-40, 15);
	if (side == 1) point.setY(-15);
	MoveCommand *move = new MoveCommand(point);
	step.setBodyCommand(move);
	mState = STATE_TELEPORTED;
      }
    } else if (mBallView->isBallKickable()) {
      // Kick ball
      AngleDeg kickdir = (agentPos.getDirectionTo(facts->getMarker(FLAG_C).getPosition()) -
                          mAgentView->getBodyDirection()).getDeg();
      KickCommand* kick = new KickCommand(facts->SP_maxpower, kickdir);
      RS_LOG ( LA_DECISION, "Goalie - kick ball in other mode: " << kickdir);
      step.setBodyCommand(kick);
      mState = STATE_NORMAL;
    }
    else if (mBallView->getPosition(NEXT).isKnown() &&
	     agentPos.isKnown() &&
	     mAgentView->getBodyDirection().isKnown()) {
      // Go to ball
      Point preferredPosition = mBallView->getPosition(NEXT).getPoint();
      AngleDeg moveDir = (agentPos.getDirectionTo(preferredPosition) -
			  mAgentView->getBodyDirection()).getDeg();

      mTurnBodyToPoint.setPosition(preferredPosition);
      mDashToPoint.setPosition(preferredPosition);

      if (abs(moveDir.getDeg()) > MAX_DASH_ANGLE &&
	  MIN_DASH_DISTANCE < agentPos.getPoint().getDistanceTo(preferredPosition) &&
	  mTurnBodyToPoint.applicable()) {
	RS_LOG (LA_DECISION, "Othermode: Turn to point" << preferredPosition << " " << abs(moveDir.getDeg()) );
	step = mTurnBodyToPoint.generateStep();
      }
      // Dash towards the point if we aren't there
      else if (MIN_DASH_DISTANCE < agentPos.getPoint().getDistanceTo(preferredPosition) &&
	       mDashToPoint.applicable()) {
	RS_LOG (LA_DECISION, "Othermode: Dash to Point " << preferredPosition);
        step = mDashToPoint.generateStep();
      }
    } else if (mSearchBall.applicable()) {
      RS_LOG(LA_DECISION, "Othermode: Search ball");
      return mSearchBall.generateStep();
    } else {
      RS_LOG(LA_DECISION, "Othermode: Turn to see");
      TurnCommand* turncmd = new TurnCommand(45);
      step.setBodyCommand(turncmd);
      return step;
    }

    // And align neck to body
    if (mAlignNeckToBody.applicable()) {
      RS_LOG(LA_DECISION, "Align neck to body");
      step.merge(mAlignNeckToBody.generateStep());
    }

    return step;
  }
  else if (mGameView->getPlayMode() == PM_PLAY_ON) {
    
    // Check if ball is catchable, if so catch it
    if (mMode >= GOALIE_MODE_2 && 
	mCatchBall.applicable()) {
      RS_LOG ( LA_DECISION, "Catch Ball");
      return mCatchBall.generateStep();
    }
    
    // If we missed a catch and the ball is kickable, kick it.
    else if (mMode >= GOALIE_MODE_2 && 
	     mBallView->isBallKickable()) {
      RS_LOG ( LA_DECISION, "Clear ball");
      AngleDeg kickdir = 0;
      if (agentPos.isKnown() && mAgentView->getBodyDirection().isKnown())
	kickdir = (agentPos.getDirectionTo(facts->getMarker(FLAG_C).getPosition()) -
		   mAgentView->getBodyDirection()).getDeg();
      KickCommand* kick = new KickCommand(facts->SP_maxpower, kickdir);
      step.setBodyCommand(kick);
      return step;
    }
    
    // Move to the preferred position.
    else if (mMode >= GOALIE_MODE_3 &&
	     mAgentView->getBodyDirection().isKnown() && 
	     agentPos.isKnown() &&
	     ballPos.isKnown()) {

      Point preferredPosition = this->preferredPosition();
      
      AngleDeg moveDir = (agentPos.getDirectionTo(preferredPosition) -
			  mAgentView->getBodyDirection()).getDeg();
      
      // Set the position in the skills
      mTurnBodyToPoint.setPosition(preferredPosition);
      mTurnBackToPoint.setPosition(preferredPosition);
      mDashToPoint.setPosition(preferredPosition);

      // Turn body if direction is greater than MAX_DASH_ANGLE forwards
      if (abs(moveDir.getDeg()) > MAX_DASH_ANGLE &&
	  MIN_DASH_DISTANCE < agentPos.getPoint().getDistanceTo(preferredPosition) &&
	  mTurnBodyToPoint.applicable()) {
	RS_LOG (LA_DECISION, "Turn to point" << preferredPosition);
	step = mTurnBodyToPoint.generateStep();
      }
      // Dash towards the point if we aren't there
      else if (MIN_DASH_DISTANCE < agentPos.getPoint().getDistanceTo(preferredPosition) &&
	       mDashToPoint.applicable()) {
	RS_LOG (LA_DECISION, "Dash to Point " << preferredPosition);
        step = mDashToPoint.generateStep();
      }
      // If we're there and turned in some strange way
      else if (mBallView->getPosition(NEXT).isKnown()) {
	RS_LOG (LA_DECISION, "Turn to ball");
	mTurnBodyToPoint.setPosition(mBallView->getPosition(NEXT).getPoint());
	if (mTurnBodyToPoint.applicable())
	  step = mTurnBodyToPoint.generateStep();
      } else {
	RS_LOG (LA_DECISION, "Turn back to goal");
	mTurnBackToPoint.setPosition(goalCenter);
	if (mTurnBackToPoint.applicable())
	  step = mTurnBackToPoint.generateStep();
      }
      
      // And turn neck to face the ball
      if (mTurnNeckToBall.applicable()) {
	RS_LOG(LA_DECISION, "Turn neck to ball");
	mTurnNeckToBall.setCommand(step.getBodyCommand());
	step.merge(mTurnNeckToBall.generateStep());
      }
      else if (mAlignNeckToBody.applicable()) {
	RS_LOG(LA_DECISION, "Align neck to body");
	step.merge(mAlignNeckToBody.generateStep());
      }

      return step;
    }
    // Look for ball.
    else if (mSearchBall.applicable()) {
      RS_LOG(LA_DECISION, "Search ball");
      return mSearchBall.generateStep();
    }
    // Look at ball.
    else if (mTurnNeckToBall.applicable()) {
      RS_LOG(LA_DECISION, "Turn neck to ball, nothing else to do");
      mTurnNeckToBall.setCommand(step.getBodyCommand());
      step.merge(mTurnNeckToBall.generateStep());
      return step;
    }
    // Bodydirection or NeckDirection unknown, do turn to see something.
    else {
      RS_LOG(LA_DECISION, "Turn to see");
      TurnCommand* turncmd = new TurnCommand(45);
      step.setBodyCommand(turncmd);
      return step;
    }
  }
  else
    return step;
}

void GoalieDecision::onAuralCoachSensorData() {
  std::string message = mAgentView->getCoachMessage();
  //std::cout << "Got coach message '" << message << "'" << std::endl;
  if ( message == "start" ) {
    std::cout << "starting test cases" << std::endl;
    mRunning = true;
    mMode = GOALIE_MODE_5;
  } else if ( message == "finished" ) {
    std::cout << "test cases finished" << std::endl;
    mRunning = false;
  } else if ( message == "success" || message == "failure" ) {
    RS_LOG ( LA_DEBUG, "last scenario was a " << message );
    mMode = NO_GOALIE_MODE;
  } else if ( message == "GMode1" ) {
    mMode = GOALIE_MODE_1;
    mNewScenario = true;
  } else if ( message == "GMode2" ) {
    mMode = GOALIE_MODE_2;
    mNewScenario = true;
  } else if ( message == "GMode3" ) {
    mMode = GOALIE_MODE_3;
    mNewScenario = true;
  } else if (message == "GMode4" ) {
    mMode = GOALIE_MODE_4;
    mNewScenario = true;
  } else if (message == "GMode5" ) {
    mMode = GOALIE_MODE_5;
    mNewScenario = true;
  }

  // Save the time when the scenario was started
  if ( mNewScenario ) {
    mNewScenarioStart = Time::currentTime;
    RS_LOG( LA_DEBUG, "new scenario started at " << mNewScenarioStart );
  }
}

void GoalieDecision::onAuralRefereeSensorData() {
}

void GoalieDecision::onAuralPlayerSensorData() {
}

void GoalieDecision::onAuralSelfSensorData() {
}

void GoalieDecision::onInit() {
  // Position the Goalie in front of the goal
  Point movePoint = Point(-45,0);
  MoveCommand* move_cmd = new MoveCommand(movePoint);
  actuatorInterface->addToPlan(move_cmd);
}

void GoalieDecision::onCommandWarning() {
  decide();
}

PointUE GoalieDecision::predictBall()
{
  PointUE ballPos = mBallView->getPosition();
  VectorUE ballSpeed = mBallView->getSpeedVector();
  PointUE agentPos = mAgentView->getPosition();
  VectorUE relBallVector = mBallView->getRelativeVector();
  Point goalCenter = facts->getOurGoal().getPosition();
  FloatUE closeOpDist
    (mPlayerView->getOpponentClosestToBall().getPosition().getDistanceTo(ballPos));

  PointUE predBallPos = ballPos;

  if (ballSpeed.isKnown() && 
      relBallVector.isKnown() &&
      closeOpDist.isKnown()) {
    //do a ball prediction to see if we want to rush out
    PointUE tmpPredBall = ballPos;
    VectorUE tmpPredBallSpeed = ballSpeed;
    Int maxPred=15;
    if(maxPred > closeOpDist.getValue()*3.0) {
      maxPred=(int)(closeOpDist.getValue()*3);
    }
    for(int a=0; a<maxPred; a++) {
      tmpPredBall += tmpPredBallSpeed;
      tmpPredBallSpeed *= facts->SP_ball_decay;
      if(goalCenter.getDistanceTo(tmpPredBall.getPoint()) < 15) {
        if (agentPos.getDistanceTo(tmpPredBall) < agentPos.getDistanceTo(predBallPos))
          predBallPos = tmpPredBall;
      }
    }
  }
  return predBallPos;
}

Point GoalieDecision::preferredPosition()
{
  PointUE agentPos = mAgentView->getPosition();
  Point goalCenter = facts->getOurGoal().getPosition();

  //Predict where to intercept ball
  PointUE predBallPos = predictBall();

  //Calculate the vector between the own goal and the ball
  VectorUE ballGoalVec;
  FloatUE ballDistance;
  if (predBallPos.isKnown()) {
    ballGoalVec = VectorUE(goalCenter, predBallPos);
    ballDistance = ballGoalVec.getLength();

    //place the goalie on a 1/0.15-radius-circle from the center of the goal
    ballGoalVec/=(ballDistance*0.15);
  }

  Point moveToPos;
  FloatUE predBallDist = predBallPos.getDistanceTo(goalCenter);
  if (mMode >= GOALIE_MODE_3 && 
      predBallPos.isKnown() && 
      (predBallDist < RUSHING_DISTANCE)) {
    moveToPos=predBallPos.getPoint();
  }
  else if (mMode >= GOALIE_MODE_4 &&
	   ballGoalVec.isKnown()) {
    //Place the goalie on circle
    moveToPos = (goalCenter+ballGoalVec).getPoint();
  }
  else if (agentPos.isKnown()) {
    // The agent is positioned the right way.
    moveToPos = agentPos.getPoint();
  }
  else {
    moveToPos = Point(0,0);
  }
  return moveToPos;
}



RS_END_NAMESPACE
