/* -*- Mode: C++ -*-
 * $Id: KickBallCloseToBody.cc,v 1.3 2005/09/08 08:58:35 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: KickBallCloseToBody.cc
 * Unitname: Framework/skill
 * $Revision: 1.3 $
 * Created by: Peter Andersson 2003-08-29
 * Last modified by $Author: frehe $ $Date: 2005/09/08 08:58:35 $
 *
 *
 * HISTORY
 *
 * $Log: KickBallCloseToBody.cc,v $
 * Revision 1.3  2005/09/08 08:58:35  frehe
 * Implemented the reset method.
 *
 * Revision 1.2  2004/08/13 15:39:28  petan
 * Moved implementation to generateStep
 *
 * Revision 1.1  2003/09/03 14:08:45  aip19
 * Updated and documented primitive skills
 *
 */

/**
 * @file KickBallCloseToBody.cc
 * Implementation file for the skill class KickBallCloseToBody.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.3 $
 * $Author: frehe $
 * $Date: 2005/09/08 08:58:35 $
 */

#include "KickBallCloseToBody.h"
#include "view_constants.h"
#include "KickCommand.h"
#include "NewCycleStrategy.h"
#include "WorldModelInterface.h"

// Used when there does not exist the necessary kick power
#include "FreezeBall.h"

RS_BEGIN_NAMESPACE


// --- CONSTRUCTORS ---

KickBallCloseToBody::KickBallCloseToBody(const AngleDeg& dir)
  : SkillWithDirection(dir),
    mAgentView(NULL),
    mBallView(NULL),
    mKickBallView(NULL)
{
  getView(mAgentView, AGENT_VIEW_ID);
  getView(mBallView, BALL_VIEW_ID);
  getView(mKickBallView, KICKBALL_VIEW_ID);
}


// --- DESTRUCTOR ---

KickBallCloseToBody::~KickBallCloseToBody()
{
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(KICKBALL_VIEW_ID);
}


// --- METHODS ---

void KickBallCloseToBody::reset()
{
}

Plan KickBallCloseToBody::generatePlan()
{
  return Plan(generateStep());
}

AgentStep KickBallCloseToBody::generateStep()
{
  // Predict next agent position
  const NewCycleStrategy* new_cycle_strategy
    = WorldModelInterface::instance()->getNewCycleStrategy();
  WorldModel predicted_wm(WorldModelInterface::instance()->getWorld(NOW));
  new_cycle_strategy->updateBeforeNewCycle(predicted_wm);

  // Now predicted_wm contains the predicted world
  AgentObject& agent = predicted_wm.getAgent();
  PointUE predictedPosition = agent.getPosition();

  // If the predicted position is unknown, get a good estimation
  if (predictedPosition.isUnknown()) {
    if  (mAgentView->getSpeedVector().isKnown())
      predictedPosition = mAgentView->getPosition() + 
	mAgentView->getSpeedVector();
    else
      predictedPosition = mAgentView->getPosition();
  }

  // Construct the relative vector from agent to wanted ball position
  const WorldFacts *facts = WorldModelInterface::instance()->getWorldFacts();
  Float length = facts->SP_player_size +
    facts->SP_ball_size +
    facts->SP_kickable_margin/6;
  AngleDeg angle = mAgentView->getBodyDirection().getDeg() + direction;

  CoordVector<Float> relativeBallVector(angle, length);

  // Get ball current position
  PointUE ballPosition = mBallView->getPosition();

  // Construct wanted ball movement vector
  CoordVector<Float> wantedBallSpeedVector((predictedPosition - ballPosition).getPoint());
  wantedBallSpeedVector += relativeBallVector;

  // Construct needed acceleration vector
  CoordVector<Float> ballAcceleration = wantedBallSpeedVector -
    mBallView->getSpeedVector().getVector();

  // The power is the length of the ballAcceleration vector
  FloatUE power = mKickBallView->getKickPowerForSpeed(ballAcceleration.getLength());

  // If the power is greater than the allowed power
  // Stop the ball using the FreezeBall skill
  FreezeBall freezeBall;
  if (power > facts->SP_maxpower && freezeBall.applicable()) {
    return freezeBall.generateStep();
  }
  else {
    angle = ballAcceleration.getAngle() - mAgentView->getBodyDirection().getAngle();

    KickCommand *command = new KickCommand(power.getValue(), angle);
    AgentStep step;
    step.setBodyCommand(command);
    return step;
  }
}


FuzzyBool KickBallCloseToBody::persistent()
{
  // Never persistent.
  return FuzzyBool(0.0);
}

FuzzyBool KickBallCloseToBody::applicable()
{
  return mAgentView->getBodyDirection().isKnown() &&
    mBallView->getPosition().isKnown() &&
    mBallView->getSpeedVector().isKnown() &&
    mBallView->isBallKickable();
}

FuzzyBool KickBallCloseToBody::succeed()
{
  // Has not succeeded yet
  return FuzzyBool(0.0);
}

FuzzyBool KickBallCloseToBody::failed()
{
  // Has not failed yet
  return FuzzyBool(0.0);
}



RS_END_NAMESPACE

