/* -*- Mode: C++ -*-
 * $Id: KickBallCloseToBody.h,v 1.4 2005/09/08 08:58:35 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: KickBallCloseToBody.h
 * Unitname: Framework/skill
 * $Revision: 1.4 $
 * Created by: Peter Andersson 2003-08-29
 * Last modified by $Author: frehe $ $Date: 2005/09/08 08:58:35 $
 *
 *
 * HISTORY
 *
 * $Log: KickBallCloseToBody.h,v $
 * Revision 1.4  2005/09/08 08:58:35  frehe
 * Implemented the reset method.
 *
 * Revision 1.3  2004/09/20 11:54:42  petan
 * Updated information about skill
 *
 * Revision 1.2  2004/08/13 15:39:28  petan
 * Moved implementation to generateStep
 *
 * Revision 1.1  2003/09/03 14:08:45  aip19
 * Updated and documented primitive skills
 *
 *
 */

/**
 * @file KickBallCloseToBody.h
 * Include file for the skill class GotoBall.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.4 $
 * $Author: frehe $
 * $Date: 2005/09/08 08:58:35 $
 */

#ifndef KICKBALLCLOSETOBODY_H
#define KICKBALLCLOSETOBODY_H


#include "SkillWithDirection.h"
#include "AgentView.h"
#include "BallView.h"
#include "KickBallView.h"

RS_BEGIN_NAMESPACE


/**
 * @class KickBallCloseToBody KickBallCloseToBody.h
 * This skill enables an agent to kick the ball close to his body.
 * It receives an angle that denotes where in a relative direction
 * to the body's direction the ball should be close. This skill can
 * favourable be used before a pass or scoring attempt with an angle
 * of 0 since the ball can be kicked with more power if it is straight
 * ahead of the agent.
 *
 * If the agent cannot kick the ball close to it's body it tries to
 * freeze the ball instead.
 *
 * This skill is a bit hard to use efficiently since it is always
 * applicable many times in a row. If it is intended to be used together
 * with a follow up kick it is good to add the check
 * if (kickBallView->kickPowerRate() < 0.023 &&
 *     kickBallCloseToBody.applicable()) {
 *    ... }
 * before it is used. Then the skill will only be used when the ball
 * is in a bad position for a kick.
 *
 * The skill is also very sensitive to the speed of the ball and
 * should be used with caution. If it is hard to know which speed the
 * ball have, this skill might fumble the ball and kick it too hard
 * in any direction.
 *
 * This skill can also be used together with a dribble behaviour to
 * keep the ball where it is wanted.
 *
 * @pre The agents body direction is known, the balls position is known,
 *      the balls speed vector is known and the ball is kickable 
 * @post The ball is in the specified angle to the agent.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.4 $
 * $Author: frehe $
 * $Date: 2005/09/08 08:58:35 $
 */
class KickBallCloseToBody : public SkillWithDirection {
public:
  /** Default constructor. Here you can initialize the skill. */
  KickBallCloseToBody(const AngleDeg& dir);

  /** Destructor should clean up after the skill. */
  virtual ~KickBallCloseToBody();

  
  /** Reset any internal state in the skill. */
  virtual void reset();
  
  /** This method is used to generate a plan for how the goal of
    the skill should be obtained by the agent. You can assume
    that the preconditions of the skill are true (this can be tested
    by calling the applicable() method, after you have implemented it). */
  virtual Plan generatePlan();

  /** Generate a step for how to accomplish the goal
      of the skill. */
  virtual AgentStep generateStep();

  /** This method should return a FuzzyBool telling the caller
    if the skill wants to be called again. */
  virtual FuzzyBool persistent();

  /** This method should implement the precondition specified in
   the description of the skill. */
  virtual FuzzyBool applicable();

  /** Return @c true (close to 1) if the skill has succeeded,
      i.e. its post conditions are satisfied. */
  virtual FuzzyBool succeed();

  /** Return @c true (close to 1) if the skill has failed,
      i.e. something went wrong. */
  virtual FuzzyBool failed();
  
private:
  /** Hide copy-constructor. */
  KickBallCloseToBody(const KickBallCloseToBody&);
  /** Hide assignment operator. */
  KickBallCloseToBody& operator= (const KickBallCloseToBody&);

  AgentView* mAgentView;
  BallView* mBallView;
  KickBallView* mKickBallView;
};


RS_END_NAMESPACE

#endif // !KICKBALLCLOSETOBODY_H
