/* -*- Mode: C++ -*-
 * $Id: TurnNeckToBall.cc,v 1.8 2005/09/29 08:27:32 petan Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintzfrehe <@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: TurnNeckToBall.cc
 * Unitname: Framework/skill
 * $Revision: 1.8 $
 * Created by: Fredrik Heintz 2003-09-02
 * Last modified by $Author: petan $ $Date: 2005/09/29 08:27:32 $
 *
 *
 * HISTORY
 *
 * $Log: TurnNeckToBall.cc,v $
 * Revision 1.8  2005/09/29 08:27:32  petan
 * Updated so that agent will align neck to body if ball is outside viewable area
 *
 * Revision 1.6  2005/09/08 08:58:36  frehe
 * Implemented the reset method.
 *
 * Revision 1.5  2004/08/18 12:03:46  petan
 * Moved initialization of AgentStep
 *
 * Revision 1.4  2004/08/13 15:39:29  petan
 * Moved implementation to generateStep
 *
 * Revision 1.3  2004/08/13 12:00:39  petan
 * Removed couts and fixed bug with turning head outside limits
 *
 * Revision 1.2  2003/11/07 15:19:28  frehe
 * Make sure the neck angles are valid.
 *
 * Revision 1.1  2003/09/02 17:59:17  frehe
 * Added TurnNeckToPoint and TurnNeckToBall.
 *
 */

/**
 * @file TurnNeckToBall.cc
 * Implementation file for the skill class TurnNeckToBall.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.8 $
 * $Author: petan $
 * $Date: 2005/09/29 08:27:32 $
 */


#include "TurnNeckToBall.h"
#include "CommandStrategy.h"
#include "NewCycleStrategy.h"
#include "WorldModelInterface.h"
#include "TurnNeckCommand.h"


RS_BEGIN_NAMESPACE


// --- CONSTRUCTORS ---

TurnNeckToBall::TurnNeckToBall(ServerCommand* cmd)
  : SkillWithCommand(cmd)
{
}


// --- DESTRUCTOR ---

void
TurnNeckToBall::reset()
{
}

TurnNeckToBall::~TurnNeckToBall()
{
}


// --- METHODS ---

Plan TurnNeckToBall::generatePlan()
{
  return Plan(generateStep());
}

AgentStep TurnNeckToBall::generateStep()
{
  const CommandStrategy* cmd_strategy
    = WorldModelInterface::instance()->getCommandStrategy();
  const NewCycleStrategy* new_cycle_strategy
    = WorldModelInterface::instance()->getNewCycleStrategy();

  WorldModel predicted_wm(WorldModelInterface::instance()->getWorld(NOW));

  // Predict the state of the world after executing the command
  if ( command != 0 ) {
    cmd_strategy->updateAfterCommand(predicted_wm, command);
  }
  new_cycle_strategy->updateBeforeNewCycle(predicted_wm);

  // Now predicted_wm contains the predicted world
  AgentObject& agent = predicted_wm.getAgent();
  AngleDegUE dir_to_pos
    = agent.getPosition().getDirectionTo(predicted_wm.getBall().getPosition());
  AngleDegUE neck_dir = agent.getNeckDirection();
  AngleDegUE turn_dir = dir_to_pos - neck_dir;

  AgentStep step;
  if ( turn_dir.isKnown() ) {
    ViewWidth width = agent.getViewWidth();
    int angle;
    switch (width) {
    case VW_NORMAL:
    case VW_UNKNOWN: angle = 90;
      break;
    case VW_WIDE: angle = 180;
      break;
    case VW_NARROW: angle = 45;
      break;
    }
    
    TurnNeckCommand *cmd;

    // If ball is outside visible area, align neck to body
    if (abs(turn_dir.getDeg()) > angle/2) {
      AngleDegUE body = agent.getBodyDirection() - agent.getNeckDirection();
      if (body.isKnown())
	cmd = new TurnNeckCommand(body.getDeg());
      else
	cmd = new TurnNeckCommand(0);
    }
    else {
    // Else turn neck to ball
      cmd = new TurnNeckCommand(minMax(WorldFacts::instance()->SP_minneckang-neck_dir.getDeg(),
				       turn_dir.getDeg(),
				       WorldFacts::instance()->SP_maxneckang-neck_dir.getDeg()));
    }
    step.setNeckCommand(cmd);
    return step;
  } else {
    return step;
  }
}

FuzzyBool TurnNeckToBall::persistent()
{
  // Never persistent.
  return FuzzyBool(0.0);
}


FuzzyBool TurnNeckToBall::applicable()
{
  // Always applicable.
  return FuzzyBool(1.0);
}

FuzzyBool TurnNeckToBall::succeed()
{
  // The skill has not succeeded yet.
  return FuzzyBool(0.0);
}

FuzzyBool TurnNeckToBall::failed()
{
  // The skill has not failed yet.
  return FuzzyBool(0.0);
}



RS_END_NAMESPACE
