/* -*- Mode: C++ -*-
 * $Id: AgentView.cc,v 1.15 2005/09/01 09:46:32 petan Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: AgentView.cc
 * Unitname: Framework/view
 * $Revision: 1.15 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: petan $ $Date: 2005/09/01 09:46:32 $
 *
 *
 * HISTORY
 *
 * $Log: AgentView.cc,v $
 * Revision 1.15  2005/09/01 09:46:32  petan
 * Added compensation for max speed in getPowerForDash
 *
 * Revision 1.14  2004/08/13 11:47:07  petan
 * Changed a if construct to use minMax instead
 *
 * Revision 1.13  2003/09/18 08:27:50  aip19
 * Fixed getPowerForDash NaN problem
 *
 * Revision 1.12  2003/09/03 14:02:45  aip19
 * Updated AgentView with getPowerForDash
 *
 * Revision 1.11  2003/09/02 17:55:27  frehe
 * Fixed problems with unknown angles.
 *
 * Revision 1.10  2003/09/02 12:52:45  aip19
 * New primitive skills checked in, not tested.
 *
 * Revision 1.9  2002/08/31 15:22:45  s02davro
 * Small changes in views
 *
 * Revision 1.8  2002/08/02 21:53:01  s02davro
 * Now works by extracting information from the AgentObject in the WorldModel, parameterView depenency removed.
 *
 * Revision 1.7  2001/10/14 21:21:44  frehe
 * Added better support for messages from self and from other players.
 *
 * Revision 1.6  2001/09/17 21:26:21  andbo928
 * Now uses the SpeedDirection value supplied by the sserver 7 protocol.
 *
 * Revision 1.5  2001/01/11 22:00:20  olalu526
 * Added the real world model with support for new simulation steps.
 *
 * Revision 1.4  2001/01/10 15:12:58  olalu526
 * Added support for WorldModel to some basic classes.
 *
 * Revision 1.3  2000/10/18 23:06:35  frehe
 * Added lastCoachMessage and lastPlayerMessage.
 *
 * Revision 1.2  2000/09/25 13:34:18  frehe
 * Made minor modifications.
 *
 * Revision 1.1  2000/09/21 13:46:45  frehe
 * Move the RoboSoc views from rsviews to the rsframework library.
 *
 * Revision 1.3  2000/09/13 00:08:05  frehe
 * Updated the views to inherit from the ViewWithHistory instead and also 
 * changed those views that used a static pointer to itself to do process 
 * the data in the sensor interface to use the new forEach methods 
 * implemented in the FrameworkSensorInterface.
 *
 * Revision 1.2  2000/09/05 21:15:57  frehe
 * Made minor changes.
 *
 * Revision 1.1  2000/09/04 15:27:18  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file AgentView.cc
 * Implementation file for the class AgentView.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.15 $
 * $Author: petan $
 * $Date: 2005/09/01 09:46:32 $
 */

#include "AgentView.h"
#include "view_constants.h"
#include "rs_error.h"
#include "rs_log.h"

RS_BEGIN_NAMESPACE

AgentView::AgentView(const ViewId id)
  : View(id) {
}

AgentView::~AgentView() {
}

bool AgentView::isAgentOutsideField() const {
  static const Float offset_x = 0.75;
  static const Float offset_y = 0.75;
  const WorldFacts* facts = worldModelInterface->getWorldFacts();
  
  PointUE apos = getPosition(NOW);
  if ( (apos.getX().isKnown() &&
	abs(apos.getX()) > (facts->SP_pitch_length/2 + offset_x )) ||
       (apos.getY().isKnown() &&
	abs(apos.getY()) > (facts->SP_pitch_width/2 + offset_y )) ) {
    return true;
  }
  return false;
}

AngleDeg AgentView::getAngleForTurn(AngleDeg angle) const {
  FloatUE turnAngleUE = angle.getDeg();
  const WorldFacts *facts = worldModelInterface->getWorldFacts();

  turnAngleUE *= (1.0 + facts->SP_inertia_moment * getSpeed());

  
  if ( turnAngleUE.isKnown() ) {
    Float turnAngle = turnAngleUE.getValue();
    if (turnAngle >= 180.0) turnAngle = 180.0;
    else if (turnAngle <= -180.0) turnAngle = -180.0;

    return AngleDeg(turnAngle);
  } else {
    /* If speed is unknown, we cannot calculate the inertia
     * then we return the original angle 
     */ 
    return angle;
  }  
}

/**
 * This method receives a relative position (x,y) to which the
 * agent wants to move and returns the power that should be
 * supplied to the dash command in order to come as close to this
 * position as possible
 * @param x The relative x coordinate
 * @param y The relative y coordinate
 * @return The power that should be supplied to the dash command
 * to come as close as possible to this point
 */
Float AgentView::getPowerForDash(Point point) const {
  const WorldFacts *facts = worldModelInterface->getWorldFacts();

  // Calculate the range to projection on body line
  CoordVector<Float> vec;
  if (getBodyDirection().isKnown()) {
    AngleDegrees<Float> angle(getBodyDirection().getAngle());
    CoordVector<Float> bodyDirection(angle, 1);
    vec = CoordVector<Float>(point).projectOn(bodyDirection);
    double range = vec.getLength();
    double d = min(range, facts->SP_player_speed_max);
    
    Float dashParameter = d;

    // Calculate current speed in the direction of the body
    if (getSpeedVector().isKnown()) {
      vec = getSpeedVector().getVector().projectOn(bodyDirection);
      dashParameter -= vec.getLength();
    }

    dashParameter /= facts->SP_dash_power_rate * getEffort().getValue();
    dashParameter = minMax(facts->SP_minpower, dashParameter, facts->SP_maxpower);

    // Compensate for current agent speed to not use too much stamina
    dashParameter = (dashParameter * (facts->SP_player_speed_max - getSpeed().getValue())) / facts->SP_player_speed_max;

    return dashParameter;
  }
  else
    return facts->SP_maxpower;
}

RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "AgentView.icc"
#endif








