/* -*- Mode: C++ -*-
 * $Id: FrameworkSensorInterface.cc,v 1.12 2004/09/01 15:51:00 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: FrameworkSensorInterface.cc
 * Unitname: BasicSystem
 * $Revision: 1.12 $
 * Created by: David Rosn 2002-06-18
 * Last modified by $Author: frehe $ $Date: 2004/09/01 15:51:00 $
 *
 *
 * HISTORY
 *
 * $Log: FrameworkSensorInterface.cc,v $
 * Revision 1.12  2004/09/01 15:51:00  frehe
 * Updated logging.
 *
 * Revision 1.11  2003/10/14 18:03:06  s02davro
 * Fixed command bug
 *
 * Revision 1.10  2003/07/22 14:47:57  s02davro
 * Playmode clockfix
 *
 * Revision 1.9  2003/07/21 08:58:24  s02davro
 * Change the communication interfaces to support v. 9
 *
 * Revision 1.8  2002/12/07 17:27:07  s02davro
 * Added support for trainer
 *
 * Revision 1.7  2002/09/10 13:37:03  s02davro
 * Fixed offside bug
 *
 * Revision 1.6  2002/08/31 15:27:02  s02davro
 * New modified startup, agent, frameworksensorinterface updated
 *
 * Revision 1.5  2002/08/02 22:40:48  s02davro
 * Uses the new Sensor version, WorldObservation and WorldStateInterface.
 *
 */

/**
 * @file FrameworkSensorInterface.cc
 * Implementation file for the class FrameworkSensorInterface.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.12 $
 * $Author: frehe $
 * $Date: 2004/09/01 15:51:00 $
 */


#include "FrameworkSensorInterface.h"
#include "rs_error.h"
#include "rs_log.h"


RS_BEGIN_NAMESPACE


FrameworkSensorInterface::FrameworkSensorInterface() 
  : worldModelInterface(WorldModelInterface::nonConstInstance()),
    obsWorld() {
}

FrameworkSensorInterface::~FrameworkSensorInterface() { 
} 

void FrameworkSensorInterface::newCycle() { 
  SensorInterface::newCycle();
  obsWorld.clear();
} 

void FrameworkSensorInterface::updateBeforeNewCycle() {
  worldModelInterface->updateBeforeNewCycle();
}

void 
FrameworkSensorInterface::see(const std::vector<ServerLineObject>& lines,
			      const std::vector<ServerMarkerObject>& markers,
			      const std::vector<ServerPlayerObject>& players,
			      const ServerBallObject& ball) {
  obsWorld.lines = lines;
  obsWorld.markers = markers;
  obsWorld.players = players;
  obsWorld.ball = ball;
}

void FrameworkSensorInterface::seeGlobal(const std::vector<CoachPlayerObject>& cplayers,
					 const CoachBallObject& cball) {
  obsWorld.cplayers = cplayers;
  obsWorld.cball = cball;
}

void FrameworkSensorInterface::sense(const ServerAgentObject& agent) {
  obsWorld.agent = agent;
}

void FrameworkSensorInterface::hearReferee(const PlayMode play_mode) {
  RS_LOG( LA_SENSOR_INTERFACE, "hearReferee" );
  
  obsWorld.playModes.push_back(play_mode);

  switch ( play_mode ) {
  case PM_OUR_KICK_OFF:
  case PM_THEIR_KICK_OFF:
  case PM_OUR_KICK_IN:
  case PM_THEIR_KICK_IN:
  case PM_OUR_FREE_KICK:
  case PM_THEIR_FREE_KICK:
  case PM_OUR_CORNER_KICK:
  case PM_THEIR_CORNER_KICK:
  case PM_OUR_GOAL_KICK:
  case PM_THEIR_GOAL_KICK:
  case PM_OUR_GOALIE_GOT_BALL:
  case PM_THEIR_GOALIE_GOT_BALL:
  case PM_PLAY_ON:
  case PM_DROP_BALL:
    clockStopped = false;
    break;
  case PM_OUR_OFFSIDE_KICK:
  case PM_THEIR_OFFSIDE_KICK:
  case PM_OUR_GOAL:
  case PM_THEIR_GOAL:
  case PM_BEFORE_KICK_OFF:
  case PM_HALF_TIME:
  case PM_TIME_OVER:
  case PM_EXTENDED_TIME:
  case PM_OUR_FOUL:
  case PM_THEIR_FOUL:
  case PM_OUR_FOUL_CHARGE:
  case PM_THEIR_FOUL_CHARGE:
  case PM_OUR_FOUL_PUSH:
  case PM_THEIR_FOUL_PUSH:
  case PM_OUR_FOUL_MULTIPLEATTACKER:
  case PM_THEIR_FOUL_MULTIPLEATTACKER:
  case PM_OUR_FOUL_BALLOUT:
  case PM_THEIR_FOUL_BALLOUT:
  case PM_OUR_FOUL_BACKPASS:
  case PM_THEIR_FOUL_BACKPASS:
  case PM_OUR_FOUL_FREEKICKFAULT:
  case PM_THEIR_FOUL_FREEKICKFAULT:
  case PM_OUR_FOUL_CATCHFAULT:
  case PM_THEIR_FOUL_CATCHFAULT:
  case PM_PAUSE:
    clockStopped = true;
    break;
  default:
    RS_ERROR("Unknown play mode in HearReferee!");
  }
}

void FrameworkSensorInterface::hearCoach(const std::string& message) {
  RS_LOG( LA_SENSOR_INTERFACE, "hearCoach: " << message );
  obsWorld.coachMessage = message;
}

void FrameworkSensorInterface::hearSelf(const std::string& message) {
  RS_LOG( LA_SENSOR_INTERFACE, "hearSelf" );
  obsWorld.ownMessage = message;
}

void FrameworkSensorInterface::hearPlayerAsCoach(const Team team,
						 const UIntUnknown nr,
						 const std::string& message) {
  RS_LOG( LA_SENSOR_INTERFACE, "hearPlayerAsCoach" );
  obsWorld.chear.push_back(CoachHearingObject(team,nr,message));
}

void FrameworkSensorInterface::hearTeammate(const UInt unum,
					    const AngleDegUE direction,
					    const std::string& message) {
  RS_LOG( LA_SENSOR_INTERFACE, "hearTeammate" );
  obsWorld.teammateMessage = message;
  obsWorld.teammateMessageDirection = direction;
  obsWorld.teammateSpeakers.push_back(unum);
}

void FrameworkSensorInterface::hearAnotherTeammate(const UInt unum) {
  RS_LOG( LA_SENSOR_INTERFACE, "hearAnotherTeammate" );
  obsWorld.teammateSpeakers.push_back(unum);
}

void FrameworkSensorInterface::hearOpponent(const AngleDegUE direction,
					    const std::string& message) {
  RS_LOG( LA_SENSOR_INTERFACE, "hearOpponent" );
  obsWorld.opponentMessage = message;
  obsWorld.opponentMessageDirection = direction;  
  obsWorld.numberOfOpponentSpeakers++;
}

void FrameworkSensorInterface::hearAnotherOpponent() {
  RS_LOG( LA_SENSOR_INTERFACE, "hearAnotherOpponent" );
  obsWorld.numberOfOpponentSpeakers++;
}  

void FrameworkSensorInterface::setLastCommand(ServerCommand* command) {
  worldModelInterface->setLastCommand(command);
} 

void FrameworkSensorInterface::setLastSensorType(const SensorType sensor_type) {
  // This method is called *after* the whole server message is parsed.
  switch ( sensor_type ) {
  case SENSOR_AURAL_COACH:
    worldModelInterface->updateAfterCoachMessage(obsWorld);
    break;
  case SENSOR_AURAL_PLAYER:
  case SENSOR_AURAL_REFEREE:
  case SENSOR_AURAL_SCORE:
  case SENSOR_AURAL_SELF:
    worldModelInterface->updateAfterHear(obsWorld);
    break;
  case SENSOR_VISUAL:
    worldModelInterface->updateAfterSee(obsWorld);
    break;
  case SENSOR_SENSE:
    worldModelInterface->updateAfterSense(obsWorld);
    break;
  case SENSOR_ACTUATORS:
    worldModelInterface->updateAfterCommand();
    break;
  case SENSOR_INIT:
    worldModelInterface->updateAfterInit();
    break;
  case SENSOR_COACH_VISUAL:
  case SENSOR_COACH_AURAL:
    worldModelInterface->updateAfterCoachInformation(obsWorld);
    break;
  case SENSOR_PARAM:
  case SENSOR_PLAYER_TYPE:
  case SENSOR_PLAYER_PARAM:
  case SENSOR_SUBSTITUTION:
  case SENSOR_OK:
    break;
  case SENSOR_ERROR:
    RS_ERROR("Error from sensors!");
    break;
  case SENSOR_UNKNOWN:
    RS_ERROR("Unknown sensor input!");
    break;
  default:
    RS_ERROR
      ("Unknown sensor type in [FrameworkSensorInterface]:SetLastSensorType.");
  }
}

RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "FrameworkSensorInterface.icc"
#endif
