/* -*- Mode: C++ -*-
 * $Id: SkillWithCommand.h,v 1.3 2004/08/18 12:01:47 petan Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: SkillWithCommand.h
 * Unitname: Framework
 * $Revision: 1.3 $
 * Created by: Fredrik Heintz 2003-09-02
 * Last modified by $Author: petan $ $Date: 2004/08/18 12:01:47 $
 *
 *
 * HISTORY
 *
 * $Log: SkillWithCommand.h,v $
 * Revision 1.3  2004/08/18 12:01:47  petan
 * Fixed minor memory leak
 *
 * Revision 1.2  2004/08/17 10:44:27  frehe
 * Added some extra const.
 *
 * Revision 1.1  2003/09/02 17:57:14  frehe
 * Added SkillWithCommand and SkillWithPositionAndCommand.
 *
 */

/**
 * @file SkillWithCommand.h
 * Include file for the class SkillWithCommand.
 *
 * @par Unit:
 *      Framework
 * $Revision: 1.3 $
 * $Author: petan $
 * $Date: 2004/08/18 12:01:47 $
 */


#ifndef SKILLWITHCOMMAND_H
#define SKILLWITHCOMMAND_H


#include "SkillWithArguments.h"


RS_BEGIN_NAMESPACE


/**
 * @class SkillWithCommand SkillWithCommand.h
 * This abstract class represent skill that uses
 * a position and a command as arguments.
 *
 * @par Unit:
 *      Framework
 * $Revision: 1.3 $
 * $Author: petan $
 * $Date: 2004/08/18 12:01:47 $
 */
class SkillWithCommand : public SkillWithArguments {
public:
  /** Constructor taking the default argument for the command
      used as arguments to the skill. */
  SkillWithCommand(const ServerCommand* const cmd);

  /** Destructor cleaning up after the skill. */
  virtual ~SkillWithCommand();


  /** Used to set the skills's arguments in a generic way.
      The command can not be set through this method. */
  virtual void setArguments(const std::vector<Float>& arguments)
     throw(TooFewArgumentsException);


  /** Used to set the skill's command argument to @a cmd. */
  virtual void setCommand(const ServerCommand* const cmd);

  /** Return the skill's command argument.
      DO NOT DELETE it's still owned by the skill! */
  virtual ServerCommand* getCommand() const;

  
  /** This method is used to generate a plan for how the goal of
    the skill should be obtained by the agent. Abstract since
    no concrete instances of this class should be created. */
  virtual Plan generatePlan() = 0;

  /** This method is used to first set the argument of the skill
    and then generate a plan for how the goal of the skill should
    be obtained by the agent. It uses the methods setCommand and
    generatePlan(), so it should not be overridden unless it should
    made unavailable. */
  virtual Plan generatePlan(const ServerCommand* const cmd);

  /** This method is used to generate the next step to execute to
      come closer to the goal of the skill */
  virtual AgentStep generateStep() = 0;

  /** This method is the same as first calling setCommand and then
      generateStep. Should not be overridden unless it should be
      made unavailable. */
  virtual AgentStep generateStep(const ServerCommand* const cmd);

  /** This method should return a FuzzyBool telling the caller
    if the skill wants to be called again. Abstract since
    no concrete instances of this class should be created. */
  virtual FuzzyBool persistent() = 0;

  /** This method should implement the precondition specified in
   the description of the skill. Abstract since no concrete instances
   of this class should be created. */
  virtual FuzzyBool applicable() = 0;

  
protected:  
  /** The command argument. */
  ServerCommand* command;

  
private:
  /** Hide copy-constructor. */
  SkillWithCommand(const SkillWithCommand&);
  /** Hide assignment operator. */
  SkillWithCommand& operator= (const SkillWithCommand&);
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "SkillWithCommand.icc"
#endif


#endif // !SKILLWITHCOMMAND_H
