/* -*- Mode: C++ -*-
 * $Id: WorldModel.h,v 1.9 2002/12/07 17:27:08 s02davro Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: WorldModel.h
 * Unitname: BasicSystem
 * $Revision: 1.9 $
 * Created by: David Rosen 2002-06-18
 * Last modified by $Author: s02davro $ $Date: 2002/12/07 17:27:08 $
 *
 *
 * HISTORY
 *
 * $Log: WorldModel.h,v $
 * Revision 1.9  2002/12/07 17:27:08  s02davro
 * Added support for trainer
 *
 * Revision 1.8  2002/09/10 13:38:04  s02davro
 * Added ClosestToBall and ClosestToAgent methods
 *
 * Revision 1.7  2002/09/03 10:57:30  s02davro
 * LastSeeTime, NextSeeTime, NextNextSeeTime
 *
 * Revision 1.6  2002/08/31 15:31:17  s02davro
 * Updated the world model
 *
 * Revision 1.5  2002/08/02 21:58:45  s02davro
 * Created to store information about the agents world model.
 *
 */

/**
 * @file WorldModel.h
 * Include file for the class WorldModel.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.9 $
 * $Author: s02davro $
 * $Date: 2002/12/07 17:27:08 $
 */


#ifndef WORLDMODEL_H
#define WORLDMODEL_H


#include "basic_system_types.h"
#include "BallObject.h"
#include "AgentObject.h"
#include "PlayerObject.h"
#include "MarkerObject.h"
#include "LineObject.h"
#include "ServerCommand.h"
#include <vector>
#include <string>
#include <list>

RS_BEGIN_NAMESPACE


/**
 * @class WorldModel WorldModel.h
 * This class contains the agens model of the world.
 * The class contains fundamental information about the agents
 * belives of the world. 
 * The agents position and speed, where the teammates
 * and opponents are, the current playmode, and the speed and position
 * of the ball are examples of information in the agens world model.
 * The Worldmodel is stored in a history in WorldModelInterface and
 * updated by the agenst Strategies.
 *
 * @todo More information
 * @todo Document interaction with Views.
 * @todo Is the const/nonconst methods necessary.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.9 $
 * $Author: s02davro $
 * $Date: 2002/12/07 17:27:08 $
 */

class WorldModel {
public:
  /** Default constructor. */
  WorldModel();  

  /** Destructor */
  virtual ~WorldModel();

  /** CopyConstructor. */
  WorldModel(const WorldModel& wm);

  /** assignment operator */
  WorldModel& operator=(const WorldModel& wm);  

  /** Add an opponent to the agents world model. */
  void addOpponent(const PlayerObject& opponent);

  /** Returns the opponents the agent knows about. */
  const std::vector<PlayerObject>& getOpponents() const;

  /** Returns the opponent closest to the agent. *
   *  If we don't known anything at all about any
   *  opponent. A PlayerObject with all values unknown
   *  will be returned. */
  const PlayerObject& getOpponentClosestToAgent() const;

  /** Returns the opponent closest to the ball. 
   *  If we don't known anything at all about any
   *  opponent. A PlayerObject with all values unknown
   *  will be returned. */
  const PlayerObject& getOpponentClosestToBall() const;

  /** Set the teammate closest to the agent. */
  void setOpponentClosestToAgent(PlayerObject *player);

  /** Set the teammate closest to the ball. */
  void setOpponentClosestToBall(PlayerObject *player);

  /** Returns the opponets to the Strategies. */
  std::vector<PlayerObject>& getOpponents();

  /** Clear all the opponents from the world model. */
  void clearOpponents();

  /** Add a teammate to the agents world model. */
  void addTeammate(const PlayerObject& teammates);

  /** Returns the teammates the agent knows about. */
  const std::vector<PlayerObject>& getTeammates() const;

  /** Returns the opponent closest to the agent.
   *  If we don't known anything at all about any
   *  opponent. A PlayerObject with all values unknown
   *  will be returned. */
  const PlayerObject& getTeammateClosestToAgent() const;

  /** Returns the opponent closest to the ball. 
   *  If we don't known anything at all about any
   *  opponent. A PlayerObject with all values unknown
   *  will be returned. */
  const PlayerObject& getTeammateClosestToBall() const;

  /** Set the teammate closest to the ball. */
  void setTeammateClosestToBall(PlayerObject *player);

  /** Set the teammate closest to the agent. */
  void setTeammateClosestToAgent(PlayerObject *player);

  /** Returns the teammates to the Strategies. */
  std::vector<PlayerObject>& getTeammates();

  /** Clear all the teammates from the world model. */
  void clearTeammates();

  /** Add a unknown player to the world model. */
  void addUnknown(const PlayerObject& unknown);

  /** Returns the unknown players. */
  const std::vector<PlayerObject>& getUnknowns() const;

  /** Returns the unknown players to the strategies. */
  std::vector<PlayerObject>& getUnknowns();

  /** Clear all the unknown player from the world model. */
  void clearUnknowns();

  /** Set the ball in the world model. */
  void setBall(const BallObject& ball);

  /** Returns the ball from the agents world model. */
  const BallObject& getBall() const;  

  /** Returns the ball to a Strategy. */
  BallObject& getBall();  

  /** Set the agents information about itself. */
  void setAgent(const AgentObject& agent);

  /** Get the agents model of itself. */
  const AgentObject& getAgent() const;

  /** Returns the agent object to a Strategy. */
  AgentObject& getAgent();

  /** Add a playmode for the current world model. */
  void addPlayMode(const PlayMode& playMode);

  /** Returns the playmodes for the current world model. */
  const std::vector<PlayMode>& getPlayModes() const;

  /** Returns the playmodes to a Strategy. */
  std::vector<PlayMode>& getPlayModes();

  /** Clear all but the last playmodes. */
  void clearPlayModes();

  /** Returns the last playmode in the worldmodel */
  const PlayMode getLastPlayMode() const;

  /** Return the current world models game time. */
  const Time& getGameTime() const;

  void setGameTime(Time time);
  
  /** Return the time for last visual information. */
  const TimeUnknown& getLastSeeTime() const;

  void setLastSeeTime(Time time);

  /** Return the time for the next visual information. */
  const TimeUnknown& getNextSeeTime() const;

  void setNextSeeTime(const TimeUnknown& time);

  /** Return the time for the second next visual information. */
  const TimeUnknown& getNextNextSeeTime() const;

  void setNextNextSeeTime(const TimeUnknown& time);
  
  void addCommand(const ServerCommand* cmd);

  // Get the commands given by the agent in this world model. */
  const std::list<const ServerCommand *>& getCommands() const;

protected:
  /** Representation of the playmodes. */
  std::vector<PlayMode> playModes;

  /** The model of the ball. */
  BallObject ball;
  
  /** The model of the agent itself. */
  AgentObject agent;

  /** The models of the opponents. */
  std::vector<PlayerObject> opponents;

  /** The opponent closest to the player. */
  PlayerObject* closestOpponent;

  /** The opponent closest to the ball. */
  PlayerObject* closestOpponentToBall;

  /** The models of the teammates. */
  std::vector<PlayerObject> teammates;

  /** The opponent closest to the player. */
  PlayerObject* closestTeammate;  

  /** The opponent closest to the ball. */
  PlayerObject* closestTeammateToBall;
  
  /** The models of the unknown players. */
  std::vector<PlayerObject> unknowns;

  /** Last heard message from ourself. */
  std::string ownMessage;

  /** Last heard message from another player. */
  std::string playerMessage;

  /** Last heard message from the coach. */
  std::string coachMessage;

  /** Represent the game time in the current world model. */
  Time gameTime;

  /** Representation of the time we last saw something. */
  TimeUnknown lastSeeTime;

  /** The next cycle we should see something. */
  TimeUnknown nextSeeTime;

  /** The second next cycle we should see something. */
  TimeUnknown nextNextSeeTime;

  /** A Unknown player for return values. */
  PlayerObject emptyPlayer;

  /** A list of the executed commands this cycle. */
  std::list<const ServerCommand*> commands;
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "WorldModel.icc"
#endif


#endif // !WORLDMODEL_H
