/* -*- Mode: C++ -*-
 * $Id: WorldModelInterface.cc,v 1.10 2004/09/08 14:43:42 petan Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: WorldModelInterface.cc
 * Unitname: BasicSystem
 * $Revision: 1.10 $
 * Created by: David Rosn 2002-07-01
 * Last modified by $Author: petan $ $Date: 2004/09/08 14:43:42 $
 *
 *
 * HISTORY
 *
 * $Log: WorldModelInterface.cc,v $
 * Revision 1.10  2004/09/08 14:43:42  petan
 * Renamed illnamed setSeeStrategy to getSeeStrategy
 *
 * Revision 1.9  2003/09/02 17:56:35  frehe
 * Added methods to get the strategies.
 *
 * Revision 1.8  2002/09/03 18:04:25  s02davro
 * Speed and NewCycle fix
 *
 * Revision 1.7  2002/09/03 10:57:30  s02davro
 * LastSeeTime, NextSeeTime, NextNextSeeTime
 *
 * Revision 1.6  2002/08/31 15:31:17  s02davro
 * Updated the world model
 *
 * Revision 1.5  2002/08/02 21:59:04  s02davro
 * History of the worldmodels and observations. Handles update of worldmodel and views.
 *
 */

/**
 * @file WorldModelInterface.cc
 * Implementation file for the class WorldModelInterface.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.10 $
 * $Author: petan $
 * $Date: 2004/09/08 14:43:42 $
 */


#include "WorldModelInterface.h"
#include "SeeStrategy.h"
#include "HearStrategy.h"
#include "CommandStrategy.h"
#include "SenseStrategy.h"
#include "InitStrategy.h"
#include "CoachMessageStrategy.h"
#include "NewCycleStrategy.h"
#include "CoachInformationStrategy.h"

RS_BEGIN_NAMESPACE

WorldModelInterface* WorldModelInterface::onlyInstance(NULL);

const WorldModelInterface* WorldModelInterface::instance() {
  if (onlyInstance == NULL)
    onlyInstance = new WorldModelInterface();
  return onlyInstance;
}

WorldModelInterface* WorldModelInterface::nonConstInstance() {
  if (onlyInstance == NULL)
    onlyInstance = new WorldModelInterface();
  return onlyInstance;
}

WorldModelInterface::WorldModelInterface() 
  : currentWorld(0),
    historyLength(DEFAULT_HISTORY_LENGTH),
    wm(DEFAULT_HISTORY_LENGTH),
    observations(DEFAULT_HISTORY_LENGTH),
    worldFacts(WorldFacts::instance()),
    viewManager(ViewManager::instance()) {
  wm[getIndex(NOW)] = WorldModel();
  wm[getIndex(NEXT)] = WorldModel();
  observations[getIndex(NOW)] = WorldObservation();  
}

WorldModelInterface::~WorldModelInterface() { 
} 

void WorldModelInterface::setWorld(const WorldModel &newWorld, const Int i) {
  wm[getIndex(i)] = newWorld;
}

void WorldModelInterface::setObservation(const WorldObservation &newObs) {
  observations[getIndex(NOW)] = newObs;
}

const WorldModel &WorldModelInterface::getWorld(const Int i) const{
  return wm[getIndex(i)];
}

WorldModel& WorldModelInterface::getWorld(const Int i) {
  return wm[getIndex(i)];
}

const WorldObservation& WorldModelInterface::getObservation(const Int i) const {
  return observations[getIndex(i)];
}					    

const WorldFacts* WorldModelInterface::getWorldFacts() const {
  return worldFacts;
}

void WorldModelInterface::setSeeStrategy(SeeStrategy *seeS) {
  seeStrategy = seeS;
}

void WorldModelInterface::setHearStrategy(HearStrategy *hearS) {
  hearStrategy = hearS;
}

void WorldModelInterface::setCommandStrategy(CommandStrategy *commandS) {
  commandStrategy = commandS;
}

void WorldModelInterface::setSenseStrategy(SenseStrategy *senseS) {
  senseStrategy = senseS;
}

void WorldModelInterface::setInitStrategy(InitStrategy *initS) {
  initStrategy = initS;
}

void 
WorldModelInterface::setCoachMessageStrategy(CoachMessageStrategy *coachS) {
  coachMessageStrategy = coachS;
}

void 
WorldModelInterface::setCoachInformationStrategy(CoachInformationStrategy 
						 *coachInfoS){
  coachInformationStrategy = coachInfoS;
}

void WorldModelInterface::setNewCycleStrategy(NewCycleStrategy *newCycleS) {
  newCycleStrategy = newCycleS;
}

const SeeStrategy* WorldModelInterface::getSeeStrategy() const {
  return seeStrategy;
}

const HearStrategy* WorldModelInterface::getHearStrategy() const {
  return hearStrategy;
}

const CommandStrategy* WorldModelInterface::getCommandStrategy() const {
  return commandStrategy;
}

const SenseStrategy* WorldModelInterface::getSenseStrategy() const {
  return senseStrategy;
}

const InitStrategy* WorldModelInterface::getInitStrategy() const {
  return initStrategy;
}

const CoachMessageStrategy* WorldModelInterface::getCoachMessageStrategy() const {
  return coachMessageStrategy;
}

const NewCycleStrategy* WorldModelInterface::getNewCycleStrategy() const {
  return newCycleStrategy;
}

const CoachInformationStrategy* WorldModelInterface::getCoachInformationStrategy() const {
  return coachInformationStrategy;
}

void WorldModelInterface::updateAfterSee(const WorldObservation& obs) {  
  setObservation(obs);
  if (seeStrategy)
    seeStrategy->updateAfterSee();
  setWorld(getWorld(), NEXT);
  if (newCycleStrategy)
    newCycleStrategy->updateBeforeNewCycle(getWorld(NEXT));
  viewManager->updateAfterSee();
}

void WorldModelInterface::updateAfterInit() {  
  if (initStrategy)
    initStrategy->updateAfterInit();
  viewManager->updateAfterInit();
}

void WorldModelInterface::updateAfterHear(const WorldObservation& obs) {  
  setObservation(obs);
  if (hearStrategy)
    hearStrategy->updateAfterHear();
  viewManager->updateAfterHear();
}

void WorldModelInterface::updateAfterCommand() {  
  if (commandStrategy) {
    commandStrategy->updateAfterCommand(getWorld(), lastCommand);
    lastCommand = 0;
  }
  setWorld(getWorld(), NEXT);
  if (newCycleStrategy)
    newCycleStrategy->updateBeforeNewCycle(getWorld(NEXT));
  viewManager->updateAfterCommand();
}

void WorldModelInterface::updateAfterSense(const WorldObservation& obs) {  
  setObservation(obs);
  if (senseStrategy)
    senseStrategy->updateAfterSense();
  setWorld(getWorld(), NEXT);
  if (newCycleStrategy)
    newCycleStrategy->updateBeforeNewCycle(getWorld(NEXT));
  viewManager->updateAfterSense();
}

void WorldModelInterface::updateAfterCoachMessage(const WorldObservation& obs) {
  setObservation(obs);
  if (coachMessageStrategy)
    coachMessageStrategy->updateAfterCoachMessage();
  setWorld(getWorld(), NEXT);
  if (newCycleStrategy)
    newCycleStrategy->updateBeforeNewCycle(getWorld(NEXT));
  viewManager->updateAfterCoachMessage();
}

void WorldModelInterface::updateBeforeNewCycle() {    
  increaseIndex();
  setWorld(getWorld(PREV));
  if (newCycleStrategy)
    newCycleStrategy->updateBeforeNewCycle(getWorld());
  getWorld().clearPlayModes();
  setWorld(getWorld(), NEXT);
  if (newCycleStrategy)
    newCycleStrategy->updateBeforeNewCycle(getWorld(NEXT));
  setObservation(WorldObservation());
  viewManager->newCycle();
  lastCommand = NULL;
}

void 
WorldModelInterface::updateAfterCoachInformation(const WorldObservation& obs) {
  setObservation(obs);
  if (coachInformationStrategy)
    coachInformationStrategy->updateAfterCoachInformation();
}

const Time& WorldModelInterface::getCurrentTime() const {
  return Time::currentTime;
}

RS_END_NAMESPACE
