/* -*- Mode: C++ -*-
 * $Id: WorldModelInterface.h,v 1.11 2004/09/08 14:43:42 petan Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: WorldModelInterface.h
 * Unitname: BasicSystem
 * $Revision: 1.11 $
 * Created by: David Rosen 2002-07-01
 * Last modified by $Author: petan $ $Date: 2004/09/08 14:43:42 $
 *
 *
 * HISTORY
 *
 * $Log: WorldModelInterface.h,v $
 * Revision 1.11  2004/09/08 14:43:42  petan
 * Renamed illnamed setSeeStrategy to getSeeStrategy
 *
 * Revision 1.10  2003/10/14 18:03:07  s02davro
 * Fixed command bug
 *
 * Revision 1.9  2003/09/02 17:56:33  frehe
 * Added methods to get the strategies.
 *
 * Revision 1.8  2002/09/03 18:04:25  s02davro
 * Speed and NewCycle fix
 *
 * Revision 1.7  2002/09/03 10:57:31  s02davro
 * LastSeeTime, NextSeeTime, NextNextSeeTime
 *
 * Revision 1.6  2002/08/31 15:31:17  s02davro
 * Updated the world model
 *
 * Revision 1.5  2002/08/02 21:59:04  s02davro
 * History of the worldmodels and observations. Handles update of worldmodel and views.
 *
 */

/**
 * @file WorldModelInterface.h
 * Include file for the class WorldModelInterface.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.11 $
 * $Author: petan $
 * $Date: 2004/09/08 14:43:42 $
 */


#ifndef WORLDMODELINTERFACE_H
#define WORLDMODELINTERFACE_H

#include "framework_types.h"
#include "WorldFacts.h"
#include "WorldObservation.h"
#include "WorldModel.h"
#include "HistoryDefault.h"
#include "ObjectCollection.h"
#include "Command.h"

#include "ViewManager.h"

RS_BEGIN_NAMESPACE


/**
 * @class WorldModelInterface WorldModelInterface.h
 * This class contains the history of the agent world model.
 * The class is also responsible for updateing the world model
 * when new information is recived in a WorldObservation object from
 * the FrameworkSensorInterface. This update of the agents current
 * worldmodel is done by calling the Strategies associated with the
 * WorldModelInterface. 
 * 
 * @todo More methods.
 * @todo Call the viewManager to update the views. After updating the
 *       world model.
 * @todo The methods getLastHearTime, getLastSeeTime? Are they used?
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.11 $
 * $Author: petan $
 * $Date: 2004/09/08 14:43:42 $
 */

class SeeStrategy;
class HearStrategy;
class CommandStrategy;
class SenseStrategy;
class InitStrategy;
class CoachMessageStrategy;
class NewCycleStrategy;
class CoachInformationStrategy;

class WorldModelInterface {
public:
  /** Static method to get a constant pointer to the WorldModelInterface.
   *  This dosn't allow changing the agents worldmodel, only using the 
   *  stored information. Recomended for skills, predicates and views. */
  static const WorldModelInterface* instance();

  /** Static method to get a pointer to the WorldModelInterface.
   *  This pointer can be used to modify the agents world model.
   *  Used in Strategy and the actual agent (to set strategies). */
  static WorldModelInterface* nonConstInstance();

  /** Destructor for the class. */
  virtual ~WorldModelInterface();
  
  //@{
  /** @name Update functions, calls the different strategies. */
  void updateAfterSee(const WorldObservation& obs);
  void updateAfterHear(const WorldObservation& obs);
  void updateAfterCommand();
  void updateAfterSense(const WorldObservation& obs);
  void updateAfterInit();
  void updateAfterCoachMessage(const WorldObservation& obs);
  void updateAfterCoachInformation(const WorldObservation& obs);
  void updateBeforeNewCycle();
  //@}

  //@{
  /** @name Set the worldmodels update strategies. */
  void setSeeStrategy(SeeStrategy *seeS);
  void setHearStrategy(HearStrategy *hearS);
  void setCommandStrategy(CommandStrategy *commandS);
  void setSenseStrategy(SenseStrategy *senseS);
  void setInitStrategy(InitStrategy *initS);
  void setCoachMessageStrategy(CoachMessageStrategy *coachS);
  void setNewCycleStrategy(NewCycleStrategy *newCycleS);
  void setCoachInformationStrategy(CoachInformationStrategy *coachInfoS);
  //@}

  //@{
  /** @name Get the worldmodels update strategies. */
  const SeeStrategy* getSeeStrategy() const;
  const HearStrategy* getHearStrategy() const;
  const CommandStrategy* getCommandStrategy() const;
  const SenseStrategy* getSenseStrategy() const;
  const InitStrategy* getInitStrategy() const;
  const CoachMessageStrategy* getCoachMessageStrategy() const;
  const NewCycleStrategy* getNewCycleStrategy() const;
  const CoachInformationStrategy* getCoachInformationStrategy() const;
  //@}

  /** Set a new current world model.
   *  For the Strategies to play with. */
  void setWorld(const WorldModel& newWorld, const Int i = NOW);

  /** Set a new observation for the current time.
   *  New information is recived from the SensorInterface. */
  void setObservation(const WorldObservation& newObs);

  void setLastCommand(ServerCommand* lastCommand);

  ServerCommand* getLastCommand() const;

  /** Returns a reference to a worldmodel (for strategies). */
  WorldModel& getWorld(const Int i = NOW);

  /** Returns a constant reference to a worldmodel. */
  const WorldModel& getWorld(const Int i = NOW) const;  

  /** Returns a constant reference to the a worldobservation. */
  const WorldObservation& getObservation(const Int i = NOW) const;  

  /** Returns a constant pointer to the facts we know about the world. */
  const WorldFacts* getWorldFacts() const;

  /** Returns the current time according to the agents world model. */
  const Time& getCurrentTime() const;

protected:
  /** The current observed time step. */
  UInt currentWorld;

  /** The length of the history. */
  UInt historyLength;

  /** Get the index associated with the time specified. */
  UInt getIndex(const Int i) const;

  /** Advance one step in time, new worldmodel, empty worldobservation. */
  void increaseIndex();
  
  //@{
  /** @name The agents strategies for updateing the world model. */
  SeeStrategy *seeStrategy;
  HearStrategy *hearStrategy;
  CommandStrategy *commandStrategy;
  InitStrategy *initStrategy;
  SenseStrategy *senseStrategy;  
  CoachMessageStrategy *coachMessageStrategy;
  NewCycleStrategy *newCycleStrategy; // Simulation?
  CoachInformationStrategy *coachInformationStrategy;
  //@}

  /** A history of the agents worldmodels. */
  ObjectCollection<WorldModel> wm;

  /** A history of the agents observations. */
  ObjectCollection<WorldObservation> observations;

  /** The facts known about the world. */
  const WorldFacts* worldFacts;  

  /** Pointer to the viewManager to order update of the views.
   *  This is called after the worldModel has been updated by the
   *  strategies, so that the views can uses the newly calculated
   *  information. */
  ViewManager* viewManager;

private:
  /** The only instance of the class, Singleton. */
  static WorldModelInterface *onlyInstance;

  /** Hidden copy-constructor. */
  WorldModelInterface(const WorldModelInterface&);  

  /** Hidden assignment. */
  WorldModelInterface& operator= (const WorldModelInterface&);

  /** Hidden constructor. */
  WorldModelInterface();  

  ServerCommand* lastCommand;
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "WorldModelInterface.icc"
#endif


#endif // !WORLDMODELINTERFACE_H
