/*!

Field Strategy - AIP10

The FieldStrategy model will approach both "knowledge representation and
reasoning" as well as team organization concepts for RoboSoc. This high
level strategic passing model, consisting of representation of the field,
broken down into different zones takes a classic divide and conquer approach
to the problem of a cohesive team strategy.

The aim of this strategy module is first to develop a model that includes
different higher level strategies in passing by setting up a few default
strategies, like the passing game over the flanks ( in you own half, try
not to pass in the middle, but rather to the left or right flank, and as
the agent comes close to the opponent goal, try to pass in the middle again).

The model works by splitting the field into a number of different 'zones'.
Each of these zones has a relation value (between 0 and 1, where 1 is the
most desirable) to every other zone on the field. To minimize the size of
the data structure required to store such relations, any two zones which do
not have a specified relation between them are taken to have a realation value
of 0.5.

In the case of GoToPassingPosition, this value represents the desirability
of a pass betwee any 2 given zones. That is, if I am standing in zone2 and can
pass to either zone4 (relation to zone2 of 0.3) or zone11 (relation to zone2
of 0.89), then zone11 is a more desirable passing option.

This strategy however is not designed to represent a few separate zones, but
the field as a whole. Through specifying relationships between a large number
of zones on the field, it is easy to see that we can effectively control the
flow of the ball when it is in our teams possession if our agents are implementing
the same FieldStrategy model.

Further to this, by decreasing the size of the zones (at this stage the minimum
size is 1m square) we are able to more precisely control the balls flow, while on
the other hand, if we increase the size of the zones we can give a more general
FieldStrategy model. (eg. Kicking towards the opponents goal is better than kicking
toward our own).

Aside from complementing GoToPassingPosition, the major advantage of this module
is it's adaptability. While the initial focus will be working with GoToPassingPosition,
it will be very easy to adapt this form of knowledge representation to any other
functionality that requires a way to model a relationship between different parts of the
field.

For any given FieldStrategy model, all zones will be rectangles f the same size, specified
at construction. This allows any Point on the field to be related easily to a zone and then
have it's realtion to any other zone returned.

FieldStrategies can be obtained by the agent in a number of ways. Firstly the agent would
be able to read in pre-designed strategies from a text file at initialization and store
these in a dynamic data structure. A more interesting approach to obtaining field
strategies is through the use of the coach during game time. As the coach has perfect
world information at all times, the ability for him to communicate this knowledge,
through up to date FieldStrategy models, to his players would be invaluable. Coaches
could also be assigned the task of choosing which of the existing FieldStrategy models
should be used by any individual agent at a given time in the game. The possibility also
exists for the agent itself to update it's existing models, or create new ones as a result
of it's experiences.
 
The result of this real time data interpretation and representation would be a more skilled
and dynamic set of FieldStrategie models. The agents could even have separate sets of
strategies for each team they compete against. If they then reuse and update this model
each time they play the team, they should learn to out perform their opposition.

The initial inspiration for our planned research was the model implemented by The ORCA
(Osnabrck RoboCup Agents). 
*/

/* -*- Mode: C++ -*-
 * $Id: FieldStrategy.h,v 1.4 2004/01/16 12:49:16 aip10 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: FieldStrategy.h
 * Unitname: Framework/strategy/
 * $Revision: 1.4 $
 * Created by: aip10 2003/11/10
 * Last modified by $Author: aip10 $ $Date: 2004/01/16 12:49:16 $
 *
 *
 * HISTORY
 *
 * $Log: FieldStrategy.h,v $
 * Revision 1.4  2004/01/16 12:49:16  aip10
 * changed descriptions
 *
 * Revision 1.3  2004/01/12 12:03:41  aip10
 * New strategy commited
 *
 *
 */

/**
 * @file FieldStrategy.h
 * Include file for the class FieldStrategy
 *
 * @par Unit:
 *      Framework/strategy/
 * $Revision: 1.4 $
 * $Author: aip10 $
 * $Date: 2004/01/16 12:49:16 $
 */


#ifndef FIELD_STRATEGY_H
#define FIELD_STRATEGY_H

#include <string>
#include <iostream>
//#include "view_constants.h"
#include "ViewManager.h"
#include "WorldFacts.h"
using namespace std;

RS_BEGIN_NAMESPACE



class FieldStrategy {

public:
	
	
	/* constructor with xSize adn ySize
	@param1 float xSize defines xSize of the strategy model
	@param2 float ySize defines ySize of the strategy model
	*/
	FieldStrategy(const string& name_, float xSize_, float ySize_);
	
	/* destructor */
	~FieldStrategy();
	
	/* getAverageRelation
	@param1 int zone defines the zone we are refering to
	@return float returns the average relation to 'zone' from all other zones
	*/	
	float getAverageRelation(int zone);
	
	/* getRelation
	@param1 Point us, the position that 'this' player is at
	@param2 Point them, the position to find the relation to
	@return float giving the value of relation us->them
	*/
	float getRelation(Point us, Point them);

	/* getRelation
	@param1 int from, the zone to get relation from
	@param2 int to, the zone to get the relation to
	@retunr float giving the value of relation from->to
	*/
	float getRelation(int from, int to);
	

	/* getBest
	@param1 Point us, the position that 'this' player is at
	@param2 Point *bestLeft, the top left corner of the best zone
	@param3 Point *bestRight, the top right corner of the best zone
	@return float giving the relation value between us->best
	@return Point *bestLeft,*bestRight giving the top left and bottom right corners
	of the best zone from position Point us
	*/
	float getBestZone(Point us, Point *bestLeft, Point *bestRight);
	
	/* getName
	@return string, the name of this FieldStrategy
	*/
	string getName();

	/* getNumZones
	@return int, the number of zones in 'this' FieldStrategy
	*/
	int getNumZones();

	/* getXSize
	@return float the Xsize of this FieldStrategy
	*/
	float getXSize();

	/* getYSize
	@return float the Ysize of this FieldStrategy
	*/
	float getYSize();
	
	
	/* getPoints
	@param1 int zone, the zone we want to retur points for
	@param2 Point **left, pointer to top left of zone
	@param3 Point **right, pointer to top right of zone
	@return **left and **right point to top left and bottom right corners of zone
	respectively
	*/
	void getPoints(int zone, Point **left, Point **right);
	
	/* setRelation
	@param1 int zone, the zone to set the relation from
	@param2 int target, the zone to set the relation to
	@param3 float value, the relation from zone -> target
	@return relation of relations[zone][target] == value
	*/
	void setRelation(int zone, int target, float value);
	
	/* getZone
	@param Point us
	@return int, the zone that 'us' falls into
	*/
	int getZone(Point us);

private:

/////////////////PARAMS///////////////////////////////////////

	/* hold the x and y size for this FieldStrategy*/
	string name;
	float xSize;
	float ySize;
	int numZonesX;
	int numZonesY;
	int numZones;
	float  **relations;

};

RS_END_NAMESPACE

#endif // !FIELD_STRATEGY_H
