/* -*- Mode: C++ -*-
 * $Id: NewSeeStrategy.cc,v 1.5 2004/09/02 08:00:04 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: NewSeeStrategy.cc
 * Unitname: BasicSystem
 * $Revision: 1.5 $
 * Created by: David Rosn 2002-07-01
 * Last modified by $Author: frehe $ $Date: 2004/09/02 08:00:04 $
 *
 *
 * HISTORY
 *
 * $Log: NewSeeStrategy.cc,v $
 * Revision 1.5  2004/09/02 08:00:04  frehe
 * Added update of goalie information.
 *
 * Revision 1.4  2003/11/05 12:31:45  aip15
 * Fixed an error in the speed vector calculation.
 *
 * Revision 1.3  2003/07/22 14:48:56  s02davro
 * Tackling + point
 *
 * Revision 1.2  2002/10/09 21:56:36  aip19
 * Bug fix for NARROW view width
 *
 * Revision 1.1  2002/09/26 15:24:59  s02davro
 * Added new see strategy, this use BasicSeeStrategy and handles player matching
 *
 *
 */

/**
 * @file NewSeeStrategy.cc
 * Implementation file for the class NewSeeStrategy.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.5 $
 * $Author: frehe $
 * $Date: 2004/09/02 08:00:04 $
 */


#include "NewSeeStrategy.h"
#include "BasicSeeStrategy.h"
#include "WorldModelInterface.h"
#include "rs_error.h"

RS_BEGIN_NAMESPACE

NewSeeStrategy::NewSeeStrategy() {
}

NewSeeStrategy::~NewSeeStrategy() { 
}

void NewSeeStrategy::updateAfterSee() {
  RS_LOG ( LA_STRATEGY, "NewSeeStrategy - start");
  world = &wmi->getWorld();
  observation = &wmi->getObservation();
  handleAgent();
  handleBall();
  handlePlayers();
  handleClosestPlayers();
  handleTime();
  RS_LOG ( LA_STRATEGY, "NewSeeStrategy - end");
}

void NewSeeStrategy::handlePlayers () {
  std::vector<PlayerObject> obsTeammates;
  std::vector<PlayerObject> obsOpponents;
  std::vector<PlayerObject> obsUnknowns;
  for (std::vector<ServerPlayerObject>::const_iterator i = 
	 observation->players.begin();
       i != observation->players.end(); ++i) {
    PlayerObject p(agent.getPosition() +
		   VectorUE(i->direction + agent.getNeckDirection(),
			    i->distance),
		   wmi->getCurrentTime());
    p.setBodyDirection(i->relBodyDirection + agent.getNeckDirection());
    p.setNeckDirection(i->relNeckDirection + agent.getNeckDirection());    
    p.setUniformNumber(i->uniformNumber);
    p.setTackling(i->tackling);
    p.setPointDirection(i->pointDirection);
    if (i->uniformNumber.isKnown()) 
      p.setUniformNumberObservationTime(wmi->getCurrentTime());    
    p.setTeam(i->team);
    p.setGoalie(i->goalie);
    if (i->distanceChange.isKnown() && i->directionChange.isKnown()) {
      // Setting the player's speed
      p.setSpeedObservationTime(wmi->getCurrentTime());
      PointUE agentPos(world->getAgent().getPosition());
      PointUE playerPos(p.getPosition());
      VectorUE agentToPlayer(agentPos, playerPos);
      PointUE relPos = playerPos - agentPos;
      PointUE unit = relPos / agentToPlayer.getLength();
      FloatUE tmp = i->directionChange.getRad() * agentToPlayer.getLength();
      VectorUE absVel(PointUE(i->distanceChange * unit.getX() - tmp*unit.getY(),
			      i->distanceChange * unit.getY() + tmp*unit.getX()));    
      // The vector is already relative...
      //absVel.rotate(world->getAgent().getNeckDirection());
      // We need to compensate for agent decay.?
      absVel += world->getAgent().getSpeedVector();
      p.setSpeedVector(absVel);
    }
    if (p.getPosition().isKnown()) {
      switch (i->team) {
      case T_OUR_TEAM:
	p.setTeamObservationTime(wmi->getCurrentTime());
	p.setSide(facts->getOurSide());	
	obsTeammates.push_back(p);
	break;
      case T_THEIR_TEAM:
	p.setTeamObservationTime(wmi->getCurrentTime());
	p.setSide(facts->getTheirSide());
	obsOpponents.push_back(p);
	break;
      case T_UNKNOWN:
      default:
	p.setSide(S_UNKNOWN);
	obsUnknowns.push_back(p);
	break;
      }
    }
  }
  mergePlayers(obsTeammates, obsOpponents, obsUnknowns);
}

void NewSeeStrategy::mergePlayers(std::vector<PlayerObject>& obsTeammates,
				    std::vector<PlayerObject>& obsOpponents,
				    std::vector<PlayerObject>& obsUnknowns) {
  std::vector<PlayerObject> teammates = world->getTeammates();
  std::vector<PlayerObject> opponents = world->getOpponents();
  std::vector<PlayerObject> unknowns = world->getUnknowns();
  
  std::vector<PlayerObject> resTeammates;
  std::vector<PlayerObject> resOpponents;
  std::vector<PlayerObject> resUnknowns;

  // Match by uniformnumber
  teamNumberMatch(obsTeammates, teammates, resTeammates);
  teamNumberMatch(obsOpponents, opponents, resOpponents);
  
  // Match closest player to player.
  unknownToNumberMatch(obsTeammates, teammates, unknowns, resTeammates);
  unknownToNumberMatch(obsOpponents, opponents, unknowns, resOpponents);

  // Try matching current unknown players to previous known players.
  numberToUnknownMatch(obsUnknowns, teammates, opponents, unknowns, 
		       resTeammates, resOpponents, resUnknowns);

  // Remove all the players that should been seen.
  removeGhosts(teammates);
  removeGhosts(opponents);
  removeGhosts(unknowns);    

  // Add All remaining players, observed and previously known
  addPlayers(obsTeammates, resTeammates);
  addPlayers(teammates, resTeammates);
  addPlayers(obsOpponents, resOpponents);
  addPlayers(opponents, resOpponents);
  addPlayers(obsUnknowns, resUnknowns);
  addPlayers(unknowns, resUnknowns);

  world->clearOpponents();
  world->clearTeammates();
  world->clearUnknowns();
  for (std::vector<PlayerObject>::iterator i = resTeammates.begin();
       i != resTeammates.end(); ++i)
    world->addTeammate(*i);

  for (std::vector<PlayerObject>::iterator i = resOpponents.begin();
       i != resOpponents.end(); ++i)
    world->addOpponent(*i);

  for (std::vector<PlayerObject>::iterator i = resUnknowns.begin();
       i != resUnknowns.end(); ++i)
    world->addUnknown(*i);
}

void NewSeeStrategy::removeGhosts(std::vector<PlayerObject>& players) {
    AngleDeg viewAngle;
    AgentObject agent = world->getAgent();
    switch (agent.getViewWidth()) {
    case VW_NARROW:
      viewAngle = 22.5;
      break;
    case VW_NORMAL:
      viewAngle = 45;
      break;
    case VW_WIDE:
      viewAngle = 90;
      break;
    default:
      viewAngle = 45;
    }
    std::vector<PlayerObject>::iterator i = players.begin();
    while (i != players.end()) {
      if ((abs(agent.getPosition().getDirectionTo(i->getPosition())
	       -agent.getBodyDirection()) < viewAngle) ||
	  (agent.getPosition().getDistanceTo(i->getPosition()) <
	   facts->SP_feel_distance))
	i = players.erase(i);
      else
	++i;
    }
}

void NewSeeStrategy::addPlayers(std::vector<PlayerObject>& players,
				std::vector<PlayerObject>& resPlayers) {
  for (std::vector<PlayerObject>::iterator i = players.begin();
       i != players.end(); ++i)
    resPlayers.push_back(*i);
}

void NewSeeStrategy::numberToUnknownMatch(std::vector<PlayerObject>& obsUnknowns,
					  std::vector<PlayerObject>& teammates,
					  std::vector<PlayerObject>& opponents,
					  std::vector<PlayerObject>& unknowns,
					  std::vector<PlayerObject>& resTeammates,
					  std::vector<PlayerObject>& resOpponents,
					  std::vector<PlayerObject>& resUnknowns) {
  std::vector<PlayerObject>::iterator j;
  std::vector<PlayerObject>::iterator i = obsUnknowns.begin();
  while (i != obsUnknowns.end()) {
    bool found = false;
    enum MatchType {M_UNKNOWN, M_TEAMMATE, M_OPPONENT} matchType = M_UNKNOWN;
    std::vector<PlayerObject>::iterator bestMatch;
    FloatUE distance;
    for (j = teammates.begin(); j != teammates.end(); ++j) {
      FloatUE tmpDistance = i->getPosition().getDistanceTo(j->getPosition());      
      if (j->getObservationTime().isKnown()) {
	Time curTime = wmi->getCurrentTime();
	Time obsTime = j->getObservationTime().getTime();
	UInt cycles = curTime.getT() - obsTime.getT();
	FloatUE distanceTo = 
	  agent.getPosition().getDistanceTo(i->getPosition());
	Float distError = 0;
	if (distanceTo.isKnown()) distError = distanceTo.getValue()*0.1;
	if (cycles == 0)
	  cycles = curTime.getS() - obsTime.getS();
	if ((cycles != 0) && ((tmpDistance / cycles)  < 
			      (facts->SP_player_speed_max + distError)) ||
	    (tmpDistance < distError)) {
	    // The position is possible according to last observation.	
	    found = true;
	    if (distance.isUnknown() || (tmpDistance < distance)) {
	      // Better match.
	      bestMatch = j;
	      distance = tmpDistance;
	      matchType = M_TEAMMATE;
	    }
	  }
      }
    }    
    for (j = opponents.begin(); j != opponents.end(); ++j) {
      FloatUE tmpDistance = i->getPosition().getDistanceTo(j->getPosition());      
      if (j->getObservationTime().isKnown()) {
	Time curTime = wmi->getCurrentTime();
	Time obsTime = j->getObservationTime().getTime();
	UInt cycles = curTime.getT() - obsTime.getT();
	FloatUE distanceTo = 
	  agent.getPosition().getDistanceTo(i->getPosition());
	Float distError = 0;
	if (distanceTo.isKnown()) distError = distanceTo.getValue()*0.1;
	if (cycles == 0)
	  cycles = curTime.getS() - obsTime.getS();
	if ((cycles != 0) && ((tmpDistance / cycles)  < 
			      (facts->SP_player_speed_max + distError)) ||
	    (tmpDistance < distError)) {
	    // The position is possible according to last observation.	
	    found = true;
	    if (distance.isUnknown() || (tmpDistance < distance)) {
	      // Better match.
	      bestMatch = j;
	      distance = tmpDistance;
	      matchType = M_OPPONENT;
	    }
	  }
      }
    }    
    for (j = unknowns.begin(); j != unknowns.end(); ++j) {
      FloatUE tmpDistance = i->getPosition().getDistanceTo(j->getPosition());      
      if (j->getObservationTime().isKnown()) {
	Time curTime = wmi->getCurrentTime();
	Time obsTime = j->getObservationTime().getTime();
	UInt cycles = curTime.getT() - obsTime.getT();
	FloatUE distanceTo = 
	  agent.getPosition().getDistanceTo(i->getPosition());
	Float distError = 0;
	if (distanceTo.isKnown()) distError = distanceTo.getValue()*0.1;
	if (cycles == 0)
	  cycles = curTime.getS() - obsTime.getS();
	if ((cycles != 0) && ((tmpDistance / cycles)  < 
			      (facts->SP_player_speed_max + distError)) ||
	    (tmpDistance < distError)) {
	    // The position is possible according to last observation.	
	    found = true;
	    if (distance.isUnknown() || (tmpDistance < distance)) {
	      // Better match.
	      bestMatch = j;
	      distance = tmpDistance;
	      matchType = M_UNKNOWN;
	    }
	  }
      }
    }
    if (found) {
      switch (matchType) {
      case M_TEAMMATE: {
	resTeammates.push_back(*i);
	resTeammates.back().setTeam(T_OUR_TEAM);
	resTeammates.back().setUniformNumber(bestMatch->getUniformNumber());
	teammates.erase(bestMatch);
	i = obsUnknowns.erase(i);
	RS_LOG ( LA_STRATEGY, "Match unknown player with teammate");
	RS_LOG ( LA_STRATEGY, "  Pos: " << resTeammates.back().getPosition());
	RS_LOG ( LA_STRATEGY, "  Nr:  " << resTeammates.back().getUniformNumber());
	break;
      }
      case M_OPPONENT: {
	resOpponents.push_back(*i);
	resOpponents.back().setTeam(T_THEIR_TEAM);
	resOpponents.back().setUniformNumber(bestMatch->getUniformNumber());
	opponents.erase(bestMatch);
	i = obsUnknowns.erase(i);
	RS_LOG ( LA_STRATEGY, "Match unknown player with opponent");
	RS_LOG ( LA_STRATEGY, "  Pos: " << resOpponents.back().getPosition());
	RS_LOG ( LA_STRATEGY, "  Nr:  " << resOpponents.back().getUniformNumber());
	break;
      }
      case M_UNKNOWN: {
	resUnknowns.push_back(*i);
	resUnknowns.back().setUniformNumber(bestMatch->getUniformNumber());
	resUnknowns.back().setUniformNumber(bestMatch->getUniformNumber());
	unknowns.erase(bestMatch);
	i = obsUnknowns.erase(i);
	RS_LOG ( LA_STRATEGY, "Match unknown player with unknown");
	RS_LOG ( LA_STRATEGY, "  Pos: " << resUnknowns.back().getPosition());
	RS_LOG ( LA_STRATEGY, "  Nr:  " << resUnknowns.back().getUniformNumber());
	break;
      }
      default:
	RS_LOG( LA_WARNING, "Warning: Matching problem!");
	++i;
      }
    }
    else
      ++i;
  }
}


void NewSeeStrategy::teamNumberMatch(std::vector<PlayerObject>& obsPlayers,
				       std::vector<PlayerObject>& players,
				       std::vector<PlayerObject>& resPlayers) {
  std::vector<PlayerObject>::iterator i = obsPlayers.begin();
  while (i != obsPlayers.end()) {
    bool found = false;
    if (i->getUniformNumber().isKnown()) {
      for (std::vector<PlayerObject>::iterator j = players.begin();
	   j != players.end(); ++j) {
	if (j->getUniformNumber() == i->getUniformNumber()) {
	  // Exact match found.
	  resPlayers.push_back(*i);
	  players.erase(j);
	  i = obsPlayers.erase(i);
	  RS_LOG ( LA_STRATEGY, "Match known player with player");
	  RS_LOG ( LA_STRATEGY, "  Pos: " << resPlayers.back().getPosition());
	  RS_LOG ( LA_STRATEGY, "  Nr:  " << resPlayers.back().getUniformNumber());
	  found = true;
	  break;
	}
      }
    }
    if (!found) ++i;
  }
}

void NewSeeStrategy::unknownToNumberMatch(std::vector<PlayerObject>& obsPlayers,
					    std::vector<PlayerObject>& players,
					    std::vector<PlayerObject>& unknowns,
					    std::vector<PlayerObject>& resPlayers) {
  std::vector<PlayerObject>::iterator i = obsPlayers.begin();
  std::vector<PlayerObject>::iterator j;
  while (i != obsPlayers.end()) {
    bool found = false;
    bool unknown = false;
    std::vector<PlayerObject>::iterator bestMatch;
    FloatUE distance;
    for (j = players.begin(); j != players.end(); ++j) {
      if (i->getUniformNumber().isKnown() && j->getUniformNumber().isKnown() &&
	  i->getUniformNumber() != j->getUniformNumber())
	// No match, skip player.
	break;
      FloatUE tmpDistance = i->getPosition().getDistanceTo(j->getPosition());      
      if (j->getObservationTime().isKnown()) {
	Time curTime = wmi->getCurrentTime();
	Time obsTime = j->getObservationTime().getTime();
	UInt cycles = curTime.getT() - obsTime.getT();
	FloatUE distanceTo = 
	  agent.getPosition().getDistanceTo(i->getPosition());
	Float distError = 0;
	if (distanceTo.isKnown()) distError = distanceTo.getValue()*0.1;
	if (cycles == 0)
	  cycles = curTime.getS() - obsTime.getS();
	if ((cycles != 0) && ((tmpDistance / cycles)  < 
			      (facts->SP_player_speed_max + distError)) ||
	    (tmpDistance < distError))
	  {
	    // The position is possible according to last observation.	
	    found = true;
	    if (distance.isUnknown() || (tmpDistance < distance)) {
	      // Better match.'
	      bestMatch = j;
	      distance = tmpDistance;
	    }
	  }
      }
    }
    for (j = unknowns.begin(); j != unknowns.begin(); ++j) {
      FloatUE tmpDistance = i->getPosition().getDistanceTo(j->getPosition());      
      if (j->getObservationTime().isKnown()) {
	Time curTime = wmi->getCurrentTime();
	Time obsTime = j->getObservationTime().getTime();
	UInt cycles = curTime.getT() - obsTime.getT();
	FloatUE distanceTo = 
	  agent.getPosition().getDistanceTo(i->getPosition());
	Float distError = 0;
	if (distanceTo.isKnown()) distError = distanceTo.getValue()*0.1;
	if (cycles == 0)
	  cycles = curTime.getS() - obsTime.getS();
	if (cycles != 0)
	if ((cycles != 0) && ((tmpDistance / cycles)  < 
			      (facts->SP_player_speed_max + distError)) ||
	    (tmpDistance < distError)) {
	    // The position is possible according to last observations.	
	    found = true;
	    if (distance.isUnknown() || (tmpDistance < distance)) {
	      // Better match.'
	      unknown = true;	
	      bestMatch = j;
	      distance = tmpDistance;
	    }
	  }
      }
    }
    if (found) {
      resPlayers.push_back(*i);
      if (bestMatch->getUniformNumber().isKnown())
	resPlayers.back().setUniformNumber(bestMatch->getUniformNumber());
      if (unknown) {
	unknowns.erase(bestMatch);
	RS_LOG ( LA_STRATEGY, "Match known player with unknown");
	RS_LOG ( LA_STRATEGY, "  Pos: " << resPlayers.back().getPosition());
	RS_LOG ( LA_STRATEGY, "  Nr:  " << resPlayers.back().getUniformNumber());
      }
      else {
	players.erase(bestMatch);
	RS_LOG ( LA_STRATEGY, "Match known player with player (no nr)");
	RS_LOG ( LA_STRATEGY, "  Pos: " << resPlayers.back().getPosition());
	RS_LOG ( LA_STRATEGY, "  Nr:  " << resPlayers.back().getUniformNumber());
      }
      i = obsPlayers.erase(i);
    }
    else 
      ++i;
  }
}				    


RS_END_NAMESPACE




