/* -*- Mode: C++ -*-
 * $Id: PersistentSeeStrategy.cc,v 1.3 2004/09/12 23:27:57 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: PersistentSeeStrategy.cc
 * Unitname: PersistentSystem
 * $Revision: 1.3 $
 * Created by: Peter Andersson 2004-08-17
 * Last modified by $Author: frehe $ $Date: 2004/09/12 23:27:57 $
 *
 *
 * HISTORY
 *
 * $Log: PersistentSeeStrategy.cc,v $
 * Revision 1.3  2004/09/12 23:27:57  frehe
 * Added a margin parameter to the constructor.
 *
 * Revision 1.2  2004/09/03 10:00:35  petan
 * Added case when ball is not where it is supposed to
 *
 * Revision 1.1  2004/08/18 12:01:23  petan
 * Added strategy for remembering positions better
 *
 *
 */

/**
 * @file PersistentSeeStrategy.cc
 * Implementation file for the class PersistentSeeStrategy.
 *
 * @par Unit:
 *      PersistentSystem
 * $Revision: 1.3 $
 * $Author: frehe $
 * $Date: 2004/09/12 23:27:57 $
 */


#include "PersistentSeeStrategy.h"
#include "WorldModelInterface.h"
#include "rs_error.h"

RS_BEGIN_NAMESPACE

PersistentSeeStrategy::PersistentSeeStrategy(int margin) :
  mMargin(margin),
  mLastNeckObservation(0),
  mLastPositionObservation(0)
{
}

PersistentSeeStrategy::~PersistentSeeStrategy() { 
}

void PersistentSeeStrategy::setUncertaintyCycleMargin(int margin)
{
  mMargin = margin;
}

void PersistentSeeStrategy::handleAgent() {  
  agent = world->getAgent();

  // Try to localize the agent neck/body direction
  AngleDegUE neckdir = computeAgentNeckDirection();
  RS_LOG ( LA_STRATEGY, "  Computed neckdir: " << neckdir);
  // If successful, we'll update, otherwise we keep old values
  if (neckdir.isKnown() || mLastNeckObservation > mMargin) {
    mLastNeckObservation = 0;
    agent.setNeckDirection(neckdir);
    AngleDegUE bodydir = neckdir - observation->agent.relNeckDirection;
    RS_LOG ( LA_STRATEGY, "  Computed bodydir: " << neckdir);
    agent.setBodyDirection(bodydir);
  }
  else {
    mLastNeckObservation++;
  }

  // Try to localize agent
  PointUE agentPos = computeAgentPosition();  
  // If successful, we'll update the agent
  if (agentPos.isKnown() || mLastPositionObservation > mMargin) {
    mLastPositionObservation = 0;
    agent.setPosition(agentPos);
    agent.setObservationTime(wmi->getCurrentTime());
  }
  else {
    mLastPositionObservation++;
  }

  if (facts->getOurSide() == S_RIGHT) {
    // Rotate the agent.
    agent.setNeckDirection(agent.getNeckDirection()+180.0);
    agent.setBodyDirection(agent.getBodyDirection()+180.0);
    agent.setPosition(-agent.getPosition());
  }

  world->setAgent(agent);
}

void PersistentSeeStrategy::handleBall() {
  RS_LOG(LA_STRATEGY, "handleBall");
  
  BallObject currentBall = world->getBall();
  BallObject ball = world->getBall();

  // The agent has seen the ball in the last cycle
  if (observation->ball.distance.isKnown()) {
    RS_LOG(LA_STRATEGY, "saw ball!");
    ball.setPosition(agent.getPosition() + 
		     VectorUE(observation->ball.direction + 
			      agent.getNeckDirection(),
			      observation->ball.distance));
    VectorUE speed1(calculateSpeed1(observation->ball.distanceChange,
				    observation->ball.directionChange,
				    ball.getPosition()));
    VectorUE speed2(calculateSpeed2(currentBall, ball));    
    if (speed1.isKnown() || speed2.isKnown()) {
      if (speed2.isKnown()) {
	RS_LOG(LA_STRATEGY, "BallSpeed (position) : " << Point(speed2.getVector()));
	ball.setSpeedVector(speed2);
      }
      if (speed1.isKnown()) {
	RS_LOG(LA_STRATEGY, "BallSpeed (dirchange): " << Point(speed1.getVector()));
	ball.setSpeedVector(speed1);
      }
    }
    ball.setObservationTime(wmi->getCurrentTime());    
  } else if ( ball.getObservationTime().isKnown() &&
	      world->getGameTime().getT() - ball.getObservationTime().getT() > facts->CP_cycles_before_unknown_ball ) {
    RS_LOG(LA_STRATEGY, "BallPosition set to unknown since I haven't seen it for "
	   << world->getGameTime().getT() << " - "
	   << ball.getObservationTime().getT() << " = "
	   << world->getGameTime().getT() - ball.getObservationTime().getT()
	   << " < " << facts->CP_cycles_before_unknown_ball
	   << " cycles.");
    ball.setPosition(PointUE());
    ball.setSpeedVector(VectorUE());	  
  } else if (agent.getPosition().isKnown() &&
             agent.getBodyDirection().isKnown() &&
             ball.getPosition().isKnown()) {
    // If we're looking at the place where the ball was supposed
    // to be, but it isn't there. Set to unknown
    AngleDegUE angleToBall = 
      agent.getPosition().getDirectionTo(ball.getPosition());
    angleToBall -= agent.getNeckDirection();

    AngleDeg viewAngle = 0;
    switch (agent.getViewWidth()) {
    case VW_NARROW:
      viewAngle = 22.5;
      break;
    case VW_NORMAL:
      viewAngle = 45;
      break;
    case VW_WIDE:
      viewAngle = 90;
      break;
    default:
      viewAngle = 45;
    }
    if (abs(angleToBall.getDeg()) < viewAngle-5) {
      ball.setPosition(PointUE());
      ball.setSpeedVector(VectorUE());	
      RS_LOG(LA_STRATEGY, "BallPosition set to unknown since body "<<
	     agent.getBodyDirection() << " neck " <<
	     agent.getNeckDirection() << " agent pos " <<
	     agent.getPosition() << " ball pos " <<
	     ball.getPosition() << " angle to ball " <<
	     angleToBall.getDeg() << " is outside view " <<
	     viewAngle << " degrees");
    }
  }

  world->setBall(ball);
}

RS_END_NAMESPACE
