/* -*- Mode: C++ -*-
 * $Id: PersistentSeeStrategy.h,v 1.3 2005/09/19 17:09:01 petan Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: PersistentSeeStrategy.h
 * Unitname: Framework/strategy
 * $Revision: 1.3 $
 * Created by: Peter Andersson 2004-08-17
 * Last modified by $Author: petan $ $Date: 2005/09/19 17:09:01 $
 *
 *
 * HISTORY
 *
 * $Log: PersistentSeeStrategy.h,v $
 * Revision 1.3  2005/09/19 17:09:01  petan
 * Updated documentation to reflect margin and CP_cycles_before_unknown_ball
 *
 * Revision 1.2  2004/09/12 23:27:57  frehe
 * Added a margin parameter to the constructor.
 *
 * Revision 1.1  2004/08/18 12:01:23  petan
 * Added strategy for remembering positions better
 *
 *
 */

/**
 * @file PersistentSeeStrategy.h
 * This strategy tries to minimize the "Goldfish" memory problems
 * of Basic and NewSeeStrategy. The main change is that if a new
 * position can't be calculated from the sensor information it
 * will retain it's old position albeit a tad more unsure.
 * If it get's too unsure, the position will become unknown.
 *
 * It also updates the ball with unknown if it is not where it
 * should be. The number of cycles before the ball is considered
 * unknown is controlled by CP_cycles_before_unknown_ball in
 * WorldFacts.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.3 $
 * $Author: petan $
 * $Date: 2005/09/19 17:09:01 $
 */


#ifndef PERSISTENTSEESTRATEGY_H
#define PERSISTENTSEESTRATEGY_H

#include "basic_system_types.h"
#include "NewSeeStrategy.h"

RS_BEGIN_NAMESPACE

/**
 * @class PersistentSeeStrategy PersistentSeeStrategy.h
 *
 * @par Unit:
 *      Framework/strategy
 * $Revision: 1.3 $
 * $Author: petan $
 * $Date: 2005/09/19 17:09:01 $
 */

class PersistentSeeStrategy : public NewSeeStrategy {
public:

  /**
   * Creates a new PersistentSeeStrategy.
   * @param margin The number of cycles before the agent forgets its position. For the ball, margin has
   * no effect, use CP_cycles_before_unknown_ball in WorldFacts.
   */
  PersistentSeeStrategy(int margin = 10);

  /**
   * Deallocates the current strategy
   */
  virtual ~PersistentSeeStrategy();

  /**
   * Sets the number of cycles before the agent's position/direction is considered unknown.
   * It does not affect the number of cycles for the ball. For this purpose, use the
   * CP_cycles_before_unknown_ball
   * @param margin The number of cycles before the agent's position/direction should be considered unknown
   */
  virtual void setUncertaintyCycleMargin(int margin);
protected:
  virtual void handleBall();
  virtual void handleAgent();
 private:
  int mMargin;

  int mLastNeckObservation;
  int mLastPositionObservation;
};

RS_END_NAMESPACE

#endif // !PERSISTENTSEESTRATEGY_H
