/* -*- Mode: C++ -*-
 * $Id: TestTeam.cc,v 1.6 2002/09/13 14:48:56 s02davro Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: TestTeam.cc
 * Unitname: Framework/test/TestTeam
 * $Revision: 1.6 $
 * Created by: Fredrik Heintz 2000-09-16
 * Last modified by $Author: s02davro $ $Date: 2002/09/13 14:48:56 $
 *
 *
 * HISTORY
 * $Log: TestTeam.cc,v $
 * Revision 1.6  2002/09/13 14:48:56  s02davro
 * FALSE -> FuzzyBool(0.0), TRUE -> FuzzyBool(1.0)
 *
 * Revision 1.5  2002/09/13 14:10:39  s02davro
 * Updated the TestTeam to use new version of RoboSoc
 *
 * Revision 1.4  2001/09/02 11:33:12  frehe
 * Fixed some incorrect include files reported by Peter Andersson and Jonas Carlsson.
 *
 * Revision 1.3  2001/09/01 15:06:52  frehe
 * Updated the file to work with the current version of RoboSoc.
 *
 * Revision 1.2  2000/09/16 18:39:41  frehe
 * Fixed the log messages.
 *
 * Revision 1.1  2000/09/16 18:37:08  frehe
 * Add a team that can be used as the basis for test teams.
 *
 */

/**
 * @file Team.cc
 * Implementation file for the agent Team.
 *
 * @par Unit:
 *      Framework/test/TestTeam
 * $Revision: 1.6 $
 * $Author: s02davro $
 * $Date: 2002/09/13 14:48:56 $
 */


#include <fstream>

#include "framework_compiler_directives.h"
#include "view_constants.h"
#include "FrameworkAgent.h"
#include "WorldModelInterface.h"
#include "TestTeamDecision.h"
#include "LogStream.h"
#include "BasicSeeStrategy.h"
#include "BasicHearStrategy.h"
#include "BasicCommandStrategy.h"
#include "BasicSenseStrategy.h"
#include "BasicInitStrategy.h"
#include "BasicCoachMessageStrategy.h"
#include "BasicNewCycleStrategy.h"
#include "BasicCoachInformationStrategy.h"

RS_BEGIN_NAMESPACE

#if defined(RS_USE_ERROR) && !defined(RS_LOG_ERROR)
extern void rs_warning(const std::string& str,
		       const std::string& file,
		       int line)
{
  std::cerr << "WARNING: " << str
	    << " at line " << line 
	    << " in file " << file
	    << std::endl;
}

extern void rs_error(const std::string& str,
		     const std::string& file,
		     int line)
{
  std::cerr << "ERROR: " << str
	    << " at line " << line 
	    << " in file " << file
	    << std::endl;
}

extern void rs_fatal(const std::string& str,
		     const std::string& file,
		     int line)
{
  std::cerr << "FATAL: " << str
	    << " at line " << line 
	    << " in file " << file
	    << std::endl;
  abort();
}
#endif

#if !defined(RS_USE_EXCEPTION) && !defined(RS_LOG_EXCEPTION)
extern void rs_throw(const std::string& exception,
		     const std::string& msg,
		     const std::string& file,
		     int line)
{
  std::cerr << "RAISED: " << exception << "(\"" << msg << "\")"
	    << " at line " << line 
	    << " in file " << file
	    << std::endl;
}
#endif


RS_END_NAMESPACE


int main(int argc, char** argv)
{
#ifdef RS_USE_NAMESPACE
  using namespace RS_NAMESPACE;
#endif
  std::ofstream logFile1("logfile.txt");
  std::ofstream logFile2("commands.txt");
  std::ofstream logFile3("sensor_data.txt");
  std::ofstream coachLog("coachLog.txt");
  
  LogStream *logStream = LogStream::instance();
  
  logStream->addLogArea(LA_ROBOSOC, LA_TOP, true, &std::cerr);
  logStream->addLogArea(LA_WARNING, LA_ROBOSOC, true, &logFile1,
			true, &std::cerr);
  logStream->addLogArea(LA_ERROR, LA_ROBOSOC, true, &logFile1,
			true, &std::cerr);
  logStream->addLogArea(LA_FATAL, LA_ROBOSOC, true, &logFile1,
			true, &std::cerr);

  logStream->addLogArea(LA_ACTUATORS, LA_ROBOSOC, false, &logFile1);
  logStream->addLogArea(LA_CONTROLLER, LA_ROBOSOC, false, &logFile1);
  logStream->addLogArea(LA_SENSORS, LA_ROBOSOC, false, &logFile1);
  logStream->addLogArea(LA_SERVER_COMMANDS, LA_ROBOSOC, false, &logFile1);
  logStream->addLogArea(LA_SENSOR_DATA, LA_ROBOSOC, false, &logFile1);
  logStream->addLogArea(LA_STRATEGY, LA_ROBOSOC, false, &logFile1);
  logStream->addLogArea(LA_SENSOR_INTERFACE, LA_ROBOSOC, false, &logFile1);
  logStream->addLogArea(LA_DECISION, LA_ROBOSOC, true, &logFile1);
  logStream->addLogArea(LA_FRAMEWORK_AGENT, LA_ROBOSOC, false, &logFile1);
  logStream->addLogArea(LA_AGENT, LA_ROBOSOC, false, &logFile1);
  logStream->addLogArea(LA_SKILL, LA_ROBOSOC, true, &logFile1);
  logStream->addLogArea(LA_COACH, LA_ROBOSOC, true, &coachLog);
  logStream->addLogArea(LA_WORLDMODEL, LA_ROBOSOC, false, &logFile1);
  logStream->addLogArea(LA_TIME, LA_ROBOSOC, false, &logFile1);


  try {
    std::cerr << "Initializing the world...\n";
    WorldFacts::init(argc, argv);

    std::cerr << "Initializing the worldModelInterface...\n";
    WorldModelInterface *wmi = WorldModelInterface::nonConstInstance();

    std::cerr << "Setting worldModel update strategies...\n";
    wmi->setSeeStrategy(new BasicSeeStrategy());
    wmi->setHearStrategy(new BasicHearStrategy());
    wmi->setCommandStrategy(new BasicCommandStrategy());
    wmi->setInitStrategy(new BasicInitStrategy());
    wmi->setSenseStrategy(new BasicSenseStrategy());
    wmi->setCoachMessageStrategy(new BasicCoachMessageStrategy());
    wmi->setNewCycleStrategy(new BasicNewCycleStrategy());
    wmi->setCoachInformationStrategy(new BasicCoachInformationStrategy());

    std::cerr << "Starting decision...\n";
    
    Decision* decision = new TestTeamDecision();    

    std::cerr << "Initializing the FrameworkAgent...\n";
    FrameworkAgent agent(decision);

    std::cerr << "Starting the agent...\n";
    agent.start();
    
  } catch (RSException &excpt) {
    std::cerr << "Caught a RoboSoc exception: " << excpt.What() << std::endl;
    return 1;
  } catch (std::exception &excpt) {
    std::cerr << "Caught a standard exception: " << excpt.what() << std::endl;
    return 1;
  } catch (...) {
    std::cerr << "Caught an unknown exception: " << std::endl;
    return 1;
  }
  
  return 0;
 
}
