/* -*- Mode: C++ -*-
 * $Id: TestTeamDecision.cc,v 1.11 2003/09/08 07:47:23 aip19 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: TestTeamDecision.cc
 * Unitname: Framework/test/TestTeam
 * $Revision: 1.11 $
 * Created by: Fredrik Heintz 2000-09-16
 * Last modified by $Author: aip19 $ $Date: 2003/09/08 07:47:23 $
 *
 *
 * HISTORY
 *
 * $Log: TestTeamDecision.cc,v $
 * Revision 1.11  2003/09/08 07:47:23  aip19
 * Updated templates to use SearchBall and made the HOWTOS a bit clearer
 *
 * Revision 1.10  2003/09/02 18:01:57  frehe
 * Updated to the new command pointer semantics. RoboSoc now takes control of the pointer when sent to the actuator interface.
 *
 * Revision 1.9  2002/09/13 15:10:19  s02davro
 * *** empty log message ***
 *
 * Revision 1.8  2002/09/13 14:48:57  s02davro
 * FALSE -> FuzzyBool(0.0), TRUE -> FuzzyBool(1.0)
 *
 * Revision 1.7  2002/09/13 14:10:39  s02davro
 * Updated the TestTeam to use new version of RoboSoc
 *
 * Revision 1.6  2001/09/26 13:02:44  frehe
 * Changed the order of include files to make circular updates less likely.
 *
 * Revision 1.5  2001/09/02 11:33:12  frehe
 * Fixed some incorrect include files reported by Peter Andersson and Jonas Carlsson.
 *
 * Revision 1.4  2001/09/01 23:54:19  frehe
 * CP_version -> SP_version
 *
 * Revision 1.3  2001/09/01 15:06:52  frehe
 * Updated the file to work with the current version of RoboSoc.
 *
 * Revision 1.2  2000/09/16 18:39:41  frehe
 * Fixed the log messages.
 *
 * Revision 1.1  2000/09/16 18:37:09  frehe
 * Add a team that can be used as the basis for test teams.
 *
 */

/**
 * @file TeamDecision.cc
 * Implementation of the class TeamDecision.
 *
 * @par Unit:
 *      Team
 * $Revision: 1.11 $
 * $Author: aip19 $
 * $Date: 2003/09/08 07:47:23 $
 */

#include "TestTeamDecision.h"
#include "GameView.h"
#include "BallView.h"
#include "AgentView.h"
#include "PlayerView.h"
#include "ActuatorInterface.h"
#include "ViewManager.h"
#include "view_constants.h"
#include "InitCommand.h"
#include "MoveCommand.h"
#include "SayCommand.h"
#include "DashCommand.h"
#include "KickCommand.h"
#include "TurnCommand.h"
#include "TurnNeckCommand.h"
#include "rs_assert.h"
#include "rs_log.h"

RS_BEGIN_NAMESPACE


TestTeamDecision::TestTeamDecision(ActuatorInterface* const actuator_interface)
  : Decision(actuator_interface),
    viewManager(ViewManager::instance()),
    worldModelInterface(WorldModelInterface::instance()),
    facts(worldModelInterface->getWorldFacts()),
    gameView(NULL),
    ballView(NULL),
    agentView(NULL),
    playerView(NULL) {
  getView(gameView, GAME_VIEW_ID);
  getView(ballView, BALL_VIEW_ID);
  getView(agentView, AGENT_VIEW_ID);
  getView(playerView, PLAYER_VIEW_ID);
}

TestTeamDecision::~TestTeamDecision() {
  viewManager->releaseView(GAME_VIEW_ID);
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(PLAYER_VIEW_ID);
}

ServerCommand* TestTeamDecision::initAgent() {
  InitCommand* init_cmd = new InitCommand(facts->CP_team_name, 
					  facts->SP_version, 
					  facts->CP_goalie);
  return init_cmd;
}

bool TestTeamDecision::start() {
  RS_REQUIRE( initialized == true );
  return true;
}

bool TestTeamDecision::decide() {
  static Plan plan;
  
  // Example of a simple decision maker using skills.
  // If the ball's position is known, the agent will run
  // To that position, and try to kick the ball towards the 
  // opponents goal.
  
  if (searchBall.applicable()) {
    RS_LOG ( LA_DECISION, "SearchBall" );
    plan = searchBall.generatePlan();
    actuatorInterface->addToPlan(plan.getNextStep());     
  }
  else if (gotoBall.applicable()) {
    RS_LOG ( LA_DECISION, "GotoBall" );
    plan = gotoBall.generatePlan();
    actuatorInterface->addToPlan(plan.getNextStep());     
  }
  else {

    // Example of more primitive decision making without using skills.
    ServerCommand* scmd = 0;
 
    if ( gameView->isKickAllowed() && (ballView->getRelativeDistance() 
				       <= facts->SP_kickable_margin)) {
      // The ball is kickable, can we kick it to the ball.      
      // Get the direction to the opponents goal.
      AngleDegUE goal_dir = agentView->getPosition().getDirectionTo
	(facts->getTheirGoal().getPosition());
      // Kick the ball towards the goal.
      scmd = new KickCommand(50,(goal_dir - 
				 agentView->getBodyDirection()).getAngle());
    }
    else {
      // The game has not begun, do nothing.
      scmd = new TurnCommand(0);
    }
    if ( scmd != 0 ) {
      // Send the command to the actuators.
      actuatorInterface->addToPlan(scmd);
    }
  }    

  return true;
}

void TestTeamDecision::onAuralCoachSensorData() {
}

void TestTeamDecision::onAuralRefereeSensorData() {
}

void TestTeamDecision::onAuralPlayerSensorData() {
}

void TestTeamDecision::onAuralSelfSensorData() {
}

void TestTeamDecision::onInit() {
  // Position the players with some distance at the beginning of game.
  Point movePoint = Point(-10,-10);
  if (agentView->getUniformNumber().isKnown()) {
    UInt nr = agentView->getUniformNumber().getValue();
    switch (nr) {      
    case 1:
      movePoint = Point(-10, 0);
      break;
    case 2:
      movePoint = Point(-5, -15);
      break;
    case 3:
      movePoint = Point(-5, 15);
      break;
    case 4:
      movePoint = Point(-25, 10);
      break;
    case 5:
      movePoint = Point(-25, -10);
      break;
    case 6:
      movePoint = Point(-25, 20);
      break;
    case 7:
      movePoint = Point(-25, -20);
      break;
    case 8:
      movePoint = Point(-45, 10);
      break;
    case 9:
      movePoint = Point(-25, -10);
      break;
    case 10:
      movePoint = Point(-25, 20);
      break;
    case 11:
      movePoint = Point(-25, -20);
      break;
    default:
      movePoint = Point(-10,-10);
    }
  }
  MoveCommand* move_cmd = new MoveCommand(movePoint);
  actuatorInterface->addToPlan(move_cmd);
}

void TestTeamDecision::onCommandWarning() {
  decide();
}

RS_END_NAMESPACE
