/* -*- Mode: C++ -*-
 * $Id: AttackerDefenderGoalieScoreScenario.cc,v 1.2 2005/08/18 20:22:35 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: AttackerDefenderGoalieScoreScenario.cc
 * Unitname: Framework/trainer
 * $Revision: 1.2 $
 * Created by: Fredrik Heintz 2004-09-03
 * Last modified by $Author: frehe $ $Date: 2005/08/18 20:22:35 $
 *
 *
 * HISTORY
 *
 * $Log: AttackerDefenderGoalieScoreScenario.cc,v $
 * Revision 1.2  2005/08/18 20:22:35  frehe
 * Added a min level to remove certain test cases.
 *
 * Revision 1.1  2005/08/13 13:36:57  frehe
 * Readded the files since the trainer seems to have been deleted from the CVS.
 *
 * Revision 1.7  2004/09/17 13:34:42  frehe
 * Fixed spelling misstake.
 *
 * Revision 1.6  2004/09/13 19:41:15  frehe
 * Fixed and to or to remove player placement bug.
 *
 * Revision 1.5  2004/09/12 23:28:17  frehe
 * Improved the player and ball positioning.
 *
 * Revision 1.4  2004/09/08 22:29:58  frehe
 * Updated to the new CoachSayCommand.
 *
 * Revision 1.3  2004/09/08 18:36:13  frehe
 * Say the type of the scenario.
 *
 * Revision 1.2  2004/09/07 17:28:16  frehe
 * Changed the design to allow reading of test cases from file.
 *
 * Revision 1.1  2004/09/03 09:44:15  frehe
 * Added new score scenario with both goalie and defenders.
 *
 */

/**
 * @file AttackerDefenderGoalieScoreScenario.cc
 * Implementation file for the class AttackerDefenderGoalieScoreScenario.
 *
 * @par Unit:
 *      Framework/trainer
 * $Revision: 1.2 $
 * $Author: frehe $
 * $Date: 2005/08/18 20:22:35 $
 */


#include "AttackerDefenderGoalieScoreScenario.h"
#include "ActuatorInterface.h"
#include "ViewManager.h"
#include "WorldModelInterface.h"
#include "WorldFacts.h"
#include "view_constants.h"
#include "PlayerView.h"
#include "TrainerView.h"
#include "GameView.h"
#include "TrainerStartCommand.h"
#include "ChangeModeCommand.h"
#include "CoachSayCommand.h"
#include "TrainerRecoverCommand.h"
#include "TrainerMoveObjectCommand.h"
#include "string_utils.h"
#include "trainer_helpers.h"


RS_BEGIN_NAMESPACE


AttackerDefenderGoalieScoreScenario::AttackerDefenderGoalieScoreScenario(int random_seed,
									 int gmode,
									 const std::vector<int>& modes,
									 int cycles)
  : Scenario(modes.size()+1),
    mViewManager(ViewManager::instance()),
    mWorldModelInterface(WorldModelInterface::instance()),
    mFacts(mWorldModelInterface->getWorldFacts()),
    mPlayerView(0),
    mTrainerView(0),
    mGameView(0),
    mGoalieMode(gmode),
    mDefenderModes(modes),
    mRunning(false),
    mCycles(cycles)
{
  getView(mPlayerView, PLAYER_VIEW_ID);
  getView(mTrainerView, TRAINER_VIEW_ID);
  getView(mGameView, GAME_VIEW_ID);

  // Initialize the random number generator
  init_random_number_generator(random_seed);
}


AttackerDefenderGoalieScoreScenario::~AttackerDefenderGoalieScoreScenario()
{
  mViewManager->releaseView(PLAYER_VIEW_ID);
  mViewManager->releaseView(TRAINER_VIEW_ID);
  mViewManager->releaseView(GAME_VIEW_ID);
}


Plan
AttackerDefenderGoalieScoreScenario::setup()
{
  Step step;

  if ( mPlayerView->getTeammates().size() == 0 ) {
    std::cout << "Not enough players on the left team!" << std::endl;
    RS_LOG( LA_DEBUG, "AttackerDefenderGoalieScoreScenario: Not enough players on the left team" );

    return Plan();
  }
  const std::vector<PlayerObject>& teammates = mPlayerView->getTeammates();
  for ( std::vector<PlayerObject>::const_iterator i = teammates.begin();
	i != teammates.end(); ++i ) {
    if ( i->getUniformNumber().isKnown()
	 && i->getUniformNumber().getValue() > 1 ) {
      step.addCommand(new TrainerMoveObjectCommand(mTrainerView->getLeftTeamName(), i->getUniformNumber().getValue(), 0, 0, 0, 0, 0));
    }
  }

  if ( mPlayerView->getOpponents().size() <= mDefenderModes.size() ) {
    std::cout << "Not enough players on the right team!" << std::endl;
    RS_LOG( LA_DEBUG, "AttackerDefenderGoalieScoreScenario: Not enough players on the right team" );

    return Plan();
  }
  const std::vector<PlayerObject>& opps = mPlayerView->getOpponents();
  for ( std::vector<PlayerObject>::const_iterator i = opps.begin();
	i != opps.end(); ++i ) {
    if ( i->getUniformNumber().isKnown()
	 && i->getUniformNumber().getValue() > (mDefenderModes.size()+1) ) {
      step.addCommand(new TrainerMoveObjectCommand(mTrainerView->getRightTeamName(), i->getUniformNumber().getValue(), 0, 0, 0, 0, 0));
    }
  }

  // Add check to make its a goalie on the right team!


  // Get the current score
  mStartScore = mTrainerView->getLeftScore();

  // Get the current time
  mStartTime = Time::currentTime.getT();

  // The player should be positioned in the penalty area
  Point player_pos = generate_point_in_penalty_area(S_RIGHT);

  // Get the direction to the goal
  Vector goal_dir(player_pos, mFacts->getTheirGoal().getPosition());

  // Make sure the angle is not too large and the distance is ok
  while ( abs(goal_dir.getAngle()) > 75
	  || goal_dir.getLength() > 15
	  || goal_dir.getLength() < 4 ) {
    player_pos = generate_point_in_penalty_area(S_RIGHT);
    goal_dir = Vector(player_pos, mFacts->getTheirGoal().getPosition());
  }
  
  AngleDeg min_angle = goal_dir.getAngle() - 45;
  AngleDeg max_angle = goal_dir.getAngle() + 45;

  // The player should be facing the goal plus/minus 45 degrees.
  AngleDeg player_dir = generate_direction_in_sector(min_angle, max_angle);

  // Make sure the ball is kickable and between the player and the goal
  Point ball_pos = player_pos + generate_kickable_ball_vector(min_angle,
							      max_angle);
  //Point ball_pos = player_pos + generate_vector_in_circle_segment(0, 0, 0.5);

  // Generate a good position for the goalie
  Point goalie_pos = generate_point_on_goal_line(S_RIGHT, mFacts->getMarker(GOAL_R).getPosition(), 15);
  //Point goalie_pos = mFacts->getTheirGoal().getPosition();
  // Move the goalie one meter from the goal line
  goalie_pos.addX(-0.5);

  
  /*
    std::cout << "Setup AttackerDefenderGoalieScoreScenario\n   player_pos = "
    << player_pos << "; ball_pos = " << ball_pos
    << "\n goalie_pos = " << goalie_pos << std::endl;
  */

  std::string goalie_mode = "GMode" + toString(mGoalieMode);

  // Send the setup commands
  step.addCommand(new TrainerRecoverCommand());
  step.addCommand(new TrainerStartCommand());
  step.addCommand(new CoachSayCommand(goalie_mode));
  step.addCommand(new ChangeModeCommand(PM_PLAY_ON));
  step.addCommand(new CoachSayCommand("score"));

  step.addCommand(new TrainerMoveObjectCommand(ball_pos));
  step.addCommand(new TrainerMoveObjectCommand(mTrainerView->getRightTeamName(), 1, goalie_pos, 180, Point(0, 0)));

  int defender = 2;
  for ( std::vector<int>::const_iterator i = mDefenderModes.begin();
	i != mDefenderModes.end(); ++i ) {
    Point defender_pos = generate_point_in_defense_sector(S_RIGHT,
							  player_pos);
    std::string dmode = "DMode" + toString(*i) + " " + toString(defender);
    //std::cout << "dmode = " << dmode << std::endl;
    step.addCommand(new CoachSayCommand(dmode));
    step.addCommand(new TrainerMoveObjectCommand(mTrainerView->getRightTeamName(), defender, defender_pos, 180, Point(0, 0)));
    ++defender;
  }

  step.addCommand(new TrainerMoveObjectCommand(mTrainerView->getLeftTeamName(), 1, player_pos, player_dir.getDeg(), Point(0, 0)));


  mRunning = true;

  return Plan(step);
}


bool
AttackerDefenderGoalieScoreScenario::running()
{
  return mRunning;
}


bool
AttackerDefenderGoalieScoreScenario::success()
{
  // If scored after the scenario started it succeeded
  if ( mRunning && mTrainerView->getLeftScore() > mStartScore ) {
    mRunning = false;
    return true;
  }
  return false;
}


bool
AttackerDefenderGoalieScoreScenario::failure()
{
  // If more than mCycles cycles since the scenario started then fail
  // or the play mode is not play on.
  if ( mRunning
       && ( Time::currentTime.getT() >= (mStartTime+mCycles)
	    || mGameView->getPlayMode() != PM_PLAY_ON ) ) {
    mRunning = false;
    return true;
  }
  return false;
}


RS_END_NAMESPACE
