/* -*- Mode: C++ -*-
 * $Id: AttackerDefendersDribbleScenario.cc,v 1.3 2005/08/18 20:45:56 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: AttackerDefendersDribbleScenario.cc
 * Unitname: Framework/trainer
 * $Revision: 1.3 $
 * Created by: Fredrik Heintz 2004-09-07
 * Last modified by $Author: frehe $ $Date: 2005/08/18 20:45:56 $
 *
 *
 * HISTORY
 *
 * $Log: AttackerDefendersDribbleScenario.cc,v $
 * Revision 1.3  2005/08/18 20:45:56  frehe
 * Added a max y dist parameter. Now 10m, maybe it should be lower.
 *
 * Revision 1.2  2005/08/18 20:22:35  frehe
 * Added a min level to remove certain test cases.
 *
 * Revision 1.1  2005/08/13 13:36:57  frehe
 * Readded the files since the trainer seems to have been deleted from the CVS.
 *
 * Revision 1.6  2004/09/17 13:34:42  frehe
 * Fixed spelling misstake.
 *
 * Revision 1.5  2004/09/16 11:38:50  frehe
 * Moved the test cases to the -40 line.
 *
 * Revision 1.4  2004/09/13 14:31:58  frehe
 * Fixed the message sent.
 *
 * Revision 1.3  2004/09/08 22:29:58  frehe
 * Updated to the new CoachSayCommand.
 *
 * Revision 1.2  2004/09/08 18:36:25  frehe
 * Say the type of the scenario.
 *
 * Revision 1.1  2004/09/07 21:25:36  frehe
 * Added a scenario for testing dribble.
 *
 */

/**
 * @file AttackerDefendersDribbleScenario.cc
 * Implementation file for the class AttackerDefendersDribbleScenario.
 *
 * @par Unit:
 *      Framework/trainer
 * $Revision: 1.3 $
 * $Author: frehe $
 * $Date: 2005/08/18 20:45:56 $
 */


#include "AttackerDefendersDribbleScenario.h"
#include "ViewManager.h"
#include "WorldModelInterface.h"
#include "WorldFacts.h"
#include "view_constants.h"
#include "PlayerView.h"
#include "TrainerView.h"
#include "GameView.h"
#include "BallView.h"
#include "TrainerStartCommand.h"
#include "ChangeModeCommand.h"
#include "CoachSayCommand.h"
#include "TrainerRecoverCommand.h"
#include "TrainerMoveObjectCommand.h"
#include "string_utils.h"
#include "trainer_helpers.h"


RS_BEGIN_NAMESPACE


AttackerDefendersDribbleScenario::AttackerDefendersDribbleScenario(int random_seed,
								   const std::vector<int>& modes,
								   int dribble_dist,
								   int min_defender_dist,
								   int max_defender_dist,
								   int max_y_dist,
								   int cycles)
  : Scenario(modes.size()),
    mViewManager(ViewManager::instance()),
    mWorldModelInterface(WorldModelInterface::instance()),
    mFacts(mWorldModelInterface->getWorldFacts()),
    mPlayerView(0),
    mTrainerView(0),
    mGameView(0),
    mBallView(0),
    mDefendersModes(modes),
    m_dribble_dist(dribble_dist),
    m_min_defender_dist(min_defender_dist),
    m_max_defender_dist(max_defender_dist),
    m_max_y_dist(max_y_dist),
    mRunning(false),
    mCycles(cycles)
{
  getView(mPlayerView, PLAYER_VIEW_ID);
  getView(mTrainerView, TRAINER_VIEW_ID);
  getView(mGameView, GAME_VIEW_ID);
  getView(mBallView, BALL_VIEW_ID);

  // Initialize the random number generator
  init_random_number_generator(random_seed);
}


AttackerDefendersDribbleScenario::~AttackerDefendersDribbleScenario()
{
  mViewManager->releaseView(PLAYER_VIEW_ID);
  mViewManager->releaseView(TRAINER_VIEW_ID);
  mViewManager->releaseView(GAME_VIEW_ID);
  mViewManager->releaseView(BALL_VIEW_ID);
}


Plan
AttackerDefendersDribbleScenario::setup()
{
  Step step;

  if ( mPlayerView->getTeammates().size() == 0 ) {
    std::cout << "Not enough players on the left team!" << std::endl;
    RS_LOG( LA_DEBUG, "AttackerDefendersDribbleScenario: Not enough players on the left team" );

    return Plan();
  }
  const std::vector<PlayerObject>& teammates = mPlayerView->getTeammates();
  for ( std::vector<PlayerObject>::const_iterator i = teammates.begin();
	i != teammates.end(); ++i ) {
    if ( i->getUniformNumber().isKnown()
	 && i->getUniformNumber().getValue() > 1 ) {
      step.addCommand(new TrainerMoveObjectCommand(mTrainerView->getLeftTeamName(), i->getUniformNumber().getValue(), 0, 0, 0, 0, 0));
    }
  }

  if ( mPlayerView->getOpponents().size() < mDefendersModes.size() ) {
    std::cout << "Not enough players on the right team!" << std::endl;
    RS_LOG( LA_DEBUG, "AttackerDefendersDribbleScenario: Not enough players on the right team" );

    return Plan();
  }
  const std::vector<PlayerObject>& opps = mPlayerView->getOpponents();
  for ( std::vector<PlayerObject>::const_iterator i = opps.begin();
	i != opps.end(); ++i ) {
    if ( i->getUniformNumber().isKnown()
	 && i->getUniformNumber().getValue() > mDefendersModes.size() ) {
      step.addCommand(new TrainerMoveObjectCommand(mTrainerView->getRightTeamName(), i->getUniformNumber().getValue(), 0, 0, 0, 0, 0));
    }
  }

  // Get the current time
  mStartTime = Time::currentTime.getT();

  // Send the setup commands
  Point player_pos = generate_point_on_line(Point(-40,-5), Point(-40,5));
  AngleDeg player_dir = generate_direction_in_sector(0, 0);
  Point ball_pos = player_pos + generate_kickable_ball_vector(-45, 45);
  m_start_x = ball_pos.getX();
  m_start_y = ball_pos.getY();

  /*
    std::cout << "Setup AttackerDefendersDribbleScenario\n   player_pos = "
    << player_pos << "; ball_pos = " << ball_pos
    << "\n _pos = " << _pos << std::endl;
  */

  step.addCommand(new TrainerRecoverCommand());
  step.addCommand(new TrainerStartCommand());
  step.addCommand(new ChangeModeCommand(PM_PLAY_ON));
  step.addCommand(new CoachSayCommand("dribble " + toString(m_dribble_dist)));

  step.addCommand(new TrainerMoveObjectCommand(ball_pos));

  int defenders = 1;
  Point upper_left(m_start_x+m_min_defender_dist, player_pos.getY()-2);
  Point lower_right(m_start_x+m_max_defender_dist, player_pos.getY()+2);
  for ( std::vector<int>::const_iterator i = mDefendersModes.begin();
	i != mDefendersModes.end(); ++i ) {
    Point defenders_pos = generate_point_in_area(upper_left, lower_right);
    std::string dmode = "DMode" + toString(*i) + " " + toString(defenders);
    step.addCommand(new CoachSayCommand(dmode));
    step.addCommand(new TrainerMoveObjectCommand(mTrainerView->getRightTeamName(),
						 defenders, defenders_pos,
						 180, Point(0, 0)));
    ++defenders;
  }

  m_player_unum = 1;
  step.addCommand(new TrainerMoveObjectCommand(mTrainerView->getLeftTeamName(),
					       m_player_unum,
					       player_pos,
					       player_dir.getDeg(),
					       Point(0, 0)));


  mRunning = true;

  return Plan(step);
}


bool
AttackerDefendersDribbleScenario::running()
{
  return mRunning;
}


bool
AttackerDefendersDribbleScenario::success()
{
  // If the ball and the player has passed the line and the ball
  // is still kickable then the scenario has succeeded.
  const PlayerObject& player = find_player();
  const BallObject& ball = mBallView->getBallObject();
  FloatUE ball_dist = player.getPosition().getDistanceTo(ball.getPosition());
  if ( mRunning
       && ( player.getPosX() > (m_start_x + m_dribble_dist) )
       && ( ball.getPosition().getX() > (m_start_x + m_dribble_dist) )
       && ball_dist.isKnown()
       && ( ball_dist.getValue() < mFacts->SP_kickable_area ) ) {
    mRunning = false;
    return true;
  }
  return false;
}


bool
AttackerDefendersDribbleScenario::failure()
{
  // If more than mCycles cycles since the scenario started then fail
  // or the play mode is not play on or the ball is more than m_max_y_dist
  // meters above or below the starting y-position.
  if ( mRunning
       && ( Time::currentTime.getT() >= (mStartTime+mCycles)
	    || mGameView->getPlayMode() != PM_PLAY_ON
	    || abs(mBallView->getBallObject().getPosition().getY() - m_start_y) > m_max_y_dist ) ) {
    mRunning = false;
    return true;
  }
  return false;
}


const PlayerObject&
AttackerDefendersDribbleScenario::find_player() const
{
  const std::vector<PlayerObject>& players = mPlayerView->getTeammates();
  for ( std::vector<PlayerObject>::const_iterator i = players.begin();
	i != players.end(); ++i ) {
    if ( i->getUniformNumber() == m_player_unum ) {
      return *i;
    }
  }
}


RS_END_NAMESPACE
