/* -*- Mode: C++ -*-
 * $Id: AttackersDefendersPassScenario.cc,v 1.5 2005/09/12 19:27:46 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: AttackersDefendersPassScenario.cc
 * Unitname: Framework/trainer
 * $Revision: 1.5 $
 * Created by: Fredrik Heintz 2004-09-07
 * Last modified by $Author: frehe $ $Date: 2005/09/12 19:27:46 $
 *
 *
 * HISTORY
 *
 * $Log: AttackersDefendersPassScenario.cc,v $
 * Revision 1.5  2005/09/12 19:27:46  frehe
 * Fixed receiver positioning.
 *
 * Revision 1.4  2005/09/12 10:45:17  frehe
 * Fixed bug
 *
 * Revision 1.3  2005/08/18 21:06:50  frehe
 * Added a max y dist parameter. Now 5m, maybe it should be dependent on the actual position of the receiver not the possible positions.
 *
 * Revision 1.2  2005/08/18 20:22:35  frehe
 * Added a min level to remove certain test cases.
 *
 * Revision 1.1  2005/08/13 13:36:58  frehe
 * Readded the files since the trainer seems to have been deleted from the CVS.
 *
 * Revision 1.7  2004/09/17 13:34:42  frehe
 * Fixed spelling misstake.
 *
 * Revision 1.6  2004/09/13 15:34:35  frehe
 * Moved the receivers a meter from the min pass distance line.
 *
 * Revision 1.5  2004/09/13 14:56:24  frehe
 * Turned the defenders in the right direction.
 *
 * Revision 1.4  2004/09/13 14:31:59  frehe
 * Fixed the message sent.
 *
 * Revision 1.3  2004/09/08 22:29:59  frehe
 * Updated to the new CoachSayCommand.
 *
 * Revision 1.2  2004/09/08 18:37:17  frehe
 * Say the type of the scenario.
 *
 * Revision 1.1  2004/09/07 23:05:36  frehe
 * Added a scenario for testing passing.
 *
 */

/**
 * @file AttackersDefendersPassScenario.cc
 * Implementation file for the class AttackersDefendersPassScenario.
 *
 * @par Unit:
 *      Framework/trainer
 * $Revision: 1.5 $
 * $Author: frehe $
 * $Date: 2005/09/12 19:27:46 $
 */


#include "AttackersDefendersPassScenario.h"
#include "ViewManager.h"
#include "WorldModelInterface.h"
#include "WorldFacts.h"
#include "view_constants.h"
#include "PlayerView.h"
#include "TrainerView.h"
#include "GameView.h"
#include "BallView.h"
#include "TrainerStartCommand.h"
#include "ChangeModeCommand.h"
#include "CoachSayCommand.h"
#include "TrainerRecoverCommand.h"
#include "TrainerMoveObjectCommand.h"
#include "string_utils.h"
#include "trainer_helpers.h"


RS_BEGIN_NAMESPACE


AttackersDefendersPassScenario::AttackersDefendersPassScenario(int random_seed,
							       int attackers,
							       const std::vector<int>& modes,
							       int min_pass_dist,
							       int max_pass_dist,
							       int min_y_dist,
							       int max_y_dist,
							       int min_defender_dist,
							       int max_defender_dist,
							       int max_y_diff,
							       int cycles)
  : Scenario(modes.size()),
    mViewManager(ViewManager::instance()),
    mWorldModelInterface(WorldModelInterface::instance()),
    mFacts(mWorldModelInterface->getWorldFacts()),
    mPlayerView(0),
    mTrainerView(0),
    mGameView(0),
    mBallView(0),
    mDefendersModes(modes),
    m_attackers(attackers),
    m_min_pass_dist(min_pass_dist),
    m_max_pass_dist(max_pass_dist),
    m_min_y_dist(min_y_dist),
    m_max_y_dist(max_y_dist),
    m_min_defender_dist(min_defender_dist),
    m_max_defender_dist(max_defender_dist),
    m_max_y_diff(max_y_diff),
    mRunning(false),
    mCycles(cycles)
{
  getView(mPlayerView, PLAYER_VIEW_ID);
  getView(mTrainerView, TRAINER_VIEW_ID);
  getView(mGameView, GAME_VIEW_ID);
  getView(mBallView, BALL_VIEW_ID);

  // Initialize the random number generator
  init_random_number_generator(random_seed);
}


AttackersDefendersPassScenario::~AttackersDefendersPassScenario()
{
  mViewManager->releaseView(PLAYER_VIEW_ID);
  mViewManager->releaseView(TRAINER_VIEW_ID);
  mViewManager->releaseView(GAME_VIEW_ID);
  mViewManager->releaseView(BALL_VIEW_ID);
}


Plan
AttackersDefendersPassScenario::setup()
{
  Step step;

  if ( mPlayerView->getTeammates().size() < m_attackers ) {
    std::cout << "Not enough players on the left team!" << std::endl;
    RS_LOG( LA_DEBUG, "AttackersDefendersPassScenario: Not enough players on the left team" );

    return Plan();
  }
  const std::vector<PlayerObject>& teammates = mPlayerView->getTeammates();
  for ( std::vector<PlayerObject>::const_iterator i = teammates.begin();
	i != teammates.end(); ++i ) {
    if ( i->getUniformNumber().isKnown()
	 && i->getUniformNumber().getValue() > m_attackers ) {
      step.addCommand(new TrainerMoveObjectCommand(mTrainerView->getLeftTeamName(), i->getUniformNumber().getValue(), 0, 0, 0, 0, 0));
    }
  }

  if ( mPlayerView->getOpponents().size() < mDefendersModes.size() ) {
    std::cout << "Not enough players on the right team!" << std::endl;
    RS_LOG( LA_DEBUG, "AttackersDefendersPassScenario: Not enough players on the right team" );

    return Plan();
  }
  const std::vector<PlayerObject>& opps = mPlayerView->getOpponents();
  for ( std::vector<PlayerObject>::const_iterator i = opps.begin();
	i != opps.end(); ++i ) {
    if ( i->getUniformNumber().isKnown()
	 && i->getUniformNumber().getValue() > mDefendersModes.size() ) {
      step.addCommand(new TrainerMoveObjectCommand(mTrainerView->getRightTeamName(), i->getUniformNumber().getValue(), 0, 0, 0, 0, 0));
    }
  }

  // Get the current time
  mStartTime = Time::currentTime.getT();

  // Send the setup commands
  Point passer_pos = generate_point_on_line(Point(-40,-5), Point(-40,5));
  AngleDeg passer_dir = generate_direction_in_sector(0, 0);
  Point ball_pos = passer_pos + generate_kickable_ball_vector(-45, 45);
  m_start_x = ball_pos.getX();
  m_start_y = ball_pos.getY();

  /*
    std::cout << "Setup AttackersDefendersPassScenario\n   passer_pos = "
    << passer_pos << "; ball_pos = " << ball_pos
    << "\n _pos = " << _pos << std::endl;
  */

  step.addCommand(new TrainerRecoverCommand());
  step.addCommand(new TrainerStartCommand());
  step.addCommand(new ChangeModeCommand(PM_PLAY_ON));
  step.addCommand(new CoachSayCommand("pass " + toString(m_min_pass_dist)
				      + " " + toString(m_max_pass_dist)));

  step.addCommand(new TrainerMoveObjectCommand(ball_pos));

  int defenders = 1;
  Point def_upper_left(m_start_x+m_min_defender_dist,
		       m_start_y-m_min_y_dist);
  Point def_lower_right(m_start_x+m_max_defender_dist,
			m_start_y+m_max_y_dist);
  for ( std::vector<int>::const_iterator i = mDefendersModes.begin();
	i != mDefendersModes.end(); ++i ) {
    Point defender_pos = generate_point_in_area(def_upper_left,
						def_lower_right);
    // The defender should be turned towards the ball
    //AngleDeg ball_dir = defender_pos.getDirectioTo(ball_pos);
    std::string dmode = "DMode" + toString(*i) + " " + toString(defenders);
    step.addCommand(new CoachSayCommand(dmode));
    step.addCommand(new TrainerMoveObjectCommand(mTrainerView->getRightTeamName(),
						 defenders, defender_pos,
						 //ball_dir.getDeg(),
						 180, Point(0, 0)));
    ++defenders;
  }

  assert( m_max_pass_dist - m_min_pass_dist > 5 );
  Point att_upper_left(m_start_x+m_min_pass_dist+5,
		       m_start_y-m_min_y_dist);
  Point att_lower_right(m_start_x+m_max_pass_dist-5,
			m_start_y+m_max_y_dist);
  for ( int attackers = 2; attackers <= m_attackers; ++attackers ) {
    Point attacker_pos = generate_point_in_area(att_upper_left,
						att_lower_right);
    step.addCommand(new TrainerMoveObjectCommand(mTrainerView->getLeftTeamName(),
						 attackers, attacker_pos,
						 180, Point(0, 0)));
  }

  m_player_unum = 1;
  step.addCommand(new TrainerMoveObjectCommand(mTrainerView->getLeftTeamName(),
					       m_player_unum,
					       passer_pos,
					       passer_dir.getDeg(),
					       Point(0, 0)));


  mRunning = true;

  return Plan(step);
}


bool
AttackersDefendersPassScenario::running()
{
  return mRunning;
}


bool
AttackersDefendersPassScenario::success()
{
  // If the ball has passed the pass line and the ball is kickable
  // by an attacker which is also on the right side of the line
  // then the scenario has succeeded.
  const BallObject& ball = mBallView->getBallObject();
  if ( mRunning
       && ball.getPosition().getX() > (m_start_x + m_min_pass_dist) ) {
    //std::cout << "ball on the right side of pass line" << std::endl;
    const std::vector<PlayerObject>& players = mPlayerView->getTeammates();
    for ( std::vector<PlayerObject>::const_iterator i = players.begin();
	  i != players.end(); ++i ) {
      if ( i->getUniformNumber() != m_player_unum
	   && i->getPosX() > (m_start_x + m_min_pass_dist) ) {
	FloatUE ball_dist = i->getPosition().getDistanceTo(ball.getPosition());
	if ( ball_dist.isKnown()
	     && ( ball_dist.getValue() < mFacts->SP_kickable_area ) ) {
	  //std::cout << "success!" << std::endl;
	  mRunning = false;
	  return true;
	  /*
	    } else {
	    std::cout << "ball is not kickable by "
	    << i->getUniformNumber() << std::endl;
	  */
	}
      }
    }
  }
  return false;
}


bool
AttackersDefendersPassScenario::failure()
{
  // If more than mCycles cycles since the scenario started then fail
  // or the play mode is not play on or the ball is too far off in the
  // y direction.
  const BallObject& ball = mBallView->getBallObject();
  if ( mRunning
       && ( Time::currentTime.getT() >= (mStartTime+mCycles)
	    || mGameView->getPlayMode() != PM_PLAY_ON
	    || ball.getPosition().getY() < m_start_y-m_min_y_dist-m_max_y_diff
	    || ball.getPosition().getY() > m_start_y+m_max_y_dist+m_max_y_diff ) ) {
    /*
      if ( mGameView->getPlayMode() != PM_PLAY_ON ) {
      std::cout << "play_mode = " << mGameView->getPlayMode() << std::endl;
      } else {
      std::cout << "timeout" << std::endl;
      }
    */
    mRunning = false;
    return true;
  }
  return false;
}


RS_END_NAMESPACE
