/* -*- Mode: C++ -*-
 * $Id: OnlyAttackerScoreScenario.cc,v 1.2 2005/08/18 20:22:35 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: OnlyAttackerScoreScenario.cc
 * Unitname: Framework/trainer
 * $Revision: 1.2 $
 * Created by: Fredrik Heintz 2004-08-31
 * Last modified by $Author: frehe $ $Date: 2005/08/18 20:22:35 $
 *
 *
 * HISTORY
 *
 * $Log: OnlyAttackerScoreScenario.cc,v $
 * Revision 1.2  2005/08/18 20:22:35  frehe
 * Added a min level to remove certain test cases.
 *
 * Revision 1.1  2005/08/13 13:36:58  frehe
 * Readded the files since the trainer seems to have been deleted from the CVS.
 *
 * Revision 1.10  2004/09/17 13:34:42  frehe
 * Fixed spelling misstake.
 *
 * Revision 1.9  2004/09/13 19:41:16  frehe
 * Fixed and to or to remove player placement bug.
 *
 * Revision 1.8  2004/09/12 23:28:17  frehe
 * Improved the player and ball positioning.
 *
 * Revision 1.7  2004/09/08 22:29:59  frehe
 * Updated to the new CoachSayCommand.
 *
 * Revision 1.6  2004/09/08 18:35:56  frehe
 * Say the type of the scenario.
 *
 * Revision 1.5  2004/09/07 17:28:16  frehe
 * Changed the design to allow reading of test cases from file.
 *
 * Revision 1.4  2004/09/02 21:48:59  frehe
 * Move players not supposed to be involved in the scenario.
 *
 * Revision 1.3  2004/09/01 09:03:21  frehe
 * Parameterized the number of cycles before failure and check the play mode in failure.
 *
 * Revision 1.2  2004/08/31 23:05:16  frehe
 * Update to using the trainer_helpers.
 *
 * Revision 1.1  2004/08/31 17:12:10  frehe
 * Added a scenario with only an attacker.
 *
 */

/**
 * @file OnlyAttackerScoreScenario.cc
 * Implementation file for the class OnlyAttackerScoreScenario.
 *
 * @par Unit:
 *      Framework/trainer
 * $Revision: 1.2 $
 * $Author: frehe $
 * $Date: 2005/08/18 20:22:35 $
 */


#include "OnlyAttackerScoreScenario.h"
#include "ViewManager.h"
#include "WorldModelInterface.h"
#include "WorldFacts.h"
#include "view_constants.h"
#include "PlayerView.h"
#include "TrainerView.h"
#include "GameView.h"
#include "TrainerStartCommand.h"
#include "ChangeModeCommand.h"
#include "CoachSayCommand.h"
#include "TrainerRecoverCommand.h"
#include "TrainerMoveObjectCommand.h"
#include "trainer_helpers.h"


RS_BEGIN_NAMESPACE


OnlyAttackerScoreScenario::OnlyAttackerScoreScenario(int random_seed,
						     int cycles)
  : Scenario(0),
    mViewManager(ViewManager::instance()),
    mWorldModelInterface(WorldModelInterface::instance()),
    mFacts(mWorldModelInterface->getWorldFacts()),
    mPlayerView(0),
    mTrainerView(0),
    mGameView(0),
    mRunning(false),
    mCycles(cycles)
{
  getView(mPlayerView, PLAYER_VIEW_ID);
  getView(mTrainerView, TRAINER_VIEW_ID);
  getView(mGameView, GAME_VIEW_ID);

  // Initialize the random number generator
  init_random_number_generator(random_seed);
}


OnlyAttackerScoreScenario::~OnlyAttackerScoreScenario()
{
  mViewManager->releaseView(PLAYER_VIEW_ID);
  mViewManager->releaseView(TRAINER_VIEW_ID);
  mViewManager->releaseView(GAME_VIEW_ID);
}


Plan
OnlyAttackerScoreScenario::setup()
{
  Step step;
  
  if ( mPlayerView->getTeammates().size() == 0 ) {
    std::cout << "Not enough players on the left team!" << std::endl;
    return Plan();
  }
  const std::vector<PlayerObject>& teammates = mPlayerView->getTeammates();
  for ( std::vector<PlayerObject>::const_iterator i = teammates.begin();
	i != teammates.end(); ++i ) {
    if ( i->getUniformNumber().isKnown()
	 && i->getUniformNumber().getValue() > 1 ) {
      step.addCommand(new TrainerMoveObjectCommand(mTrainerView->getLeftTeamName(), i->getUniformNumber().getValue(), 0, 0, 0, 0, 0));
    }
  }

  const std::vector<PlayerObject>& opps = mPlayerView->getOpponents();
  for ( std::vector<PlayerObject>::const_iterator i = opps.begin();
	i != opps.end(); ++i ) {
    if ( i->getUniformNumber().isKnown() ) {
      step.addCommand(new TrainerMoveObjectCommand(mTrainerView->getRightTeamName(), i->getUniformNumber().getValue(), 0, 0, 0, 0, 0));
    }
  }

  /*
  int left_start_x = mFacts->SP_pitch_length/2;
  int max_x_diff = mFacts->SP_penalty_area_length;
  int left_start_y = mFacts->SP_penalty_area_width/2;
  int max_y_diff = mFacts->SP_penalty_area_width;
  int ball_pos_x = left_start_x - max_x_diff*drand48();
  int ball_pos_y = left_start_y - max_y_diff*drand48();
  int player_pos_x = left_start_x - max_x_diff*drand48();
  int player_pos_y = left_start_y - max_y_diff*drand48();
  */

  // Get the current score
  mStartScore = mTrainerView->getLeftScore();

  // Get the current time
  mStartTime = Time::currentTime.getT();

  // The player should be positioned in the penalty area
  Point player_pos = generate_point_in_penalty_area(S_RIGHT);

  // Get the direction to the goal
  Vector goal_dir(player_pos, mFacts->getTheirGoal().getPosition());

  // Make sure the angle is not too large and the distance is ok
  while ( abs(goal_dir.getAngle()) > 75
	  || goal_dir.getLength() > 15
	  || goal_dir.getLength() < 4 ) {
    player_pos = generate_point_in_penalty_area(S_RIGHT);
    goal_dir = Vector(player_pos, mFacts->getTheirGoal().getPosition());
  }
  
  AngleDeg min_angle = goal_dir.getAngle() - 45;
  AngleDeg max_angle = goal_dir.getAngle() + 45;

  // The player should be facing the goal plus/minus 45 degrees.
  AngleDeg player_dir = generate_direction_in_sector(min_angle, max_angle);

  // Make sure the ball is kickable and between the player and the goal
  Point ball_pos = player_pos + generate_kickable_ball_vector(min_angle,
							      max_angle);
  //Point ball_pos = player_pos + generate_vector_in_circle_segment(0, 0, 0.5);
  
  /*
    std::cout << "Setup OnlyAttackerScoreScenario\n   player_pos = "
    << player_pos << "; ball_pos = " << ball_pos << std::endl;
  */
  
  // Send the setup commands
  step.addCommand(new TrainerRecoverCommand());
  step.addCommand(new TrainerStartCommand());
  step.addCommand(new ChangeModeCommand(PM_PLAY_ON));
  step.addCommand(new CoachSayCommand("score"));

  step.addCommand(new TrainerMoveObjectCommand(ball_pos));
  step.addCommand(new TrainerMoveObjectCommand(mTrainerView->getLeftTeamName(), 1, player_pos, player_dir.getDeg(), Point(0, 0)));
  
  mRunning = true;

  return Plan(step);
}


bool
OnlyAttackerScoreScenario::running()
{
  return mRunning;
}


bool
OnlyAttackerScoreScenario::success()
{
  // If scored after the scenario started it succeeded
  if ( mRunning && mTrainerView->getLeftScore() > mStartScore ) {
    mRunning = false;
    return true;
  }
  return false;
}


bool
OnlyAttackerScoreScenario::failure()
{
  // If more than mCycles cycles since the scenario started then fail
  if ( mRunning &&
       ( Time::currentTime.getT() >= (mStartTime+mCycles)
	 || mGameView->getPlayMode() != PM_PLAY_ON ) ) {
    mRunning = false;
    return true;
  }
  return false;
}


RS_END_NAMESPACE
