/* -*- Mode: C++ -*-
 * $Id: TestCase.cc,v 1.4 2005/09/12 19:28:11 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: TestCase.cc
 * Unitname: Framework/trainer
 * $Revision: 1.4 $
 * Created by: Fredrik Heintz 2004-08-31
 * Last modified by $Author: frehe $ $Date: 2005/09/12 19:28:11 $
 *
 *
 * HISTORY
 *
 * $Log: TestCase.cc,v $
 * Revision 1.4  2005/09/12 19:28:11  frehe
 * Added a delay between success/failure and next scenario.
 *
 * Revision 1.3  2005/09/08 14:17:58  frehe
 * Send messages to start and end a test case as well as report success or failure on each scenario.
 *
 * Revision 1.2  2005/08/18 20:22:35  frehe
 * Added a min level to remove certain test cases.
 *
 * Revision 1.1  2005/08/13 13:36:58  frehe
 * Readded the files since the trainer seems to have been deleted from the CVS.
 *
 * Revision 1.5  2004/09/07 17:28:17  frehe
 * Changed the design to allow reading of test cases from file.
 *
 * Revision 1.4  2004/09/05 19:38:54  frehe
 * Improved statistics printing.
 *
 * Revision 1.3  2004/09/01 22:02:31  frehe
 * Added possibility to add more than one scenario.
 *
 * Revision 1.2  2004/08/31 23:04:44  frehe
 * Fixed off by one bug in print_statistics.
 *
 * Revision 1.1  2004/08/31 17:12:26  frehe
 * Updated to the new test case scenario wrapper.
 *
 */

/**
 * @file TestCase.cc
 * Implementation file for the class TestCase.
 *
 * @par Unit:
 *      Framework/trainer
 * $Revision: 1.4 $
 * $Author: frehe $
 * $Date: 2005/09/12 19:28:11 $
 */


#include "TestCase.h"
#include "Scenario.h"
#include "ActuatorInterface.h"
#include "CoachSayCommand.h"
#include <iostream>
#include <iomanip>


RS_BEGIN_NAMESPACE


TestCase::TestCase(const std::string& name,
		   int iterations,
		   int verbose)
  : verbose_(verbose),
    name_(name),
    min_level_(0),
    iterations_(iterations),
    iterations_digits_(0),
    iteration_(0),
    cycles_since_finished_(0),
    successes_(),
    current_scenario_(0),
    scenarios_(),
    actuator_interface_(0)
{
  int tmp = iterations_;
  while ( tmp >= 1 ) {
    ++iterations_digits_;
    tmp /= 10;
  }
}


TestCase::~TestCase()
{
  while ( !scenarios_.empty() ) {
    delete scenarios_.front();
    scenarios_.pop_front();
  }
}


void
TestCase::add_scenario(Scenario* scenario)
{
  if ( scenario != 0 && scenario->level() >= min_level_ ) {
    scenarios_.push_back(scenario);
  }
}


bool
TestCase::start(int min_level)
{
  min_level_ = min_level;

  // Remove all scenarios that are too low level
  while ( !scenarios_.empty()
	  && scenarios_.front()->level() < min_level_ ) {
    delete scenarios_.front();
    scenarios_.pop_front();
  }

  if ( current_scenario_ == 0
       && !scenarios_.empty() ) {
    current_scenario_ = scenarios_.front();
    scenarios_.pop_front();
  }

  if ( iterations_ > 0 && current_scenario_ != 0 ) {
    iteration_ = 1;
    successes_.push_back(0);
    if ( setup(current_scenario_) ) {
      actuator_interface_->addToPlan(new CoachSayCommand("start"));
      return true;
    }
  }

  return false;
}


bool
TestCase::setup(Scenario* scenario)
{
  //std::cout << "setup scenario..." << std::endl;
  Plan plan = scenario->setup();
  if ( !plan.empty() ) {
    while ( !plan.empty() ) {
      actuator_interface_->addToPlan(plan.getNextStep());
      plan.removeNextStep();
    }
    return true;
  } else {
    return false;
  }
}


bool
TestCase::running()
{
  return iteration_ > 0;
}


bool
TestCase::finished()
{
  // If we haven't started or don't have a scenario we are finished
  if ( iteration_ == 0 || current_scenario_ == 0 ) {
    return true;
  }

  // Check if the current scenario is finished or not
  if ( current_scenario_->success() && cycles_since_finished_ == 0 ) {
    ++successes_.back();
    if ( verbose_ > 1 ) {
      std::cout << "success, " << std::setw(iterations_digits_)
		<< successes_.back() << "/" << std::setw(iterations_digits_)
		<< std::left << iteration_ << " = " << std::setw(3)
		<< std::right << int((100*successes_.back())/iteration_)
		<< "% success" << std::endl;
    }
    actuator_interface_->addToPlan(new CoachSayCommand("success"));
    //std::cout << "sent success" << std::endl;
    cycles_since_finished_ = 1;
  } else if ( current_scenario_->failure() && cycles_since_finished_ == 0 ) {
    if ( verbose_ > 1 ) {
      std::cout << "failure, " << std::setw(iterations_digits_)
		<< successes_.back() << "/" << std::setw(iterations_digits_)
		<< std::left << iteration_ << " = " << std::setw(3)
		<< std::right << int((100*successes_.back())/iteration_)
		<< "% success" << std::endl;
    }
    actuator_interface_->addToPlan(new CoachSayCommand("failure"));
    //std::cout << "sent failure" << std::endl;
    cycles_since_finished_ = 1;
  } else if ( cycles_since_finished_ > 0 ) {
    //std::cout << "waiting before setup... cycles_since_finished_ = " << cycles_since_finished_ << std::endl;
    ++cycles_since_finished_;
  } else {
  }

  if ( !current_scenario_->running() && cycles_since_finished_ > 5 ) {
    cycles_since_finished_ = 0;
    ++iteration_;
    if ( iteration_ > iterations_ ) {
      // Remove all scenarios that are too low level
      while ( !scenarios_.empty()
	      && scenarios_.front()->level() < min_level_ ) {
	delete scenarios_.front();
	scenarios_.pop_front();
      }
      if ( scenarios_.empty() ) {
	actuator_interface_->addToPlan(new CoachSayCommand("finished"));
	print_statistics(std::cout);
	return true;
      } else {
	if ( verbose_ > 0 ) {
	}

	// Get the next scenario
	current_scenario_ = scenarios_.front();
	scenarios_.pop_front();
	iteration_ = 1;
	successes_.push_back(0);
	return !setup(current_scenario_);
      }
    } else {
      return !setup(current_scenario_);
    }
  }

  return false;
}


void
TestCase::print_statistics(std::ostream& os)
{
  int s = 0;
  int t = 0;
  int w1 = 0;
  int tmp = successes_.size();
  while ( tmp >= 1 ) {
    ++w1;
    tmp /= 10;
  }
  for ( std::vector<int>::const_iterator i = successes_.begin();
	i != successes_.end(); ++i ) {
    ++s;
    os << "scenario " << std::setw(w1) << s << ": "
       << std::setw(iterations_digits_) << *i << " successes = "
       << std::setw(3) << int((*i * 100)/iterations_) << "% success"
       << std::endl;
    t += *i;
  }
  os << "in summary " << t << "/" << iterations_*s
     << " = " << int((100*t)/(iterations_*s)) << "% success"
     << std::endl;
}


void
TestCase::set_actuator_interface(ActuatorInterface* const actuator_interface)
{
  actuator_interface_ = actuator_interface;
}


RS_END_NAMESPACE
