/* -*- Mode: C++ -*-
 * $Id: TestCase.h,v 1.3 2005/09/12 19:28:11 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: TestCase.h
 * Unitname: Framework/trainer
 * $Revision: 1.3 $
 * Created by: Fredrik Heintz 2004-08-31
 * Last modified by $Author: frehe $ $Date: 2005/09/12 19:28:11 $
 *
 *
 * HISTORY
 *
 * $Log: TestCase.h,v $
 * Revision 1.3  2005/09/12 19:28:11  frehe
 * Added a delay between success/failure and next scenario.
 *
 * Revision 1.2  2005/08/18 20:22:35  frehe
 * Added a min level to remove certain test cases.
 *
 * Revision 1.1  2005/08/13 13:36:58  frehe
 * Readded the files since the trainer seems to have been deleted from the CVS.
 *
 * Revision 1.4  2004/09/07 17:28:17  frehe
 * Changed the design to allow reading of test cases from file.
 *
 * Revision 1.3  2004/09/05 19:38:54  frehe
 * Improved statistics printing.
 *
 * Revision 1.2  2004/09/01 22:02:31  frehe
 * Added possibility to add more than one scenario.
 *
 * Revision 1.1  2004/08/31 17:12:26  frehe
 * Updated to the new test case scenario wrapper.
 *
 */

/**
 * @file TestCase.h
 * Include file for the class TestCase.
 *
 * @par Unit:
 *      Framework/trainer
 * $Revision: 1.3 $
 * $Author: frehe $
 * $Date: 2005/09/12 19:28:11 $
 */


#ifndef TEST_CASE_H
#define TEST_CASE_H


#include "machine_dependent_compiler_directives.h"
#include <string>
#include <deque>
#include <vector>
#include <iosfwd>


RS_BEGIN_NAMESPACE


class Scenario;
class ActuatorInterface;
  

/**
 * @class TestCase TestCase.h
 * Representation of test cases for the RoboSoc trainer.
 * A test case is one or more scenarios which are ran a number
 * of iterations. When the test case is done some statistics are
 * presented.
 *
 * @par Unit:
 *      Framework/trainer
 * $Revision: 1.3 $
 * $Author: frehe $
 * $Date: 2005/09/12 19:28:11 $
 */
class TestCase {
public:
  /**
   * Create a test case called @a name running
   * its scenarios @a iterations times.
   * Set the verbosity level to @a verbose.
   */
  TestCase(const std::string& name, int iterations, int verbose);

  /// Destroys the test_case.
  ~TestCase();


  /// Add a new scenario. The test case takes control of @a scenario.
  void add_scenario(Scenario* scenario);


  /**
   * Start the test case, run all scenarios with level greater
   * than or equal to @a min_level, return true if successful.
   */
  bool start(int min_level);


  /// Return true iff the test case has been started
  bool running();
  

  /// Return true iff the test case is finished.
  bool finished();


  /// Print the current statistics.
  void print_statistics(std::ostream& os);


  /// Set the actuator interface used to send commands
  void set_actuator_interface(ActuatorInterface* const actuator_interface);


protected:
  bool setup(Scenario* scenario);
  

private:
  /// The verbosity level of the test case
  int verbose_;
  
  /// The name of the test case
  std::string name_;

  /// The minimum level used in this test case
  int min_level_;

  /// The total number of iterations in this test case
  int iterations_;

  /// The number of digits in iterations_
  int iterations_digits_;

  /// The number of iterations ran so far
  int iteration_;

  /// The number of cycles since the last scenario was finished
  int cycles_since_finished_;

  /// The number of successes so far for each scenario
  std::vector<int> successes_;

  /// The scenarios of the test case
  std::deque<Scenario*> scenarios_;

  /// The current scenario
  Scenario* current_scenario_;

  /// The actuator interface
  ActuatorInterface* actuator_interface_;
};


RS_END_NAMESPACE


#endif // !TEST_CASE_H
