/* -*- Mode: C++ -*-
 * $Id: ThreeOnThreeScenario.cc,v 1.2 2005/08/18 20:22:35 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: ThreeOnThreeScenario.cc
 * Unitname: Framework/trainer
 * $Revision: 1.2 $
 * Created by: Fredrik Heintz 2004-09-29
 * Last modified by $Author: frehe $ $Date: 2005/08/18 20:22:35 $
 *
 *
 * HISTORY
 *
 * $Log: ThreeOnThreeScenario.cc,v $
 * Revision 1.2  2005/08/18 20:22:35  frehe
 * Added a min level to remove certain test cases.
 *
 * Revision 1.1  2005/08/13 13:36:58  frehe
 * Readded the files since the trainer seems to have been deleted from the CVS.
 *
 * Revision 1.1  2004/09/29 20:24:22  frehe
 * Added support for three on three scenario.
 *
 */

/**
 * @file ThreeOnThreeScenario.cc
 * Implementation file for the class ThreeOnThreeScenario.
 *
 * @par Unit:
 *      Framework/trainer
 * $Revision: 1.2 $
 * $Author: frehe $
 * $Date: 2005/08/18 20:22:35 $
 */


#include "ThreeOnThreeScenario.h"
#include "ActuatorInterface.h"
#include "ViewManager.h"
#include "WorldModelInterface.h"
#include "WorldFacts.h"
#include "view_constants.h"
#include "PlayerView.h"
#include "TrainerView.h"
#include "GameView.h"
#include "TrainerStartCommand.h"
#include "ChangeModeCommand.h"
#include "CoachSayCommand.h"
#include "TrainerRecoverCommand.h"
#include "TrainerMoveObjectCommand.h"
#include "string_utils.h"
#include "trainer_helpers.h"

#include <cassert>


RS_BEGIN_NAMESPACE


ThreeOnThreeScenario::ThreeOnThreeScenario(int random_seed,
					   const std::vector<Point>& attackers_pos,
					   const std::vector<Point>& defenders_pos,
					   int cycles)
  : Scenario(defenders_pos.size()),
    mViewManager(ViewManager::instance()),
    mWorldModelInterface(WorldModelInterface::instance()),
    mFacts(mWorldModelInterface->getWorldFacts()),
    mPlayerView(0),
    mTrainerView(0),
    mGameView(0),
    mAttackersPos(attackers_pos),
    mDefendersPos(defenders_pos),
    mRunning(false),
    mCycles(cycles)
{
  getView(mPlayerView, PLAYER_VIEW_ID);
  getView(mTrainerView, TRAINER_VIEW_ID);
  getView(mGameView, GAME_VIEW_ID);

  // Initialize the random number generator
  init_random_number_generator(random_seed);
}


ThreeOnThreeScenario::~ThreeOnThreeScenario()
{
  mViewManager->releaseView(PLAYER_VIEW_ID);
  mViewManager->releaseView(TRAINER_VIEW_ID);
  mViewManager->releaseView(GAME_VIEW_ID);
}


Plan
ThreeOnThreeScenario::setup()
{
  Step step;

  if ( mPlayerView->getTeammates().size() < mAttackersPos.size() ) {
    std::cout << "Not enough players on the left team!" << std::endl;
    RS_LOG( LA_DEBUG, "ThreeOnThreeScenario: Not enough players on the left team" );

    return Plan();
  }
  const std::vector<PlayerObject>& teammates = mPlayerView->getTeammates();
  for ( std::vector<PlayerObject>::const_iterator i = teammates.begin();
	i != teammates.end(); ++i ) {
    if ( i->getUniformNumber().isKnown()
	 && i->getUniformNumber().getValue() > mAttackersPos.size() ) {
      step.addCommand(new TrainerMoveObjectCommand(mTrainerView->getLeftTeamName(), i->getUniformNumber().getValue(), 0, 0, 0, 0, 0));
    }
  }

  if ( mPlayerView->getOpponents().size() < mDefendersPos.size() ) {
    std::cout << "Not enough players on the right team!" << std::endl;
    RS_LOG( LA_DEBUG, "ThreeOnThreeScenario: Not enough players on the right team" );

    return Plan();
  }
  const std::vector<PlayerObject>& opps = mPlayerView->getOpponents();
  for ( std::vector<PlayerObject>::const_iterator i = opps.begin();
	i != opps.end(); ++i ) {
    if ( i->getUniformNumber().isKnown()
	 && i->getUniformNumber().getValue() > mDefendersPos.size() ) {
      step.addCommand(new TrainerMoveObjectCommand(mTrainerView->getRightTeamName(), i->getUniformNumber().getValue(), 0, 0, 0, 0, 0));
    }
  }

  // Get the current score
  mStartScore = mTrainerView->getLeftScore();

  // Get the current time
  mStartTime = Time::currentTime.getT();

  // Place the ball infront of one of the attackers
  int unum = generate_integer_in_interval(1, mAttackersPos.size());
  assert( unum > 0 ); assert( unum <= mAttackersPos.size() );
  std::cout << "ball at " << unum << std::endl;
  Point ball_pos;
  
  // Send the setup commands
  step.addCommand(new TrainerRecoverCommand());
  step.addCommand(new TrainerStartCommand());

  // Setup the attackers
  int attacker = 1;
  for ( std::vector<Point>::const_iterator p = mAttackersPos.begin();
	p != mAttackersPos.end(); ++p ) {
    Point player_pos = *p;
    while ( in_penalty_area(player_pos, S_RIGHT) ) {
      player_pos = generate_point_in_area(Point(30,-30), Point(40,30));
    }
    if ( unum == attacker ) {
      ball_pos = player_pos + generate_kickable_ball_vector(-45,45);
    }
    step.addCommand(new TrainerMoveObjectCommand(mTrainerView->getLeftTeamName(), attacker, player_pos, 0, Point(0, 0)));
    ++attacker;
  }

  // Setup the defenders
  int defender = 1;
  for ( std::vector<Point>::const_iterator p = mDefendersPos.begin();
	p != mDefendersPos.end(); ++p ) {
    Point player_pos = *p;
    while ( !in_penalty_area(player_pos, S_RIGHT) ) {
      player_pos = generate_point_in_penalty_area(S_RIGHT);
    }
    step.addCommand(new TrainerMoveObjectCommand(mTrainerView->getRightTeamName(), defender, player_pos, 180, Point(0, 0)));
    ++defender;
  }

  // Setup the ball
  step.addCommand(new TrainerMoveObjectCommand(ball_pos));

  // Start the scenario
  step.addCommand(new ChangeModeCommand(PM_PLAY_ON));
  step.addCommand(new CoachSayCommand("start"));

  mRunning = true;

  return Plan(step);
}


bool
ThreeOnThreeScenario::running()
{
  return mRunning;
}


bool
ThreeOnThreeScenario::success()
{
  // If scored after the scenario started it succeeded
  if ( mRunning && mTrainerView->getLeftScore() > mStartScore ) {
    mRunning = false;
    return true;
  }
  return false;
}


bool
ThreeOnThreeScenario::failure()
{
  // If more than mCycles cycles since the scenario started then fail
  // or the play mode is not play on.
  if ( mRunning
       && ( Time::currentTime.getT() >= (mStartTime+mCycles)
	    || mGameView->getPlayMode() != PM_PLAY_ON ) ) {
    mRunning = false;
    return true;
  }
  return false;
}


RS_END_NAMESPACE
