/* -*- Mode: C++ -*-
 * $Id: TrainerDecision.cc,v 1.3 2005/09/15 09:32:39 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: TrainerDecision.cc
 * Unitname: Framework/trainer
 * $Revision: 1.3 $
 * Created by: Fredrik Heintz 2004-08-23
 * Last modified by $Author: frehe $ $Date: 2005/09/15 09:32:39 $
 *
 *
 * HISTORY
 *
 * $Log: TrainerDecision.cc,v $
 * Revision 1.3  2005/09/15 09:32:39  frehe
 * Checked in commented out test code.
 *
 * Revision 1.2  2005/08/18 20:22:35  frehe
 * Added a min level to remove certain test cases.
 *
 * Revision 1.1  2005/08/13 13:36:58  frehe
 * Readded the files since the trainer seems to have been deleted from the CVS.
 *
 * Revision 1.10  2004/09/07 17:28:17  frehe
 * Changed the design to allow reading of test cases from file.
 *
 * Revision 1.9  2004/09/05 20:59:59  frehe
 * Updated the test cases.
 *
 * Revision 1.8  2004/09/03 09:43:56  frehe
 * Added new scenarios with defenders.
 *
 * Revision 1.7  2004/09/02 21:49:43  frehe
 * Wait for play mode before start scenario and added more scenarios.
 *
 * Revision 1.6  2004/09/01 22:02:32  frehe
 * Added possibility to add more than one scenario.
 *
 * Revision 1.5  2004/09/01 09:04:04  frehe
 * Use the new AttackerGoalieScoreScenario.
 *
 * Revision 1.4  2004/08/31 17:12:26  frehe
 * Updated to the new test case scenario wrapper.
 *
 * Revision 1.3  2004/08/31 12:21:43  frehe
 * First version with external scenarios.
 *
 * Revision 1.2  2004/08/30 14:52:02  frehe
 * Simple trainer with one fixed scenario, score from random position in penalty area.
 *
 * Revision 1.1  2004/08/23 20:46:44  frehe
 * First version of trainer.
 *
 */

/**
 * @file TrainerDecision.cc
 * Implementation of the class TrainerDecision.
 *
 * @par Unit:
 *      Framework/trainer
 * $Revision: 1.3 $
 * $Author: frehe $
 * $Date: 2005/09/15 09:32:39 $
 */


#include "TrainerDecision.h"
#include "GameView.h"
#include "BallView.h"
#include "AgentView.h"
#include "PlayerView.h"
#include "TrainerView.h"
#include "ActuatorInterface.h"
#include "ViewManager.h"
#include "view_constants.h"
#include "TrainerInitCommand.h"
#include "TrainerStartCommand.h"
#include "ChangeModeCommand.h"
#include "TrainerRecoverCommand.h"
#include "TrainerMoveObjectCommand.h"
#include "TrainerGetTeamnamesCommand.h"
#include "EarCommand.h"
#include "EyeCommand.h"
#include "OnlyAttackerScoreScenario.h"
#include "AttackerGoalieScoreScenario.h"
#include "AttackerDefenderGoalieScoreScenario.h"
#include "TestCase.h"
#include "rs_assert.h"
#include "rs_log.h"

#include <cstdlib>


RS_BEGIN_NAMESPACE


TrainerDecision::TrainerDecision(TestCase* test_case,
				 int min_level,
				 ActuatorInterface* const actuator_interface)
  : Decision(actuator_interface),
    viewManager(ViewManager::instance()),
    worldModelInterface(WorldModelInterface::instance()),
    facts(worldModelInterface->getWorldFacts()),
    gameView(NULL),
    ballView(NULL),
    agentView(NULL),
    playerView(NULL),
    trainerView(NULL),
    min_level_(min_level),
    currentTestCase(test_case),
    got_visual_data(false)
{
  getView(gameView, GAME_VIEW_ID);
  getView(ballView, BALL_VIEW_ID);
  getView(agentView, AGENT_VIEW_ID);
  getView(playerView, PLAYER_VIEW_ID);
  getView(trainerView, TRAINER_VIEW_ID);

  // Set the actuator interface and random seed of the test case
  if ( currentTestCase != 0 ) {
    currentTestCase->set_actuator_interface(actuatorInterface);
  }
}


TrainerDecision::~TrainerDecision()
{
  viewManager->releaseView(GAME_VIEW_ID);
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(PLAYER_VIEW_ID);
  viewManager->releaseView(TRAINER_VIEW_ID);
}


ServerCommand*
TrainerDecision::initAgent()
{
  TrainerInitCommand* init_cmd = new TrainerInitCommand(facts->SP_version);
  return init_cmd;
}


bool
TrainerDecision::init(ActuatorInterface* const actuator_interface)
{
  Decision::init(actuator_interface);
  
  // Set the actuator interface and random seed of the test case
  if ( currentTestCase != 0 ) {
    currentTestCase->set_actuator_interface(actuatorInterface);
  }
}


bool
TrainerDecision::start()
{
  RS_REQUIRE( initialized == true );
  return true;
}


bool
TrainerDecision::decide()
{
  std::cout << Time::currentTime << ": decide" << std::endl;
  return false;
}


void
TrainerDecision::onAuralSensorData()
{
  //std::cout << Time::currentTime << ": onAuralSensorData()" << std::endl;
}


void
TrainerDecision::onAuralCoachSensorData()
{
  std::cout << Time::currentTime << ": onAuralCoachSensorData()" << std::endl;
}


void
TrainerDecision::onAuralRefereeSensorData()
{
  //std::cout << Time::currentTime << ": onAuralRefereeSensorData()" << std::endl;
  if ( !got_visual_data && gameView->getPlayMode() == PM_PLAY_ON ) {
    //std::cout << "play mode = PLAY_ON!" << std::endl;
    got_visual_data = true;
  }
}


void
TrainerDecision::onAuralPlayerSensorData()
{
  std::cout << Time::currentTime << ": onAuralPlayerSensorData()" << std::endl;
}


void
TrainerDecision::onAuralSelfSensorData()
{
}


void
TrainerDecision::onInit()
{
  //std::cout << Time::currentTime << ": onInit()" << std::endl;
  actuatorInterface->addToPlan(new EyeCommand(true));
  actuatorInterface->addToPlan(new EarCommand(true));
  actuatorInterface->addToPlan(new TrainerGetTeamnamesCommand());
  actuatorInterface->addToPlan(new ChangeModeCommand(PM_PLAY_ON));
}


void
TrainerDecision::onPhysicalSensorData()
{
  std::cout << Time::currentTime << ": onPhysicalSensorData()" << std::endl;
}


void
TrainerDecision::onVisualCoachSensorData()
{
  /*
  static Time lastSeeTime = 0;
  if ( lastSeeTime != 0 && lastSeeTime != Time::currentTime && lastSeeTime != ( Time::currentTime - 1) ) {
    std::cout << "Missed see time, last see = " << lastSeeTime
	      << "; current time = " << Time::currentTime
	      << std::endl;
  }
  lastSeeTime = Time::currentTime;
  */

  /*
  std::cout << Time::currentTime << ": ball pos = "
	    << ballView->getPosition() << std::endl;
  */
  
  if ( currentTestCase == 0 ) {
    std::cout << "No test case!" << std::endl;
    abort();
  } else {
    if ( !currentTestCase->running() ) {
      if ( got_visual_data ) {
	// Start the current test case
	currentTestCase->start(min_level_);
      }
    } else {
      if ( currentTestCase->finished() ) {
	//currentTestCase->print_statistics(std::cout);
	// Remove the current test case
	delete currentTestCase;
	currentTestCase = 0;
	// Get the next test case...
	std::cout << "Finished, no more test cases" << std::endl;
	abort();
      }
    }
  }
}


void
TrainerDecision::onVisualSensorData()
{
  //std::cout << Time::currentTime << ": onVisualSensorData()" << std::endl;
}


void
TrainerDecision::onBeforeSensorData()
{
  //std::cout << Time::currentTime << ": onBeforeSensorData()" << std::endl;
}


void
TrainerDecision::onAfterSensorData()
{
  //std::cout << Time::currentTime <<  ": onAfterSensorData()" << std::endl;

  /*
    std::cout << "Left side = " << S_LEFT
    << "; Right side = " << S_RIGHT << std::endl;
    std::cout << "Our side: " << gameView->getOurSide() << std::endl;
    std::cout << "Our team: " << gameView->getOurTeamName() << std::endl;
    std::cout << "Their side: " << gameView->getTheirSide() << std::endl;
    std::cout << "Their team: " << gameView->getTheirTeamName() << std::endl;
  */
}


void
TrainerDecision::onBeforeTick()
{
  //std::cout << Time::currentTime << ": onBeforeTick()" << std::endl;
}


void
TrainerDecision::onAfterTick()
{
  //std::cout << Time::currentTime << ": onAfterTick()" << std::endl;
}


void
TrainerDecision::onNewCycle(const bool forced)
{
  /*
    struct timeval now;
    gettimeofday(&now, NULL);
    int time_diff = now.tv_usec - timeOfLastCycle.tv_usec;
    if ( time_diff < 0 ) time_diff += 1000000;
    std::cout << Time::currentTime << ": onNewCycle(" << forced << ") time since last cycle = " << time_diff << std::endl;
    timeOfLastCycle = now;
  */
  /*
    static int x = 0;
    static int y = 0;
    actuatorInterface->addToPlan(new TrainerMoveObjectCommand(x++, y++));
  */

  /*
    std::cout << Time::currentTime << ": "
    << playerView->getOpponents().size() << " opponents; "
    << playerView->getTeammates().size() << " teammates; "
    << playerView->getUnknownPlayers().size() << " unknown players."
    << std::endl;
  */

  /*
  if ( currentTestCase != 0 ) {
    currentTestCase->print_statistics(std::cout);
  }
  */
}


bool
TrainerDecision::onEstimatedNewCycle(UInt16 estimated_time, UInt16 ticks)
{
  /*
    struct timeval now;
    gettimeofday(&now, NULL);
    int time_diff = now.tv_usec - timeOfLastCycle.tv_usec;
    if ( time_diff < 0 ) time_diff += 1000000;

    std::cout << Time::currentTime << ": onEstimatedNewCycle("
    << estimated_time << ", " << ticks << ") time_diff = "
    << time_diff << std::endl;
  */

  return false;
}


void
TrainerDecision::onCommandWarning() {
  decide();
}


RS_END_NAMESPACE
