/* -*- Mode: C++ -*-
 * $Id: trainer.cc,v 1.2 2005/08/18 20:22:36 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: trainer.cc
 * Unitname: Framework/trainer
 * $Revision: 1.2 $
 * Created by: Fredrik Heintz 2004-08-23
 * Last modified by $Author: frehe $ $Date: 2005/08/18 20:22:36 $
 *
 *
 * HISTORY
 * $Log: trainer.cc,v $
 * Revision 1.2  2005/08/18 20:22:36  frehe
 * Added a min level to remove certain test cases.
 *
 * Revision 1.1  2005/08/13 13:36:59  frehe
 * Readded the files since the trainer seems to have been deleted from the CVS.
 *
 * Revision 1.5  2004/09/07 17:28:17  frehe
 * Changed the design to allow reading of test cases from file.
 *
 * Revision 1.4  2004/09/05 20:45:13  frehe
 * Added support for CP_no_logging.
 *
 * Revision 1.3  2004/08/31 12:21:43  frehe
 * First version with external scenarios.
 *
 * Revision 1.2  2004/08/30 14:52:02  frehe
 * Simple trainer with one fixed scenario, score from random position in penalty area.
 *
 * Revision 1.1  2004/08/23 20:46:45  frehe
 * First version of trainer.
 *
 */

/**
 * @file trainer.cc
 * The main function for the trainer.
 *
 * @par Unit:
 *      Framework/trainer
 * $Revision: 1.2 $
 * $Author: frehe $
 * $Date: 2005/08/18 20:22:36 $
 */


#include <fstream>

#include "FrameworkAgent.h"
#include "WorldModelInterface.h"
#include "TrainerDecision.h"
#include "LogStream.h"
#include "BasicSeeStrategy.h"
#include "BasicHearStrategy.h"
#include "BasicCommandStrategy.h"
#include "BasicSenseStrategy.h"
#include "BasicInitStrategy.h"
#include "BasicCoachMessageStrategy.h"
#include "BasicNewCycleStrategy.h"
#include "BasicCoachInformationStrategy.h"
#include "TrainerLexer.hpp"
#include "TrainerParser.hpp"
#include "TestCase.h"


RS_BEGIN_NAMESPACE


#if defined(RS_USE_ERROR) && !defined(RS_LOG_ERROR)
extern void rs_warning(const std::string& str,
		       const std::string& file,
		       int line)
{
  std::cerr << "WARNING: " << str
	    << " at line " << line
	    << " in file " << file
	    << std::endl;
}

extern void rs_error(const std::string& str,
		     const std::string& file,
		     int line)
{
  std::cerr << "ERROR: " << str
	    << " at line " << line
	    << " in file " << file
	    << std::endl;
}

extern void rs_fatal(const std::string& str,
		     const std::string& file,
		     int line)
{
  std::cerr << "FATAL: " << str
	    << " at line " << line
	    << " in file " << file
	    << std::endl;
  abort();
}
#endif


#if !defined(RS_USE_EXCEPTION) && !defined(RS_LOG_EXCEPTION)
extern void rs_throw(const std::string& exception,
		     const std::string& msg,
		     const std::string& file,
		     int line)
{
  std::cerr << "RAISED: " << exception << "(\"" << msg << "\")"
	    << " at line " << line
	    << " in file " << file
	    << std::endl;
}
#endif


RS_END_NAMESPACE


int main(int argc, char** argv)
{
#ifdef RS_USE_NAMESPACE
  using namespace RS_NAMESPACE;
#endif
  
  std::cerr << "Initializing the world..." << std::endl;
  WorldFacts::init(argc, argv);
  // Set the trainer parameter to true
  WorldFacts::nonConstInstance()->CP_trainer = true;

  // Get the minimum level from the command line
  int min_level = 0;
  for ( int i = 0; i < argc; ++i ) {
    if ( strcmp(argv[i], "-min_level") == 0 ) {
      if ( argc > i ) {
	min_level = atoi(argv[i+1]);
      }
    }
  }
  std::cout << "min_level = " << min_level << std::endl;
  
  std::ofstream logFile1("trainer_logfile.txt");
  LogStream *logStream = LogStream::instance();

  if ( WorldFacts::instance()->CP_no_logging ) {
    logStream->addLogArea(LA_ROBOSOC, LA_TOP, false, &std::cerr);
  } else {
    logStream->addLogArea(LA_ROBOSOC, LA_TOP, true, &std::cerr);
    logStream->addLogArea(LA_WARNING, LA_ROBOSOC, true, &logFile1,
			  true, &std::cerr);
    logStream->addLogArea(LA_ERROR, LA_ROBOSOC, true, &logFile1,
			  true, &std::cerr);
    logStream->addLogArea(LA_FATAL, LA_ROBOSOC, true, &logFile1,
			  true, &std::cerr);

    logStream->addLogArea(LA_ACTUATORS, LA_ROBOSOC, false, &logFile1);
    logStream->addLogArea(LA_CONTROLLER, LA_ROBOSOC, false, &logFile1);
    logStream->addLogArea(LA_SENSORS, LA_ROBOSOC, true, &logFile1);
    logStream->addLogArea(LA_SERVER_COMMANDS, LA_ROBOSOC, true, &logFile1);
    logStream->addLogArea(LA_SENSOR_DATA, LA_ROBOSOC, true, &logFile1);
    logStream->addLogArea(LA_STRATEGY, LA_ROBOSOC, false, &logFile1);
    logStream->addLogArea(LA_SENSOR_INTERFACE, LA_ROBOSOC, false, &logFile1);
    logStream->addLogArea(LA_DECISION, LA_ROBOSOC, true, &logFile1);
    logStream->addLogArea(LA_FRAMEWORK_AGENT, LA_ROBOSOC, false, &logFile1);
    logStream->addLogArea(LA_AGENT, LA_ROBOSOC, false, &logFile1);
    logStream->addLogArea(LA_SKILL, LA_ROBOSOC, true, &logFile1);
    logStream->addLogArea(LA_COACH, LA_ROBOSOC, true, &logFile1);
    logStream->addLogArea(LA_WORLDMODEL, LA_ROBOSOC, false, &logFile1);
    logStream->addLogArea(LA_TIME, LA_ROBOSOC, false, &logFile1);
  }

  try {
    std::cerr << "Initializing the world model interface..." << std::endl;
    WorldModelInterface *wmi = WorldModelInterface::nonConstInstance();

    std::cerr << "Setting world model update strategies..." << std::endl;
    wmi->setSeeStrategy(new BasicSeeStrategy());
    wmi->setHearStrategy(new BasicHearStrategy());
    wmi->setCommandStrategy(new BasicCommandStrategy());
    wmi->setInitStrategy(new BasicInitStrategy());
    wmi->setSenseStrategy(new BasicSenseStrategy());
    wmi->setCoachMessageStrategy(new BasicCoachMessageStrategy());
    wmi->setNewCycleStrategy(new BasicNewCycleStrategy());
    wmi->setCoachInformationStrategy(new BasicCoachInformationStrategy());

    // Create a string stream to use as the input for the lexer
    std::stringstream input;

    // Create the trainer lexer
    TrainerLexer lexer(std::cin);
    lexer.setFilename("<stdin>");

    // Create the trainer parser
    TrainerParser parser(lexer);
    parser.setFilename("<stdin>");
    parser.random_seed(WorldFacts::instance()->CP_random_seed);

    // Read the test case from the input file
    TestCase* test_case = parser.test_case();

    std::cerr << "Starting decision..." << std::endl;
    Decision* decision = new TrainerDecision(test_case, min_level);

    std::cerr << "Initializing the FrameworkAgent..." << std::endl;
    FrameworkAgent trainer(decision);

    std::cerr << "Starting the trainer...\n";
    trainer.start();

  } catch (RSException &excpt) {
    std::cerr << "Caught a RoboSoc exception: " << excpt.What() << std::endl;
    return 1;
  } catch (std::exception &excpt) {
    std::cerr << "Caught a standard exception: " << excpt.what() << std::endl;
    return 1;
  } catch (...) {
    std::cerr << "Caught an unknown exception: " << std::endl;
    return 1;
  }

  return 0;
}
