/* -*- Mode: C++ -*-
 * $Id: trainer_helpers.cc,v 1.2 2005/08/18 20:22:33 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: trainer_helpers.cc
 * Unitname: Framework/trainer
 * $Revision: 1.2 $
 * Created by: Fredrik Heintz 2004-08-31
 * Last modified by $Author: frehe $ $Date: 2005/08/18 20:22:33 $
 *
 *
 * HISTORY
 *
 * $Log: trainer_helpers.cc,v $
 * Revision 1.2  2005/08/18 20:22:33  frehe
 * Improved the generate_point_on_goal_line function.
 *
 * Revision 1.1  2005/08/13 13:36:59  frehe
 * Readded the files since the trainer seems to have been deleted from the CVS.
 *
 * Revision 1.5  2004/09/29 20:24:23  frehe
 * Added support for three on three scenario.
 *
 * Revision 1.4  2004/09/03 09:42:21  frehe
 * Added new helper for generating defensive positions.
 *
 * Revision 1.3  2004/09/01 12:50:01  frehe
 * Added point in scoring sector generator.
 *
 * Revision 1.2  2004/09/01 08:59:32  frehe
 * Added more helpers.
 *
 * Revision 1.1  2004/08/31 23:05:31  frehe
 * First version of the trainer helper functionc.
 *
 */

/**
 * @file trainer_helpers.cc
 * Implementation file for the RoboSoc trainer utilities.
 *
 * @par Unit:
 *      Framework/trainer
 * $Revision: 1.2 $
 * $Author: frehe $
 * $Date: 2005/08/18 20:22:33 $
 */


#include "trainer_helpers.h"
#include "WorldFacts.h"
#include <cstdlib>
#include <cassert>


RS_BEGIN_NAMESPACE


bool
in_penalty_area(const Point& pos, const Side& side)
{
  if ( side == S_LEFT ) {
    return ( pos.getX() < -36.0
	     && pos.getX() > -52.5
	     && abs(pos.getY()) < 20.16 );
  } else {
    return ( pos.getX() > 36.0
	     && pos.getX() < 52.5
	     && abs(pos.getY()) < 20.16 );
  }
}


void
init_random_number_generator(int seed)
{
  srand48(seed);
}


Int
generate_integer_in_interval(Int min, Int max)
{
  Int dist = max - min;
  return Int(min + dist*drand48());
}


Float
generate_distance_in_interval(Float min_dist, Float max_dist)
{
  Float dist_diff = max_dist - min_dist;
  return min_dist + dist_diff*drand48();
}


AngleDeg
generate_direction_in_sector(const AngleDeg& min_angle,
			     const AngleDeg& max_angle)
{
  Float angle_diff = max_angle.getDeg() - min_angle.getDeg();
  return AngleDeg(min_angle.getDeg() + angle_diff*drand48());
}


Point
generate_point_on_line(const Point& left, const Point& right)
{
  return left + (left.getVectorTo(right) * drand48());
}


Point
generate_point_in_area(const Point& upper_left,
		       const Point& lower_right)
{
  Float x_dist = lower_right.getX() - upper_left.getX();
  Float y_dist = lower_right.getY() - upper_left.getY();
  return Point(upper_left.getX() + x_dist*drand48(),
	       upper_left.getY() + y_dist*drand48());
}


Point
generate_point_in_parallellogram(const Point& upper_left,
				 const Point& upper_right,
				 const Point& lower_left,
				 const Point& lower_right)
{
  Vector ul_to_ur = upper_left.getVectorTo(upper_right);

  // Generate point in an area of the right size
  Point res = generate_point_in_area(upper_left, lower_right);
  //std::cout << "point = " << res << std::flush;
      
  // Find out the shape of the area
  if ( ul_to_ur.getAngle() < 0 ) {
    // The right side is the longest

    // If the point is in the right area then return it
    // otherwise move it to the upper left triangle.
    Vector ll_to_res = lower_left.getVectorTo(res);
    Vector ll_to_lr = lower_left.getVectorTo(lower_right);
    /*
    std::cout << " ll_to_res = " << ll_to_res.getAngle()
	      << " ll_to_lr = " << ll_to_lr.getAngle() << std::flush;
    */
    if ( ll_to_res.getAngle() > ll_to_lr.getAngle() ) {
      // The point is outside the area, move it
      // Reflect the vector in the longest side
      ll_to_res.setAngle(180.0 - ll_to_res.getAngle());
      // Add the vector to upper_right
      res = upper_right + ll_to_res;
      //std::cout << " outside new point = " << res << std::endl;
    } else {
      //std::cout << " inside" << std::endl;
    }

    // Make sure the answer is within the area
    ll_to_res = lower_left.getVectorTo(res);
    assert ( ll_to_res.getAngle() <= ll_to_lr.getAngle() );
  } else {
    // The left side is the longest

    // If the point is in the right area then return it
    // otherwise move it to the lower right triangle.
    Vector ul_to_res = upper_left.getVectorTo(res);
    Vector ul_to_ur = upper_left.getVectorTo(upper_right);
    /*
    std::cout << " ul_to_res = " << ul_to_res.getAngle()
	      << " ul_to_ur = " << ul_to_ur.getAngle() << std::flush;
    */
    if ( ul_to_res.getAngle() < ul_to_ur.getAngle() ) {
      // The point is outside the area, move it
      // Reflect the vector in the longest side
      ul_to_res.setAngle(180.0 - ul_to_res.getAngle());
      // Add the vector to lower_right
      res = lower_right + ul_to_res;
      //std::cout << " outside new point = " << res << std::endl;
    } else {
      //std::cout << " inside" << std::endl;
    }

    // Make sure the answer is within the area
    ul_to_res = upper_left.getVectorTo(res);
    assert ( ul_to_res.getAngle() >= ul_to_ur.getAngle() );
  }

  return res;
}


Point
generate_point_in_scoring_sector(const Side& side)
{
  static const WorldFacts* facts = WorldFacts::instance();

  if ( side == S_LEFT ) {
    // The parallellogram of the left penalty area
    static Point left_upper_left(facts->getMarker(FLAG_GLT).getPosX() + 1,
				 facts->getMarker(FLAG_GLT).getPosY());
    static Point left_upper_right(facts->getMarker(FLAG_PLT).getPosition());
    static Point left_lower_left(facts->getMarker(FLAG_GLB).getPosX() + 1,
				 facts->getMarker(FLAG_GLB).getPosY());
    static Point left_lower_right(facts->getMarker(FLAG_PLB).getPosition());
    
    return generate_point_in_parallellogram(left_upper_left, left_upper_right,
					    left_lower_left, left_lower_right);
  } else if ( side == S_RIGHT ) {
    // The parallellogram of the right penalty area
    static Point right_upper_left(facts->getMarker(FLAG_PRT).getPosition());
    static Point right_upper_right(facts->getMarker(FLAG_GRT).getPosX() - 1,
				   facts->getMarker(FLAG_GRT).getPosY());
    static Point right_lower_left(facts->getMarker(FLAG_PRB).getPosition());
    static Point right_lower_right(facts->getMarker(FLAG_GRB).getPosX() - 1,
				   facts->getMarker(FLAG_GRB).getPosY());

    return generate_point_in_parallellogram(right_upper_left,
					    right_upper_right,
					    right_lower_left,
					    right_lower_right);
  } else {
    std::cout << "UNKNOWN SIDE!" << std::endl;
    return Point();
  }
}


Point
generate_point_in_defense_sector(const Side& side, const Point& pos)
{
  static const WorldFacts* facts = WorldFacts::instance();

  // THESE ARE NOT PARALLELLOGRAMS!!!
  if ( side == S_LEFT ) {
    // The parallellogram of the left penalty area
    static Point left_upper_left(facts->getMarker(FLAG_GLT).getPosX() + 1,
				 facts->getMarker(FLAG_GLT).getPosY());
    //Point left_upper_right(pos.getX() - 1, pos.getY() - 1);
    Point left_upper_right(max(pos.getX() - 2,
			       facts->getMarker(FLAG_GLT).getPosX() + 2),
			   facts->getMarker(GOAL_L).getPosY() - 2);
    static Point left_lower_left(facts->getMarker(FLAG_GLB).getPosX() + 1,
				 facts->getMarker(FLAG_GLB).getPosY());
    //Point left_lower_right(pos.getX() - 1, pos.getY() + 1);
    Point left_lower_right(max(pos.getX() - 2,
			       facts->getMarker(FLAG_GLT).getPosX() + 2),
			   facts->getMarker(GOAL_L).getPosY() + 2);
    
    return generate_point_in_parallellogram(left_upper_left, left_upper_right,
					    left_lower_left, left_lower_right);
  } else if ( side == S_RIGHT ) {
    // The parallellogram of the right penalty area
    //Point right_upper_left(pos.getX() + 1, pos.getY() - 1);
    Point right_upper_left(min(pos.getX() + 2,
			       facts->getMarker(FLAG_GRT).getPosX() - 2),
			   facts->getMarker(GOAL_R).getPosY() - 2);
    static Point right_upper_right(facts->getMarker(FLAG_GRT).getPosX() - 1,
				   facts->getMarker(FLAG_GRT).getPosY());
    //Point right_lower_left(pos.getX() + 1, pos.getY() + 1);
    Point right_lower_left(min(pos.getX() + 2,
			       facts->getMarker(FLAG_GRT).getPosX() - 2),
			   facts->getMarker(GOAL_R).getPosY() + 2);
    static Point right_lower_right(facts->getMarker(FLAG_GRB).getPosX() - 1,
				   facts->getMarker(FLAG_GRB).getPosY());

    return generate_point_in_parallellogram(right_upper_left,
					    right_upper_right,
					    right_lower_left,
					    right_lower_right);
  } else {
    std::cout << "UNKNOWN SIDE!" << std::endl;
    return Point();
  }
}


Point
generate_point_in_penalty_area(const Side& side)
{
  static const WorldFacts* facts = WorldFacts::instance();

  if ( side == S_LEFT ) {
    // The limits of the left penalty area
    static Point left_upper_left(facts->getMarker(GOAL_L).getPosX(),
				 facts->getMarker(FLAG_PLT).getPosY());
    static Point left_lower_right(facts->getMarker(FLAG_PLB).getPosition());
    
    return generate_point_in_area(left_upper_left, left_lower_right);
  } else if ( side == S_RIGHT ) {
    // The limits of the right penalty area
    static Point right_upper_left(facts->getMarker(FLAG_PRT).getPosition());
    static Point right_lower_right(facts->getMarker(GOAL_R).getPosX(),
				   facts->getMarker(FLAG_PRB).getPosY());

    return generate_point_in_area(right_upper_left, right_lower_right);
  } else {
    std::cout << "UNKNOWN SIDE!" << std::endl;
    return Point();
  }
}


Point
generate_point_in_pitch()
{
  static const WorldFacts* facts = WorldFacts::instance();

  // The limits of the pitch
  static Point upper_left(facts->getMarker(FLAG_LT).getPosition());
  static Point lower_right(facts->getMarker(FLAG_RB).getPosition());

  return generate_point_in_area(upper_left, lower_right);
}


Point
generate_point_on_goal_line(const Side& side,
			    const Point& pos,
			    Float max_dist)
{
  static const WorldFacts* facts = WorldFacts::instance();

  // The left side goal line
  static Point upper_left(facts->getMarker(FLAG_GLT).getPosition());
  static Point lower_left(facts->getMarker(FLAG_GLB).getPosition());

  // The right side goal line
  static Point upper_right(facts->getMarker(FLAG_GRT).getPosition());
  static Point lower_right(facts->getMarker(FLAG_GRB).getPosition());

  if ( side == S_LEFT ) {
    Point res = generate_point_on_line(upper_left, lower_left);
    while ( abs(pos.getY() - res.getY()) > max_dist ) {
      res = generate_point_on_line(upper_left, lower_left);
    }
    return res;
  } else if ( side == S_RIGHT ) {
    Point res = generate_point_on_line(upper_right, lower_right);
    while ( abs(pos.getY() - res.getY()) > max_dist ) {
      res = generate_point_on_line(upper_right, lower_right);
    }
    return res;
  } else {
    std::cout << "UNKNOWN SIDE!" << std::endl;
    return Point();
  }
}


Vector
generate_vector_in_circle_segment(const AngleDeg& min_angle,
				  const AngleDeg& max_angle,
				  Float min_length,
				  Float max_length)
{
  return Vector(generate_direction_in_sector(min_angle, max_angle),
		generate_distance_in_interval(min_length, max_length));
}


Vector
generate_kickable_ball_vector()
{
  static const WorldFacts* facts = WorldFacts::instance();
  static Float max_kickable_dist = facts->SP_kickable_margin;
  static Float min_kickable_dist = facts->SP_ball_size + facts->SP_player_size;
  return generate_vector_in_circle_segment(-180, 180,
					   min_kickable_dist,
					   max_kickable_dist);
}


Vector
generate_kickable_ball_vector(const AngleDeg& min_angle,
			      const AngleDeg& max_angle)
{
  static const WorldFacts* facts = WorldFacts::instance();
  static Float max_kickable_dist = facts->SP_kickable_margin;
  static Float min_kickable_dist = facts->SP_ball_size + facts->SP_player_size;
  return generate_vector_in_circle_segment(min_angle, max_angle,
					   min_kickable_dist,
					   max_kickable_dist);
}


RS_END_NAMESPACE
