/* -*- Mode: C++ -*-
 * $Id: trainer_helpers.h,v 1.2 2005/08/18 20:22:33 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: trainer_helpers.h
 * Unitname: Framework/trainer
 * $Revision: 1.2 $
 * Created by: Fredrik Heintz 2004-08-31
 * Last modified by $Author: frehe $ $Date: 2005/08/18 20:22:33 $
 *
 *
 * HISTORY
 *
 * $Log: trainer_helpers.h,v $
 * Revision 1.2  2005/08/18 20:22:33  frehe
 * Improved the generate_point_on_goal_line function.
 *
 * Revision 1.1  2005/08/13 13:36:59  frehe
 * Readded the files since the trainer seems to have been deleted from the CVS.
 *
 * Revision 1.6  2004/09/29 20:24:23  frehe
 * Added support for three on three scenario.
 *
 * Revision 1.5  2004/09/03 09:42:21  frehe
 * Added new helper for generating defensive positions.
 *
 * Revision 1.4  2004/09/01 12:51:46  frehe
 * Fixed bug in comment.
 *
 * Revision 1.3  2004/09/01 12:50:01  frehe
 * Added point in scoring sector generator.
 *
 * Revision 1.2  2004/09/01 08:59:32  frehe
 * Added more helpers.
 *
 * Revision 1.1  2004/08/31 23:05:31  frehe
 * First version of the trainer helper functionc.
 *
 */

/**
 * @file trainer_helpers.h
 * Include file for the RoboSoc trainer utilities.
 *
 * @par Unit:
 *      Framework/trainer
 * $Revision: 1.2 $
 * $Author: frehe $
 * $Date: 2005/08/18 20:22:33 $
 */


#ifndef TRAINER_HELPERS_H
#define TRAINER_HELPERS_H


#include "library_types.h"
#include "soccer_types.h"


RS_BEGIN_NAMESPACE


/// Return true iff @a pos is in the penalty area on @a side
bool in_penalty_area(const Point& pos, const Side& side);


/// Initialize the random number generator with @a seed.
void init_random_number_generator(int seed);


/// Generate a random integer between @a min and @a max.
Int generate_integer_in_interval(Int min, Int max);


/// Generate a random distance between @a min_dist and @a max_dist.
Float generate_distance_in_interval(Float min_dist, Float max_dist);


/// Generate a random direction between @a min_angle and @a max_angle.
AngleDeg generate_direction_in_sector(const AngleDeg& min_angle,
				      const AngleDeg& max_angle);


/// Generate a random point on the lined defined by @a left and @a right.
Point generate_point_on_line(const Point& left, const Point& right);

/**
 * Generate a random point in area defined by @a upper_left
 * and @a lower_right.
 */
Point generate_point_in_area(const Point& upper_left,
			     const Point& lower_right);

/**
 * Generate a random point in the parallellogram defined
 * by upper_left, upper_right, lower_left, lower_right.
 */
Point generate_point_in_parallellogram(const Point& upper_left,
				       const Point& upper_right,
				       const Point& lower_left,
				       const Point& lower_right);

/**
 * Generate a random point in the scoring sector on @a side.
 * The scoring sector is defined as the area in the polygon
 * FLAG_PRT, FLAG_GRT - 1m x-dir, FLAG_GRB - 1m x-dir, FLAG_PRB,
 * for the right side of the field. The left is symmetric.
 */
Point generate_point_in_scoring_sector(const Side& side);

/**
 * Generate a random point in the defensice sector on @a side
 * when defending against a player at position @a pos.
 * The scoring sector is defined as the area in the polygon
 * pos + 1m x-dir -1m y-dir, FLAG_GRT - 1m x-dir,
 * FLAG_GRB - 1m x-dir, pos +1m x-dir +1m y-dir,
 * for the right side of the field. The left is symmetric.
 */
Point generate_point_in_defense_sector(const Side& side, const Point& pos);

/// Generate a random point in the penalty area on @a side.
Point generate_point_in_penalty_area(const Side& side);

/// Generate a random point in the pitch
Point generate_point_in_pitch();

/**
 * Generate a random point on the goalie line on @a side which
 * is at most @a max_dist meters in Y-direction from @a pos.
 */
Point generate_point_on_goal_line(const Side& side,
				  const Point& pos,
				  Float max_dist);


/**
 * Generate a random vector with an angle between @a min_angle and
 * @a max_angle with a length between @a min_length and @a max_length.
 */
Vector generate_vector_in_circle_segment(const AngleDeg& min_angle,
					 const AngleDeg& max_angle,
					 Float min_length,
					 Float max_length);

/// Generate a random ball vector which is kickable
Vector generate_kickable_ball_vector();

/**
 * Generate a random ball vector with an angle between @a min_angle
 * and @a max_angle where the ball is kickable.
 */
Vector generate_kickable_ball_vector(const AngleDeg& min_angle,
				     const AngleDeg& max_angle);


RS_END_NAMESPACE


#endif // TRAINER_HELPERS_H
