/* -*- Mode: C++ -*-
 * $Id: KickBallView.cc,v 1.15 2003/09/02 12:52:46 aip19 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: KickBallView.cc
 * Unitname: Framework/view
 * $Revision: 1.15 $
 * Created by: Fredrik Heintz 2000-09-12
 * Last modified by $Author: aip19 $ $Date: 2003/09/02 12:52:46 $
 *
 *
 * HISTORY
 *
 * $Log: KickBallView.cc,v $
 * Revision 1.15  2003/09/02 12:52:46  aip19
 * New primitive skills checked in, not tested.
 *
 * Revision 1.14  2003/08/27 09:25:43  petan
 * Fixed warnings at compile time
 *
 * Revision 1.13  2002/09/04 13:49:48  s02davro
 * *** empty log message ***
 *
 * Revision 1.12  2002/08/31 15:36:23  s02davro
 * Small modifications of views updated to the new world model
 *
 * Revision 1.11  2002/08/02 23:16:06  s02davro
 * Now works with the WorldModelInterface and new views
 *
 * Revision 1.10  2001/09/01 23:50:51  frehe
 * Updated the file for the new implementation of getValue and getAngle.
 *
 * Revision 1.9  2001/08/23 00:32:07  frehe
 * Removed dependencies.
 *
 * Revision 1.8  2000/10/28 20:22:02  mikbr212
 * changed initBall: now ballSpeed is retrieved from commandView if not known in ballView
 *
 * Revision 1.7  2000/10/26 16:00:45  mikbr212
 * Fixed bug in kickPowerRate: kickable_area to kickable_margin
 *
 * Revision 1.6  2000/10/09 18:46:15  mikbr212
 * Fixed serious wrong-calculations in getKickVector
 *
 * Revision 1.5  2000/10/09 10:17:44  mikbr212
 * added comments
 *
 * Revision 1.4  2000/10/08 20:05:01  mikbr212
 * Removed spam messages
 *
 * Revision 1.3  2000/10/08 12:57:28  mikbr212
 * removed comment
 *
 * Revision 1.2  2000/10/07 19:42:01  olalu526
 * Uncommented part using line.
 *
 * Revision 1.1  2000/10/07 14:29:25  mikbr212
 * created
 *
 */

/**
 * @file KickBallView.cc
 * Implementation file for the view class KickBallView.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.15 $
 * $Author: aip19 $
 * $Date: 2003/09/02 12:52:46 $
 */


#include "KickBallView.h"
//#include <math.h>

RS_BEGIN_NAMESPACE

/**
 * Constructor. Use getView instead.
 * @see getView
 */
KickBallView::KickBallView(const ViewId id)
  : View(id),
    facts(worldModelInterface->getWorldFacts()) {
}

KickBallView::~KickBallView() {
}

/**
 * Calculates the kickPowerRate for given ballPosition.
 * Should only be used when ball is inside kickable_area.
 */
FloatUE KickBallView::kickPowerRate()  {
  VectorUE ballVector = getRelativeVector();
  if (ballVector.isKnown())
    return facts->SP_kick_power_rate *
        (1.0 - 
	 0.25 * (abs(ballVector.getAngle().getDeg()) / 180.0) - 
	 0.25 * ((ballVector.getLength() 
			    - facts->SP_player_size 
			    - facts->SP_ball_size)
	 / facts->SP_kickable_margin));
  return FloatUE();
}

/**
 * This method receives the desired speed that the agent wants
 * to give the ball on a kick and returns the kick power
 * that needs to be supplied to the kick command in order
 * to achieve this.
 *
 * Note: The argument can be larger than the maximum kick power,
 * in this way it is possible to identify situations where
 * the ball cannot reach the given speed.
 * @param speed The speed the agent wants to reach
 * @return The argument to the kick command
 */
FloatUE KickBallView::getKickPowerForSpeed(FloatUE speed) {
  return speed / kickPowerRate();
}

/**
 * This method returns the speed that has to be giver to the ball
 * on a kick such that it has a remaining speed equal to destinationSpeed
 * after travelling a certain distance
 *
 * @param distance The distance the ball should travel
 * @param destinationSpeed The speed the ball should hold at the destination
 * @return The speed that should be given to the ball
 */
FloatUE KickBallView::getKickSpeedToTravel(FloatUE distance, FloatUE destinationSpeed) {
  double decayRatio = 1/facts->SP_ball_decay;

  double n = (log ((distance.getValue()*(decayRatio - 1))/destinationSpeed.getValue())+1) /
    log(decayRatio);

  return destinationSpeed / pow(facts->SP_ball_decay, n);
}

/**
 * Calculates the kick vector to get the desired ball speed.
 * @param desBallSpeed The desired ball speed
 * @param makKick Restricts how hard the kick can be.
 * @returns the kick vector, length is the ball acceleration.
 */

VectorUE KickBallView::getKickVector(const Vector& desBallSpeed,
				     Float def_maxKick)  {
  VectorUE ballVector = getRelativeVector();
  FloatUE maxKick = def_maxKick * kickPowerRate();
  VectorUE ballSpeed = getBall().getSpeedVector();
  VectorUE kickVec;
  
  if (!(ballVector.isUnknown() || 
	ballSpeed.isUnknown()))
    if((desBallSpeed - ballSpeed).getLength() > maxKick) {
      // I can't shoot to desired ball speed
      kickVec = -ballSpeed;
      FloatUE powerLeft = (maxKick*maxKick) - 
	(kickVec.getLength()*kickVec.getLength());
      if (powerLeft < 0) {
	// No power to kick the ball in the desired directoin.
	// Stop it.
      } else {
	kickVec += VectorUE(desBallSpeed.getAngle() + 
			    (desBallSpeed.getAngle() - ballSpeed.getAngle()),
			    sqrt(powerLeft));
      }
    } else {
      // Just kick!
      kickVec = desBallSpeed - ballSpeed;
    }  
  return kickVec;
}

/**
 * Checks if we can shoot to desired ball speed.
 * Checks if current ball speed is too high and if we will
 * kick through ourself.
 * @param desBallSpeed The desired ball speed.
 * @param maxKick Restricts how hard the kick should be.
 * @returns True if it is possible.
 */
bool KickBallView::canShootToSpeedVector(const Vector& desBallSpeed,
					 Float maxKick)  {
  VectorUE ballPos = getRelativeVector();
  VectorUE ballSpeed = getBall().getSpeedVector();

  if (ballPos.isKnown() && ballSpeed.isKnown())
    return false;
  
  if(ballPos.getLength() > facts->SP_kickable_area)
    return false;
  
  // Check if new ballPos will be inside player
  // Note: it is possible to shoot through myself
  if((ballPos + desBallSpeed).getLength() < (facts->SP_player_size
					     + facts->SP_ball_size))
    return false; 
  
  return ((desBallSpeed - ballSpeed).getLength() <= 
	  facts->SP_maxpower * kickPowerRate());
}

VectorUE KickBallView::getRelativeVector() const {
  return (getAgent().getPosition().getVectorTo(getBall().getPosition())).rotate
    (-getAgent().getBodyDirection());
}

const BallObject& KickBallView::getBall() const {
  return worldModelInterface->getWorld().getBall();
}

const AgentObject& KickBallView::getAgent() const {
  return worldModelInterface->getWorld().getAgent();
}

inline void KickBallView::myUpdateAfterCoachMessage() {}

inline void KickBallView::myNewCycle() {}

inline void KickBallView::myUpdateAfterInit() {}

inline void KickBallView::myUpdateAfterSee() {}

inline void KickBallView::myUpdateAfterHear() {}

inline void KickBallView::myUpdateAfterSense() {}

inline void KickBallView::myUpdateAfterCommand() {}

RS_END_NAMESPACE



