/* -*- Mode: C++ -*-
 * $Id: OffsideTrapView.h,v 1.8 2003/10/12 21:53:38 aip4 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: OffsideTrapView.h
 * Unitname: Framework/view
 * $Revision: 1.8 $
 * Created by: Mikael Asplund
 * Last modified by $Author: aip4 $ $Date: 2003/10/12 21:53:38 $
 *
 *
 * HISTORY
 *
 * $Log: OffsideTrapView.h,v $
 * Revision 1.8  2003/10/12 21:53:38  aip4
 * Commenting
 *
 * Revision 1.7  2003/10/12 21:50:38  aip4
 * Commenting
 *
 * Revision 1.6  2003/10/11 18:39:07  aip4
 * Comment update
 *
 * Revision 1.5  2003/10/11 17:37:06  aip4
 * Major update
 *
 * Revision 1.4  2003/10/11 14:44:27  aip4
 * unsuitable update
 *
 * Revision 1.3  2003/10/11 13:51:37  aip4
 * unsuitable stup
 *
 * Revision 1.2  2003/10/11 13:45:47  aip4
 * More than one to put pressure
 *
 * Revision 1.1  2003/09/29 01:17:30  aip4
 * Update after deletion
 *
 * Revision 1.1  2002/10/22 14:42:16  aip4
 * Offside view created, only basic testing performed
 *
 */

/**
 * @file OffsideTrapView.h
 * Include file for the view class OffsideTrapView. 
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.8 $
 * $Author: aip4 $
 * $Date: 2003/10/12 21:53:38 $
 */


#ifndef OFFSIDETRAPVIEW_H
#define OFFSIDETRAPVIEW_H


#include "View.h"
#include "AgentView.h"
#include "BallView.h"
#include "PlayerView.h"
#include "view_constants.h"
#include "WorldModelInterface.h"

RS_BEGIN_NAMESPACE

class BallView;
class AgentView;

/**
 * @class OffsideTrapView OffsideTrapView.h
 * This view is intended be used to spot situations where an offside trap
 * is suitable. It can be used by all players. It is used by the skills 
 * InitiateOffsideTrap and ExecuteOffsideTrap.
 *
 * @todo 
 *
 * - The member canPressureOpponent should be implemented, as for
 *   now it returns always true.
 *
 * - suitable can be refined to handle take the speed vector of the ball
 *   in to account.
 *
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.8 $
 * $Author: aip4 $
 * $Date: 2003/10/12 21:53:38 $
 */
class OffsideTrapView : public View {
public:

  /** 
   * Construct a view with the ViewId @a id.
   * @param id The view id.
   * @param maxTrapDist This is the maximum distance a defender should 
   * run to make an offside trap.
   * The default value is 25 meters.
   */
  OffsideTrapView(const ViewId id, Float maxTrapDist=25.0);

  /** Destroys the view. */
  ~OffsideTrapView();

  /** 
   * Checks whether there are any defenders that can put 
   * pressure on the ball keeper. 
   * @return True if there are defenders that can pressure the ball keeper.
   */  
  bool canPressureOpponent();

  /** 
   * Get the players that can put pressure on the
   * ball keeper 
   * @param pressurePlayers A vector that will be filled with up to two 
   * PlayerObjects. Both players will have a x-coordinate with is smaller than
   * that of the ball. The first player will have a y-coordinate which is larger than
   * that of the ball and the second will have a y-coordinate which is smaller than
   * that of the ball.
   */
  void getPlayersToSetPress(std::vector<PlayerObject>& pressurePlayers);

  /** 
   * Checks if opponent is ball keeper. 
   * @return True if the opponent has got the ball.
   */
  bool opponentHasBall();

  /** 
   * Get the position of the opponent closest to our goal.
   * @return The x-coordinate of the opponent closest to our goal.
   */
  FloatUE getOpponentPosClosestToGoal();

  /** 
   * Get the second last defender (including the goalie) 
   * @return The x-coordinate of the teammate next closest to our goal
   * (including the goalie).
   */
  FloatUE getTeammatePosNextClosestToGoal();
  
  /** 
   * Checks whether it is suitbale to start an offside trap. This is 
   * considered suitable if these conditions are met:
   * 
   * - There is an opponent on our half of the field
   * - The ball is behind the opponent which is closest to our goal
   * - The ball is controlled by an opponent
   * - Our second last defender is close enough to the x-coordinate 
   *   to pass
   * - The opponent is not already offside
   *
   * @return True if it is suitable to start an offside trap.
   */
  bool suitable();
  
  /** 
   * Checks wheter it is unsuitable to continue with the trap.
   * It is considered unsuitable if one of these conditions are met:
   *
   * - The ball has pass the given x-coordinate
   * - One defender has more than #maxTrapDistance to run.
   *
   * @param xcoord X-coordinate which ball must not have passed.
   * @return True if it is unsuitable to continue with the trap.
   */
  bool unsuitable(int xcoord);
	 
protected:  
  virtual void myNewCycle();
  virtual void myUpdateAfterCoachMessage();
  virtual void myUpdateAfterInit();
  virtual void myUpdateAfterSee();
  virtual void myUpdateAfterHear();
  virtual void myUpdateAfterSense();
  virtual void myUpdateAfterCommand();
  
  const WorldFacts* facts;
  
  /** The agents view of other players */
  PlayerView *playerView;

  /** The agents view of the ball */
  BallView *ballView;

  /** The agents view of itself */
  AgentView *agentView;  
private:
  // Hide copy-constructor and assignment.
  OffsideTrapView(const OffsideTrapView&);
  OffsideTrapView& operator= (const OffsideTrapView&);  

  // Max distance that a defender will run in an offside trap
  Float maxTrapDistance;

};

RS_END_NAMESPACE


#endif // !OFFSIDETRAPVIEW_H
