/* -*- Mode: C++ -*-
 * $Id: PassView.h,v 1.14 2004/01/12 12:01:11 aip10 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: PassView.h
 * Unitname: Framework/view
 * $Revision: 1.14 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: aip10 $ $Date: 2004/01/12 12:01:11 $
 *
 *
 * HISTORY
 *
 * $Log: PassView.h,v $
 * Revision 1.14  2004/01/12 12:01:11  aip10
 * FieldStrategy.cc
 *
 * Revision 1.13  2003/12/02 13:34:59  frehe
 * Removed member opponent and used a const reference directly when needed.
 *
 * Revision 1.12  2003/12/01 19:56:22  aip10
 * MoveWithBallToPosition removed and added to skill GoToPassingPosition
 *
 * Revision 1.11  2003/10/12 19:17:59  aip10
 * documentation updated
 *
 * Revision 1.10  2003/10/12 17:04:40  aip10
 * New evaluations added
 *
 * Revision 1.9  2003/10/12 14:05:23  aip10
 * Modified pointers
 *
 * Revision 1.8  2003/10/11 10:41:00  aip10
 * Move To Better position implemented
 *
 * Revision 1.7  2003/10/07 09:13:10  aip19
 * Fixed linkage errors and compile errors
 *
 * Revision 1.6  2003/10/06 12:12:18  aip10
 * Added secondary evaluation
 *
 * Revision 1.5  2003/09/28 21:55:59  frehe
 * Removed virtual functions without implementation.
 *
 * Revision 1.4  2003/09/28 21:21:56  aip10
 * Modified the virtual functions
 *
 * Revision 1.3  2003/09/28 21:15:42  aip10
 * Virtual functions modified
 *
 * Revision 1.2  2003/09/28 20:39:34  frehe
 * Made minor changes to make it compile without warnings.
 *
 * Revision 1.1  2003/09/28 20:00:24  aip10
 * PassView.cc
 *
 * Revision 1.1  2003/09/22 18:00:00  ginba729
 * Small changes in views
 *
 */

/**
 * @file PassView.h
 * Include file for the class PassView.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.14 $
 * $Author: aip10 $
 * $Date: 2004/01/12 12:01:11 $
 */


#ifndef PASS_VIEW_H
#define PASS_VIEW_H

#include "View.h"
#include "AgentView.h"
#include "PlayerView.h"
#include "WorldFacts.h"
#include "BallView.h"
#include "PlayerObject.h"
#include "../strategy/FieldStrategy.h"
#include "Plan.h"


RS_BEGIN_NAMESPACE


/**
 * @class PassView PassView.h
 * This view is used to keep track of the ball.
 *
 * @todo Is the view needed?
 * @todo Information from kickview could be included.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.14 $
 * $Author: aip10 $
 * $Date: 2004/01/12 12:01:11 $
 */
typedef struct {
  float value;
  PlayerObject *player;
} EvaluatedPlayer;

typedef std::vector <EvaluatedPlayer> SortedVector;
typedef std::vector <PlayerObject> PlayerVector;

static const float Area = 10;
static const float threshold = 0.5;
static const float moveAngle = 10;
static const float moveDistance = 1;

class PassView : public View {
public:
  /** Create a pass view with ViewId @a id */
  PassView(const ViewId id);

  
  /** Destroys the view */ 
  PassView::~PassView();
  
  /**
   * isAvailablePlayer()
   * @return: bool value, true if teammate available to pass to, else false
   */
  bool isAvailablePlayer();
	
  /**
   * passPartner()
   * @param: PlayerObject bestPartner
   * @param: FieldStrategy current Strategy
   * @return: bestPartner points to best pass partner, return float evaluation
   * of this partner between 0..1 where 1==perfect partner.
   */
  float passPartner(PlayerObject **bestPartner, FieldStrategy *strategy=NULL);

  /** Finds the best position where the player should move, that it would be better to pass to pass partner.
   * moveToBetterPosition()
   * @param: Takes a PlayerOject* to evaluate
   * @return: coordinates where to move
   *  
   */
  Point moveToBetterPosition(PlayerObject *player);

  
protected:
  /**
   * primaryEvaluation()
   * @return: returns a sorted cector of PlayerObjects with the most
   *  desirable at the front
   */
  SortedVector primaryEvaluation();

  /** secondaryEvaluation()evaluates in a thorough evaluation the best passpartner from a set of candidates.
   * set to best pass partner and returns a float evaluation of	this partner between 0..1 where 1==perfect partner. 
   * @param: Takes a sorted PlayerVector and a PlayerObject.	
   * @return: float the value of the best player's evaluation. */
  float secondaryEvaluation(SortedVector candidates, PlayerObject **bestPlayer);
	
  /**
   * playerEvaluation()	
   * @param: Takes a PlayerObject* to evaluate	
   * @return: returns a float evaluation of PlayerObject
   * between 0..1 where 1==perfect partner.
   */
  const float playerEvaluation(PlayerObject *player);

  /**
   * fieldGain()
   * @param: Takes a PlayerOject* to evaluate
   * @param: Point reference position ( usually the position of the player willing to pass).
   * @return: returns a normalised projection of the vector from 'this' player
   *  to PlayerObject *player on the vector from 'this' player to the opposition goal
   */
  float fieldGain(PlayerObject *player, Point position);

  /** markedEval()checks upon how well a certain player is marked.
   * The marking is determined by how many opponents are in the vincinity of a player and how close they are.
   * @param: PlayerObject player* to evaluate.
   * @param: bool Thorough if flag is set, the angle between opponent and vector from agent to pass partner is also considered.
   * @param: Point reference position ( usually the position of the player willing to pass).
   * @return: returns a number from 0 to 1 which indicates how close are enemies to the player. */
  
  
   float markedEval(PlayerObject *player, bool Thorough, Point position);

   /** partnerBodyDirection()checks how well the player is aligend to the traverse direction of the ball.
    * @param PlayerObject the player to be evaluated.
    * @param the anlge difference between body and balldirection, scaled between 0...1 . */
   float partnerBodyDirection(PlayerObject *player);

  
  /** partnerViewDivergence()checks how well a player looks into the direction where from a pass will come.
   * @param PlayerObject the player to be evaluated.
   * @return float angledifference between view direction and source-of-pass directions, caled between 0...1 . */
   
	float partnerViewDivergence(PlayerObject *player);


 
 /** strategyModel()
   * checks upon how desirable a pass to a player in a certain zone is.
   *  the size of the zones and the desirabilities are stored in the current strategy model.
   * @param: Takes a PlayerOject* to evaluate
   * @return: returns a number from 0 to 1 which indicates the desirability of a pass from our 'zone' to the.
   *  'zone' of the PlayerObject* according the the current FieldStrategy *strategy. A return of 1 is most desirable. */
   
  float strategyModel(PlayerObject *player);



  //@{
  /** @name Update functions. */

  virtual void myNewCycle() {}
  virtual void myUpdateAfterCoachMessage() {}
  virtual void myUpdateAfterInit() {}
  virtual void myUpdateAfterSee() {}
  virtual void myUpdateAfterHear() {}
  virtual void myUpdateAfterSense() {}
  virtual void myUpdateAfterCommand() {}

  //@}


  /**
   * ATTRIBUTES	
   */
  PlayerView *playerView;

  /** A pointer to the agent view. */
  AgentView* agentView;

  /** A pointer to the ball view */
  BallView *ballView;

  /** The known facts of the world. (old ParamenterView) */
  const WorldFacts* facts;

  /** A pointer to field strategy */
  FieldStrategy *strategy;

  //PlayerVector opponents;

private:

  /** Hidden copy-constructor. */
  PassView(const PassView&);

  /** Hidden assignment operator. */
  PassView& operator= (const PassView&);  
  
  
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "PassView.icc"
#endif


#endif // !PASS_VIEW_H
