/* -*- Mode: C++ -*-
 * $Id: AgentStep.cc,v 1.5 2004/08/18 12:00:30 petan Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: AgentStep.cc
 * Unitname: Library
 * $Revision: 1.5 $
 * Created by: Peter Andersson 2004-08-11
 * Last modified by $Author: petan $ $Date: 2004/08/18 12:00:30 $
 *
 *
 * HISTORY
 *
 * $Log: AgentStep.cc,v $
 * Revision 1.5  2004/08/18 12:00:30  petan
 * Updated with a print function and some gets
 *
 * Revision 1.4  2004/08/17 16:10:17  frehe
 * Fixed numerous memory leaks and problems.
 *
 * Revision 1.3  2004/08/16 13:11:41  petan
 * Made functions virtual to seamlessly transfer command to AgentStep
 *
 * Revision 1.2  2004/08/13 15:37:58  petan
 * Fixed potential memory leak in merge
 *
 * Revision 1.1  2004/08/13 13:53:42  petan
 * Added AgentStep as a simpler alternative to Step
 *
 */

/**
 * @file AgentStep.cc
 * Implementation file for the class AgentStep.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.5 $
 * $Author: petan $
 * $Date: 2004/08/18 12:00:30 $
 */

#include <algorithm>
#include <functional>
#include "AgentStep.h"
#include "ServerCommand.h"
#include "rs_assert.h"

RS_BEGIN_NAMESPACE


struct add_command : public std::binary_function<AgentStep*, const Command*, void> {
  void operator()(AgentStep* step, const Command* const command) const
  {
    if ( command != 0 ) {
      step->addCommand(command->clone());
    }
  }
};


// --- CONSTRUCTORS ---

AgentStep::AgentStep()
  : Step(),
    mBodyCommand(0),
    mNeckCommand(0),
    mSayCommand(0),
    mPointToCommand(0),
    mChangeModeCommand(0),
    mChangeViewCommand(0)
{
}


AgentStep::AgentStep(const AgentStep& obj)
  : Step(obj),
    mBodyCommand(0),
    mNeckCommand(0),
    mSayCommand(0),
    mPointToCommand(0),
    mChangeModeCommand(0),
    mChangeViewCommand(0)
{
  *this = obj;
}


// --- DESTRUCTOR ---

AgentStep::~AgentStep()
{
  // Clean up the old data
  makeEmpty();
}


// --- OPERATORS ---

AgentStep& AgentStep::operator=(const AgentStep& obj)
{
  if ( &obj != this ) {
    // Clean up the old data
    makeEmpty();

    // Copy the new data from obj
    if ( obj.mBodyCommand != 0 ) {
      mBodyCommand = obj.mBodyCommand->clone();
    }

    if ( obj.mNeckCommand != 0 ) {
      mNeckCommand = obj.mNeckCommand->clone();
    }

    if ( obj.mSayCommand != 0 ) {
      mSayCommand = obj.mSayCommand->clone();
    }

    if ( obj.mPointToCommand != 0 ) {
      mPointToCommand = obj.mPointToCommand->clone();
    }

    if ( obj.mChangeModeCommand != 0 ) {
      mChangeModeCommand = obj.mChangeModeCommand->clone();
    }

    if ( obj.mChangeViewCommand != 0 ) {
      mChangeViewCommand = obj.mChangeViewCommand->clone();
    }

    for_each(obj.commandSequence.begin(),
	     obj.commandSequence.end(),
	     bind1st(add_command(), this));
  }

  return *this;
}


// --- OTHER METHODS ---

Command* AgentStep::getNextCommand() const
{
  Command* cmd = 0;
  if (mBodyCommand != 0) {
    cmd = mBodyCommand->clone();
  }
  else if (mNeckCommand != 0) {
    cmd = mNeckCommand->clone();
  }
  else if (mSayCommand != 0) {
    cmd = mSayCommand->clone();
  }
  else if (mPointToCommand != 0) {
    cmd = mPointToCommand->clone();
  }
  else if (mChangeModeCommand != 0) {
    cmd = mChangeModeCommand->clone();
  }
  else if (mChangeViewCommand != 0) {
    cmd = mChangeViewCommand->clone();
  }
  else if ( !commandSequence.empty() ) {
    cmd = commandSequence.front()->clone();
  }
  return cmd;
}

AgentStep& AgentStep::removeNextCommand()
{
  if (mBodyCommand != 0) {
    delete mBodyCommand;
    mBodyCommand = 0;
  }
  else if (mNeckCommand != 0) {
    delete mNeckCommand;
    mNeckCommand = 0;
  }
  else if (mSayCommand != 0) {
    delete mSayCommand;
    mSayCommand = 0;
  }
  else if (mPointToCommand != 0) {
    delete mPointToCommand;
    mPointToCommand = 0;
  }
  else if (mChangeModeCommand != 0) {
    delete mChangeModeCommand;
    mChangeModeCommand = 0;
  }
  else if (mChangeViewCommand != 0) {
    delete mChangeViewCommand;
    mChangeViewCommand = 0;
  }
  else if ( !commandSequence.empty() ) {
    Command* cmd = commandSequence.front();
    commandSequence.pop_front();
    delete cmd;
  }
  return *this;
}

AgentStep& AgentStep::makeEmpty()
{
  while ( !empty() ) {
    removeNextCommand();
  }
  return *this;
}

void AgentStep::addMiscellaneousCommand(const Command* const command)
{
  addCommand(command);
}

void AgentStep::setBodyCommand(const TurnCommand* const command)
{
  if ( command != 0 ) {
    if (mBodyCommand != 0) delete mBodyCommand;
    mBodyCommand = command->clone();
  }
}

void AgentStep::setBodyCommand(const DashCommand* const command)
{
  if ( command != 0 ) {
    if (mBodyCommand != 0) delete mBodyCommand;
    mBodyCommand = command->clone();
  }
}

void AgentStep::setBodyCommand(const KickCommand* const command)
{
  if ( command != 0 ) {
    if (mBodyCommand != 0) delete mBodyCommand;
    mBodyCommand = command->clone();
  }
}

void AgentStep::setBodyCommand(const CatchCommand* const command)
{
  if ( command != 0 ) {
    if (mBodyCommand != 0) delete mBodyCommand;
    mBodyCommand = command->clone();
  }
}

void AgentStep::setBodyCommand(const TackleCommand* const command)
{
  if ( command != 0 ) {
    if (mBodyCommand != 0) delete mBodyCommand;
    mBodyCommand = command->clone();
  }
}

void AgentStep::setBodyCommand(const MoveCommand* const command)
{
  if ( command != 0 ) {
    if (mBodyCommand != 0) delete mBodyCommand;
    mBodyCommand = command->clone();
  }
}

void AgentStep::setNeckCommand(const TurnNeckCommand* const command)
{
  if ( command != 0 ) {
    if (mNeckCommand != 0) delete mNeckCommand;
    mNeckCommand = command->clone();
  }
}

void AgentStep::setSayCommand(const SayCommand* const command)
{
  if ( command != 0 ) {
    if (mSayCommand != 0) delete mSayCommand;
    mSayCommand = command->clone();
  }
}
  
void AgentStep::setPointToCommand(const PointToCommand* const command)
{
  if ( command != 0 ) {
    if (mPointToCommand != 0) delete mPointToCommand;
    mPointToCommand = command->clone();
  }
}

void AgentStep::setChangeViewCommand(const ChangeViewCommand* const command)
{
  if ( command != 0 ) {
    if (mChangeViewCommand != 0) delete mChangeViewCommand;
    mChangeViewCommand = command->clone();
  }
}

void AgentStep::setChangeModeCommand(const ChangeModeCommand* const command)
{
  if ( command != 0 ) {
    if (mChangeModeCommand != 0) delete mChangeModeCommand;
    mChangeModeCommand = command->clone();
  }
}

void AgentStep::merge(const AgentStep& agentStep)
{
  if (agentStep.mBodyCommand != 0) {
    if (mBodyCommand != 0) {
      delete mBodyCommand;
    }
    mBodyCommand = agentStep.mBodyCommand->clone();
  }
  if (agentStep.mNeckCommand != 0) {
    if (mNeckCommand != 0) {
      delete mNeckCommand;
    }
    mNeckCommand = agentStep.mNeckCommand->clone();
  }
  if (agentStep.mSayCommand != 0) {
    if (mSayCommand != 0) {
      delete mSayCommand;
    }
    mSayCommand = agentStep.mSayCommand->clone();
  }
  if (agentStep.mPointToCommand != 0) {
    if (mPointToCommand != 0) {
      delete mPointToCommand;
    }
    mPointToCommand = agentStep.mPointToCommand->clone();
  }
  if (agentStep.mChangeModeCommand != 0) {
    if (mChangeModeCommand != 0) {
      delete mChangeModeCommand;
    }
    mChangeModeCommand = agentStep.mChangeModeCommand->clone();
  }
  if (agentStep.mChangeViewCommand != 0) {
    if (mChangeViewCommand != 0) {
      delete mChangeViewCommand;
    }
    mChangeViewCommand = agentStep.mChangeViewCommand->clone();
  }

  // Copy miscellaneous commands
  for_each(agentStep.commandSequence.begin(),
	   agentStep.commandSequence.end(),
	   bind1st(add_command(), this));
}

void AgentStep::print() const
{
  std::cout << "AgentStep:" << std::endl;
  if (mBodyCommand != 0)
    std::cout << mBodyCommand->serverString() << std::endl;
  if (mNeckCommand != 0)
    std::cout << mNeckCommand->serverString() << std::endl;
  if (mSayCommand != 0)
    std::cout << mSayCommand->serverString() << std::endl;
  if (mPointToCommand != 0)
    std::cout << mPointToCommand->serverString() << std::endl;
  if (mChangeModeCommand != 0)
    std::cout << mChangeModeCommand->serverString() << std::endl;
  if (mChangeViewCommand != 0)
    std::cout << mChangeViewCommand->serverString() << std::endl;

  // FIXME: print stuff in the commandSequence too

  std::cout.flush();
}


RS_END_NAMESPACE


