/* -*- Mode: C++ -*-
 * $Id: AgentStep.h,v 1.5 2004/08/18 12:00:30 petan Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: AgentStep.h
 * Unitname: Library
 * $Revision: 1.5 $
 * Created by: Peter Andersson 2004-08-11
 * Last modified by $Author: petan $ $Date: 2004/08/18 12:00:30 $
 *
 *
 * HISTORY
 *
 * $Log: AgentStep.h,v $
 * Revision 1.5  2004/08/18 12:00:30  petan
 * Updated with a print function and some gets
 *
 * Revision 1.4  2004/08/17 16:10:17  frehe
 * Fixed numerous memory leaks and problems.
 *
 * Revision 1.3  2004/08/16 14:15:08  frehe
 * Fixed incorrectly ordered const.
 *
 * Revision 1.2  2004/08/16 14:09:43  petan
 * Added getBodyCommand
 *
 * Revision 1.1  2004/08/13 13:53:42  petan
 * Added AgentStep as a simpler alternative to Step
 *
 *
 */

/**
 * @file AgentStep.h
 * Include file for the class AgentStep.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.5 $
 * $Author: petan $
 * $Date: 2004/08/18 12:00:30 $
 */


#ifndef AGENTSTEP_H
#define AGENTSTEP_H


#include <deque>

#include "machine_dependent_compiler_directives.h"
#include "TurnCommand.h"
#include "DashCommand.h"
#include "KickCommand.h"
#include "CatchCommand.h"
#include "TackleCommand.h"
#include "MoveCommand.h"

#include "TurnNeckCommand.h"
#include "SayCommand.h"
#include "PointToCommand.h"
#include "ChangeModeCommand.h"
#include "ChangeViewCommand.h"

#include "Command.h"
#include "Step.h"

RS_BEGIN_NAMESPACE

/**
 * @class AgentStep AgentStep.h
 * Used to represent a step in a plan. A step is a set of commands
 * that should and could be executed by the soccer server and
 * the agent in a single simulator cycle.
 *
 * <!--@pre The preconditions for constructing the class are ...
 * @post The postconditions after constructing the class are...
 * @invariant A step contains at least one command.
 *
 * @todo Things that should be done... -->
 *
 * @par Unit:
 *      Library
 * $Revision: 1.5 $
 * $Author: petan $
 * $Date: 2004/08/18 12:00:30 $
 */
class AgentStep : public Step {
public:
  /** The default constructor. 
   *  The step will take control of the command so don't delete it.
   *  If @a command is NULL the create an empty
   *  step, otherwise create a step with the @a command command. */
  AgentStep();

  /** The copy constructor. */
  AgentStep(const AgentStep& obj);

  /** The destructor. */
  virtual ~AgentStep();

  /** The assignment operator.
   *  @param obj is the object to be assigned from.
   *  @return A reference to the the newly assigned object.
   *  @post The object is an exact copy of @a obj. */
  AgentStep& operator=(const AgentStep& obj);

  /** Adds a miscellaneous command to the step. */
  void addMiscellaneousCommand(const Command* const command);

  /** Adds a body command to the step, only one body command
   *  can be executed in a step. The command will replace
   *  prior body commands.
   */
  void setBodyCommand(const TurnCommand* const command);
  void setBodyCommand(const DashCommand* const command);
  void setBodyCommand(const KickCommand* const command);
  void setBodyCommand(const CatchCommand* const command);
  void setBodyCommand(const TackleCommand* const command);
  void setBodyCommand(const MoveCommand* const command);

  /** The AgentStep will still own the command, so do not
      delete */
  ServerCommand* getBodyCommand() const;

  /** Adds a neck command to the step, it will replace prior
   *  neck commands 
   */
  void setNeckCommand(const TurnNeckCommand* const command);

  /** Gets the neck command as set in this object. 0 if no
      command has been set. */
  TurnNeckCommand* getNeckCommand() const;

  /** Adds a say command to the step, it will replace prior
   *  say commands
   */
  void setSayCommand(const SayCommand* const command);
  
  /** Adds a point to command to the step, it will replace prior
   *  point to commands
   */
  void setPointToCommand(const PointToCommand* const command);

  /** Adds a change mode command to the step, it will replace prior
   *  change mode commands
   */
  void setChangeViewCommand(const ChangeViewCommand* const command);

  /** Adds a change view command to the step, it will replace prior
   *  change view commands
   */
  void setChangeModeCommand(const ChangeModeCommand* const command);

  /** Merges this object with the supplied agentStep, all variables
   *  supplied with agentStep will override variables defined in this
   *  object.
   */
  void merge(const AgentStep& agentStep);

  /** Get a copy of the first command in the sequence.
   *  Remember to delete the command object otherwise you will
   *  have a memory leak.
   *  If the step is empty then NULL is returned. */
  Command* getNextCommand() const;

  /** Remove the first command in the sequence, if the step is not empty. */
  AgentStep& removeNextCommand();
 
  /** Test if the step contains no commands.
   *  @returns @c true iff the step is empty. */
  bool empty() const;

  /** Removes all the commands stored in the step. */
  AgentStep& makeEmpty();

  /** Prints the AgentStep to std::cout */
  void print() const;

protected:
  ServerCommand *mBodyCommand;
  TurnNeckCommand *mNeckCommand;
  SayCommand *mSayCommand;
  PointToCommand *mPointToCommand;
  ChangeModeCommand *mChangeModeCommand;
  ChangeViewCommand *mChangeViewCommand;

private:

};


RS_END_NAMESPACE

  
/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "AgentStep.icc"
#endif


#endif // !STEP_H
