/* -*- Mode: C++ -*-
 * $Id: AngleDegreesError.h,v 1.4 2005/09/19 17:00:06 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: AngleDegreesError.h
 * Unitname: Library
 * $Revision: 1.4 $
 * Created by: David Rosn 2002-06-19
 * Last modified by $Author: frehe $ $Date: 2005/09/19 17:00:06 $
 *
 *
 * HISTORY
 *
 * $Log: AngleDegreesError.h,v $
 * Revision 1.4  2005/09/19 17:00:06  frehe
 * Fixed bugs and made WithError work.
 *
 * Revision 1.3  2005/08/13 13:31:17  frehe
 * Fixed several errors to make it compile with gcc-3.4.3
 *
 * Revision 1.2  2002/08/31 14:50:26  s02davro
 * Modifications, corrections of angle classes
 *
 * Revision 1.1  2002/08/02 21:31:23  s02davro
 * Created the class to be able to represent AngleDegree with a error interval.
 *
 *
 */

/**
 * @file AngleDegreesError.h
 * Include file for the class CAngleDegreesError.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.4 $
 * $Author: frehe $
 * $Date: 2005/09/19 17:00:06 $
 */


#ifndef ANGLEDEGREESERROR_H
#define ANGLEDEGREESERROR_H

#include <WithError.h>
#include <AngleDegrees.h>

/** String to use as the header when outputing a AngleDegreesError value. */
#define ANGLEERROR_HEAD std::string("#AErr")

RS_BEGIN_NAMESPACE

/**
 * @class AngleDegreesError AngleDegreesError.h
 * Template class for representing angles with an error value.
 * The template argument is passed to AngleDegrees.
 *
 * @todo Take a think about whether it's good to add a getValue method.
 * @todo Check if all constructors are needed. 
 * @todo Check error values in constructors and during operations.
 * @todo Handle error with nonlinjar functions (sin, cos ...).
 *
 * @par Unit:
 *      Library
 * $Revision: 1.4 $
 * $Author: frehe $
 * $Date: 2005/09/19 17:00:06 $
 */

template <typename T> 
class AngleDegreesError : public WithError<AngleDegrees<T> > {
public:    
  /** Default constructor, the value and errors is AngleDegrees<T>(). */
  AngleDegreesError();

  /** Construct a AngleDegreesError with no error and value ang. */
  AngleDegreesError(const T& ang);

  /** Construct a AngleDegreesError from a WithError object.
   *  The three values from the WithError object is used to contruct
   *  three AngleDegrees object. */
  AngleDegreesError(const WithError<T>& ang);

  /** Construct a AngleDegreesError from a AngleDegree object, no error. */  
  AngleDegreesError(const AngleDegrees<T>& ang);

  /** Construct a object based on the value and error in ang */
  AngleDegreesError(const WithError<AngleDegrees<T> >& ang);

  /** Copyconstructor for the class */
  AngleDegreesError(const AngleDegreesError<T>& ang);

  /** Construct a object using a basevalue and a specific type of error.      
   *  @see WithError for more information. */
  AngleDegreesError(const ErrorType c, const T& value, 
		    const Float64 quant_step1 = 0.1, const T& distance = T(),
		    const Float64 quant_step2 = 0.02);
  
  /** Construct a object using specified value and errors. */
  AngleDegreesError(const T& value, const T& plus, const T& minus);
  
  /** Assignment operator. */
  AngleDegreesError<T>& operator= (const AngleDegreesError<T>& op);

  /** @name Arithmetic operators. */
  //@{
  AngleDegreesError<T>& operator+= (const AngleDegreesError<T>& op);
  AngleDegreesError<T>& operator-= (const AngleDegreesError<T>& op);
  AngleDegreesError<T>& operator*= (const AngleDegreesError<T>& op);
  AngleDegreesError<T>& operator/= (const AngleDegreesError<T>& op);
  AngleDegreesError<T> operator+ (const AngleDegreesError<T>& op) const;
  AngleDegreesError<T> operator- (const AngleDegreesError<T>& op) const;
  AngleDegreesError<T> operator* (const AngleDegreesError<T>& op) const;
  AngleDegreesError<T> operator/ (const AngleDegreesError<T>& op) const;
  AngleDegreesError<T> operator- () const;  
  //@}
  
  /** @name Increment and decrement operators. */
  //@{
  AngleDegreesError<T>& operator++ ();
  AngleDegreesError<T>& operator-- ();
  //@}

  /** @name Comparison operators. */
  //@{
  bool operator== (const AngleDegreesError<T>& op) const;
  bool operator!= (const AngleDegreesError<T>& op) const;
  bool operator< (const AngleDegreesError<T>& op) const;
  bool operator<= (const AngleDegreesError<T>& op) const;
  bool operator> (const AngleDegreesError<T>& op) const;
  bool operator>= (const AngleDegreesError<T>& op) const;
  bool operator!  () const;
  //@}

  /** Improve the value using another value with error.
   *  @see WithError for more information. */
  AngleDegreesError<T>& improve (const AngleDegreesError<T>& op);

  /** Disimprove the value using another value with error. 
   *  @see WithError for more information. */
  AngleDegreesError<T>& disimprove (const AngleDegreesError<T>& op);

  /** Get the angle in radians. */
  T getDeg() const;

  /** Get the angle in degrees. */
  T getRad() const;

  /** Get the angle striped from error information. */
  AngleDegrees<T> getAngle() const;

#if 0  
  /** Get the minimum angle of the object. */
  AngleDegrees<T> getMin() const;

  /** Get the maximum angle of the object. */
  AngleDegrees<T> getMax() const;

  /** Get the plus angle, getAngle + getPlus = getMax */
  AngleDegrees<T> getPlus() const;

  /** Get the minus angle, getAngle + getMinus = getMin */
  AngleDegrees<T> getMinus() const;
#endif
};

template<typename T>
AngleDegreesError<T>::AngleDegreesError() {
}

template<typename T>
AngleDegreesError<T>::AngleDegreesError(const T& def_value) 
  : WithError<AngleDegrees<T> >(def_value) {
}

template<typename T>
AngleDegreesError<T>::AngleDegreesError(const WithError<T>& ang) 
  : WithError<AngleDegrees<T> >(AngleDegrees<T>(ang.getValue()),
				AngleDegrees<T>(ang.getPlus()),
				AngleDegrees<T>(ang.getMinus()), T()) {
}

template<typename T>
AngleDegreesError<T>::AngleDegreesError(const AngleDegrees<T>& def_value) 
  : WithError<AngleDegrees<T> >(def_value) {
}

template<typename T>
AngleDegreesError<T>::AngleDegreesError(const WithError<AngleDegrees<T> >& def_value) 
  : WithError<AngleDegrees<T> >(def_value) {
}

template <typename T>
AngleDegreesError<T>::AngleDegreesError(const AngleDegreesError<T>& def_value) 
  : WithError<AngleDegrees<T> >(def_value) {
}

template <typename T>
AngleDegreesError<T>::AngleDegreesError(const ErrorType c, 
					const T& def_value, 
					const Float64 quant_step1, 
					const T& def_distance,
					const Float64 quant_step2)
  : WithError<AngleDegrees<T> >(c, def_value, quant_step1, def_distance, quant_step2) {
  /*
  WithError<T> tmp(c, def_value, quant_step1, def_distance, quant_step2);
  WithError<AngleDegrees<T> > tmp2(AngleDegrees<T>(tmp.getValue()),
				   AngleDegrees<T>(tmp.getPlus()),
				   AngleDegrees<T>(tmp.getMinus()));
  WithError<AngleDegrees<T> >::operator=(tmp2);
  */
}

template<typename T>
AngleDegreesError<T>::AngleDegreesError(const T& def_value, 
					const T& def_plus, 
					const T& def_minus) 
  : WithError<AngleDegrees<T> >(def_value, def_plus, def_minus, T()) {
}

template<typename T>
AngleDegreesError<T>& AngleDegreesError<T>::operator=(const AngleDegreesError<T>& obj) {  
  if (obj != *this) {
    WithError<AngleDegrees<T> >::operator=(obj);
  }
  return *this;
}

template<typename T>
inline AngleDegreesError<T>& 
AngleDegreesError<T>::operator+= (const AngleDegreesError<T>& op) {
  WithError<AngleDegrees<T> >::operator+= (op);
  return *this;
}

template<typename T>
inline AngleDegreesError<T>& 
AngleDegreesError<T>::operator-= (const AngleDegreesError<T>& op) {
  WithError<AngleDegrees<T> >::operator-= (op);
  return *this;
}

template<typename T>
inline AngleDegreesError<T>& 
AngleDegreesError<T>::operator*= (const AngleDegreesError<T>& op) {
  WithError<AngleDegrees<T> >::operator*= (op);
  return *this;
}

template<typename T>
inline AngleDegreesError<T>& 
AngleDegreesError<T>::operator/= (const AngleDegreesError<T>& op) {
  WithError<AngleDegrees<T> >::operator/= (op);
  return *this;
}

template<typename T>
inline AngleDegreesError<T> 
AngleDegreesError<T>::operator+ (const AngleDegreesError<T>& op) const {
  return AngleDegreesError<T>(WithError<AngleDegrees<T> >::operator+ (op));
}

template<typename T>
inline AngleDegreesError<T> 
AngleDegreesError<T>::operator- (const AngleDegreesError<T>& op) const {
  return AngleDegreesError<T>(WithError<AngleDegrees<T> >::operator- (op));
}

template<typename T>
inline AngleDegreesError<T> 
AngleDegreesError<T>::operator* (const AngleDegreesError<T>& op) const {
  return AngleDegreesError<T>(WithError<AngleDegrees<T> >::operator* (op));
}

template<typename T>
inline AngleDegreesError<T> 
AngleDegreesError<T>::operator/ (const AngleDegreesError<T>& op) const {
  return AngleDegreesError<T>(WithError<AngleDegrees<T> >::operator/ (op));
}

template<typename T>
inline bool 
AngleDegreesError<T>::operator== (const AngleDegreesError<T>& op) const {
  if ( &op != this)
    return WithError<AngleDegrees<T> >::operator== (op);
  return true;
}

template<typename T>
inline bool 
AngleDegreesError<T>::operator!= (const AngleDegreesError<T>& op) const {
  return !(*this == op);
}

template<typename T>
inline bool 
AngleDegreesError<T>::operator< (const AngleDegreesError<T>& op) const {
  return WithError<AngleDegrees<T> >::operator< (op);
}

template<typename T>
inline bool 
AngleDegreesError<T>::operator<= (const AngleDegreesError<T>& op) const {
  return WithError<AngleDegrees<T> >::operator<= (op);
}

template<typename T>
inline bool 
AngleDegreesError<T>::operator> (const AngleDegreesError<T>& op) const {
  return WithError<AngleDegrees<T> >::operator> (op);
}

template<typename T>
inline bool 
AngleDegreesError<T>::operator>= (const AngleDegreesError<T>& op) const {
  return WithError<AngleDegrees<T> >::operator>= (op);
}

template<typename T>
inline bool AngleDegreesError<T>::operator! () const {
  return WithError<AngleDegrees<T> >::operator! ();
}

template<typename T>
inline AngleDegreesError<T>& AngleDegreesError<T>::operator++ () {
  WithError<AngleDegrees<T> >::operator++ ();
  return *this;
}

template<typename T>
inline AngleDegreesError<T>& AngleDegreesError<T>::operator-- () {
  WithError<AngleDegrees<T> >::operator-- ();
  return *this;
}

template<typename T>
inline AngleDegreesError<T> AngleDegreesError<T>::operator- () const {
  return AngleDegreesError<T>(WithError<AngleDegrees<T> >::operator- ());
}

template<typename T>
inline AngleDegreesError<T>& 
AngleDegreesError<T>::improve (const AngleDegreesError<T>& op) {
  WithError<AngleDegrees<T> >::improve(op);
  return *this;
}

template<typename T>
inline AngleDegreesError<T>& 
AngleDegreesError<T>::disimprove (const AngleDegreesError<T>& op) {
  WithError<AngleDegrees<T> >::disimprove(op);
  return *this;
}

template<typename T>
inline AngleDegrees<T> AngleDegreesError<T>::getAngle() const {
  return WithError<AngleDegrees<T> >::getValue();
}

/*
template<typename T>
inline AngleDegrees<T> AngleDegreesError<T>::getMax() const {
  return getMax();
}

template<typename T>
inline AngleDegrees<T> AngleDegreesError<T>::getMin() const {
  return getMin();
}

template<typename T>
inline AngleDegrees<T> AngleDegreesError<T>::getPlus() const {
  return plus;
}

template<typename T>
inline AngleDegrees<T> AngleDegreesError<T>::getMinus() const {
  return minus;
}
*/

template<typename T>
inline T AngleDegreesError<T>::getDeg() const {
  return getAngle().getDeg();
}

template<typename T>
inline T AngleDegreesError<T>::getRad() const {
  return getAngle().getRad();
}

template<typename T>
inline AngleDegreesError<T> operator- (const T& op1, 
				       const AngleDegreesError<T>& op2) {
  return (-op2) + op1;
}

template<typename T>
inline AngleDegreesError<T> operator- (const WithError<T>& op1, 
				       const AngleDegreesError<T>& op2) {
  return (-op2) + op1;
}

template<typename T>
inline AngleDegreesError<T> operator- (const AngleDegrees<T>& op1, 
				       const AngleDegreesError<T>& op2) {
  return (-op2) + op1;
}

template<typename T>
inline AngleDegreesError<T> operator- (const WithError<AngleDegrees<T> >& op1, 
				       const AngleDegreesError<T>& op2) {
  return (-op2) + op1;
}

template<typename T>
inline AngleDegreesError<T> operator+ (const T& op1, 
				       const AngleDegreesError<T>& op2) {
  return op2 + op1;
}

template<typename T>
inline AngleDegreesError<T> operator+ (const WithError<T>& op1, 
				       const AngleDegreesError<T>& op2) {
  return op2 + op1;
}

template<typename T>
inline AngleDegreesError<T> operator+ (const AngleDegrees<T>& op1, 
				       const AngleDegreesError<T>& op2) {
  return op2 + op1;
}

template<typename T>
inline AngleDegreesError<T> operator+ (const WithError<AngleDegrees<T> >& op1, 
				       const AngleDegreesError<T>& op2) {
  return op2 + op1;
}

template<typename T>
inline AngleDegreesError<T> operator* (const T& op1, 
				       const AngleDegreesError<T>& op2) {
  return op2 * op1;
}

template<typename T>
inline AngleDegreesError<T> operator* (const WithError<T>& op1, 
				       const AngleDegreesError<T>& op2) {
  return op2 * op1;
}

template<typename T>
inline AngleDegreesError<T> operator* (const AngleDegrees<T>& op1, 
				       const AngleDegreesError<T>& op2) {
  return op2 * op1;
}

template<typename T>
inline AngleDegreesError<T> operator* (const WithError<AngleDegrees<T> >& op1, 
				       const AngleDegreesError<T>& op2) {
  return op2 * op1;
}

template<typename T>
inline AngleDegreesError<T> operator/ (const T& op1, 
				       const AngleDegreesError<T>& op2) {
  return AngleDegreesError<T>(op1) / op2;
}

template<typename T>
inline AngleDegreesError<T> operator/ (const WithError<T>& op1, 
				       const AngleDegreesError<T>& op2) {
  return AngleDegreesError<T>(op1) / op2;
}

template<typename T>
inline AngleDegreesError<T> operator/ (const AngleDegrees<T>& op1, 
				       const AngleDegreesError<T>& op2) {
  return AngleDegreesError<T>(op1) / op2;
}

template<typename T>
inline AngleDegreesError<T> operator/ (const WithError<AngleDegrees<T> >& op1, 
				       const AngleDegreesError<T>& op2) {
  return AngleDegreesError<T>(op1) / op2;
}

template<typename T>
inline bool operator==(const T& op1, 
		       const AngleDegreesError<T> op2) {
  return op2 == op1;
}

template<typename T>
inline bool operator==(const WithError<T>& op1, 
		       const AngleDegreesError<T> op2) {
  return op2 == op1;
}

template<typename T>
inline bool operator==(const AngleDegrees<T>& op1, 
		       const AngleDegreesError<T> op2) {
  return op2 == op1;
}

template<typename T>
inline bool operator==(const WithError<AngleDegrees<T> >& op1, 
		       const AngleDegreesError<T> op2) {
  return op2 == op1;
}

template<typename T>
inline bool operator!=(const T& op1, 
		       const AngleDegreesError<T> op2) {
  return op2 != op1;

}
template<typename T>
inline bool operator!=(const WithError<T>& op1, 
		       const AngleDegreesError<T> op2) {
  return op2 != op1;
}

template<typename T>
inline bool operator!=(const AngleDegrees<T>& op1, 
		       const AngleDegreesError<T> op2) {
  return op2 != op1;
}

template<typename T>
inline bool operator!=(const WithError<AngleDegrees<T> >& op1, 
		       const AngleDegreesError<T> op2) {
  return op2 != op1;
}

template<typename T>
inline bool operator<(const T& op1, const AngleDegreesError<T> op2) {
  return op2 > op1;
}

template<typename T>
inline bool operator<(const WithError<T>& op1, 
		      const AngleDegreesError<T> op2) {
  return op2 > op1;
}

template<typename T>
inline bool operator<(const AngleDegrees<T>& op1, 
		      const AngleDegreesError<T> op2) {
  return op2 > op1;
}

template<typename T>
inline bool operator<(const WithError<AngleDegrees<T> >& op1, 
		      const AngleDegreesError<T> op2) {
  return op2 > op1;
}

template<typename T>
inline bool operator<=(const T& op1, const AngleDegreesError<T> op2) {
  return op2 >= op1;
}

template<typename T>
inline bool operator<=(const WithError<T>& op1, 
		       const AngleDegreesError<T> op2) {
  return op2 >= op1;
}

template<typename T>
inline bool operator<=(const AngleDegrees<T>& op1, 
		       const AngleDegreesError<T> op2) {
  return op2 >= op1;
}

template<typename T>
inline bool operator<=(const WithError<AngleDegrees<T> >& op1, 
		       const AngleDegreesError<T> op2) {
  return op2 >= op1;
}

template<typename T>
inline bool operator>(const T& op1, const AngleDegreesError<T> op2) {
  return op2 < op1;

}

template<typename T>
inline bool operator>(const WithError<T>& op1, 
		      const AngleDegreesError<T> op2) {
  return op2 < op1;
}

template<typename T>
inline bool operator>(const AngleDegrees<T>& op1, 
		      const AngleDegreesError<T> op2) {
  return op2 < op1;
}

template<typename T>
inline bool operator>(const WithError<AngleDegrees<T> >& op1, 
		      const AngleDegreesError<T> op2) {
  return op2 < op1;
}

template<typename T>
inline bool operator>=(const T& op1, const AngleDegreesError<T> op2) {
  return op2 <= op1;
}

template<typename T>
inline bool operator>=(const WithError<T>& op1, 
		       const AngleDegreesError<T> op2) {
  return op2 <= op1;
}

template<typename T>
inline bool operator>=(const AngleDegrees<T>& op1, 
		       const AngleDegreesError<T> op2) {
  return op2 <= op1;
}

template<typename T>
inline bool operator>=(const WithError<AngleDegrees<T> >& op1, 
		       const AngleDegreesError<T> op2) {
  return op2 <= op1;
}

template<typename T>
inline std::ostream& operator<<(std::ostream& os, 
				const AngleDegreesError<T>& obj) {
#ifdef ANGLEERROR_HEAD
  if (obj.getPlus() != T() || obj.getMinus() != T())    
    return os << ANGLEERROR_HEAD << "(" << toString(obj.getAngle()) << " ("
	      << toString(obj.getPlus()) << ", " 
	      << toString(obj.getMinus()) << ")";
  return os << ANGLEERROR_HEAD << "(" << toString(obj.getAngle()) << ")";
#else
  return os << "(" << toString(obj.getAngle()) << ")";
#endif  
}

template<typename T>
inline const std::string toString(const AngleDegreesError<T>& obj) {
#ifdef ANGLEERROR_HEAD
  if (obj.getPlus() != T() || obj.getMinus() != T())    
    return ANGLEERROR_HEAD + "(" + toString(obj.getAngle()) + " (" +
      toString(obj.getPlus()) + ", -" + 
      toString(obj.getMinus()) + ")";
  return ANGLEERROR_HEAD + "(" << toString(obj.getAngle()) << ")";
#else
  return "(" + toString(obj.getAngle()) + ")";
#endif
}

template<typename T>
inline WithError<T> sin(const AngleDegreesError<T>& x)
{
  T val = sin(x.getAngle());
  T plus = sin(x.getMax()) - val;
  T minus = sin(x.getMin()) - val;
  //T plus = sin(x.getPlus());
  //T minus = sin(x.getMinus());
  // Ugly as hack!
  if ( plus < T() ) {
    plus = -plus;
  }
  if ( minus > T() ) {
    minus = -minus;
  }
  if ( plus < T() || minus > T() ) {
    std::cout << "sin(" << x << ") gave value = " << val
	      << "; plus = " << plus << "; minus = " << minus
	      << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( plus >= T() && minus <= T() );
#endif
  return WithError<T>(val, plus, minus, T());
}

template<typename T>
inline WithError<T> cos(const AngleDegreesError<T>& x) {
  T val = cos(x.getAngle());
  T plus = cos(x.getMax()) - val;
  T minus = cos(x.getMin()) - val;
  //T plus = cos(x.getPlus());
  //T minus = cos(x.getMinus());
  // Ugly as hack!
  if ( plus < T() ) {
    plus = -plus;
  }
  if ( minus > T() ) {
    minus = -minus;
  }
  if ( plus < T() || minus > T() ) {
    std::cout << "cos(" << x << ") gave value = " << val
	      << "; plus = " << plus << "; minus = " << minus
	      << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( plus >= T() && minus <= T() );
#endif
  return WithError<T>(val, plus, minus, T());
}

template<typename T>
inline WithError<T> tan(const AngleDegreesError<T>& x) {
  T val = tan(x.getAngle());
  T plus = tan(x.getMax()) - val;
  T minus = tan(x.getMin()) - val;
  //T plus = tan(x.getPlus());
  //T minus = tan(x.getMinus());
  // Ugly as hack!
  if ( plus < T() ) {
    plus = -plus;
  }
  if ( minus > T() ) {
    minus = -minus;
  }
  if ( plus < T() || minus > T() ) {
    std::cout << "tan(" << x << ") gave value = " << val
	      << "; plus = " << plus << "; minus = " << minus
	      << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( plus >= T() && minus <= T() );
#endif
  return WithError<T>(val, plus, minus, T());
}

#ifdef RS_TRIG_IN_DEG

template<typename T>
inline AngleDegrees<T> ASin(const WithError<T>& op) {
  return AngleDegrees<T>(ASin(op.getValue()));
}

template<typename T>
inline AngleDegrees<T> ACos(const WithError<T>& op) {
  return AngleDegrees<T>(ACos(op.getValue()));
}

template<typename T>
inline AngleDegrees<T> ATan(const WithError<T>& op) {
  return AngleDegrees<T>(ATan(op.getValue()));
}

template<typename T>
inline AngleDegrees<T> ATan2(const WithError<T>& op1, 
			     const WithError<T>& op2) {
  return AngleDegrees<T>(ATan2(op1.getValue(), op2.getValue()));
}

#endif // RS_TRIG_IN_DEG

RS_END_NAMESPACE

#endif // !WITHERROR_H

