/* -*- Mode: C++ -*-
 * $Id: AngleDegreesUE.h,v 1.5 2005/09/19 17:00:06 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: AngleDegreesUE.h
 * Unitname: Library
 * $Revision: 1.5 $
 * Created by: David Rosn 2002-06-19
 * Last modified by $Author: frehe $ $Date: 2005/09/19 17:00:06 $
 *
 *
 * HISTORY
 *
 * $Log: AngleDegreesUE.h,v $
 * Revision 1.5  2005/09/19 17:00:06  frehe
 * Fixed bugs and made WithError work.
 *
 * Revision 1.4  2005/08/13 13:31:17  frehe
 * Fixed several errors to make it compile with gcc-3.4.3
 *
 * Revision 1.3  2002/09/30 10:45:55  s02davro
 * Fixed AngleDegUE bug
 *
 * Revision 1.2  2002/08/31 14:50:26  s02davro
 * Modifications, corrections of angle classes
 *
 * Revision 1.1  2002/08/02 21:31:48  s02davro
 * Created the class to be able to represent a AngleDegreeUnknown with a error interval.
 *
 *
 */

/**
 * @file AngleDegreesUE.h
 * Include file for the template AngleDegreesUE.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.5 $
 * $Author: frehe $
 * $Date: 2005/09/19 17:00:06 $
 */


#ifndef ANGLEDEGUE_H
#define ANGLEDEGUE_H

#include "AngleDegrees.h"
#include "AngleDegreesUnknown.h"
#include "AngleDegreesError.h"
#include "UnknownValueException.h"
#include "WithUE.h"
#include "rs_throw.h"

#define AUE_HEAD std::string("#AUE")

RS_BEGIN_NAMESPACE


/**
 * @class AngleDegreesUE AngleDegreesUE.h
 * Template class for angles in degrees that can be unknown and have an error.
 * The template argument is forwarded to AngleDegrees.
 * The only demands on the template argument T is that the trigonometric
 * and arithmetic functions are defined on it.
 *
 * @todo Check if all constructors are needed.
 *       Implement getValue method?
 *       Testing!
 *
 * @par Unit:
 *      Library
 * $Revision: 1.5 $
 * $Author: frehe $
 * $Date: 2005/09/19 17:00:06 $
 */

template<typename T>
class AngleDegreesUE : public WithUE<AngleDegrees<T> > { 
public:
  /** Default constructor, the value is set to unknown. */
  AngleDegreesUE();

  /** Construct a AngleDegreesUE with no error and value set to ang. */
  AngleDegreesUE(const T& def_value);

  /** Construct a AngleDegreesUE for a WithError object. */
  AngleDegreesUE(const WithUnknown<T>& def_value);

  /** Construct a AngleDegreeUE with no error and value set to obj. */
  AngleDegreesUE(const AngleDegrees<T>& obj);

  /** Construct a known or unknown AngleDegreeUE based on obj. */
  AngleDegreesUE(const AngleDegreesUnknown<T>& obj);

  /** Construct a AngleDegreeUE with errors from obj. */
  AngleDegreesUE(const AngleDegreesError<T>& obj);

  /** Copyconstrctor for the class. */
  AngleDegreesUE(const AngleDegreesUE<T>& obj);  

  /** Construct a AngleDegreeUE from a WithUE<T> object. */
  AngleDegreesUE(const WithUE<T>& obj);

  /** Construct a AngleDegreeUE from a WithUE<AngleDegrees<T>> object. */
  AngleDegreesUE(const WithUE<AngleDegrees<T> >& obj);

  /** Construct a object using a basevalue and a errortype.
   *  @see WithError for more information. */
  AngleDegreesUE(const ErrorType c, const T& value,
		 const Float64 quant_step1 = 0.1, const T& distance = T(),
		 const Float64 quant_step2 = 0.02);

  /** Construct a object using specified value and errors. */
  AngleDegreesUE(const T& def_value, const T& def_plus, const T& def_minus);  

  /** Assignment operator. */
  AngleDegreesUE<T>& operator= (const AngleDegreesUE<T>& op);  

  /** @name Arithmetic operators. */
  //@{
  AngleDegreesUE<T>& operator+= (const AngleDegreesUE<T>& op);
  AngleDegreesUE<T>& operator-= (const AngleDegreesUE<T>& op);
  AngleDegreesUE<T>& operator*= (const AngleDegreesUE<T>& op);
  AngleDegreesUE<T>& operator/= (const AngleDegreesUE<T>& op);
  AngleDegreesUE<T> operator+ (const AngleDegreesUE<T>& op) const;
  AngleDegreesUE<T> operator- (const AngleDegreesUE<T>& op) const;
  AngleDegreesUE<T> operator* (const AngleDegreesUE<T>& op) const;
  AngleDegreesUE<T> operator/ (const AngleDegreesUE<T>& op) const;
  AngleDegreesUE<T> operator- () const;
  //@}

  /** @name Increment and decrement operators. */
  //@{
  AngleDegreesUE<T>& operator++ ();
  AngleDegreesUE<T>& operator-- ();
  //@}

  /** @name Comparison operators. */
  //@{
  bool operator== (const AngleDegreesUE<T>& op) const;
  bool operator!= (const AngleDegreesUE<T>& op) const;
  bool operator< (const AngleDegreesUE<T>& op) const;
  bool operator<= (const AngleDegreesUE<T>& op) const;
  bool operator> (const AngleDegreesUE<T>& op) const;
  bool operator>= (const AngleDegreesUE<T>& op) const;
  bool operator!  () const;
  //@}
  
  /** Improve the value using another value with error.
   *  @see WithError for more information. */
  AngleDegreesUE<T>& improve(const AngleDegreesUE<T>& op);

  /** Disimprove the value using another value with error. 
   *  @see WithError for more information. */
  AngleDegreesUE<T>& disimprove(const AngleDegreesUE<T>& op);

  /** Make the angle unknown. */
  AngleDegreesUE<T>& makeUnknown();

#if 0
  /** Check if the angle is known. */
  bool isKnown() const;

  /** Check if the angle is unknown. */
  bool isUnknown() const;
#endif

  /** Get the angle striped from error information. 
   *  @throws UnknownValueException if the angle is unknown. */
  AngleDegrees<T> getAngle() const throw(UnknownValueException);

  /** Get the angle in radians.
   *  @throws UnknownValueException if the angle is unknown. */
  T getRad() const throw(UnknownValueException);

  /** Get the angle in degrees.
   *  @throws UnknownValueException if the angle is unknown. */
  T getDeg() const throw(UnknownValueException);

#if 0
  /** Get the angle with error information.
   *  @throws UnknownValueException if the angle is unknown. */
  WithError<AngleDegrees<T> > getValueWithError() const 
    throw(UnknownValueException);

  /** Strip the AngleDegreesUE from error information. */
  WithUnknown<AngleDegrees<T> > getValueWithUnknown() const;
#endif
  
  /** Returns a AngleDegreesError<T> object. 
   *  @throws UnknownValueException if the angle is unknown. */
  AngleDegreesError<T> getAngleWithError() const 
    throw(UnknownValueException);

  /** Returns a AngleDegreesUnknown<T> object. Strips the error information. */
  AngleDegreesUnknown<T> getAngleWithUnknown() const;

#if 0  
  /** Get the minimum angle.
   *  @throws UnknownValueException if the angle is unknown. */
  AngleDegrees<T> getMin() const throw(UnknownValueException);
  
  /** Get the maximum angle.
   *  @throws UnknownValueException if the angle is unknown. */
  AngleDegrees<T> getMax() const throw(UnknownValueException);

  /** Get the plus angle. 
   *  @throws UnknownValueException if the angle is unknown.
   *  getPlus = getMax - getAngle */  
  AngleDegrees<T> getPlus() const throw(UnknownValueException);

  /** Get the minus angle.
   *  @throws UnknownValueException if the angle is unknown.
   *  getMinus = getMin - getAngle */
  AngleDegrees<T> getMinus() const throw(UnknownValueException);
#endif
};

template<typename T>
AngleDegreesUE<T>::AngleDegreesUE()
  : WithUE<AngleDegrees<T> >() {
}

template<typename T>
AngleDegreesUE<T>::AngleDegreesUE(const T& def_value)
  : WithUE<AngleDegrees<T> >(def_value) {
}

template<typename T>
AngleDegreesUE<T>::AngleDegreesUE(const WithUnknown<T>& def_value)
  : WithUE<AngleDegrees<T> >() {
  if (def_value.isKnown()) {
    WithUE<AngleDegrees<T> >::operator=(def_value.getValue());
  }
}

template<typename T>
AngleDegreesUE<T>::AngleDegreesUE(const AngleDegrees<T>& obj) 
  : WithUE<AngleDegrees<T> >(obj) {
}

template<typename T>
AngleDegreesUE<T>::AngleDegreesUE(const AngleDegreesUnknown<T>& ang)
  : WithUE<AngleDegrees<T> >() {
  if (ang.isKnown()) {
    WithUE<AngleDegrees<T> >::operator=(WithError<AngleDegrees<T> >(ang.getAngle()));
  }
}

template<typename T>
AngleDegreesUE<T>::AngleDegreesUE(const AngleDegreesError<T>& obj) 
  : WithUE<AngleDegrees<T> >(obj.getAngle(), obj.getPlus(), obj.getMinus()) {
}

template<typename T>
AngleDegreesUE<T>::AngleDegreesUE(const AngleDegreesUE<T>& op) 
  : WithUE<AngleDegrees<T> >(op) {
}

template<typename T>
AngleDegreesUE<T>::AngleDegreesUE(const WithUE<AngleDegrees<T> >& obj)
  : WithUE<AngleDegrees<T> >(obj) {
}

template<typename T>
AngleDegreesUE<T>::AngleDegreesUE(const WithUE<T>& obj)
  : WithUE<AngleDegrees<T> >() {
  if (obj.isKnown()) {
    WithUE<AngleDegrees<T> >::operator=(WithError<AngleDegrees<T> >(obj.getValue(),
								    obj.getPlus(), 
								    obj.getMinus(), T()));
  }
}

template<typename T>
AngleDegreesUE<T>::AngleDegreesUE(const ErrorType c, 
				  const T& def_value,
				  const Float64 quant_step1,
				  const T& distance,
				  const Float64 quant_step2)
  : WithUE<AngleDegrees<T> >(c, def_value, quant_step1, distance, quant_step2) {
  /*
  WithError<T> tmp(c, def_value, quant_step1, distance, quant_step2);
  value = WithError<AngleDegrees<T> >(tmp.getValue(), 
				      tmp.getPlus(), 
				      tmp.getMinus());
  */
}

template<typename T>
AngleDegreesUE<T>::AngleDegreesUE(const T& def_value, const T& def_plus,
				  const T& def_minus)
   : WithUE<AngleDegrees<T> >(def_value, def_plus, def_minus) {
}

template<typename T>
AngleDegreesUE<T>& AngleDegreesUE<T>::operator= (const AngleDegreesUE<T>& op) {
  if (&op != this) {
    if (op.isKnown()) {
      WithUE<AngleDegrees<T> >::operator=(op.getValueWithError());
    } else {
      WithUE<AngleDegrees<T> >::makeUnknown();
    }
  }
  return *this;  
}

template<typename T>
inline AngleDegreesUE<T>& 
AngleDegreesUE<T>::operator+= (const AngleDegreesUE<T>& op) {
  WithUE<AngleDegrees<T> >::operator+= (op);
  return *this;
}

template<typename T>
inline AngleDegreesUE<T>& 
AngleDegreesUE<T>::operator-= (const AngleDegreesUE<T>& op) {
  WithUE<AngleDegrees<T> >::operator-= (op);
  return *this;
}

template<typename T>
inline AngleDegreesUE<T>& 
AngleDegreesUE<T>::operator*= (const AngleDegreesUE<T>& op) {
  WithUE<AngleDegrees<T> >::operator*= (op);
  return *this;
}

template<typename T>
inline AngleDegreesUE<T>& 
AngleDegreesUE<T>::operator/= (const AngleDegreesUE<T>& op) {
  WithUE<AngleDegrees<T> >::operator/= (op);
  return *this;
}

template<typename T>
inline AngleDegreesUE<T> 
AngleDegreesUE<T>::operator+ (const AngleDegreesUE<T>& op) const {
  return AngleDegreesUE<T>(WithUE<AngleDegrees<T> >::operator+ (op));
}

template<typename T>
inline AngleDegreesUE<T> 
AngleDegreesUE<T>::operator- (const AngleDegreesUE<T>& op) const {
  return AngleDegreesUE<T>(WithUE<AngleDegrees<T> >::operator- (op));
}

template<typename T>
inline AngleDegreesUE<T> 
AngleDegreesUE<T>::operator* (const AngleDegreesUE<T>& op) const {
  return AngleDegreesUE<T>(WithUE<AngleDegrees<T> >::operator* (op));
}

template<typename T>
inline AngleDegreesUE<T> 
AngleDegreesUE<T>::operator/ (const AngleDegreesUE<T>& op) const {
  return AngleDegreesUE<T>(WithUE<AngleDegrees<T> >::operator/ (op));
}

template<typename T>
inline bool 
AngleDegreesUE<T>::operator== (const AngleDegreesUE<T>& op) const {
  if ( &op != this)
    return WithUE<AngleDegrees<T> >::operator== (op);
  return true;
}

template<typename T>
inline bool 
AngleDegreesUE<T>::operator!= (const AngleDegreesUE<T>& op) const {
  return !(*this == op);
}

template<typename T>
inline bool 
AngleDegreesUE<T>::operator< (const AngleDegreesUE<T>& op) const {
  return WithUE<AngleDegrees<T> >::operator< (op);
}

template<typename T>
inline bool 
AngleDegreesUE<T>::operator<= (const AngleDegreesUE<T>& op) const {
  return WithUE<AngleDegrees<T> >::operator<= (op);
}

template<typename T>
inline bool 
AngleDegreesUE<T>::operator> (const AngleDegreesUE<T>& op) const {
  return WithUE<AngleDegrees<T> >::operator> (op);
}

template<typename T>
inline bool 
AngleDegreesUE<T>::operator>= (const AngleDegreesUE<T>& op) const {
  return WithUE<AngleDegrees<T> >::operator>= (op);
}

template<typename T>
inline bool AngleDegreesUE<T>::operator! () const {
  return WithUE<AngleDegrees<T> >::operator! ();
}

template<typename T>
inline AngleDegreesUE<T>& AngleDegreesUE<T>::operator++ () {
  WithUE<AngleDegrees<T> >::operator++ ();
  return *this;
}

template<typename T>
inline AngleDegreesUE<T>& AngleDegreesUE<T>::operator-- () {
  WithUE<AngleDegrees<T> >::operator-- ();
  return *this;
}

template<typename T>
inline AngleDegreesUE<T> AngleDegreesUE<T>::operator- () const {
  return AngleDegreesUE<T>(WithUE<AngleDegrees<T> >::operator- ());
}

template<typename T>
inline AngleDegreesUE<T>& 
AngleDegreesUE<T>::improve(const AngleDegreesUE<T>& op) {
  WithUE<AngleDegrees<T> >::improve(op);
  return *this;
}

template<typename T>
inline AngleDegreesUE<T>& 
AngleDegreesUE<T>::disimprove(const AngleDegreesUE<T>& op) {
  WithUE<AngleDegrees<T> >::disimprove(op);
  return *this;
}

template<typename T>
inline AngleDegreesUE<T>& AngleDegreesUE<T>::makeUnknown() {
  WithUE<AngleDegrees<T> >::makeUnknown();
  return *this;
}

/*
template<typename T>
inline bool AngleDegreesUE<T>::isKnown() const {
  return WithUE<AngleDegrees<T> >::isKnown();
}

template<typename T>
inline bool AngleDegreesUE<T>::isUnknown() const {
  return WithUE<AngleDegrees<T> >::isUnknown();
}
*/

template<typename T>
inline T AngleDegreesUE<T>::getRad() const throw(UnknownValueException) {
  return getAngle().getRad();
}

template<typename T>
inline T AngleDegreesUE<T>::getDeg() const throw(UnknownValueException) {
  return getAngle().getDeg();
}

template<typename T>
inline AngleDegrees<T> 
AngleDegreesUE<T>::getAngle() const throw(UnknownValueException) {
  return WithUE<AngleDegrees<T> >::getValue();
}

/*
template<typename T>
inline AngleDegrees<T> 
AngleDegreesUE<T>::getMin() const throw(UnknownValueException) {
  return getValueWithError().getMin();
}

template<typename T>
inline AngleDegrees<T> 
AngleDegreesUE<T>::getMax() const throw(UnknownValueException) {
  return getValueWithError().getMax();
}

template<typename T>
inline AngleDegrees<T> 
AngleDegreesUE<T>::getPlus() const throw(UnknownValueException) {
  return getValueWithError().getPlus();
}

template<typename T>
inline AngleDegrees<T> 
AngleDegreesUE<T>::getMinus() const throw(UnknownValueException) {
  return getValueWithError().getMinus();
}

template<typename T>
inline WithError<AngleDegrees<T> > 
AngleDegreesUE<T>::getValueWithError() const throw(UnknownValueException) {
  return WithUE<AngleDegrees<T> >::getValueWithError();
}

template<typename T>
inline WithUnknown<AngleDegrees<T> > 
AngleDegreesUE<T>::getValueWithUnknown() const {
  return WithUE<AngleDegrees<T> >::getValueWithUnknown();
}
*/

template<typename T>
inline AngleDegreesError<T> 
AngleDegreesUE<T>::getAngleWithError() const throw(UnknownValueException) {
  return WithUE<AngleDegrees<T> >::getValueWithError();
}

template<typename T>
inline AngleDegreesUnknown<T> 
AngleDegreesUE<T>::getAngleWithUnknown() const {
  return WithUE<AngleDegrees<T> >::getValueWithUnknown();
}

template<typename T>
inline AngleDegreesUE<T> operator- (const T& op1, 
				    const AngleDegreesUE<T>& op2) {
  return (-op2) + op1;
}

template<typename T>
inline AngleDegreesUE<T> operator- (const WithUnknown<T>& op1, 
				    const AngleDegreesUE<T>& op2) {
  return (-op2) + op1;
}

template<typename T>
inline AngleDegreesUE<T> operator- (const AngleDegrees<T>& op1, 
				    const AngleDegreesUE<T>& op2) {
  return (-op2) + op1;
}

template<typename T>
inline AngleDegreesUE<T> operator- (const AngleDegreesUnknown<T>& op1, 
				    const AngleDegreesUE<T>& op2) {
  return (-op2) + op1;
}

template<typename T>
inline AngleDegreesUE<T> operator- (const AngleDegreesError<T>& op1, 
				    const AngleDegreesUE<T>& op2) {
  return (-op2) + op1;
}

template<typename T>
inline AngleDegreesUE<T> operator- (const WithUE<AngleDegrees<T> >& op1, 
				    const AngleDegreesUE<T>& op2) {
  return (-op2) + op1;
}

template<typename T>
inline AngleDegreesUE<T> operator+ (const T& op1, 
				    const AngleDegreesUE<T>& op2) {
  return op2 + op1;
}

template<typename T>
inline AngleDegreesUE<T> operator+ (const WithUnknown<T>& op1,
				    const AngleDegreesUE<T>& op2) {
  return op2 + op1;
}

template<typename T>
inline AngleDegreesUE<T> operator+ (const AngleDegrees<T>& op1,
				    const AngleDegreesUE<T>& op2) {
  return op2 + op1;
}

template<typename T>
inline AngleDegreesUE<T> operator+ (const AngleDegreesUnknown<T>& op1,
				    const AngleDegreesUE<T>& op2) {
  return op2 + op1;
}

template<typename T>
inline AngleDegreesUE<T> operator+ (const AngleDegreesError<T>& op1,
				    const AngleDegreesUE<T>& op2) {
  return op2 + op1;
}

template<typename T>
inline AngleDegreesUE<T> operator+ (const WithUE<AngleDegrees<T> >& op1,
				    const AngleDegreesUE<T>& op2) {
  return op2 + op1;
}

template<typename T>
inline AngleDegreesUE<T> operator* (const T& op1,
				    const AngleDegreesUE<T>& op2) {
  return op2 * op1;
}

template<typename T>
inline AngleDegreesUE<T> operator* (const WithUnknown<T>& op1,
				    const AngleDegreesUE<T>& op2) {
  return op2 * op1;
}

template<typename T>
inline AngleDegreesUE<T> operator* (const AngleDegrees<T>& op1,
				    const AngleDegreesUE<T>& op2) {
  return op2 * op1;
}

template<typename T>
inline AngleDegreesUE<T> operator* (const AngleDegreesUnknown<T>& op1,
				    const AngleDegreesUE<T>& op2) {
  return op2 * op1;
}

template<typename T>
inline AngleDegreesUE<T> operator* (const AngleDegreesError<T>& op1,
				    const AngleDegreesUE<T>& op2) {
  return op2 * op1;
}

template<typename T>
inline AngleDegreesUE<T> operator* (const WithUE<AngleDegrees<T> >& op1,
				    const AngleDegreesUE<T>& op2) {
  return op2 * op1;
}

template<typename T>
inline AngleDegreesUE<T> operator/ (const T& op1,
				    const AngleDegreesUE<T>& op2) {
  return AngleDegreesUE<T>(op1) / op2;
}

template<typename T>
inline AngleDegreesUE<T> operator/ (const WithUnknown<T>& op1,
				    const AngleDegreesUE<T>& op2) {
  return AngleDegreesUE<T>(op1) / op2;
}

template<typename T>
inline AngleDegreesUE<T> operator/ (const AngleDegrees<T>& op1,
				    const AngleDegreesUE<T>& op2) {
  return AngleDegreesUE<T>(op1) / op2;
}

template<typename T>
inline AngleDegreesUE<T> operator/ (const AngleDegreesUnknown<T>& op1,
				    const AngleDegreesUE<T>& op2) {
  return AngleDegreesUE<T>(op1) / op2;
}

template<typename T>
inline AngleDegreesUE<T> operator/ (const AngleDegreesError<T>& op1,
				    const AngleDegreesUE<T>& op2) {
  return AngleDegreesUE<T>(op1) / op2;
}

template<typename T>
inline AngleDegreesUE<T> operator/ (const WithUE<AngleDegrees<T> >& op1,
				    const AngleDegreesUE<T>& op2) {
  return AngleDegreesUE<T>(op1) / op2;
}

template<typename T>
inline bool operator==(const T& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 == op1;
}

template<typename T>
inline bool operator==(const WithUnknown<T>& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 == op1;
}

template<typename T>
inline bool operator==(const AngleDegrees<T>& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 == op1;
}

template<typename T>
inline bool operator==(const AngleDegreesUnknown<T>& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 == op1;
}

template<typename T>
inline bool operator==(const AngleDegreesError<T>& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 == op1;
}

template<typename T>
inline bool operator==(const WithUE<AngleDegrees<T> >& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 == op1;
}

template<typename T>
inline bool operator!=(const T& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 != op1;
}

template<typename T>
inline bool operator!=(const WithUnknown<T>& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 != op1;
}

template<typename T>
inline bool operator!=(const AngleDegrees<T>& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 != op1;
}

template<typename T>
inline bool operator!=(const AngleDegreesUnknown<T>& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 != op1;
}

template<typename T>
inline bool operator!=(const AngleDegreesError<T>& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 != op1;
}

template<typename T>
inline bool operator!=(const WithUE<AngleDegrees<T> >& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 != op1;
}

template<typename T>
inline bool operator<(const T& op1,
		      const AngleDegreesUE<T> op2) {
  return op2 > op1;
}

template<typename T>
inline bool operator<(const WithUnknown<T>& op1,
		      const AngleDegreesUE<T> op2) {
  return op2 > op1;
}

template<typename T>
inline bool operator<(const AngleDegrees<T>& op1,
		      const AngleDegreesUE<T> op2) {
  return op2 > op1;
}

template<typename T>
inline bool operator<(const AngleDegreesUnknown<T>& op1,
		      const AngleDegreesUE<T> op2) {
  return op2 > op1;
}

template<typename T>
inline bool operator<(const AngleDegreesError<T>& op1,
		      const AngleDegreesUE<T> op2) {
  return op2 > op1;
}

template<typename T>
inline bool operator<(const WithUE<AngleDegrees<T> >& op1,
		      const AngleDegreesUE<T> op2) {
  return op2 > op1;
}

template<typename T>
inline bool operator<=(const T& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 >= op1;
}

template<typename T>
inline bool operator<=(const WithUnknown<T>& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 >= op1;
}

template<typename T>
inline bool operator<=(const AngleDegrees<T>& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 >= op1;
}

template<typename T>
inline bool operator<=(const AngleDegreesUnknown<T>& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 >= op1;
}

template<typename T>
inline bool operator<=(const AngleDegreesError<T>& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 >= op1;
}

template<typename T>
inline bool operator<=(const WithError<AngleDegrees<T> >& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 >= op1;
}

template<typename T>
inline bool operator>(const T& op1,
		      const AngleDegreesUE<T> op2) {
  return op2 < op1;
}

template<typename T>
inline bool operator>(const WithError<T>& op1,
		      const AngleDegreesUE<T> op2) {
  return op2 < op1;
}

template<typename T>
inline bool operator>(const AngleDegrees<T>& op1,
		      const AngleDegreesUE<T> op2) {
  return op2 < op1;
}

template<typename T>
inline bool operator>(const AngleDegreesUnknown<T>& op1,
		      const AngleDegreesUE<T> op2) {
  return op2 < op1;
}

template<typename T>
inline bool operator>(const AngleDegreesError<T>& op1,
		      const AngleDegreesUE<T> op2) {
  return op2 < op1;
}

template<typename T>
inline bool operator>(const WithUE<AngleDegrees<T> >& op1,
		      const AngleDegreesUE<T> op2) {
  return op2 < op1;
}

template<typename T>
inline bool operator>=(const T& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 <= op1;
}

template<typename T>
inline bool operator>=(const WithUnknown<T>& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 <= op1;
}

template<typename T>
inline bool operator>=(const AngleDegrees<T>& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 <= op1;
}

template<typename T>
inline bool operator>=(const AngleDegreesUnknown<T>& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 <= op1;
}

template<typename T>
inline bool operator>=(const AngleDegreesError<T>& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 <= op1;
}

template<typename T>
inline bool operator>=(const WithUE<AngleDegrees<T> >& op1,
		       const AngleDegreesUE<T> op2) {
  return op2 <= op1;
}

template<typename T>
inline std::ostream& operator <<(std::ostream& os, 
				 const AngleDegreesUE<T>& ang) {
#ifdef AUE_HEAD
  if (ang.isKnown())    
    return os << AUE_HEAD << "(" << toString(ang.getValueWithError()) << ")";
  else
    return os << AUE_HEAD << "(" << toString(ang.getValueWithUnknown()) << ")";
#else
  if (ang.isKnown())
    return os << toString(ang.getValueWithError());
  else
    return os << toString(ang.getValueWithUnknown());    
#endif
}


template<typename T>
inline const std::string toString(const AngleDegreesUE<T>& ang) {
#ifdef AUE_HEAD
  if (ang.isKnown())    
    return AUE_HEAD + "(" + toString(ang.getValueWithError()) + ")";
  else
    return AUE_HEAD + "(" + toString(ang.getValueWithUnknown()) + ")";
#else
  if (ang.isKnown())
    return toString(ang.getValueWithError());
  else
    return toString(ang.getValueWithUnknown());    
#endif
}

template<typename T>
inline WithUE<T> sin(const AngleDegreesUE<T>& x) {
  if (x.isUnknown()) 
    return WithUE<T>();
  else 
    return WithUE<T>(sin(x.getAngleWithError()));
}

template<typename T>
inline WithUE<T> cos(const AngleDegreesUE<T>& x) {
  if (x.isUnknown()) 
    return WithUE<T>();
  else 
    return WithUE<T>(cos(x.getAngleWithError()));
}

template<typename T>
inline WithUE<T> tan(const AngleDegreesUE<T>& x) {
  if (x.isUnknown()) 
    return WithUE<T>();
  else 
    return WithUE<T>(tan(x.getAngleWithError()));
}

#ifdef RS_TRIG_IN_DEG

template<typename T>
inline AngleDegreesUnknown<T> ASin(const WithUE<T>& op) {
  if ( op.isUnknown() )
    return AngleDegreesUnknown<T>();
  else
    return AngleDegreesUnknown<T>(ASin(op.getValue()));
}

template<typename T>
inline AngleDegreesUnknown<T> ACos(const WithUE<T>& op) {
  if ( op.isUnknown() )
    return AngleDegreesUnknown<T>();
  else
    return AngleDegreesUnknown<T>(ACos(op.getValue()));
}

template<typename T>
inline AngleDegreesUnknown<T> ATan(const WithUE<T>& op) {
  if ( op.isUnknown() )
    return AngleDegreesUnknown<T>();
  else
    return AngleDegreesUnknown<T>(ATan(op.getValue()));
}

template<typename T>
inline AngleDegreesUnknown<T> ATan2(const WithUE<T>& op1,
				    const WithUE<T>& op2) {
  if ( op1.isUnknown() ||
       op2.isUnknown() )
    return AngleDegreesUnknown<T>();
  else
    return AngleDegreesUnknown<T>(ATan2(op1.getValue(), op2.getValue()));
}

#endif // RS_TRIG_IN_DEG

RS_END_NAMESPACE

#endif // !ANGLEDEGREESUE_H
