/* -*- Mode: C++ -*-
 * $Id: CoordPoint.h,v 1.9 2002/09/04 12:31:34 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: CoordPoint.h
 * Unitname: Library
 * $Revision: 1.9 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2002/09/04 12:31:34 $
 *
 *
 * HISTORY
 *
 * $Log: CoordPoint.h,v $
 * Revision 1.9  2002/09/04 12:31:34  frehe
 * Removed unnecessary parameter names.
 *
 * Revision 1.8  2002/08/31 14:51:26  s02davro
 * Modifications and corrections of point classes.
 *
 * Revision 1.7  2002/08/02 21:33:15  s02davro
 * Added documentation. Restructing.
 *
 * Revision 1.6  2001/08/29 21:08:27  frehe
 * Replaces class T with typename T where T can be a built in type.
 *
 * Revision 1.5  2000/09/20 15:17:41  frehe
 * Added documentation.
 *
 * Revision 1.4  2000/09/18 15:20:04  frehe
 * Due to a problem with namespaces the name for atan2, atan, acos and 
 * asin is changed to ATan2, ATan, ASin and ACos.
 *
 * Revision 1.3  2000/09/04 15:21:48  frehe
 * Changed the name of the trigonometric functions to the same as in the 
 * standard math library (sin, cos, tan, asin, acos, atan, and atan2).
 *
 * Revision 1.2  2000/09/01 03:23:51  frehe
 * Made minor changes.
 *
 * Revision 1.1  2000/09/01 01:24:37  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 *
 */

/**
 * @file CoordPoint.h
 * Include file for the template CoordPoint.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.9 $
 * $Author: frehe $
 * $Date: 2002/09/04 12:31:34 $
 */


#ifndef COORDPOINT_H
#define COORDPOINT_H

#include <iostream>

#include "math_utils.h"
#include "AngleDegrees.h"
#include "CoordVector.h"


/** String to use as the header when outputing a CoordPoint. */
#define POINT_HEAD std::string("#Point")

RS_BEGIN_NAMESPACE

template <typename T> class CoordVector;

/**
 * @class CoordPoint CoordPoint.h
 * Template class for representing a cartesian two dimensional
 * coordinate. The X and Y values are of type T.
 *
 * @todo More testing.
 * @todo Check if all constructors and methods are needed.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.9 $
 * $Author: frehe $
 * $Date: 2002/09/04 12:31:34 $
 */
template<typename T> class CoordPoint {
public:
  /** Default constructor, the point is (T(), T()) */
  CoordPoint();

  /** Construct a point using coordinates (x,y) */
  CoordPoint(const T& x, const T& y);

  /** Copyconstructor */
  CoordPoint(const CoordPoint<T>&);
  
  /** Create a point using a start point and a vector */
  CoordPoint(const CoordPoint<T>&, const CoordVector<T>&);

  /** Explicit constructor for converting a vector to a point.
   *  This is actually the same as using:
   *  CoordPoint<T>(CoordPoint<T>(), vector) since the vector is
   *  assumed is from origo. */
  explicit CoordPoint(const CoordVector<T>&);

  /** Assignment operator. */
  CoordPoint<T>& operator=(const CoordPoint<T>&);

  /** @name Arithmetic operators. */  
  //@{
  CoordPoint<T> operator- () const;
  CoordPoint<T> operator- (const CoordPoint<T>&) const;
  CoordPoint<T> operator- (const CoordVector<T>&) const;
  CoordPoint<T> operator+ (const CoordPoint<T>&) const;
  CoordPoint<T> operator+ (const CoordVector<T>&) const;
  CoordPoint<T> operator* (const T&) const;
  CoordPoint<T> operator* (const CoordPoint<T>&) const;
  CoordPoint<T> operator/ (const T&) const;
  CoordPoint<T> operator/ (const CoordPoint<T>&) const;
  CoordPoint<T>& operator-=(const CoordPoint<T>&);
  CoordPoint<T>& operator-=(const CoordVector<T>&);
  CoordPoint<T>& operator+=(const CoordPoint<T>&);
  CoordPoint<T>& operator+=(const CoordVector<T>&);
  CoordPoint<T>& operator*=(const T&);
  CoordPoint<T>& operator*=(const CoordPoint<T>&);
  CoordPoint<T>& operator/=(const T&);
  CoordPoint<T>& operator/=(const CoordPoint<T>&);
  //@}

  /** @name Comparison operators. */
  //@{
  bool operator! () const;
  bool operator==(const CoordPoint<T>& pt) const;
  bool operator!=(const CoordPoint<T>& pt) const;
  //@}  

  /** Get the x coordinate of the point. */
  T getX() const;

  /** Get the y coordinate of the point. */
  T getY() const;

  /** Add to the x coordinate of the point. */
  CoordPoint<T>& addX(const T& pt_x);

  /** Add to the x coordinate of the point. */
  CoordPoint<T>& addY(const T& pt_y);

  /** Add to x and y coordinate of the point. */
  CoordPoint<T>& addXY(const T& pt_x, const T& pt_y);

  /** Set the x coordinate of the point. */
  CoordPoint<T>& setX(const T& pt_x);

  /** Set the y coordinate of the point. */
  CoordPoint<T>& setY(const T& pt_y);

  /** Set the x and y coordinate of the point. */
  CoordPoint<T>& set (const T& pt_x, const T& pt_y);
  
  /** Get a vector between this point and point given in call.
   *  Very usefull! */        
  CoordVector<T> getVectorTo(const CoordPoint<T>& pt) const;

  /** Get the direction between this point and point given in call.
   *  Very usefull! */
  AngleDegrees<T> getDirectionTo(const CoordPoint<T>& pt) const;

  /** Get the distance between this point and point given in call.
   *  Very usefull! */
  T getDistanceTo(const CoordPoint<T>& pt) const;  
protected:
  /** The actual representation of the x coordinate value. */
  T x;

  /** The actual representation of the y coordinate value. */
  T y;  
};

template<typename T>
CoordPoint<T>::CoordPoint()
  : x(),
    y() {
}

template<typename T>
CoordPoint<T>::CoordPoint(const T& pt_x, const T& pt_y)
  : x(pt_x),
    y(pt_y) {
}

template<typename T>
CoordPoint<T>::CoordPoint(const CoordPoint<T>& pt)
  : x(pt.x),
    y(pt.y) {
}

template<typename T>
CoordPoint<T>::CoordPoint(const CoordPoint<T>& pt, const CoordVector<T>& v) {
  *this = pt + CoordPoint<T>(v);
}

template<typename T>
CoordPoint<T>::CoordPoint(const CoordVector<T>& v) {
  x = v.getLength() * cos(v.getAngle());
  y = v.getLength() * sin(v.getAngle());
}


template<typename T>
inline CoordPoint<T>& CoordPoint<T>::addX (const T& pt_x) {
  x += pt_x;
  return *this;
}

template<typename T>
inline CoordPoint<T>& CoordPoint<T>::addY (const T& pt_y) {
  y += pt_y;
  return *this;
}

template<typename T>
CoordPoint<T>& CoordPoint<T>::operator =(const CoordPoint<T>& pt) {
  if ( this != &pt ) {
    y = pt.y;
    x = pt.x;
  }
  return *this;
}

template<typename T>
inline CoordPoint<T>& CoordPoint<T>::addXY (const T& pt_x, const T& pt_y) {
  x += pt_x;
  y += pt_y;
  return *this;
}

template<typename T>
inline T CoordPoint<T>::getX() const {
  return x;
}

template<typename T>
inline T CoordPoint<T>::getY() const {
  return y;
}

template<typename T>
inline CoordPoint<T>& CoordPoint<T>::setX(const T& pt_x) {
  x = pt_x;
  return *this;
}

template<typename T>
inline CoordPoint<T>& CoordPoint<T>::setY(const T& pt_y) {
  y = pt_y;
  return *this;
}

template<typename T>
inline CoordPoint<T>& CoordPoint<T>::set(const T& pt_x, const T& pt_y) {
  x = pt_x;
  y = pt_y;
  return *this;
}

template<typename T>
inline AngleDegrees<T> 
CoordPoint<T>::getDirectionTo(const CoordPoint<T>& pt) const {
  return ATan2((pt.y - y), (pt.x - x));
}

template<typename T>
inline T CoordPoint<T>::getDistanceTo(const CoordPoint<T>& pt) const {
  return sqrt(sqr(pt.x - x) + sqr(pt.y - y));
}

template<typename T>
inline CoordVector<T> 
CoordPoint<T>::getVectorTo(const CoordPoint<T>& pt) const {
  return CoordVector<T>(*this, pt);
}

template<typename T>
inline CoordPoint<T> CoordPoint<T>::operator -() const {
  return CoordPoint<T>(-x,-y);
}

template<typename T>
inline CoordPoint<T> CoordPoint<T>::operator -(const CoordVector<T>& v) const {
  return *this + CoordPoint<T>(-v);
}

template<typename T>
inline CoordPoint<T> CoordPoint<T>::operator -(const CoordPoint<T>& pt) const {
  return CoordPoint<T>(x - pt.x, y - pt.y);
}

template<typename T>
inline CoordPoint<T> CoordPoint<T>::operator +(const CoordVector<T>& v) const {
  return *this + CoordPoint<T>(v);
}

template<typename T>
inline CoordPoint<T> CoordPoint<T>::operator +(const CoordPoint<T>& pt) const {
  return CoordPoint<T>(x + pt.x, y + pt.y);
}

template<typename T>
inline CoordPoint<T> CoordPoint<T>::operator *(const T& a) const {
  return CoordPoint<T>(x * a, y * a);
}

template<typename T>
inline CoordPoint<T> CoordPoint<T>::operator *(const CoordPoint<T>& pt) const {
  return CoordPoint<T>(x * pt.x, y * pt.y);
}


template<typename T>
inline CoordPoint<T> CoordPoint<T>::operator /(const T& a) const {
#if CHECK_DIVISION_BY_ZERO == 1
  if (a == T(0))
    return COORDPOINT_DIVISION_BY_ZERO_RESULT;
  else
#endif
    return CoordPoint<T>(x / a, y / a);
}

template<typename T>
inline CoordPoint<T> CoordPoint<T>::operator /(const CoordPoint<T>& pt) const {
#if CHECK_DIVISION_BY_ZERO == 1
  if ( pt.x == T(0) ) {
    if ( pt.y == T(0) ) {
      return CoordPoint_DIVISION_BY_ZERO_RESULT;
    } else {
      return CoordPoint<T>(T_DIVISION_BY_ZERO_RESULT, y / pt.y);
    }
  } else {
    if ( pt.y == T(0) ) {
      return CoordPoint<T>(x / pt.x, T_DIVISION_BY_ZERO_RESULT);
    } else {
#endif
      return CoordPoint<T>(x / pt.x, y / pt.y);
#if CHECK_DIVISION_BY_ZERO == 1      
    }
  }
#endif
}

template<typename T>
inline CoordPoint<T>& CoordPoint<T>::operator-=(const CoordPoint<T>& pt) {
  x -= pt.x;
  y -= pt.y;
  return *this;
}

template<typename T>
inline CoordPoint<T>& CoordPoint<T>::operator-=(const CoordVector<T>& vec) {
  *this = *this - vec;
  return *this;
}

template<typename T>
inline CoordPoint<T>& CoordPoint<T>::operator+=(const CoordPoint<T>& pt) {
  x += pt.x;
  y += pt.y;
  return *this;
}

template<typename T>
inline CoordPoint<T>& CoordPoint<T>::operator+=(const CoordVector<T>& vec) {
  *this = *this + vec;
  return *this;
}

template<typename T>
inline CoordPoint<T>& CoordPoint<T>::operator*=(const T& a) {
  x *= a;
  y *= a;
  return *this;
}

template<typename T>
inline CoordPoint<T>& CoordPoint<T>::operator*=(const CoordPoint<T>& pt) {
  x *= pt.x;
  y *= pt.y;
  return *this;
}

template<typename T>
inline CoordPoint<T>& CoordPoint<T>::operator/=(const T& a) {
#if CHECK_DIVISION_BY_ZERO == 1
  if ( a == T(0)) {
    x = T_DIVISION_BY_ZERO_RESULT;
    y = T_DIVISION_BY_ZERO_RESULT;
  } else {
#endif
    x /= a;
    y /= a;
#if CHECK_DIVISION_BY_ZERO == 1
  }
#endif
  return *this;
}

template<typename T>
inline CoordPoint<T>& CoordPoint<T>::operator/=(const CoordPoint<T>& pt) {
#if CHECK_DIVISION_BY_ZERO == 1
  if ( pt.x == T(0) )
    x = T_DIVISION_BY_ZERO_RESULT;
  else
#endif
    x /= pt.x;
  
#if CHECK_DIVISION_BY_ZERO == 1
  if ( pt.y == T(0) )
    y = T_DIVISION_BY_ZERO_RESULT;
  else
#endif
    y /= pt.y;

  return *this;
}

template<typename T>
inline bool CoordPoint<T>::operator !() const {
  return (!y) && (!x);
}

template<typename T>
inline bool CoordPoint<T>::operator ==(const CoordPoint<T>& pt) const {
  return (y == pt.y) && (x == pt.x);
}

template<typename T>
inline bool CoordPoint<T>::operator !=(const CoordPoint<T>& pt) const {
  return (y != pt.y) || (x != pt.x);
}

template<typename T>
inline CoordPoint<T> operator *(const T& a, const CoordPoint<T>& pt) {
  return pt * a;
}

template<typename T>
inline const std::string toString(const CoordPoint<T>& pt) {
#ifdef POINT_HEAD
  return POINT_HEAD + "(" + pt.getX() + ", " + pt.getY() + ")";
#else
  return "(" + pt.getX() + ", " + pt.getY() + ")";
#endif
}

template<typename T>
inline std::ostream& operator<<(std::ostream& os, const CoordPoint<T>& pt) {
#ifdef POINT_HEAD
  return os << POINT_HEAD << "(" << pt.getX() << ", " << pt.getY() << ")";
#else
  return os << "(" << pt.getX() << ", " << pt.getY() << ")";
#endif
}

// Extend this to be the point on the line between pt1 and pt2 
// with a certain ratio 1:1 is the middle point?
// CoordPoint<T> MiddlePoint(const CoordPoint<T>&, const CoordPoint<T>&);
template<typename T>
inline CoordPoint<T> middlePoint(const CoordPoint<T>& pt1,
				 const CoordPoint<T>& pt2) {
  return pt1 + CoordPoint<T>((pt2.getX() - pt1.getX())/2,
			     (pt2.getY() - pt1.getY())/2);
}

RS_END_NAMESPACE


#endif // COORDPOINT_H
