/* -*- Mode: C++ -*-
 * $Id: CoordPointUE.h,v 1.2 2002/08/31 14:51:26 s02davro Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: CoordPointUE.h
 * Unitname: Library
 * $Revision: 1.2 $
 * Created by: David Rosn 2002-06.19
 * Last modified by $Author: s02davro $ $Date: 2002/08/31 14:51:26 $
 *
 *
 * HISTORY
 *
 * $Log: CoordPointUE.h,v $
 * Revision 1.2  2002/08/31 14:51:26  s02davro
 * Modifications and corrections of point classes.
 *
 * Revision 1.1  2002/08/02 21:33:38  s02davro
 * Created the class to be able to represent a Point with an error. (Errors not handled yet).
 *
 *
 */

/**
 * @file CoordPointUE.h
 * Include file for the template CoordPointUE.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.2 $
 * $Author: s02davro $
 * $Date: 2002/08/31 14:51:26 $
 */


#ifndef COORDPOINTUE_H
#define COORDPOINTUE_H

#include "WithUE.h"
#include "AngleDegreesUE.h"
#include "CoordPointUnknown.h"
#include "CoordVectorUnknown.h"
#include "CoordVectorUE.h"

#define POINTUE_HEAD std::string("#PointUE")

RS_BEGIN_NAMESPACE

template<typename T> class CoordVectorUE;

/**
 * @class CoordPointUE CoordPointUE.h
 * Template class for representing a cartesian two dimensional
 * coordinate that can be unknown and have a confidence factor.
 * The X and Y values are of type WithUE<T>.
 *
 * @todo Handle the errors using circle and polygons.
 * @todo Testing.
 * @todo Check constructors and methods.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.2 $
 * $Author: s02davro $
 * $Date: 2002/08/31 14:51:26 $
 */

template<typename T> class CoordPointUE {
public:
  /** Default constructor, the point is set to unknown.*/
  CoordPointUE();

  /** Construct a point using coordinates (x,y). */
  CoordPointUE(const T& x, const T& y);
  /** Cunstruct a point using the (unknown or known) coordinates with errors. */
  CoordPointUE(const WithUE<T>& x, const WithUE<T>& y);

  /** Create a CoordPointUE using a known point. */
  CoordPointUE(const CoordPoint<T>& pt);

  /** Create a CoordPointUE using a known or unknown point. */
  CoordPointUE(const CoordPointUnknown<T>& ptu);

  /** Copyconstructor. */
  CoordPointUE(const CoordPointUE<T>& ptue);

  /** @name Create a CoordPointUE using a start point an a vector. */
  //@{
  CoordPointUE(const CoordPoint<T>& pt, const CoordVectorUE<T>& vec);
  CoordPointUE(const CoordPointUnknown<T>& pt, const CoordVectorUE<T>& vec);
  CoordPointUE(const CoordPointUE<T>& pt, const CoordVector<T>& vec); 
  CoordPointUE(const CoordPointUE<T>& pt, const CoordVectorUnknown<T>& vec); 
  CoordPointUE(const CoordPointUE<T>& pt, const CoordVectorUE<T>& vec); 
  //@}

  /** @name Explicit constructors for convertion a vector to a point
   *  This is actually the same as using:
   *  CoordPointUE<T>(CoordPointUE<T>(), vector) since the vector is
   *  assumed is from origo. */
  //@{
  explicit CoordPointUE(const CoordVectorUE<T>& vec);
  explicit CoordPointUE(const CoordVectorUnknown<T>& vec);
  explicit CoordPointUE(const CoordVector<T>& vec);
  //@}

  /** Assignment operator. */
  CoordPointUE<T>& operator=(const CoordPointUE<T>& pt);

  /** @name Arithmetic operators. */
  //@{
  CoordPointUE<T> operator- () const;
  CoordPointUE<T> operator- (const CoordPointUE<T>& pt) const;
  CoordPointUE<T> operator- (const CoordVectorUE<T>& vec) const;
  CoordPointUE<T> operator+ (const CoordPointUE<T>& pt) const;
  CoordPointUE<T> operator+ (const CoordVectorUE<T>& vec) const;
  CoordPointUE<T> operator* (const WithUE<T>& val) const;
  CoordPointUE<T> operator* (const CoordPointUE<T>& pt) const;
  CoordPointUE<T> operator/ (const WithUE<T>& val) const;
  CoordPointUE<T> operator/ (const CoordPointUE<T>& pt) const;

  CoordPointUE<T>& operator-=(const CoordPointUE<T>& pt);
  CoordPointUE<T>& operator-=(const CoordVectorUE<T>& vec);
  CoordPointUE<T>& operator+=(const CoordPointUE<T>& pt);
  CoordPointUE<T>& operator+=(const CoordVectorUE<T>& vec);
  CoordPointUE<T>& operator*=(const WithUE<T>& val);
  CoordPointUE<T>& operator*=(const CoordPointUE<T>& pt);
  CoordPointUE<T>& operator/=(const WithUE<T>& val);
  CoordPointUE<T>& operator/=(const CoordPointUE<T>& pt);
  //@}

  /** @name Comparison operators. */
  //@{
  bool operator ! () const;
  bool operator ==(const CoordPointUE<T>& pt) const;
  bool operator !=(const CoordPointUE<T>& pt) const;
  //@}

  /** Make the point unknown. */
  CoordPointUE<T>& makeUnknown();

  /** Check if the point is known. */
  bool isKnown() const;

  /** Check if the point is unknown. */
  bool isUnknown() const;

  /** Get the point striped from error information.
   *  @throws UnknownValueException if the point is unknown. */
  CoordPoint<T> getPoint() const throw(UnknownValueException);  
  
  /** Get the x coordinate of the point. */
  WithUE<T> getX() const;

  /** Get the y coordinate of the point. */
  WithUE<T> getY() const;

  /** Add to the x coordinate of the point. */
  CoordPointUE<T>& addX(const WithUE<T>& pt_x);

  /** Add to the x coordinate of the point. */
  CoordPointUE<T>& addY(const WithUE<T>& pt_y);

  /** Add to x and y coordinate of the point. */
  CoordPointUE<T>& addXY(const WithUE<T>& pt_x,
			 const WithUE<T>& pt_y);

  /** Set the x coordinate of the point. */
  CoordPointUE<T>& setX(const WithUE<T>& pt_x);

  /** Set the y coordinate of the point. */
  CoordPointUE<T>& setY(const WithUE<T>& pt_y);

  /** Set the x and y coordinate of the point. */
  CoordPointUE<T>& set (const WithUE<T>& pt_x, const WithUE<T>& pt_y);
  
  /** Get a vector between this point and point given in call.
   *  Very usefull! */        
  CoordVectorUE<T> getVectorTo(const CoordPointUE<T>& pt) const;

  /** Get the direction between this point and point given in call.
   *  Very usefull! */
  AngleDegreesUE<T> getDirectionTo(const CoordPointUE<T>& pt) const;

  /** Get the distance between this point and point given in call.
   *  Very usefull! */
  WithUE<T> getDistanceTo(const CoordPointUE<T>& pt) const;
  
protected:
  /** The actual representation of the x coordinate value. */
  WithUE<T> x;
  
  /** The actual representation of the y coordinate value. */
  WithUE<T> y;  
};

template<typename T>
CoordPointUE<T>::CoordPointUE()
  : x(),
    y() {
}

template<typename T>
CoordPointUE<T>::CoordPointUE(const T& pt_x, const T& pt_y)
  : x(pt_x),
    y(pt_y) {
}

template<typename T>
CoordPointUE<T>::CoordPointUE(const WithUE<T>& pt_x, const WithUE<T>& pt_y)
  : x(pt_x),
    y(pt_y) {
}

template<typename T>
CoordPointUE<T>::CoordPointUE(const CoordPoint<T>& pt)
  : x(pt.getX()),
    y(pt.getY()) {
}

template<typename T>
CoordPointUE<T>::CoordPointUE(const CoordPointUnknown<T>& pt)
  : x(pt.getX()),
    y(pt.getY()) {
}

template<typename T>
CoordPointUE<T>::CoordPointUE(const CoordPointUE<T>& pt)
  : x(pt.x),
    y(pt.y) {
}


template<typename T>
CoordPointUE<T>::CoordPointUE(const CoordPoint<T>& pt, 
			      const CoordVectorUE<T>& v) {
  *this = pt + v;
}

template<typename T>
CoordPointUE<T>::CoordPointUE(const CoordPointUnknown<T>& pt, 
			      const CoordVectorUE<T>& v){
  *this = pt + v;
}

template<typename T>
CoordPointUE<T>::CoordPointUE(const CoordPointUE<T>& pt,
			      const CoordVector<T>& v) {
  *this = pt + v;
}

template<typename T>
CoordPointUE<T>::CoordPointUE(const CoordPointUE<T>& pt,
			      const CoordVectorUnknown<T>& v) {
  *this = pt + v;
}

template<typename T>
CoordPointUE<T>::CoordPointUE(const CoordPointUE<T>& pt,
			      const CoordVectorUE<T>& v) {
  *this = pt + v;
}

template<typename T>
CoordPointUE<T>::CoordPointUE(const CoordVector<T>& v) {
  x = v.getLength() * cos(v.getAngle());
  y = v.getLength() * sin(v.getAngle());
}

template<typename T>
CoordPointUE<T>::CoordPointUE(const CoordVectorUnknown<T>& v) {
  x = v.getLength() * cos(v.getAngle());
  y = v.getLength() * sin(v.getAngle());
}

template<typename T>
CoordPointUE<T>::CoordPointUE(const CoordVectorUE<T>& v) {
  x = v.getLength() * cos(v.getAngle());
  y = v.getLength() * sin(v.getAngle());
}

template<typename T>
inline CoordPointUE<T>& 
CoordPointUE<T>::operator =(const CoordPointUE<T>& pt) {
  if ( this != &pt ) {
    y = pt.y;
    x = pt.x;
  }
  return *this;
}

template<typename T>
inline bool CoordPointUE<T>::isKnown() const {
  return x.isKnown() && y.isKnown();
}

template<typename T>
inline bool CoordPointUE<T>::isUnknown() const {
  return x.isUnknown() || y.isUnknown();
}

template<typename T>
inline CoordPointUE<T>& CoordPointUE<T>::makeUnknown() {
  x.makeUnknown();
  y.makeUnknown();
  return *this;
}

template<typename T> 
inline CoordPoint<T> 
CoordPointUE<T>::getPoint() const throw(UnknownValueException) {
  if ( isUnknown() ) {
    RS_THROW(UnknownValueException, 
	     "Tried to take the value of an unknown point");
    return CoordPoint<T>();
  }  else
    return CoordPoint<T>(x.getValue(), y.getValue());
}


template<typename T>
inline CoordPointUE<T>& CoordPointUE<T>::addX (const WithUE<T>& pt_x) {
  x += pt_x;
  return *this;
}

template<typename T>
inline CoordPointUE<T>& CoordPointUE<T>::addY (const WithUE<T>& pt_y) {
  y += pt_y;
  return *this;
}

template<typename T>
inline CoordPointUE<T>& CoordPointUE<T>::addXY (const WithUE<T>& pt_x,
						const WithUE<T>& pt_y) {
  x += pt_x;
  y += pt_y;
  return *this;
}

template<typename T>
inline WithUE<T> CoordPointUE<T>::getX() const {
  return x;
}

template<typename T>
inline WithUE<T> CoordPointUE<T>::getY() const {
  return y;
}

template<typename T>
inline CoordPointUE<T>& CoordPointUE<T>::setX(const WithUE<T>& pt_x) {
  x = pt_x;
  return *this;
}

template<typename T>
inline CoordPointUE<T>& CoordPointUE<T>::setY(const WithUE<T>& pt_y) {
  y = pt_y;
  return *this;
}

template<typename T>
inline CoordPointUE<T>& CoordPointUE<T>::set(const WithUE<T>& pt_x,
					     const WithUE<T>& pt_y) {
  x = pt_x;
  y = pt_y;
  return *this;
}

template<typename T>
inline AngleDegreesUE<T> 
CoordPointUE<T>::getDirectionTo(const CoordPointUE<T>& pt) const {
  if ( pt.isUnknown() || isUnknown())
    return AngleDegreesUE<T>();
  else
    return ATan2(pt.y - y, pt.x - x);
}

template<typename T>
inline WithUE<T> 
CoordPointUE<T>::getDistanceTo(const CoordPointUE<T>& pt) const {
  return sqrt(sqr(pt.x - x) + sqr(pt.y - y));
}


template<typename T>
inline CoordVectorUE<T> 
CoordPointUE<T>::getVectorTo(const CoordPointUE<T>& pt) const {
  return CoordVectorUE<T>(*this, pt);
}

template<typename T>
inline CoordPointUE<T> CoordPointUE<T>::operator -() const {
  return CoordPointUE<T>(-x,-y);
}

template<typename T>
inline CoordPointUE<T> 
CoordPointUE<T>::operator -(const CoordPointUE<T>& pt) const {
  return CoordPointUE<T>(x - pt.x, y - pt.y);
}

template<typename T>
inline CoordPointUE<T> 
CoordPointUE<T>::operator -(const CoordVectorUE<T>& v) const {
  return *this + CoordPointUE<T>(-v);
}

template<typename T>
inline CoordPointUE<T> 
CoordPointUE<T>::operator +(const CoordPointUE<T>& pt) const {
  return CoordPointUE<T>(x + pt.x, y + pt.y);
}

template<typename T>
inline CoordPointUE<T> 
CoordPointUE<T>::operator+(const CoordVectorUE<T>& v) const {
  return *this + CoordPointUE<T>(v);
}

template<typename T>
inline CoordPointUE<T> 
CoordPointUE<T>::operator *(const WithUE<T>& a) const {
  return CoordPointUE<T>(x * a, y * a);
}

template<typename T>
inline CoordPointUE<T> 
CoordPointUE<T>::operator *(const CoordPointUE<T>& pt) const {
  return CoordPointUE<T>(x * pt.x, y * pt.y);
}


template<typename T>
inline CoordPointUE<T> CoordPointUE<T>::operator /(const WithUE<T>& a) const {
#if CHECK_DIVisION_BY_ZERO == 1
  if (a == T(0))
    return CoordPointUE_DIVisION_BY_ZERO_RESULT;
  else
#endif
    return CoordPointUE<T>(x / a, y / a);
}

template<typename T>
inline CoordPointUE<T> 
CoordPointUE<T>::operator /(const CoordPointUE<T>& pt) const {
#if CHECK_DIVisION_BY_ZERO == 1
  if ( pt.x == T(0) ) {
    if ( pt.y == T(0) ) {
      return CoordPointUE_DIVisION_BY_ZERO_RESULT;
    } else {
      return CoordPointUE<T>(T_DIVisION_BY_ZERO_RESULT, y / pt.y);
    }
  } else {
    if ( pt.y == T(0) ) {
      return CoordPointUE<T>(x / pt.x, T_DIVisION_BY_ZERO_RESULT);
    } else {
#endif
      return CoordPointUE<T>(x / pt.x, y / pt.y);
#if CHECK_DIVisION_BY_ZERO == 1      
    }
  }
#endif
}

template<typename T>
inline CoordPointUE<T>& 
CoordPointUE<T>::operator-=(const CoordPointUE<T>& pt) {
  x -= pt.x;
  y -= pt.y;
  return *this;
}

template<typename T>
inline CoordPointUE<T>& 
CoordPointUE<T>::operator-=(const CoordVectorUE<T>& vec) {
  *this = *this - vec;
  return *this;
}

template<typename T>
inline CoordPointUE<T>& 
CoordPointUE<T>::operator+=(const CoordPointUE<T>& pt) {
  x += pt.x;
  y += pt.y;
  return *this;
}

template<typename T>
inline CoordPointUE<T>& 
CoordPointUE<T>::operator+=(const CoordVectorUE<T>& vec) {
  *this = *this + vec;
  return *this;
}

template<typename T>
inline CoordPointUE<T>& CoordPointUE<T>::operator*=(const WithUE<T>& a) {
  x *= a;
  y *= a;
  return *this;
}

template<typename T>
inline CoordPointUE<T>& 
CoordPointUE<T>::operator*=(const CoordPointUE<T>& pt) {
  x *= pt.x;
  y *= pt.y;
  return *this;
}

template<typename T>
inline CoordPointUE<T>& CoordPointUE<T>::operator/=(const WithUE<T>& a) {
#if CHECK_DIVisION_BY_ZERO == 1
  if ( a == T(0)) {
    x = T_DIVisION_BY_ZERO_RESULT;
    y = T_DIVisION_BY_ZERO_RESULT;
  } else {
#endif
    x /= a;
    y /= a;
#if CHECK_DIVisION_BY_ZERO == 1
  }
#endif
  return *this;
}

template<typename T>
inline CoordPointUE<T>& 
CoordPointUE<T>::operator/=(const CoordPointUE<T>& pt) {
#if CHECK_DIVisION_BY_ZERO == 1
  if ( pt.x == T(0) )
    x = T_DIVisION_BY_ZERO_RESULT;
  else
#endif
    x /= pt.x;
  
#if CHECK_DIVisION_BY_ZERO == 1
  if ( pt.y == T(0) )
    y = T_DIVisION_BY_ZERO_RESULT;
  else
#endif
    y /= pt.y;

  return *this;
}

template<typename T>
inline bool CoordPointUE<T>::operator !() const {
  return (!y) && (!x);
}

template<typename T>
inline bool CoordPointUE<T>::operator ==(const CoordPointUE<T>& pt) const {
  return (y == pt.y) && (x == pt.x);
}

template<typename T>
inline bool CoordPointUE<T>::operator !=(const CoordPointUE<T>& pt) const {
  return (y != pt.y) || (x != pt.x);
}

template<typename T>
inline CoordPointUE<T> operator -(const CoordPoint<T>& pt1,
				  const CoordPointUE<T>& pt2) {
  return (-pt2) + pt1;
}

template<typename T>
inline CoordPointUE<T> operator -(const CoordPointUnknown<T>& pt1,
				  const CoordPointUE<T>& pt2) {
  return (-pt2) + pt1;
}

template<typename T>
inline CoordPointUE<T> operator +(const CoordPoint<T>& pt1,
				  const CoordPointUE<T>& pt2) {
  return pt2 + pt1;
}

template<typename T>
inline CoordPointUE<T> operator +(const CoordPointUnknown<T>& pt1,
				  const CoordPointUE<T>& pt2) {
  return pt2 + pt1;
}

template<typename T>
inline CoordPointUE<T> operator *(const T& a, const CoordPointUE<T>& pt) {
  return pt * a;
}

template<typename T>
inline CoordPointUE<T> operator *(const WithUE<T>& a, 
				  const CoordPointUE<T>& pt) {
  return pt * a;
}

template<typename T>
inline CoordPointUE<T> operator *(const CoordPoint<T>& pt1,
				  const CoordPointUE<T>& pt2) {
  return pt2 * pt1;
}

template<typename T>
inline CoordPointUE<T> operator *(const CoordPointUnknown<T>& pt1,
				  const CoordPointUE<T>& pt2) {
  return pt2 * pt1;
}

template<typename T>
inline CoordPointUE<T> operator /(const CoordPoint<T>& pt1, 
				  const CoordPointUE<T>& pt2) {
  if ( pt2.isUnknown() )
    return CoordPointUE<T>();
  else
    return pt1 / pt2.getPoint();
}

template<typename T>
inline CoordPointUE<T> operator /(const CoordPointUnknown<T>& pt1, 
				  const CoordPointUE<T>& pt2) {
  if ( pt2.isUnknown() )
    return CoordPointUE<T>();
  else
    return pt1 / pt2.getPoint();
}

template<typename T>
inline bool operator ==(const CoordPoint<T>& pt1,
			const CoordPointUE<T>& pt2) {
  return pt2 == pt1;
}

template<typename T>
inline bool operator ==(const CoordPointUnknown<T>& pt1,
			const CoordPointUE<T>& pt2) {
  return pt2 == pt1;
}

template<typename T>
inline bool operator !=(const CoordPoint<T>& pt1,
			const CoordPointUE<T>& pt2) {
  return pt2 != pt1;
}

template<typename T>
inline bool operator !=(const CoordPointUnknown<T>& pt1,
			const CoordPointUE<T>& pt2) {
  return pt2 != pt1;
}

template<typename T>
inline const std::string toString(const CoordPointUE<T>& pt) {
#ifdef POINTUE_HEAD
  return POINTUE_HEAD + "(" + toString(pt.getX()) + ", " + 
    toString(pt.getY()) + ")";
#else
  return "(" + toString(pt.getX()) + ", " + toString(pt.getY()) + ")";
#endif
}

template<typename T>
inline std::ostream& operator<<(std::ostream& os, const CoordPointUE<T>& pt) {
#ifdef POINTUE_HEAD
  return os << POINTUE_HEAD << "(" << pt.getX() << ", " << pt.getY() << ")";
#else
  return os << "(" << pt.getX() << ", " << pt.getY() << ")";
#endif
}


// Extend this to be the point on the line between pt1 and pt2 
// with a certain ratio 1:1 is the middle point?
template<typename T>
inline CoordPointUE<T> middlePoint(const CoordPointUE<T>& pt1,
				   const CoordPointUE<T>& pt2) {
  return pt1 + CoordPointUE<T>((pt2.getX() - pt1.getX())/2,
			       (pt2.getY() - pt1.getY())/2);
}


RS_END_NAMESPACE


#endif // !COORDPOINTUE_H
