/* -*- Mode: C++ -*-
 * $Id: CoordPointUnknown.h,v 1.7 2002/08/31 14:51:26 s02davro Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: CoordPointUnknown.h
 * Unitname: Library
 * $Revision: 1.7 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: s02davro $ $Date: 2002/08/31 14:51:26 $
 *
 *
 * HISTORY
 *
 * $Log: CoordPointUnknown.h,v $
 * Revision 1.7  2002/08/31 14:51:26  s02davro
 * Modifications and corrections of point classes.
 *
 * Revision 1.6  2002/08/02 21:33:56  s02davro
 * Added documentation. Restructing.
 *
 * Revision 1.5  2001/08/29 21:08:28  frehe
 * Replaces class T with typename T where T can be a built in type.
 *
 * Revision 1.4  2000/09/18 15:20:05  frehe
 * Due to a problem with namespaces the name for atan2, atan, acos and asin 
 * is changed to ATan2, ATan, ASin and ACos.
 *
 * Revision 1.3  2000/09/04 15:21:49  frehe
 * Changed the name of the trigonometric functions to the same as in the 
 * standard math library (sin, cos, tan, asin, acos, atan, and atan2).
 *
 * Revision 1.2  2000/09/01 03:23:52  frehe
 * Made minor changes.
 *
 * Revision 1.1  2000/09/01 01:24:38  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 *
 */

/**
 * @file CoordPointUnknown.h
 * Include file for the template CoordPointUnknown.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.7 $
 * $Author: s02davro $
 * $Date: 2002/08/31 14:51:26 $
 */


#ifndef COORDPOINTUNKNOWN_H
#define COORDPOINTUNKNOWN_H

#include <iostream>

#include "math_utils.h"
#include "AngleDegrees.h"
#include "WithUnknown.h"
#include "CoordPoint.h"
#include "CoordVector.h"
#include "CoordVectorUnknown.h"
#include "UnknownValueException.h"
#include "rs_throw.h"

#define POINTUNKNOWN_HEAD std::string("#PointUnknown")

RS_BEGIN_NAMESPACE

template<typename T> class CoordVectorUnknown;

/**
 * @class CoordPointUnknown CoordPointUnknown.h
 * Template class for representing a cartesian two dimensional
 * coordinate that can be unknown.
 * The X and Y values are of type WithUnknown<T>.
 *
 *
 * @par Unit:
 *      Library
 * $Revision: 1.7 $
 * $Author: s02davro $
 * $Date: 2002/08/31 14:51:26 $
 */
template<typename T> class CoordPointUnknown {
public:
  /** Default constructor, the point is set to unknown. */
  CoordPointUnknown();
  
  /** Construct a point using coordinates (x,y) */
  CoordPointUnknown(const T& x, const T& y);

  /** Construct a point using (known or unknown) coordinates (x,y) */
  CoordPointUnknown(const WithUnknown<T>& x, const WithUnknown<T>& y);

  /** Construct a CoordPointUnknown using a known point */
  CoordPointUnknown(const CoordPoint<T>& pt);

  /** Copyconstructor */
  CoordPointUnknown(const CoordPointUnknown<T>& pt);

  /** @name Create a CoordPointUnknown using a start point an a vector. */
  //@{
  CoordPointUnknown(const CoordPoint<T>& pt,
		    const CoordVectorUnknown<T>& vec);
  CoordPointUnknown(const CoordPointUnknown<T>& pt,
		    const CoordVector<T>& vec);
  CoordPointUnknown(const CoordPointUnknown<T>& pt,
		    const CoordVectorUnknown<T>& vec);
  //@}

  /** Explicit constructors for convertion a vector to a point
   *  This is actually the same as using:
   *  CoordPointUnknown<T>(CoordPointUnknown<T>(), vector) since the vector is
   *  assumed is from origo. */
  //@{
  explicit CoordPointUnknown(const CoordVectorUnknown<T>& vec);
  explicit CoordPointUnknown(const CoordVector<T>& vec);
  //@}
 
  /** Assignment operator. */
  CoordPointUnknown<T>& operator =(const CoordPointUnknown<T>& pt);

  /** @name Arithmetic operators. */
  //@{
  CoordPointUnknown<T> operator- () const;
  CoordPointUnknown<T> operator- (const CoordPointUnknown<T>& pt) const;
  CoordPointUnknown<T> operator- (const CoordVectorUnknown<T>& vec) const;
  CoordPointUnknown<T> operator+ (const CoordPointUnknown<T>& pt) const;
  CoordPointUnknown<T> operator+ (const CoordVectorUnknown<T>& vec) const;
  CoordPointUnknown<T> operator* (const WithUnknown<T>& val) const;
  CoordPointUnknown<T> operator* (const CoordPointUnknown<T>& pt) const;
  CoordPointUnknown<T> operator/ (const WithUnknown<T>& val) const;
  CoordPointUnknown<T> operator/ (const CoordPointUnknown<T>& pt) const;
  CoordPointUnknown<T>& operator-=(const CoordPointUnknown<T>& pt);
  CoordPointUnknown<T>& operator-=(const CoordVectorUnknown<T>& vec);
  CoordPointUnknown<T>& operator+=(const CoordPointUnknown<T>& pt);
  CoordPointUnknown<T>& operator+=(const CoordVectorUnknown<T>& vec);
  CoordPointUnknown<T>& operator*=(const WithUnknown<T>& val);
  CoordPointUnknown<T>& operator*=(const CoordPointUnknown<T>& pt);
  CoordPointUnknown<T>& operator/=(const WithUnknown<T>& val);
  CoordPointUnknown<T>& operator/=(const CoordPointUnknown<T>& pt);
  //@}

  /** @name Comparison operators. */
  //@{
  bool operator ! () const;
  bool operator ==(const CoordPointUnknown<T>& pt) const;
  bool operator !=(const CoordPointUnknown<T>& pt) const;
  //@}

  /** Make the point unknown. */
  CoordPointUnknown<T>& makeUnknown();

  /** Check if the point is known. */
  bool isKnown() const;

  /** Check if the point is unknown. */
  bool isUnknown() const;

  /** Get the point.
   *  @throws UnknownValueException if the point is unknown. */
  CoordPoint<T> getPoint() const throw(UnknownValueException);

  /** Get the x coordinate of the point. */
  WithUnknown<T> getX() const;

  /** Get the y coordinate of the point. */
  WithUnknown<T> getY() const;

  /** Add to the x coordinate of the point. */
  CoordPointUnknown<T>& addX(const WithUnknown<T>& pt_x);

  /** Add to the x coordinate of the point. */
  CoordPointUnknown<T>& addY(const WithUnknown<T>& pt_y);

  /** Add to x and y coordinate of the point. */
  CoordPointUnknown<T>& addXY(const WithUnknown<T>& pt_x,
			      const WithUnknown<T>& pt_y);

  /** Set the x coordinate of the point. */
  CoordPointUnknown<T>& setX(const WithUnknown<T>& pt_x);

  /** Set the y coordinate of the point. */
  CoordPointUnknown<T>& setY(const WithUnknown<T>& pt_y);

  /** Set the x and y coordinate of the point. */
  CoordPointUnknown<T>& set (const WithUnknown<T>& pt_x, 
			     const WithUnknown<T>& pt_y);
  
  /** Get a vector between this point and point given in call.
   *  Very usefull! */        
  CoordVectorUnknown<T> getVectorTo(const CoordPointUnknown<T>& pt) const;

  /** Get the direction between this point and point given in call.
   *  Very usefull! */
  AngleDegreesUnknown<T> getDirectionTo(const CoordPointUnknown<T>& pt) const;

  /** Get the distance between this point and point given in call.
   *  Very usefull! */  
  WithUnknown<T> getDistanceTo(const CoordPointUnknown<T>& pt) const;
  
protected:
  /** The actual representation of the x coordinate value. */
  WithUnknown<T> x;

  /** The actual representation of the y coordinate value. */
  WithUnknown<T> y;  
};

template<typename T>
CoordPointUnknown<T>::CoordPointUnknown() {
}

template<typename T>
CoordPointUnknown<T>::CoordPointUnknown(const WithUnknown<T>& pt_x, 
					const WithUnknown<T>& pt_y)
  : x(pt_x),
    y(pt_y) {
}

template<typename T>
CoordPointUnknown<T>::CoordPointUnknown(const T& pt_x, const T& pt_y)
  : x(pt_x),
    y(pt_y) {
}

template<typename T>
CoordPointUnknown<T>::CoordPointUnknown(const CoordPoint<T>& pt)
  : x(pt.getX()),
    y(pt.getY()) {
}

template<typename T>
CoordPointUnknown<T>::CoordPointUnknown(const CoordPointUnknown<T>& pt)
  : x(pt.x),
    y(pt.y) {
}

template<typename T>
CoordPointUnknown<T>::CoordPointUnknown(const CoordPoint<T>& pt, 
					const CoordVectorUnknown<T>& v) {
  *this = pt + v;
}

template<typename T>
CoordPointUnknown<T>::CoordPointUnknown(const CoordPointUnknown<T>& pt, 
					const CoordVector<T>& v) {
  *this = pt + v;
}

template<typename T>
CoordPointUnknown<T>::CoordPointUnknown(const CoordPointUnknown<T>& pt, 
					const CoordVectorUnknown<T>& v) {
  *this = pt + v;
}

template<typename T>
CoordPointUnknown<T>::CoordPointUnknown(const CoordVector<T>& v) {
  x = v.getLength() * cos(v.getAngle());
  y = v.getLength() * sin(v.getAngle());
}

template<typename T>
CoordPointUnknown<T>::CoordPointUnknown(const CoordVectorUnknown<T>& v) {
  x = v.getLength() * cos(v.getAngle());
  y = v.getLength() * sin(v.getAngle());
}

template<typename T>
CoordPointUnknown<T>& 
CoordPointUnknown<T>::operator =(const CoordPointUnknown<T>& pt) {
  if ( this != &pt ) {
    y = pt.y;
    x = pt.x;
  }
  return *this;
}

template<typename T>
inline bool CoordPointUnknown<T>::isKnown() const {
  return x.isKnown() && y.isKnown();
}

template<typename T>
inline bool CoordPointUnknown<T>::isUnknown() const {
  return x.isUnknown() && y.isUnknown();
}

template<typename T>
CoordPointUnknown<T>& CoordPointUnknown<T>::makeUnknown() {
  x.makeUnknown();
  y.makeUnknown();
  return *this;
}

template <typename T> 
inline CoordPoint<T> 
CoordPointUnknown<T>::getPoint() const throw(UnknownValueException) {
  if ( isUnknown() ) {
    RS_THROW(UnknownValueException, "Tried to take the value of an unknown point");
    return CoordPoint<T>();
  } else
    return CoordPoint<T>(x.getValue(), y.getValue());
}

template<typename T>
inline CoordPointUnknown<T>& 
CoordPointUnknown<T>::addX (const WithUnknown<T>& pt_x) {
  x += pt_x;
  return *this;
}

template<typename T>
inline CoordPointUnknown<T>& 
CoordPointUnknown<T>::addY (const WithUnknown<T>& pt_y) {
  y += pt_y;
  return *this;
}

template<typename T>
inline CoordPointUnknown<T>& 
CoordPointUnknown<T>::addXY (const WithUnknown<T>& pt_x,
			     const WithUnknown<T>& pt_y) {
  x += pt_x;
  y += pt_y;
  return *this;
}

template<typename T>
inline WithUnknown<T> CoordPointUnknown<T>::getX() const {
  return x;
}

template<typename T>
inline WithUnknown<T> CoordPointUnknown<T>::getY() const {
  return y;
}

template<typename T>
inline CoordPointUnknown<T>& 
CoordPointUnknown<T>::setX(const WithUnknown<T>& pt_x) {
  x = pt_x;
  return *this;
}

template<typename T>
inline CoordPointUnknown<T>& 
CoordPointUnknown<T>::setY(const WithUnknown<T>& pt_y) {
  y = pt_y;
  return *this;
}

template<typename T>
inline CoordPointUnknown<T>& 
CoordPointUnknown<T>::set(const WithUnknown<T>& pt_x, 
			  const WithUnknown<T>& pt_y) {
  x = pt_x;
  y = pt_y;
  return *this;
}

template<typename T>
AngleDegreesUnknown<T> 
CoordPointUnknown<T>::getDirectionTo(const CoordPointUnknown<T>& pt) const {
  if ( pt.isUnknown() )
    return AngleDegreesUnknown<T>();
  else
    return ATan2((pt.y - y), (pt.x - x));
}

template<typename T>
inline WithUnknown<T> 
CoordPointUnknown<T>::getDistanceTo(const CoordPointUnknown<T>& pt) const {
  return sqrt(sqr(pt.x - x) + sqr(pt.y - y));
}

template<typename T>
CoordVectorUnknown<T> 
CoordPointUnknown<T>::getVectorTo(const CoordPointUnknown<T>& pt) const {
  return CoordVectorUnknown<T>(*this, pt);
}

template<typename T>
inline CoordPointUnknown<T> CoordPointUnknown<T>::operator -() const {
  return CoordPointUnknown<T>(-x,-y);
}

template<typename T>
CoordPointUnknown<T> 
CoordPointUnknown<T>::operator-(const CoordVectorUnknown<T>& v) const {
  return *this + CoordPointUnknown<T>(-v);
}

template<typename T>
inline CoordPointUnknown<T> 
CoordPointUnknown<T>::operator -(const CoordPointUnknown<T>& pt) const {
  return CoordPointUnknown<T>(x - pt.x, y - pt.y);
}

template<typename T>
inline CoordPointUnknown<T> 
CoordPointUnknown<T>::operator +(const CoordPointUnknown<T>& pt) const {
  return CoordPointUnknown<T>(x + pt.x, y + pt.y);
}

template<typename T>
CoordPointUnknown<T> 
CoordPointUnknown<T>::operator+(const CoordVectorUnknown<T>& v) const {
  return *this + CoordPointUnknown<T>(v);
}

template<typename T>
inline CoordPointUnknown<T> 
CoordPointUnknown<T>::operator *(const CoordPointUnknown<T>& pt) const {
  return CoordPointUnknown<T>(x * pt.x, y * pt.y);
}

template<typename T>
inline CoordPointUnknown<T> 
CoordPointUnknown<T>::operator *(const WithUnknown<T>& a) const {
  return CoordPointUnknown<T>(x * a, y * a);
}

template<typename T>
inline CoordPointUnknown<T> 
CoordPointUnknown<T>::operator /(const CoordPointUnknown<T>& pt) const {
#if CHECK_DIVisION_BY_ZERO == 1
  if ( pt.x == T(0) ) {
    if ( pt.y == T(0) ) {
      return CoordPointUnknown_DIVisION_BY_ZERO_RESULT;
    } else {
      return CoordPointUnknown<T>(T_DIVisION_BY_ZERO_RESULT, y / pt.y);
    }
  } else {
    if ( pt.y == T(0) ) {
      return CoordPointUnknown<T>(x / pt.x, T_DIVisION_BY_ZERO_RESULT);
    } else {
#endif
      return CoordPointUnknown<T>(x / pt.x, y / pt.y);
#if CHECK_DIVisION_BY_ZERO == 1      
    }
  }
#endif
}

template<typename T>
inline CoordPointUnknown<T> 
CoordPointUnknown<T>::operator /(const WithUnknown<T>& a) const {
#if CHECK_DIVisION_BY_ZERO == 1
  if (a == T(0))
    return CoordPointUnknown_DIVisION_BY_ZERO_RESULT;
  else
#endif
    return CoordPointUnknown<T>(x / a, y / a);
}

template<typename T>
inline CoordPointUnknown<T>& 
CoordPointUnknown<T>::operator-=(const CoordPointUnknown<T>& pt) {
  x -= pt.x;
  y -= pt.y;
  return *this;
}

template<typename T>
inline CoordPointUnknown<T>& 
CoordPointUnknown<T>::operator-=(const CoordVectorUnknown<T>& vec) {
  *this = *this - vec;
  return *this;
}

template<typename T>
inline CoordPointUnknown<T>& 
CoordPointUnknown<T>::operator+=(const CoordVectorUnknown<T>& vec) {
  *this = *this + vec;
  return *this;
}

template<typename T>
inline CoordPointUnknown<T>& 
CoordPointUnknown<T>::operator+=(const CoordPointUnknown<T>& pt) {
  x += pt.x;
  y += pt.y;
  return *this;
}

template<typename T>
inline CoordPointUnknown<T>& 
CoordPointUnknown<T>::operator*=(const WithUnknown<T>& a) {
  x *= a;
  y *= a;
  return *this;
}

template<typename T>
inline CoordPointUnknown<T>& 
CoordPointUnknown<T>::operator*=(const CoordPointUnknown<T>& pt) {
  x *= pt.x;
  y *= pt.y;
  return *this;
}

template<typename T>
CoordPointUnknown<T>& 
CoordPointUnknown<T>::operator/=(const WithUnknown<T>& a) {
#if CHECK_DIVisION_BY_ZERO == 1
  if ( a == T(0)) {
    x = T_DIVisION_BY_ZERO_RESULT;
    y = T_DIVisION_BY_ZERO_RESULT;
  } else {
#endif
    x /= a;
    y /= a;
#if CHECK_DIVisION_BY_ZERO == 1
  }
#endif
  return *this;
}

template<typename T>
inline CoordPointUnknown<T>& 
CoordPointUnknown<T>::operator/=(const CoordPointUnknown<T>& pt) {
#if CHECK_DIVisION_BY_ZERO == 1
  if ( pt.x == T(0) )
    x = T_DIVisION_BY_ZERO_RESULT;
  else
#endif
    x /= pt.x;
  
#if CHECK_DIVisION_BY_ZERO == 1
  if ( pt.y == T(0) )
    y = T_DIVisION_BY_ZERO_RESULT;
  else
#endif
    y /= pt.y;

  return *this;
}

template<typename T>
inline bool CoordPointUnknown<T>::operator !() const {
  return (!y) && (!x);
}

template<typename T>
inline bool 
CoordPointUnknown<T>::operator ==(const CoordPointUnknown<T>& pt) const {
  return (y == pt.y) && (x == pt.x);
}

template<typename T>
inline bool 
CoordPointUnknown<T>::operator !=(const CoordPointUnknown<T>& pt) const {
  return (y != pt.y) || (x != pt.x);
}


template<typename T>
inline CoordPointUnknown<T> operator -(const CoordPoint<T>& pt1,
				       const CoordPointUnknown<T>& pt2) {
  return (-pt2) + pt1;
}


template<typename T>
inline CoordPointUnknown<T> operator +(const CoordPoint<T>& pt1,
				       const CoordPointUnknown<T>& pt2) {
  return pt2 + pt1;
}

template<typename T>
inline CoordPointUnknown<T> operator *(const T& a,
				       const CoordPointUnknown<T>& pt) {
  return pt * a;
}

template<typename T>
inline CoordPointUnknown<T> operator *(const WithUnknown<T>& a,
				       const CoordPointUnknown<T>& pt) {
  return pt * a;
}

template<typename T>
inline CoordPointUnknown<T> operator *(const CoordPoint<T>& pt1,
				       const CoordPointUnknown<T>& pt2) {
  return pt2 * pt1;
}


template<typename T>
CoordPointUnknown<T> operator /(const CoordPoint<T>& pt1,
				const CoordPointUnknown<T>& pt2) {
  if ( pt2.isUnknown() )
    return CoordPointUnknown<T>();
  else
    return pt1 / pt2.getPoint();
}

template<typename T>
inline bool operator ==(const CoordPoint<T>& pt1,
			const CoordPointUnknown<T>& pt2) {
  return pt2 == pt1;
}

template<typename T>
inline bool operator !=(const CoordPoint<T>& pt1,
			const CoordPointUnknown<T>& pt2) {
  return pt2 != pt1;
}

template<typename T>
inline const std::string toString(const CoordPointUnknown<T>& pt) {
#ifdef POINTUNKNOWN_HEAD
  return POINTUNKNOWN_HEAD + "(" + toString(pt.getX()) + ", " + 
    toString(pt.getY()) + ")";
#else
  return "(" + toString(pt.getX()) + ", " + toString(pt.getY()) + ")";
#endif
}

template<typename T>
inline std::ostream& operator<<(std::ostream& os, 
				const CoordPointUnknown<T>& pt) {
#ifdef POINTUNKNOWN_HEAD
  return os << POINTUNKNOWN_HEAD << "(" << pt.getX() << ", " 
	    << pt.getY() << ")";
#else
  return os << "(" << pt.getX() << ", " << pt.getY() << ")";
#endif
}

// Extend this to be the point on the line between pt1 and pt2 
// with a certain ratio 1:1 is the middle point?
template<typename T>
inline CoordPointUnknown<T> middlePoint(const CoordPointUnknown<T>& pt1,
					const CoordPointUnknown<T>& pt2) {
  return pt1 + CoordPointUnknown<T>((pt2.getX() - pt1.getX())/2,
				    (pt2.getY() - pt1.getY())/2);
}


RS_END_NAMESPACE


#endif // COORDPOINTUNKNOWN_H
