/* -*- Mode: C++ -*-
 * $Id: CoordVector.h,v 1.12 2004/09/09 15:12:44 petan Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: CoordVector.h
 * Unitname: Library
 * $Revision: 1.12 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: petan $ $Date: 2004/09/09 15:12:44 $
 *
 *
 * HISTORY
 *
 * $Log: CoordVector.h,v $
 * Revision 1.12  2004/09/09 15:12:44  petan
 * Implemented missing implementation of angleBetween
 *
 * Revision 1.11  2003/09/03 13:54:35  aip19
 * Added projectOn-function
 *
 * Revision 1.10  2002/12/07 17:31:00  s02davro
 * InitCommand handle trainer
 *
 * Revision 1.9  2002/09/04 12:31:35  frehe
 * Removed unnecessary parameter names.
 *
 * Revision 1.8  2002/08/31 14:51:54  s02davro
 * Modifications and corrections of point classes.
 *
 * Revision 1.7  2002/08/02 21:34:15  s02davro
 * Added documentation.
 *
 * Revision 1.6  2001/08/29 21:08:28  frehe
 * Replaces class T with typename T where T can be a built in type.
 *
 * Revision 1.5  2000/09/20 15:12:52  frehe
 * Added documentation.
 *
 * Revision 1.4  2000/09/18 15:20:05  frehe
 * Due to a problem with namespaces the name for atan2, atan, acos and asin 
 * is changed to ATan2, ATan, ASin and ACos.
 *
 * Revision 1.3  2000/09/04 15:21:49  frehe
 * Changed the name of the trigonometric functions to the same as in the 
 * standard math library (sin, cos, tan, asin, acos, atan, and atan2).
 *
 * Revision 1.2  2000/09/01 03:23:52  frehe
 * Made minor changes.
 *
 * Revision 1.1  2000/09/01 01:24:38  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 *
 */

/**
 * @file CoordVector.h
 * Include file for the template CoordVector.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.12 $
 * $Author: petan $
 * $Date: 2004/09/09 15:12:44 $
 */

#ifndef COORDVECTOR_H
#define COORDVECTOR_H

#include <iostream>
#include <string>
#include "AngleDegrees.h"
#include "CoordPoint.h"

/** String to use as the header when outputing a Vector. */
#define VECTOR_HEAD std::string("#Vector")

RS_BEGIN_NAMESPACE

template<typename T> class CoordPoint;

/**
 * @class CoordVector CoordVector.h
 * Template class for representing a two dimensional mathematical vector.
 * The distance and direction are of type T.
 * T is the base type for the angle: AngleDegrees<T> and the distance T.
 *
 * @todo Think about the method names. Fix it?
 *
 * @par Unit:
 *      Library
 * $Revision: 1.12 $
 * $Author: petan $
 * $Date: 2004/09/09 15:12:44 $
 */
template<typename T> class CoordVector {
public:
  /** Default constructor, the zerovector. */
  CoordVector();

  /** Construct a vector using a direction and a distance. */
  CoordVector(const T& dir, const T& dist);

  /** Construct a vector using a angle (as direction) and a distance. */
  CoordVector(const AngleDegrees<T>& ang, const T& dist);

  /** Copyconstructor. */
  CoordVector(const CoordVector& vec);
  
  /** Construct a vector based on the distance and angle between two points. */
  CoordVector(const CoordPoint<T>& pt1, const CoordPoint<T>& pt2);

  /** Explicit constructor for converting a point to a vector. */
  explicit CoordVector(const CoordPoint<T>& pt);

  /** Assignment operator. */
  CoordVector<T>& operator =(const CoordVector<T>& vec);

  /** @name Arithmetic operators. */
  //@{
  CoordVector<T> operator- () const;
  CoordVector<T> operator+ (const CoordVector<T>& vec) const; 
  CoordVector<T> operator- (const CoordVector<T>& vec) const;
  CoordVector<T> operator* (const T& val) const;
  CoordVector<T> operator/ (const T& val) const;  
  CoordVector<T>& operator-=(const CoordVector<T>& vec);
  CoordVector<T>& operator+=(const CoordVector<T>& vec);
  CoordVector<T>& operator*=(const T& val);
  CoordVector<T>& operator/=(const T& val);
  //@}

  /** @name Comparison operators. */
  //@{
  /** Check if the vector is the zero vector? */
  bool     operator ! () const;

  /** Are the vectors equal or not. */
  bool     operator ==(const CoordVector<T>& vec) const;

  /** Are the vectors equal or not. */
  bool     operator !=(const CoordVector<T>& vec) const;

  /** Compares the length of the vectors. */
  bool     operator < (const CoordVector<T>& vec) const;

  /** Compares the length of the vectors. */
  bool     operator <=(const CoordVector<T>& vec) const;

  /** Compares the length of the vectors. */
  bool     operator > (const CoordVector<T>& vec) const;

  /** Compares the length of the vectors. */
  bool     operator >=(const CoordVector<T>& vec) const;
  //@}

  /** Adds to the angle of the vector. */
  CoordVector<T>& rotate(const AngleDegrees<T>& ang);
  
  /** Adds to the distance of the vector. */
  CoordVector<T>& addLength (const T& val);

  /** Returns the length of the vector. */
  T getLength() const;
  
  /** Set the length of the vector. */
  CoordVector<T>& setLength(const T& val);

  /** Returns the angle of the vector. */
  AngleDegrees<T> getAngle() const;

  /** Set the angle of the vector. */
  CoordVector<T>& setAngle(const AngleDegrees<T>& ang);

  /** Get the cross product from two CoordVectors. */
  T crossProduct(const CoordVector<T>& v) const;

  /** Get the dot product from two CoordVectors. */
  T dotProduct(const CoordVector<T>& v) const;

  /** Projects vector onto another vector */
  CoordVector<T> projectOn(const CoordVector<T>& v) const;
  
  /** Get the absolute angle between two CoordVectors. */
  T angleBetween(const CoordVector<T>& v) const;
  
protected:
  /** The actual representation of the direction of the vector. */
  AngleDegrees<T> dir;

  /** The actual representation of the distance of the vector. */
  T dist;
};

template <typename T>
CoordVector<T>::CoordVector()
  : dir(),
    dist() {
}

template <typename T>
CoordVector<T>::CoordVector(const T& vdir, const T& vdist)
  : dir(vdir),
    dist(vdist) {
  if (dist < T(0)) {
    dir -= AngleDegrees<T>(180);
    dist = -dist;
  }
}

template <typename T>
CoordVector<T>::CoordVector(const AngleDegrees<T>& vdir, const T& vdist)
  : dir(vdir),
    dist(vdist) {
  if (dist < T(0)) {
    dir -= AngleDegrees<T>(180);
    dist = -dist;
  }
}

template <typename T>
CoordVector<T>::CoordVector(const CoordVector<T>& v)
  : dir(v.dir),
    dist(v.dist) {
}

template <typename T>
CoordVector<T>::CoordVector(const CoordPoint<T>& pt1, 
			    const CoordPoint<T>& pt2) {
  if (pt2 == pt1) {
    dir = AngleDegrees<T>(0);
    dist = T(0);
  } else {
    dir = ATan2(pt2.getY()-pt1.getY(), pt2.getX()-pt1.getX());
    dist = sqrt(sqr(pt1.getX()-pt2.getX()) + sqr(pt1.getY()-pt2.getY()));
  }
}

template <typename T>
CoordVector<T>::CoordVector(const CoordPoint<T>& pt) {
  if ( pt.getX() == T(0) &&
       pt.getY() == T(0) ) {
    dir = AngleDegrees<T>(0);
    dist = T(0);
  } else {
    dir = ATan2(pt.getY(), pt.getX());
    dist = sqrt(sqr(pt.getX()) + sqr(pt.getY()));
  }
}

template <typename T>
CoordVector<T>& CoordVector<T>::operator =(const CoordVector<T>& v) {
  if (this != &v) {
    dist = v.dist;
    dir = v.dir;
  }
  return *this;
}
 
template <typename T>
inline CoordVector<T>& CoordVector<T>::setLength(const T& len) {
  dist = len;
  return *this;
}

template <typename T>
inline CoordVector<T>& CoordVector<T>::setAngle(const AngleDegrees<T>& ang) {
  dir = ang;
  return *this;
}

template <typename T>
inline CoordVector<T>& CoordVector<T>::rotate(const AngleDegrees<T>& vdir) {
  dir += vdir;
  return *this;
}

template <typename T>
inline CoordVector<T>& CoordVector<T>::addLength(const T& vdist) {
  dist += vdist;
  return *this;
}

template <typename T>
inline T CoordVector<T>::getLength() const {
  return dist;
}

template <typename T>
inline AngleDegrees<T> CoordVector<T>::getAngle() const {
  return dir;
}

template <typename T>
inline CoordVector<T> CoordVector<T>::operator -() const {
  return CoordVector<T>(dir-AngleDegrees<T>(180), dist);
}

template <typename T>
CoordVector<T> CoordVector<T>::operator +(const CoordVector<T>& v) const {
  return CoordVector<T>(CoordPoint<T>(v) + CoordPoint<T>(*this));
}

template <typename T>
inline CoordVector<T> 
CoordVector<T>::operator -(const CoordVector<T>& v) const {
  return *this+CoordVector<T>(v.dir-AngleDegrees<T>(180), v.dist);
}

template <typename T>
inline CoordVector<T> CoordVector<T>::operator *(const T& a) const {
  return CoordVector<T>(dir, dist * a);
}

template <typename T>
inline CoordVector<T> CoordVector<T>::operator /(const T& a) const {
#if CHECK_DIVISION_BY_ZERO == 1
  if (a == T(0))
    return COORDVECTOR_DIVISION_BY_ZERO_RESULT;
  else
#endif
    return CoordVector<T>(dir, dist / a);
}

template <typename T>
inline CoordVector<T>& CoordVector<T>::operator-=(const CoordVector<T>& v) {
  return *this = *this - v;
}

template <typename T>
inline CoordVector<T>& CoordVector<T>::operator+=(const CoordVector<T>& v) {
  return *this = *this + v;
}

template <typename T>
inline CoordVector<T>& CoordVector<T>::operator*=(const T& a) {
  dist *= a;
  return *this;
}

template <typename T>
inline CoordVector<T>& CoordVector<T>::operator/=(const T& a) {
#if CHECK_DIVISION_BY_ZERO == 1
  if (a == T(0))
    dist = T_DIVISION_BY_ZERO_RESULT;
  else
#endif
    dist /= a;  
  return *this;
}

template <typename T>
inline bool CoordVector<T>::operator ==(const CoordVector<T>& v) const {
  return (dist == v.dist) && (dir == v.dir);
}

template <typename T>
inline bool CoordVector<T>::operator !=(const CoordVector<T>& v) const {
  return (dist != v.dist) || (dir != v.dir);
}

template <typename T>
inline bool CoordVector<T>::operator < (const CoordVector<T>& v) const {
  return (dist <  v.dist);
}

template <typename T>
inline bool CoordVector<T>::operator <=(const CoordVector<T>& v) const {
  return (dist <= v.dist);
}

template <typename T>
inline bool CoordVector<T>::operator > (const CoordVector<T>& v) const {
  return (dist >  v.dist);
}

template <typename T>
inline bool CoordVector<T>::operator >=(const CoordVector<T>& v) const {
  return (dist >= v.dist);
}

template <typename T>
inline bool CoordVector<T>::operator !() const {
  return (!dist) && (!dir);
}

template <typename T>
inline CoordVector<T> operator *(const T& a, const CoordVector<T>& vec) {
  return vec * a;
}

template <typename T>
inline T CoordVector<T>::crossProduct(const CoordVector<T>& v) const {
  CoordPoint<T> p1(this);
  CoordPoint<T> p2(v);
  return p1.getX()*p2.getY() - p1.getY()*p2.getX();
}

template <typename T>
inline T CoordVector<T>::dotProduct(const CoordVector<T>& v) const {
  CoordPoint<T> p1(*this);
  CoordPoint<T> p2(v);
  return p1.getX()*p2.getX() + p1.getY()*p2.getY();
}

template <typename T>
inline T CoordVector<T>::angleBetween(const CoordVector<T>& v) const {
  return abs((this->getAngle() - v.getAngle()).getDeg());
}

template <typename T>
inline  CoordVector<T> CoordVector<T>::projectOn(const CoordVector<T>& v) const {
  CoordVector<T> vec = (this->dotProduct(v)/v.dotProduct(v))*v;
  return vec;
}

template <typename T>
inline const std::string ToString(const CoordVector<T>& v) {
#ifdef VECTOR_HEAD
  return VECTOR_HEAD + "(" + ToString(v.getAngle()) + ", " + 
    ToString(v.getLength()) + ")";
#else
  return "(" + ToString(v.getAngle()) + ", " + ToString(v.getLength()) + ")";
#endif
}

template <typename T>
inline std::ostream& operator<<(std::ostream& os, const CoordVector<T>& v) {
#ifdef VECTOR_HEAD
  return os << VECTOR_HEAD << "(" << v.getAngle() << ", " 
	    << v.getLength() << ")";
#else
  return os << "(" << v.getAngle() << ", " << v.getLength() << ")";
#endif
}

RS_END_NAMESPACE


#endif  // !COORDVECTOR_H
