/* -*- Mode: C++ -*-
 * $Id: CoordVectorUE.h,v 1.2 2002/08/31 14:51:54 s02davro Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: CoordVectorUE.h
 * Unitname: Library
 * $Revision: 1.2 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: s02davro $ $Date: 2002/08/31 14:51:54 $
 *
 *
 * HISTORY
 *
 * $Log: CoordVectorUE.h,v $
 * Revision 1.2  2002/08/31 14:51:54  s02davro
 * Modifications and corrections of point classes.
 *
 * Revision 1.1  2002/08/02 21:34:40  s02davro
 * Created the class to be able to represent a Vector with an error. (Errors not handled yet).
 *
 *
 */

/**
 * @file CoordVectorUE.h
 * Include file for the template CoordVectorUE.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.2 $
 * $Author: s02davro $
 * $Date: 2002/08/31 14:51:54 $
 */


#ifndef COORDVECTORUE_H
#define COORDVECTORUE_H

#include "library_types_with_ue.h"
#include "CoordVector.h"
#include "WithUE.h"
#include "AngleDegreesUE.h"
#include "CoordPointUE.h"

/** String to use as the header when outputing a VectorUE. */
#define VECTORUE_HEAD std::string("#VectorUE")

RS_BEGIN_NAMESPACE

template<typename T> class CoordPointUE;

/**
 * @class CoordVectorUE CoordVectorUE.h
 * Template class for representing a two dimensional mathematical vector
 * that can be unknown and have an error.
 * The distance and direction are of type WithUE<T>.
 * T is the base type for the angle: AngleDegreesUE<T> and the
 * distance WithUE<T>.
 *
 * @todo Handle the errors using circle and polygons.
 * @todo Testing.
 * @todo Check methods. All isn't needed.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.2 $
 * $Author: s02davro $
 * $Date: 2002/08/31 14:51:54 $
 */

template<typename T> class CoordVectorUE {
public:

  /** Default constructor, the vector is set to unknown.*/
  CoordVectorUE();

  /** Construct a vector using a direction and a distance. */
  CoordVectorUE(const T& dir, const T& dist);

  /** Construct a vector using a angle and a distance. */
  CoordVectorUE(const AngleDegrees<T>& ang, const T& dist);

  /** Construct a vector using (known or unknown) angle and distance. */
  CoordVectorUE(const AngleDegreesUnknown<T>& ang, 
		const WithUnknown<T>& dist);

  /** Construct a vector using (unknown) angle and distance with errors. */
  CoordVectorUE(const AngleDegreesUE<T>& ang,
		const WithUE<T>& dist);

  /** Construct a vector using another vector. */
  CoordVectorUE(const CoordVector<T>& vec);

  /** Construct a vector using another (unknown) vector. */
  CoordVectorUE(const CoordVectorUnknown<T>& vec);

  /** Copyconstructor. */
  CoordVectorUE(const CoordVectorUE<T>& vec);

  /** Construct a vector using two (unknown) points with error. */
  CoordVectorUE(const CoordPointUE<T>& pt1, const CoordPointUE<T>& pt2);

  /** @name Explicit constructors for converting a point to a vector. */  
  //@{
  /** Using a known point. */
  explicit CoordVectorUE(const CoordPoint<T>& pt);

  /** Using a (unknown) point. */
  explicit CoordVectorUE(const CoordPointUnknown<T>& pt);

  /** Using a (unknown) point with error. */
  explicit CoordVectorUE(const CoordPointUE<T>& pt);
  //@}

  /** Assignment operator. */
  CoordVectorUE<T>& operator =(const CoordVectorUE<T>& vec);
  
  /** @name Arithmetic operators. */
  //@{
  CoordVectorUE<T> operator-() const;
  CoordVectorUE<T>& operator-=(const CoordVectorUE<T>& vec);
  CoordVectorUE<T>& operator+=(const CoordVectorUE<T>& vec);
  CoordVectorUE<T>& operator*=(const WithUE<T>& val);
  CoordVectorUE<T>& operator/=(const WithUE<T>& val);
  CoordVectorUE<T> operator-(const CoordVectorUE<T>& vec) const;
  CoordVectorUE<T> operator+(const CoordVectorUE<T>& vec) const; 
  CoordVectorUE<T> operator*(const WithUE<T>& val) const;
  CoordVectorUE<T> operator/(const WithUE<T>& val) const;
  //@}

  /** @name Comparison operators. 
   *  See CoordVector for mor information. */
  //@{
  bool     operator ! () const;
  bool     operator ==(const CoordVectorUE<T>& vec) const;
  bool     operator !=(const CoordVectorUE<T>& vec) const;
  bool     operator < (const CoordVectorUE<T>& vec) const;
  bool     operator <=(const CoordVectorUE<T>& vec) const;
  bool     operator > (const CoordVectorUE<T>& vec) const;
  bool     operator >=(const CoordVectorUE<T>& vec) const;
  //@}

  /** Make the vector unknown. */
  CoordVectorUE<T>& makeUnknown();

  /** Check if the vector is known. */
  bool isKnown() const;

  /** Check if the vector is unknown. */
  bool isUnknown() const;
  
  /** Get the vector striped from error information.
   *  If the point is unknown an UnknownValueException is thrown */
  CoordVector<T> getVector() const throw(UnknownValueException);

  /** Adds to the angle of the vector. */
  CoordVectorUE<T>& rotate(const AngleDegreesUE<T>& ang);

  /** Adds to the distance of the vector. 
   *  Bad name! I should have thought this was multiplication not addition. */
  CoordVectorUE<T>& scale (const WithUE<T>& val);

  /** Returns the length of the vector. */
  WithUE<T> getLength() const;

  /** Set the length of the vector. */
  CoordVectorUE<T>& setLength(const WithUE<T>& vval);

  /** Returns the angle of the vector. */
  AngleDegreesUE<T> getAngle() const;

  /** Set the angle of the vector. */
  CoordVectorUE<T>& setAngle(const AngleDegreesUE<T>& ang);

protected:
  /** The actual representation of the direction of the vector. */
  AngleDegreesUE<T> dir;

  /** The actual representation of the distance of the vector. */
  WithUE<T> dist;
};

template <typename T>
CoordVectorUE<T>::CoordVectorUE()
  : dir(),
    dist() {
}

template <typename T>
CoordVectorUE<T>::CoordVectorUE(const T& vdir, const T& vdist)
  : dir(vdir),
    dist(vdist) {
  if (dist < WithUE<T>(0)) {
    dir -= AngleDegreesUE<T>(180);
    dist = -dist;
  }
}

template <typename T>
CoordVectorUE<T>::CoordVectorUE(const AngleDegrees<T>& vdir,
				const T& vdist)
  : dir(vdir),
    dist(vdist) {
  if (dist < WithUE<T>(0)) {
    dir -= AngleDegreesUE<T>(180);
    dist = -dist;
  }
}


template <typename T>
CoordVectorUE<T>::CoordVectorUE(const AngleDegreesUnknown<T>& vdir,
				const WithUnknown<T>& vdist)
  : dir(vdir),
    dist(vdist) {
  if (dist < WithUE<T>(0)) {
    dir -= AngleDegreesUE<T>(180);
    dist = -dist;
  }
}

template <typename T>
CoordVectorUE<T>::CoordVectorUE(const AngleDegreesUE<T>& vdir,
				const WithUE<T>& vdist)
  : dir(vdir),
    dist(vdist) {
  if (dist < WithUE<T>(0)) {
    dir -= AngleDegreesUE<T>(180);
    dist = -dist;
  }
}

template <typename T>
CoordVectorUE<T>::CoordVectorUE(const CoordVector<T>& v)
  : dir(v.getAngle()),
    dist(v.getLength()) {
}

template <typename T>
CoordVectorUE<T>::CoordVectorUE(const CoordVectorUnknown<T>& v)
  : dir(v.getAngle()),
    dist(v.getLength()) {
}

template <typename T>
CoordVectorUE<T>::CoordVectorUE(const CoordVectorUE<T>& v)
  : dir(v.dir),
    dist(v.dist) {
}

template <typename T>
CoordVectorUE<T>::CoordVectorUE(const CoordPointUE<T>& pt1,
				const CoordPointUE<T>& pt2) {
  if (pt1.isUnknown() || pt2.isUnknown()) {     
    dir = AngleDegreesUE<T>();
    dist = WithUE<T>();
  } else if  (pt2 == pt1) {
    dir = AngleDegreesUE<T>(0);
    dist = WithUE<T>(0);
  } else {
    dir = ATan2(pt2.getY()-pt1.getY(),pt2.getX()-pt1.getX());      
    dist = sqrt(sqr(pt1.getX()-pt2.getX()) + sqr(pt1.getY()-pt2.getY()));
  }
}

template <typename T>
CoordVectorUE<T>::CoordVectorUE(const CoordPoint<T>& pt) {
  if ( pt.getX() == 0 &&
       pt.getY() == 0) {
    dir = AngleDegreesUE<T>(0);
    dist = WithUE<T>(0);
  } else {
    dir = ATan2(pt.getY(), pt.getX());
    dist = sqrt(sqr(pt.getX()) + sqr(pt.getY()));
  }
}

template <typename T>
CoordVectorUE<T>::CoordVectorUE(const CoordPointUnknown<T>& pt) {
  if ( pt.isUnknown() ) {
    dir = AngleDegreesUE<T>();
    dist = WithUE<T>();
  }
  else if ( pt.getX() == WithUE<T>(0) &&
	    pt.getY() == WithUE<T>(0) ) {
    dir = AngleDegreesUE<T>(0);
    dist = WithUE<T>(0);
  } else {
    dir = ATan2(pt.getY(), pt.getX());
    dist = sqrt(sqr(pt.getX()) + sqr(pt.getY()));
  }
}

template <typename T>
CoordVectorUE<T>::CoordVectorUE(const CoordPointUE<T>& pt) {
  if (pt.isUnknown()) {
    dir = AngleDegreesUE<T>();
    dist = WithUE<T>();
  }
  else if ( pt.getX() == WithUE<T>(0) &&
	    pt.getY() == WithUE<T>(0) ) {
    dir = AngleDegreesUE<T>(0);
    dist = WithUE<T>(0);
  } else {
    dir = ATan2(pt.getY(), pt.getX());
    dist = sqrt(sqr(pt.getX()) + sqr(pt.getY()));
  }
}

template <typename T>
CoordVectorUE<T>& CoordVectorUE<T>::operator =(const CoordVectorUE<T>& v) {
  if (this != &v) {
    dist = v.dist;
    dir = v.dir;
  }
  return *this;
}

template<typename T>
bool CoordVectorUE<T>::isKnown() const {
  return dir.isKnown() && dist.isKnown();
}

template<typename T>
bool CoordVectorUE<T>::isUnknown() const {
  return dir.isUnknown() && dist.isUnknown();
}

template<typename T>
CoordVectorUE<T>& CoordVectorUE<T>::makeUnknown() {
  dir.makeUnknown();
  dist.makeUnknown();
  return *this;
}

template <typename T> 
CoordVector<T> 
CoordVectorUE<T>::getVector() const throw(UnknownValueException) {
  if ( isUnknown() ) {
    RS_THROW(UnknownValueException, 
	     "Tried to take the value of an unknown vector");
    return CoordVector<T>();
  } else
    return CoordVector<T>(dir.getAngle(), dist.getValue());
}

template<typename T>
WithUE<T> CoordVectorUE<T>::getLength() const {
  return dist;
}

template<typename T>
CoordVectorUE<T>& CoordVectorUE<T>::setLength(const WithUE<T>& d) {
  dist = d;
  return *this;
}

template<typename T>
AngleDegreesUE<T> CoordVectorUE<T>::getAngle() const {
  return dir;
}

template<typename T>
CoordVectorUE<T>& CoordVectorUE<T>::setAngle(const AngleDegreesUE<T>& a) {
  dir = a;
  return *this;
}

template <typename T>
CoordVectorUE<T>& CoordVectorUE<T>::rotate(const AngleDegreesUE<T>& vdir) {
  dir += vdir;
  return *this;
}

template <typename T>
CoordVectorUE<T>& CoordVectorUE<T>::scale(const WithUE<T>& vdist) {
  dist += vdist;
  return *this;
}

template <typename T>
CoordVectorUE<T> CoordVectorUE<T>::operator -() const {
  return CoordVectorUE<T>(dir-AngleDegrees<T>(180), dist);
}

template <typename T>
CoordVectorUE<T>& CoordVectorUE<T>::operator-=(const CoordVectorUE<T>& v) {
  return *this = *this - v;
}

template <typename T>
CoordVectorUE<T>& CoordVectorUE<T>::operator+=(const CoordVectorUE<T>& v) {
  return *this = *this + v;
}

template <typename T>
CoordVectorUE<T>& CoordVectorUE<T>::operator*=(const WithUE<T>& a) {
  dist *= a;
  return *this;
}

template <typename T>
CoordVectorUE<T>& CoordVectorUE<T>::operator/=(const WithUE<T>& a) {
#if CHECK_DIVisION_BY_ZERO == 1
  if (a == T(0))
    dist = T_DIVisION_BY_ZERO_RESULT;
  else
#endif
    dist /= a;  
  return *this;
}
template <typename T>
CoordVectorUE<T> 
CoordVectorUE<T>::operator -(const CoordVectorUE<T>& v) const {
  return *this+CoordVectorUE<T>(v.getAngle()-AngleDegrees<T>(180), 
				v.getLength());
}

template <typename T>
CoordVectorUE<T> 
CoordVectorUE<T>::operator +(const CoordVectorUE<T>& v) const {
  return CoordVectorUE<T>(CoordPointUE<T>(v) + CoordPointUE<T>(*this));
}

template <typename T>
CoordVectorUE<T> CoordVectorUE<T>::operator *(const WithUE<T>& a) const {
  return CoordVectorUE<T>(dir, dist * a);
}
template <typename T>
CoordVectorUE<T> CoordVectorUE<T>::operator /(const WithUE<T>& a) const {
#if CHECK_DIVisION_BY_ZERO == 1
  if (a == T(0))
    return COORDVECTOR_DIVisION_BY_ZERO_RESULT;
  else
#endif
    return CoordVectorUE<T>(dir, dist / a);
}

template <typename T>
bool CoordVectorUE<T>::operator ==(const CoordVectorUE<T>& v) const {
  return (dist == v.dist) && (dir == v.dir);
}

template <typename T>
bool CoordVectorUE<T>::operator !=(const CoordVectorUE<T>& v) const {
  return (dist != v.dist) || (dir != v.dir);
}

template <typename T>
bool CoordVectorUE<T>::operator < (const CoordVectorUE<T>& v) const {
  return (dist <  v.dist);
}

template <typename T>
bool CoordVectorUE<T>::operator <=(const CoordVectorUE<T>& v) const {
  return (dist <= v.dist);
}

template <typename T>
bool CoordVectorUE<T>::operator > (const CoordVectorUE<T>& v) const {
  return (dist >  v.dist);
}

template <typename T>
bool CoordVectorUE<T>::operator >=(const CoordVectorUE<T>& v) const {
  return (dist >= v.dist);
}

template <typename T>
bool CoordVectorUE<T>::operator !() const {
  return (!dist) && (!dir);
}


template <typename T>
CoordVectorUE<T> operator -(const CoordVector<T>& v1,
			    const CoordVectorUE<T>& v2) {
  return (-v2) + v1;
}

template <typename T>
CoordVectorUE<T> operator -(const CoordVectorUnknown<T>& v1,
			    const CoordVectorUE<T>& v2) {
  return (-v2) + v1;
}

template <typename T>
CoordPointUE<T> operator -(const CoordPoint<T>& pt,
			   const CoordVectorUE<T>& vec) {
  return CoordPointUE<T>(pt) - vec;
}

template <typename T>
CoordPointUE<T> operator -(const CoordPointUnknown<T>& pt,
			   const CoordVectorUE<T>& vec) {
  return CoordPointUE<T>(pt) - vec;
}

template <typename T>
CoordVectorUE<T> operator +(const CoordVector<T>& v1,
			    const CoordVectorUE<T>& v2) {
  return v2 + v1;
}

template <typename T>
CoordVectorUE<T> operator +(const CoordVectorUnknown<T>& v1,
			    const CoordVectorUE<T>& v2) {
  return v2 + v1;
}

template<typename T>
CoordPointUE<T> operator +(const CoordPoint<T>& pt,
			   const CoordVectorUE<T>& vec) {
  return CoordPointUE<T>(pt) + vec;
}

template<typename T>
CoordPointUE<T> operator +(const CoordPointUnknown<T>& pt,
			   const CoordVectorUE<T>& vec) {
  return CoordPointUE<T>(pt) + vec;
}

template <typename T>
CoordVectorUE<T> operator *(const T& a, const CoordVectorUE<T>& vec) {
  return vec * a;
}

template <typename T>
CoordVectorUE<T> operator *(const WithUnknown<T>& a,
			    const CoordVectorUE<T>& vec) {
  return vec * a;
}

template <typename T>
CoordVectorUE<T> operator *(const WithUE<T>& a,
			    const CoordVectorUE<T>& vec) {
  return vec * a;
}

template<typename T>
bool operator==(const CoordVector<T>& v1,
		const CoordVectorUE<T>& v2) {
  return v2 == v1;
}

template<typename T>
bool operator==(const CoordVectorUnknown<T>& v1,
		const CoordVectorUE<T>& v2) {
  return v2 == v1;
}

template<typename T>
bool operator!=(const CoordVector<T>& v1,
		const CoordVectorUE<T>& v2) {
  return v2 != v1;
}

template<typename T>
bool operator!=(const CoordVectorUnknown<T>& v1,
		const CoordVectorUE<T>& v2) {
  return v2 != v1;
}

template<typename T>
bool operator<(const CoordVector<T>& v1,
	       const CoordVectorUE<T>& v2) {
  return v2 >= v1;
}

template<typename T>
bool operator<(const CoordVectorUnknown<T>& v1,
	       const CoordVectorUE<T>& v2) {
  return v2 >= v1;
}

template<typename T>
bool operator<=(const CoordVector<T>& v1,
		const CoordVectorUE<T>& v2) {
  return v2 > v1;
}

template<typename T>
bool operator<=(const CoordVectorUnknown<T>& v1,
		const CoordVectorUE<T>& v2) {
  return v2 > v1;
}

template<typename T>
bool operator>(const CoordVector<T>& v1,
	       const CoordVectorUE<T>& v2) {
  return v2 <= v1;
}

template<typename T>
bool operator>(const CoordVectorUnknown<T>& v1,
	       const CoordVectorUE<T>& v2) {
  return v2 <= v1;
}

template<typename T>
bool operator>=(const CoordVector<T>& v1,
		const CoordVectorUE<T>& v2) {
  return v2 < v1;
}

template<typename T>
bool operator>=(const CoordVectorUnknown<T>& v1,
		const CoordVectorUE<T>& v2) {
  return v2 < v1;
}

template <typename T>
const std::string toString(const CoordVectorUE<T>& v) {
#ifdef VECTORUE_HEAD
  return VECTORUE_HEAD + "(" + v.getAngle() + ", " + v.getLength() + ")";
#else
  return "(" + v.getAngle() + ", " + v.getLength() + ")";
#endif
}

template <typename T>
std::ostream& operator<<(std::ostream& os, const CoordVectorUE<T>& v) {
#ifdef VECTORUE_HEAD
  return os << VECTORUE_HEAD << "(" << v.getAngle() << ", "
	    << v.getLength() << ")";
#else
  return os << "(" << v.getAngle() << ", " << v.getLength() << ")";
#endif
}


RS_END_NAMESPACE


#endif // !COORDVECTORUE_H
