/* -*- Mode: C++ -*-
 * $Id: CoordVectorUnknown.h,v 1.7 2002/08/31 14:51:54 s02davro Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: CoordVectorUnknown.h
 * Unitname: Library
 * $Revision: 1.7 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: s02davro $ $Date: 2002/08/31 14:51:54 $
 *
 *
 * HISTORY
 *
 * $Log: CoordVectorUnknown.h,v $
 * Revision 1.7  2002/08/31 14:51:54  s02davro
 * Modifications and corrections of point classes.
 *
 * Revision 1.6  2002/08/02 21:34:54  s02davro
 * Added documentation.
 *
 * Revision 1.5  2001/08/29 21:08:28  frehe
 * Replaces class T with typename T where T can be a built in type.
 *
 * Revision 1.4  2000/09/18 15:20:06  frehe
 * Due to a problem with namespaces the name for atan2, atan, acos and asin 
 * is changed to ATan2, ATan, ASin and ACos.
 *
 * Revision 1.3  2000/09/04 15:21:50  frehe
 * Changed the name of the trigonometric functions to the same as in the 
 * standard math library (sin, cos, tan, asin, acos, atan, and atan2).
 *
 * Revision 1.2  2000/09/01 03:23:53  frehe
 * Made minor changes.
 *
 * Revision 1.1  2000/09/01 01:24:39  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 *
 */

/**
 * @file CoordVectorUnknown.h
 * Include file for the template CoordVectorUnknown.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.7 $
 * $Author: s02davro $
 * $Date: 2002/08/31 14:51:54 $
 */


#ifndef COORDVECTORUNKNOWN_H
#define COORDVECTORUNKNOWN_H

#include <iostream>

#include "math_utils.h"
#include "CoordVector.h"
#include "WithUnknown.h"
#include "AngleDegreesUnknown.h"
#include "CoordPointUnknown.h"
#include "UnknownValueException.h"
#include "rs_throw.h"

/** String to use as the header when outputing a VectorUnknown. */
#define VECTORUNKNOWN_HEAD std::string("#VectorUnknown")

RS_BEGIN_NAMESPACE

template<typename T> class CoordPointUnknown;

/**
 * @class CoordVectorUnknown CoordVectorUnknown.h
 * Template class for representing a two dimensional mathematical vector
 * that can be unknown.
 * The distance and direction are of type WithUnknown<T>.
 * T is the base type for the angle: AngleDegreesUnknown<T> and the
 * distance WithUnknown<T>.
 *
 *
 * @par Unit:
 *      Library
 * $Revision: 1.7 $
 * $Author: s02davro $
 * $Date: 2002/08/31 14:51:54 $
 */
template<typename T> class CoordVectorUnknown {
public:
  /** Default constructor, the angle and distance is unknown. */
  CoordVectorUnknown();

  /** Construct a vector using a direction and a distance */
  CoordVectorUnknown(const T& dir, const T& dist);

  /** Construct a vector using a angle and a distance. */
  CoordVectorUnknown(const AngleDegrees<T>& dir, const T& dist);
  
  /** Construct a vector using a unknown angle and a distance. */
  CoordVectorUnknown(const AngleDegreesUnknown<T>& dir,
		     const T& dist);
  
  /** Construct a vector using a unknown angle and a unknown distance. */
  CoordVectorUnknown(const AngleDegreesUnknown<T>& dir,
		     const WithUnknown<T>& dist);

  /** Copyconstructor. */
  CoordVectorUnknown(const CoordVectorUnknown<T>& vec);

  /** Construct a vector using a known vector. */
  CoordVectorUnknown(const CoordVector<T>& vec);

  /** Construct a vector based on two (known or unknown) points. */
  CoordVectorUnknown(const CoordPointUnknown<T>& pt1,
		     const CoordPointUnknown<T>& pt2);

  /** @name Explicit constructors for coverting a point to a vector. */
  //@{
  /** Using a known point. */
  explicit CoordVectorUnknown(const CoordPoint<T>& pt);
      
  /** Using a (unknown) point. */
  explicit CoordVectorUnknown(const CoordPointUnknown<T>& pt);
  //@}

  /** Assignment operator. */
  CoordVectorUnknown<T>& operator=(const CoordVectorUnknown<T>& pt);

  /** @name Arithmetic operators. */
  //@{
  CoordVectorUnknown<T> operator-() const;
  CoordVectorUnknown<T> operator-(const CoordVectorUnknown<T>& vec) const;
  CoordVectorUnknown<T> operator+(const CoordVectorUnknown<T>& vec) const; 
  CoordVectorUnknown<T> operator*(const WithUnknown<T>& val) const;
  CoordVectorUnknown<T> operator/(const WithUnknown<T>& val) const;  
  CoordVectorUnknown<T>& operator-=(const CoordVectorUnknown<T>& vec);
  CoordVectorUnknown<T>& operator+=(const CoordVectorUnknown<T>& vec);
  CoordVectorUnknown<T>& operator*=(const WithUnknown<T>& val);
  CoordVectorUnknown<T>& operator/=(const WithUnknown<T>& val);
  //@}

  /** @name Comparison operators. 
   *  See CoordVector for more information. */
  //@{
  bool operator ! () const;
  bool operator ==(const CoordVectorUnknown<T>& vec) const;
  bool operator !=(const CoordVectorUnknown<T>& vec) const;
  bool operator < (const CoordVectorUnknown<T>& vec) const;
  bool operator <=(const CoordVectorUnknown<T>& vec) const;
  bool operator > (const CoordVectorUnknown<T>& vec) const;
  bool operator >=(const CoordVectorUnknown<T>& vec) const;
  //@}

  /** Make the vector unknown. */
  CoordVectorUnknown<T>& makeUnknown();

  /** Check if the vector is known. */
  bool isKnown() const;

  /** Check if the vector is unknown. */
  bool isUnknown() const;

  /** Get the vector.
   *  If the vector is unknown an UnknownValueException is thrown. */
  CoordVector<T> getVector() const throw(UnknownValueException);
  
  /** Adds to the angle of the vector. */
  CoordVectorUnknown<T>& rotate(const AngleDegreesUnknown<T>& ang);

  /** Adds to the distance of the vector. 
   *  Bad name! I should have thought this was multiplication not addition. */
  CoordVectorUnknown<T>& scale (const WithUnknown<T>& val);

  /** Returns the length of the vector. */
  WithUnknown<T> getLength() const;

  /** Set the length of the vector. */
  CoordVectorUnknown<T>& setLength(const WithUnknown<T>& val);

  /** Returns the angle of the vector. */
  AngleDegreesUnknown<T> getAngle() const;

  /** Set the angle of the vector. */
  CoordVectorUnknown<T>& setAngle(const AngleDegreesUnknown<T>& ang);

protected:
  /** The actual representation of the direction of the vector. */
  AngleDegreesUnknown<T> dir;

  /** The actual representation of the distance of the vector. */
  WithUnknown<T> dist;
};

template <typename T>
CoordVectorUnknown<T>::CoordVectorUnknown()
  : dir(),
    dist() {
}

template <typename T>
CoordVectorUnknown<T>::CoordVectorUnknown(const T& vdir,
					  const T& vdist)
  : dir(vdir),
    dist(vdist) {
  if (dist < WithUnknown<T>(0)) {
    dir -= AngleDegreesUnknown<T>(180);
    dist = -dist;
  }
}

template <typename T>
CoordVectorUnknown<T>::CoordVectorUnknown(const AngleDegrees<T>& vdir,
					  const T& vdist)
  : dir(vdir),
    dist(vdist) {
  if (dist < WithUnknown<T>(0)) {
    dir -= AngleDegreesUnknown<T>(180);
    dist = -dist;
  }
}

template <typename T>
CoordVectorUnknown<T>::CoordVectorUnknown(const AngleDegreesUnknown<T>& vdir,
					  const T& vdist)
  : dir(vdir),
    dist(vdist) {
  if (dist < WithUnknown<T>(0)) {
    dir -= AngleDegreesUnknown<T>(180);
    dist = -dist;
  }
}

template <typename T>
CoordVectorUnknown<T>::CoordVectorUnknown(const AngleDegreesUnknown<T>& vdir,
					  const WithUnknown<T>& vdist)
  : dir(vdir),
    dist(vdist) {
  if (dist < WithUnknown<T>(0)) {
    dir -= AngleDegreesUnknown<T>(180);
    dist = -dist;
  }
}

template <typename T>
CoordVectorUnknown<T>::CoordVectorUnknown(const CoordVector<T>& v)
  : dir(v.getAngle()),
    dist(v.getLength()) {
}

template <typename T>
CoordVectorUnknown<T>::CoordVectorUnknown(const CoordVectorUnknown<T>& v)
  : dir(v.dir),
    dist(v.dist) {
}

template <typename T>
CoordVectorUnknown<T>::CoordVectorUnknown(const CoordPointUnknown<T>& pt1,
					  const CoordPointUnknown<T>& pt2) {
  if (pt2 == pt1) {
    dir = AngleDegreesUnknown<T>(0);
    dist = WithUnknown<T>(0);
  } else {
    dir = ATan2(pt2.getY()-pt1.getY(), pt2.getX()-pt1.getX());
    dist = sqrt(sqr(pt1.getX()-pt2.getX()) + sqr(pt1.getY()-pt2.getY()));
  }
}

template <typename T>
CoordVectorUnknown<T>::CoordVectorUnknown(const CoordPoint<T>& pt) 
  : dir(ATan2(pt.getY(), pt.getX())),
    dist(sqrt(sqr(pt.getX()) + sqr(pt.getY()))) {}

template <typename T>
CoordVectorUnknown<T>::CoordVectorUnknown(const CoordPointUnknown<T>& pt) {
  if ( pt.getX() == WithUnknown<T>(0) &&
       pt.getY() == WithUnknown<T>(0) ) {
    dir = AngleDegreesUnknown<T>(0);
    dist = WithUnknown<T>(0);
  } else {
    dir = ATan2(pt.getY(), pt.getX());
    dist = sqrt(sqr(pt.getX()) + sqr(pt.getY()));
  }
}

template <typename T>
CoordVectorUnknown<T>& 
CoordVectorUnknown<T>::operator =(const CoordVectorUnknown<T>& v) {
  if (this != &v) {
    dist = v.dist;
    dir = v.dir;
  }
  return *this;
}

template<typename T>
inline bool CoordVectorUnknown<T>::isKnown() const {
  return dir.isKnown() && dist.isKnown();
}

template<typename T>
inline bool CoordVectorUnknown<T>::isUnknown() const {
  return dir.isUnknown() && dist.isUnknown();
}

template<typename T>
CoordVectorUnknown<T>& CoordVectorUnknown<T>::makeUnknown() {
  dir.makeUnknown();
  dist.makeUnknown();
  return *this;
}

template <typename T> 
inline CoordVector<T> 
CoordVectorUnknown<T>::getVector() const throw(UnknownValueException) {
  if ( isUnknown() ) {
    RS_THROW(UnknownValueException, 
	     "Tried to take the value of an unknown vector");
    return CoordVector<T>();
  } else
    return CoordVector<T>(dir.getAngle(), dist.getValue());
}

template <typename T>
inline CoordVectorUnknown<T> CoordVectorUnknown<T>::operator -() const {
  return CoordVectorUnknown<T>(dir-AngleDegrees<T>(180), dist);
}

template <typename T>
inline CoordVectorUnknown<T> 
CoordVectorUnknown<T>::operator -(const CoordVectorUnknown<T>& v) const {
  return *this+CoordVectorUnknown<T>(v.dir-AngleDegrees<T>(180), v.dist);
}

template <typename T>
inline CoordVectorUnknown<T> 
CoordVectorUnknown<T>::operator +(const CoordVectorUnknown<T>& v) const {
  return CoordVectorUnknown<T>(CoordPointUnknown<T>(v) + 
			       CoordPointUnknown<T>(*this));
}

template <typename T>
inline CoordVectorUnknown<T> 
CoordVectorUnknown<T>::operator *(const WithUnknown<T>& a) const {
  return CoordVectorUnknown<T>(dir, dist * a);
}

template <typename T>
inline CoordVectorUnknown<T> 
CoordVectorUnknown<T>::operator /(const WithUnknown<T>& a) const {
#if CHECK_DIVisION_BY_ZERO == 1
  if (a == T(0))
    return COORDVECTOR_DIVisION_BY_ZERO_RESULT;
  else
#endif
    return CoordVectorUnknown<T>(dir, dist / a);
}

template <typename T>
inline CoordVectorUnknown<T>& 
CoordVectorUnknown<T>::operator-=(const CoordVectorUnknown<T>& v) {
  return *this = *this - v;
}

template <typename T>
inline CoordVectorUnknown<T>& 
CoordVectorUnknown<T>::operator+=(const CoordVectorUnknown<T>& v) {
  return *this = *this + v;
}

template <typename T>
inline CoordVectorUnknown<T>& 
CoordVectorUnknown<T>::operator*=(const WithUnknown<T>& a) {
  dist *= a;
  return *this;
}

template <typename T>
inline CoordVectorUnknown<T>& 
CoordVectorUnknown<T>::operator/=(const WithUnknown<T>& a) {
#if CHECK_DIVisION_BY_ZERO == 1
  if (a == T(0))
    dist = T_DIVisION_BY_ZERO_RESULT;
  else
#endif
    dist /= a;  
  return *this;
}

template <typename T>
inline bool 
CoordVectorUnknown<T>::operator ==(const CoordVectorUnknown<T>& v) const {
  return (dist == v.dist) && (dir == v.dir);
}

template <typename T>
inline bool 
CoordVectorUnknown<T>::operator !=(const CoordVectorUnknown<T>& v) const {
  return (dist != v.dist) || (dir != v.dir);
}

template <typename T>
inline bool 
CoordVectorUnknown<T>::operator < (const CoordVectorUnknown<T>& v) const {
  return (dist <  v.dist);
}

template <typename T>
inline bool 
CoordVectorUnknown<T>::operator <=(const CoordVectorUnknown<T>& v) const {
  return (dist <= v.dist);
}

template <typename T>
inline bool 
CoordVectorUnknown<T>::operator > (const CoordVectorUnknown<T>& v) const {
  return (dist >  v.dist);
}

template <typename T>
inline bool 
CoordVectorUnknown<T>::operator >=(const CoordVectorUnknown<T>& v) const {
  return (dist >= v.dist);
}

template <typename T>
inline bool CoordVectorUnknown<T>::operator !() const {
  return (!dist) && (!dir);
}

template<typename T>
inline WithUnknown<T> CoordVectorUnknown<T>::getLength() const {
  return dist;
}

template<typename T>
inline CoordVectorUnknown<T>& 
CoordVectorUnknown<T>::setLength(const WithUnknown<T>& d) {
  dist = d;
  return *this;
}

template<typename T>
inline AngleDegreesUnknown<T> CoordVectorUnknown<T>::getAngle() const {
  return dir;
}

template<typename T>
inline CoordVectorUnknown<T>& 
CoordVectorUnknown<T>::setAngle(const AngleDegreesUnknown<T>& a) {
  dir = a;
  return *this;
}

template <typename T>
inline CoordVectorUnknown<T>& 
CoordVectorUnknown<T>::rotate(const AngleDegreesUnknown<T>& vdir) {
  dir += vdir;
  return *this;
}

template <typename T>
inline CoordVectorUnknown<T>& 
CoordVectorUnknown<T>::scale(const WithUnknown<T>& vdist) {
  dist += vdist;
  return *this;
}

template <typename T>
inline CoordVectorUnknown<T> operator -(const CoordVector<T>& v1,
					const CoordVectorUnknown<T>& v2) {
  return (-v2) + v1;
}

template <typename T>
inline CoordPointUnknown<T> operator -(const CoordPoint<T>& pt,
				       const CoordVectorUnknown<T>& vec) {
  return CoordPointUnknown<T>(pt) - vec;
}
 
template <typename T>
inline CoordVectorUnknown<T> operator +(const CoordVector<T>& v1,
					const CoordVectorUnknown<T>& v2) {
  return v2 + v1;
}

template<typename T>
inline CoordPointUnknown<T> operator +(const CoordPoint<T>& pt,
				       const CoordVectorUnknown<T>& vec) {
  return CoordPointUnknown<T>(pt) - vec;
}

template <typename T>
inline CoordVectorUnknown<T> operator *(const T& a,
					const CoordVectorUnknown<T>& vec) {
  return vec * a;
}

template <typename T>
inline CoordVectorUnknown<T> operator *(const WithUnknown<T>& a,
					const CoordVectorUnknown<T>& vec) {
  return vec * a;
}

template<typename T>
inline bool operator==(const CoordVector<T>& v1,
		       const CoordVectorUnknown<T>& v2) {
  return v2 == v1;
}

template<typename T>
inline bool operator!=(const CoordVector<T>& v1,
		       const CoordVectorUnknown<T>& v2) {
  return v2 != v1;
}

template<typename T>
inline bool operator<(const CoordVector<T>& v1,
		      const CoordVectorUnknown<T>& v2) {
  return v2 >= v1;
}

template<typename T>
inline bool operator<=(const CoordVector<T>& v1,
		       const CoordVectorUnknown<T>& v2) {
  return v2 > v1; 
}

template<typename T>
inline bool operator>(const CoordVector<T>& v1,
		      const CoordVectorUnknown<T>& v2) {
  return v2 <= v1;
}

template<typename T>
inline bool operator>=(const CoordVector<T>& v1,
		       const CoordVectorUnknown<T>& v2) {
  return v2 < v1;
}

template <typename T>
std::string toSTring(const CoordVectorUnknown<T>& v) {
#ifdef VECTORUNKNOWN_HEAD
  return VECTORUNKNOWN_HEAD + "(" + v.getAngle() + ", " + v.getLength() + ")";
#else
  return "(" + v.getAngle() + ", " + v.getLength() + ")";
#endif
}

template <typename T>
inline std::ostream& operator<<(std::ostream& os, 
				const CoordVectorUnknown<T>& v) {
#ifdef VECTORUNKNOWN_HEAD
  return os << VECTORUNKNOWN_HEAD << "(" << v.getAngle() << ", "
	    << v.getLength() << ")";
#else
  return os << "(" << v.getAngle() << ", " << v.getLength() << ")";
#endif
}

RS_END_NAMESPACE


#endif // COORDVECTORUNKNOWN_H
