/* -*- Mode: C++ -*-
 * $Id: MobileObject.h,v 1.4 2002/09/03 10:56:54 s02davro Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: MobileObject.h
 * Unitname: Library
 * $Revision: 1.4 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: s02davro $ $Date: 2002/09/03 10:56:54 $
 *
 *
 * HISTORY
 *
 * $Log: MobileObject.h,v $
 * Revision 1.4  2002/09/03 10:56:54  s02davro
 * Minor update
 *
 * Revision 1.3  2002/08/31 14:46:01  s02davro
 * Small changes in the game objects.
 *
 * Revision 1.2  2002/08/02 21:36:10  s02davro
 * Change types, inheritance, removed some methods.
 *
 * Revision 1.1  2000/09/01 03:24:31  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 * Revision 1.1.1.1  2000/08/31 13:26:04  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file MobileObject.h
 * Include file for the class MobileObject.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.4 $
 * $Author: s02davro $
 * $Date: 2002/09/03 10:56:54 $
 */


#ifndef MOBILEOBJECT_H
#define MOBILEOBJECT_H

#include "library_types_with_ue.h"

RS_BEGIN_NAMESPACE


/**
 * @class MobileObject MobileObject.h
 * Class used as the base class of all mobile objects,
 * i.e. those objects that can move during a game.
 * The observationtime and speedobservationtime is the the
 * time when the agent last saw the object.
 *
 * @todo Make a toString and operator<<
 *
 * @par Unit:
 *      Library
 * $Revision: 1.4 $
 * $Author: s02davro $
 * $Date: 2002/09/03 10:56:54 $
 */
class MobileObject {
public:
  /** Default constructor, position and speed is set to unknown as default. */
  MobileObject(const PointUE& def_point = PointUE(),
	       const TimeUnknown& def_time = TimeUnknown(),
	       const VectorUE& def_speed = VectorUE(),
	       const TimeUnknown& def_speed_time = TimeUnknown(),
	       const VectorUE& def_acceleration = VectorUE());

  /** Copyconstructor. */
  MobileObject(const MobileObject& obj);

  /** Assignment operator. */
  MobileObject& operator= (const MobileObject& obj);

  /** Set the objects position. */
  void setPosition(const PointUE& pos);

  /** Get the x coordinate of the object (can be unknown and have an error). */
  FloatUE getPosX() const;

  /** Get the y coordinate of the object (can be unknown and have an error). */
  FloatUE getPosY() const;

  /** Get the (unknown) position of the object. */
  PointUE getPosition() const;

  /** Set the time for positionobservation. */
  void setObservationTime(const TimeUnknown& time);

  /** Get the time for last observation of position. */
  TimeUnknown getObservationTime() const;

  /** Set the speed of the object. */
  void setSpeedVector(const VectorUE& v);

  /** Get the speed of the object. */
  VectorUE getSpeedVector() const;

  /** Set the speedamounth. */
  void setSpeed(const FloatUE new_speed);

  /** Get the speedamounth. */
  FloatUE getSpeed() const;

  /** Set the speed direction. */
  void setSpeedDirection(const AngleDegUE& ang);

  /** Get the speed direction. */
  AngleDegUE getSpeedDirection() const;

  /** Set the time for last observation of speed. */
  void setSpeedObservationTime(const TimeUnknown& time);

  /** Returns the time for last observation of speed. */
  TimeUnknown getSpeedObservationTime() const;
  
  /** Set the current acceleration of the object.
   *  Used when players are dashing and ball is kicked.
   *  The acceleration is added to the speed when we enter a
   *  new cycle. */
  void setAcceleration(const VectorUE& v);

  /** Get the current acceleration of the object. */
  VectorUE getAcceleration() const;

protected:
  /** The representation of the position, can be unknown and have an error. */
  PointUE position;

  /** The time when we last saw the object. */
  TimeUnknown observationTime;

  /** The representation of the speed, can be unknown and have an error. */
  VectorUE speed;

  /** The time when we last observed the speed of the object. */
  TimeUnknown speedObservationTime;

  /** The current acceleration of the object. */
  VectorUE acceleration;
};



RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "MobileObject.icc"
#endif


#endif // !MOBILEOBJECT_H
