/* -*- Mode: C++ -*-
 * $Id: Time.h,v 1.8 2003/12/02 16:15:42 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Time.h
 * Unitname: Library
 * $Revision: 1.8 $
 * Created by: Fredrik Heintz 2000-xx-xx
 * Last modified by $Author: frehe $ $Date: 2003/12/02 16:15:42 $
 *
 *
 * HISTORY
 *
 * $Log: Time.h,v $
 * Revision 1.8  2003/12/02 16:15:42  frehe
 * Fixed comments.
 *
 * Revision 1.7  2002/09/03 10:54:04  s02davro
 * Updated Time and TimeUnknown
 *
 * Revision 1.6  2002/08/31 14:55:32  s02davro
 * Small modifications
 *
 * Revision 1.5  2002/08/02 21:36:58  s02davro
 * Added documentation, changed types.
 *
 * Revision 1.4  2001/09/03 19:38:59  frehe
 * Removed dependencies on <iostream>.
 *
 * Revision 1.3  2000/09/20 15:06:07  frehe
 * Added documentation.
 *
 * Revision 1.2  2000/09/03 20:05:51  frehe
 * Made minor changes.
 *
 * Revision 1.1  2000/09/01 01:24:41  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 * Revision 1.1.1.1  2000/08/31 13:26:04  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file Time.h
 * Include file for the class Time.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.8 $
 * $Author: frehe $
 * $Date: 2003/12/02 16:15:42 $
 */

#ifndef TIME_H
#define TIME_H

#include <iosfwd>

#include "library_types.h"

/** String to use as the header when outputing a Time value. */
#define TIME_HEAD std::string("#Time")

RS_BEGIN_NAMESPACE

/**
 * @class Time Time.h
 * Class used to represent the soccer server time.
 * This class is based on CMUnited's implementation created by P. Stone & Co.
 * The time is a tuple (T, S) where T is the current servertime and
 * S is the cyclenumber in T when time is stopped.
 *
 *
 * @todo Maybe this should be a singleton.
 * @todo Move some attributes to private section.
 * @todo Is lastStopClockTime and secondLastStopClockTime needed?
 *
 * @par Unit:
 *      Library
 * $Revision: 1.8 $
 * $Author: frehe $
 * $Date: 2003/12/02 16:15:42 $
 */

class Time {
public:
  Time(const UInt vt = 0);
  
  /** Default constructor */
  Time(const UInt vt, const UInt vs);

  /** Copyconstructor. */
  Time(const Time& obj);

  /** Assignment operator */
  Time& operator = (const Time&);

  /** @name Arithmetic operators */
  //@{
  Time  operator - (const Time&) const;
  Time  operator - (const UInt) const;
  Time  operator + (const Time&) const;
  Time  operator + (const UInt) const;
  Time& operator ++();
  Time& operator --();
  Time& operator -=(const Time&);
  Time& operator -=(const UInt);
  Time& operator +=(const Time&);
  Time& operator +=(const UInt);
  //@}

  /** @name Comparison operators. */
  //@{
  bool operator ! () const;
  bool operator ==(const Time&) const;
  bool operator ==(const UInt) const;
  bool operator !=(const Time&) const;
  bool operator !=(const UInt) const;
  bool operator < (const Time&) const;
  bool operator < (const UInt) const;
  bool operator <=(const Time&) const;
  bool operator <=(const UInt) const;
  bool operator > (const Time&) const;
  bool operator > (const UInt) const;
  bool operator >=(const Time&) const;
  bool operator >=(const UInt) const;
  //@}

  /** Get the current game time. */
  UInt getT() const;

  /** Set the current game time. */
  Time& setT(const UInt);

  /** Get the cycle time, the number of cycles the clock has been stopped. */
  UInt getS() const;

  /** Set the number of cycles the time has been stopped. */
  Time& setS(const UInt);

  /** Check if the game time is stoped. */
  static bool isClockStopped();

  /** Stop the clock. */
  static void stopClock();

  /** Start the clock. */
  static void startClock();
  
  /** Get the current time. */
  static Time currentTime;

  /** Is the clocked stopped. 
   *  This should maybe be moved to private section. */
  static bool clockStopped;

  /** The last time the clock was stopped. */
  static Time lastStopClockTime;

  /** The second last time the clock was stopped. */
  static Time secondLastStopClockTime;
  
private:
  /** Time from the server, this is the game time. */
  UInt t;
  /**  Stopped clock cycles.
   *   This is how many server cycles the clock has been stopped. */
  UInt s;
};


std::ostream& operator<<(std::ostream&  os, const Time& ct);


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "Time.icc"
#endif


#endif // !TIME_H
