/* -*- Mode: C++ -*-
 * $Id: TrainerMoveObjectCommand.h,v 1.3 2005/11/15 10:27:55 petan Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: TrainerMoveObjectCommand.h
 * Unitname: Library
 * $Revision: 1.3 $
 * Created by: Fredrik Heintz 2004-08-23
 * Last modified by $Author: petan $ $Date: 2005/11/15 10:27:55 $
 *
 *
 * HISTORY
 *
 * $Log: TrainerMoveObjectCommand.h,v $
 * Revision 1.3  2005/11/15 10:27:55  petan
 * Updated documentation for trainer commands.
 *
 * Revision 1.2  2004/08/26 17:08:42  frehe
 * Fixed implementation of moving players.
 *
 * Revision 1.1  2004/08/23 20:45:29  frehe
 * Added trainer commands.
 *
 */

/**
 * @file TrainerMoveObjectCommand.h
 * Include file for the class TrainerMoveObjectCommand.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.3 $
 * $Author: petan $
 * $Date: 2005/11/15 10:27:55 $
 */


#ifndef TRAINER_MOVE_OBJECT_COMMAND_H
#define TRAINER_MOVE_OBJECT_COMMAND_H


#include "TrainerCommand.h"
#include "soccer_types.h"


RS_BEGIN_NAMESPACE


/**
 * @class TrainerMoveObjectCommand TrainerMoveObjectCommand.h
 * Class used to represent the trainer command
 * (move OBJECT X Y [VDIR [VEL_X VEL_Y]]).
 * (move OBJECT X Y [VDIR [VELX VELY ]]) This command will move OBJECT, which may be a player or the ball
 * (see SoccerServer manual for format information), to absolute position (X, Y). If VDIR is specified,
 * it will also change its absolute facing direction to VDIR (this only matters for players). Additionally,
 * if VELX and VELY are specified, the object's velocity will be set accordingly. The trainer always
 * uses left-hand coordinates. Possible replies by the soccerserver:
 * <ol><li> (ok move)<br> The command succeeded.</li>
 * <li> (error illegal object form) <br> The OBJECT specification was not valid.</li>
 * <li> (error illegal command form) <br> The position, direction, and/or velocity specification was not valid.</li>
 * </ol>
 * @par Unit:
 *      Library
 * $Revision: 1.3 $
 * $Author: petan $
 * $Date: 2005/11/15 10:27:55 $
 */
class TrainerMoveObjectCommand : public TrainerCommand {
public:
  /** Create a move ball to @a x @a y trainer command. */
  TrainerMoveObjectCommand(Float x, Float y);

  /** Create a move ball to @a pos trainer command. */
  TrainerMoveObjectCommand(const Point& pos);

  /** Create a move ball to @a x @a y with speed @a vel_x and @a vel_y
      trainer command. */
  TrainerMoveObjectCommand(Float x, Float y,
			   Float vel_x, Float vel_y);

  /** Create a move ball to @a pos with speed @a vel trainer command. */
  TrainerMoveObjectCommand(const Point& pos, const Point& vel);

  
  /** Create a move player @a unum from @a team @a side to @a pos command. */
  TrainerMoveObjectCommand(const std::string& team, UInt unum,
			   const Point& pos);

  /** Create a move player @a unum from @a team @a side to @a x @a y command. */
  TrainerMoveObjectCommand(const std::string& team, UInt unum,
			   Float x, Float y);

  /**
   * Create a move player @a unum from @a team @a side to @a pos and
   * direction @a dir command.
   */
  TrainerMoveObjectCommand(const std::string& team, UInt unum,
			   const Point& pos, Float dir);

  /**
   * Create a move player @a unum from @a team @a side to @a x @a y and
   * direction @a dir command.
   */
  TrainerMoveObjectCommand(const std::string& team, UInt unum,
			   Float x, Float y, Float dir);

  /**
   * Create a move player @a unum from @a team @a side to @a pos,
   * direction @a dir, and velocity @a vel command.
   */
  TrainerMoveObjectCommand(const std::string& team, UInt unum,
			   const Point& pos, Float dir,
			   const Point& vel);

  /**
   * Create a move player @a unum from @a team @a side to @a x @a y,
   * direction @a dir, and velocity @a vel_x @a vel_y command.
   */
  TrainerMoveObjectCommand(const std::string& team, UInt unum,
			   Float x, Float y, Float dir,
			   Float vel_x, Float vel_y);

  
  /** Create an arbitrary move object command. */
  TrainerMoveObjectCommand(const std::string& cmd);
  

  /** Destructor. */
  ~TrainerMoveObjectCommand();

  
  /** Returns the server string that will execute the MoveObjectCommand
      at the soccer server. */
  const std::string serverString() const;

  
  /** Returns a new copy of the command.
      REMEMBER TO DELETE IT, OTHERWISE YOU WILL HAVE A MEMORY LEAK!!! */
  TrainerMoveObjectCommand* clone() const;


private:
  bool ball_;
  std::string team_;
  UInt unum_;
  Float x_;
  Float y_;
  bool vel_;
  Float vel_x_;
  Float vel_y_;
  bool dir_;
  Float bdir_;
  bool cmd_;
  std::string cmd_string_;
};


RS_END_NAMESPACE


#endif // !TRAINER_MOVE_OBJECT_COMMAND_H
