/* -*- Mode: C++ -*-
 * $Id: WithError.h,v 1.4 2005/09/19 17:00:06 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: WithError.h
 * Unitname: Library
 * $Revision: 1.4 $
 * Created by: David Rosn 2002-06-19
 * Last modified by $Author: frehe $ $Date: 2005/09/19 17:00:06 $
 *
 *
 * HISTORY
 *
 * $Log: WithError.h,v $
 * Revision 1.4  2005/09/19 17:00:06  frehe
 * Fixed bugs and made WithError work.
 *
 * Revision 1.3  2002/10/01 16:08:39  s02davro
 * Bugfix of / and *
 *
 * Revision 1.3 2002/09/30 20:00:00 davpo603
 * Corrected bugs in * and /
 *
 * $Log: WithError.h,v $
 * Revision 1.4  2005/09/19 17:00:06  frehe
 * Fixed bugs and made WithError work.
 *
 * Revision 1.3  2002/10/01 16:08:39  s02davro
 * Bugfix of / and *
 *
 * Revision 1.2  2002/08/31 14:56:49  s02davro
 * Corrected some small bugs
 *
 * Revision 1.1  2002/08/02 21:37:28  s02davro
 * Created template class to handle values with error.
 *
 *
 */

/**
 * @file WithError.h
 * Include file for the class CWithError.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.4 $
 * $Author: frehe $
 * $Date: 2005/09/19 17:00:06 $
 */


#ifndef WITHERROR_H
#define WITHERROR_H

#include <string>
#include <iostream>
#include "rs_error.h"
#include "string_utils.h"
#include "machine_dependent_types.h"
#include "math_utils.h"
#include "quantize_utils.h"

/** String to use as the header when outputing a WithError value. */
#define ERROR_HEAD std::string("#Err")

//#define INVARIANT_CHECK


RS_BEGIN_NAMESPACE

/**
 * @class WithError WithError.h
 * Template class for representing basic types that can have an error.
 * The template argument only defines what type should be annotated
 * with a possibility to have an error.
 * The only demands on the template argument T is that the arithmetic
 * functions are defined on it.
 * The dequantize (invQuatize) methods are partly stolen from
 * UvA Trilearn (C) Remco de Boer & Jelle Kok.
 *
 * @todo Testing
 * @todo Represent the error with one value, not two.
 * @todo Check the range of errors when doing operations.
 * @todo Should the comparison operators be implemented in a different way?
 * @todo + - values in invQuantize need to be checked for sanity.
 * @todo Sin, cos and other nonlinear functions need to be handled.
 * @todo Some arithmetic functions are missing.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.4 $
 * $Author: frehe $
 * $Date: 2005/09/19 17:00:06 $
 */

enum ErrorType {
  ERROR_QUANTIZE,
  ERROR_DISTANCE,
  ERROR_DISTCHANGE,
  ERROR_NORMAL,
  ERROR_NONE
};

template <typename T> class WithError {
public:
  /** Default constructor, no error, value set to T(). */
  WithError();

  /** Construct a object using specified value, no error. */
  WithError(const T& def_value);

  /** Construct object using a errortype according to server noise.
   *  This can be used to easy get the error interval a value from
   *  the soccerserver can be in. The third parameter is used as the
   *  quantize step, the forth as a second basevalue when calculating
   *  the distance change error. The fifth is the second quantize step
   *  in this last case. */
  WithError(const ErrorType c, const T& value,
	    const Float64 quant_step1 = 0.1, const T& distance = T(),
	    const Float64 quant_step2 = 0.02);

  /** Construct an object using specifed value and error values. */
  WithError(const T& value, const T& plus, const T& minus, const T& extra_dummy_arg);

  /** Copyconstructor. */
  WithError(const WithError<T>& obj);

  /** @name Assignment operators. */
  //@{
  WithError<T>& operator= (const T& op);
  WithError<T>& operator= (const WithError<T>& op);
  //@}

  /** @name Arithmetic operators. */
  //@{
  WithError<T>& operator++ ();
  WithError<T>& operator-- ();
  WithError<T>& operator+= (const WithError<T>& op);
  WithError<T>& operator+= (const T& c);
  WithError<T>& operator-= (const WithError<T>& op);
  WithError<T>& operator-= (const T& c);
  WithError<T>& operator*= (const WithError<T>& op);
  WithError<T>& operator*= (const T& c);
  WithError<T>& operator/= (const WithError<T>& op);
  WithError<T>& operator/= (const T& c);
  WithError<T> operator- () const;
  WithError<T> operator+ (const WithError<T>& op) const;
  WithError<T> operator+ (const T& c) const;
  WithError<T> operator- (const WithError<T>& op) const;
  WithError<T> operator- (const T& c) const;
  WithError<T> operator* (const WithError<T>& op) const;
  WithError<T> operator* (const T& c) const;
  WithError<T> operator/ (const WithError<T>& op) const;
  WithError<T> operator/ (const T& c) const;
  //@}

  /** @name Comparison operators.
   *  These functions do not care about the error values.
   *  Maybe they should in the future. */
  //@{
  bool operator!  () const;
  bool operator== (const T& op) const;
  bool operator== (const WithError<T>& op) const;
  bool operator!= (const T& op) const;
  bool operator!= (const WithError<T>& op) const;
  bool operator<  (const T& op) const;
  bool operator<  (const WithError<T>& op) const;
  bool operator<= (const T& op) const;
  bool operator<= (const WithError<T>& op) const;
  bool operator>  (const T& op) const;
  bool operator>  (const WithError<T>& op) const;
  bool operator>= (const T& op) const;
  bool operator>= (const WithError<T>& op) const;
  //@}

  /** If the intervals represented in the two objects overlap the
   *  resulting object will be set to the intersection between this two
   *  intervals and the value as the middle.
   *  Example: 45 (2, -1), 45 (1, -0) --(improve)--> 45.5 (0.5, -0.5) */
  WithError<T>& improve (const WithError<T>& op);

  /** The resulting object will be set to the union between the two intervals
   *  Example: 45 (2, -2), 43 (2, -2) --(disimprove)--> 44 (3, -3) */
  WithError<T>& disimprove (const WithError<T>& op);

  /** Get the value stripped from error information. */
  T getValue() const;

  /** Get the minimum value according to the error interval. */
  T getMin() const;

  /** Get the maximum value according to the error interval. */
  T getMax() const;

  /** Get the plus value.
   *  getPlus = getMax - getValue */
  T getPlus() const;

  /** Get the minus value.
   *  getMinus = getMin - getValue */
  T getMinus() const;

protected:
  /** Dequantize the value to get a error interval. */
  void quantizeError(const T def_value, const Float64 quant_step);

  /** Dequantize the value as a distance to get a error interval. */
  void distanceError(const T def_value, const Float64 quant_step);

  /** Dequantize as a distchange value to get a error interval. */
  void distchangeError(const T def_value,
		       const Float64 quant_step1,
		       const T def_distance,
		       const Float64 quant_step2);

  /** Set the errors to none. */
  void noError();

  /** Normalerror for future use. Not yet implemented. */
  void normalError(const T def_value);

  /** @name Helper functions */
  //@{
  T min(T t1, T t2, T t3, T t4) const;
  T max(T t1, T t2, T t3, T t4) const;
  T min(T t1, T t2) const;
  T max(T t1, T t2) const;
  //@}

  /** Representation of the value. */
  T value;

  /** Representation of the upper bound in the error interval. */
  T plus;

  /** Representation of the lower bound in the error interval. */
  T minus;
};

template <typename T>
WithError<T>::WithError(const ErrorType c, const T& def_value,
			const Float64 quant_step1, const T& def_distance,
			const Float64 quant_step2)
  : value(def_value),
    plus(),
    minus()
{
  switch (c) {
  case ERROR_QUANTIZE: {
    quantizeError(def_value, quant_step1);
    break;
  }
  case ERROR_DISTANCE: {
    distanceError(def_value, quant_step1);
    break;
  }
  case ERROR_DISTCHANGE: {
    distchangeError(def_value, quant_step1, def_distance, quant_step2);
    break;
  }
  case ERROR_NORMAL: {
    normalError(def_value);
    break;
  }
  case ERROR_NONE: {
    noError();
    break;
  }
  default: {
    noError();
    break;
  }
  }
  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in WithError(" << c << ", " << def_value << ", " << quant_step1
	      << ", " << def_distance << ", " << quant_step2 << ") = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
}

template <typename T>
WithError<T>::WithError(const T& def_value)
  : value(def_value),
    plus(),
    minus()
{
  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in WithError(" << def_value << ") = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
}

template <typename T>
WithError<T>::WithError()
  :value(),
   plus(),
   minus()
{
  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in WithError() = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
}

template <typename T>
WithError<T>::WithError(const T& def_value, const T& def_plus,
			const T& def_minus, const T& extra_dummy_arg)
  : value(def_value),
    plus(def_plus),
    minus(def_minus)
{
  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in WithError(" << def_value << ", " << def_plus
	      << ", " << def_minus << ") = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
}

template <typename T>
WithError<T>::WithError(const WithError<T>& obj)
  : value(obj.value),
    plus(obj.plus),
    minus(obj.minus)
{
  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in WithError(" << obj << ") = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
}

template <typename T>
WithError<T>& WithError<T>::operator= (const WithError<T>& c)
{
  if ( &c != this) {
    value = c.value;
    plus = c.plus;
    minus = c.minus;
  }
  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in op=(" << c << ") = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
  return *this;
}

template <typename T>
WithError<T>& WithError<T>::operator= (const T& c)
{
  value = c;
  plus = T();
  minus = T();
  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in op=(" << c << ") = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
  return *this;
}

template <typename T>
WithError<T>& WithError<T>::operator++ ()
{
  ++value;
  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in op++() = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
  return *this;
}


template <typename T>
WithError<T>& WithError<T>::operator-- ()
{
  --value;
  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in op--() = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
  return *this;
}

template <typename T>
WithError<T>& WithError<T>::operator+= (const T& c)
{
  value += c;
  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in op+=(" << c << ") = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
  return *this;
}

template <typename T>
WithError<T>& WithError<T>::operator+= (const WithError<T>& c)
{
  value += c.value;
  plus += c.plus;
  minus += c.minus;
  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in op+=(" << c << ") = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
  return *this;
}

template <typename T>
WithError<T>& WithError<T>::operator-= (const T& c)
{
  value -= c;
  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in op-=(" << c << ") = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
  return *this;
}

template <typename T>
WithError<T>& WithError<T>::operator-= (const WithError<T>& c)
{
  value -= c.value;
  plus += c.plus;
  minus += c.minus;
  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in op-=(" << c << ") = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
  return *this;
}

template <typename T>
WithError<T>& WithError<T>::operator*= (const T& c)
{
  value *= c;
  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in op*=(" << c << ") = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
  return *this;
}

template <typename T>
WithError<T>& WithError<T>::operator*= (const WithError<T>& c)
{
  T new_value = value * c.value;
  T new_plus = max(getMax() * c.getMax(),
		   getMax() * c.getMin(),
		   getMin() * c.getMax(),
		   getMin() * c.getMin()) - new_value;
  T new_minus = min(getMax() * c.getMax(),
		    getMax() * c.getMin(),
		    getMin() * c.getMax(),
		    getMin() * c.getMin()) - new_value;

  /*
  if ( isnan(new_value) || isnan(new_plus) || isnan(new_minus) ) {
    std::cout << *this << " *= " << c << " gave:" << std::endl;
    std::cout << "getMax() * c.getMax() = " << getMax() * c.getMax() << std::endl;
    std::cout << "getMax() * c.getMin() = " << getMax() * c.getMin() << std::endl;
    std::cout << "getMin() * c.getMax() = " << getMin() * c.getMax() << std::endl;
    std::cout << "getMin() * c.getMin() = " << getMin() * c.getMin() << std::endl;
    std::cout << "max(...) = "
	      << max(getMax() * c.getMax(),
		     getMax() * c.getMin(),
		     getMin() * c.getMax(),
		     getMin() * c.getMin()) << std::endl;
    std::cout << "min(...) = "
	      << min(getMax() * c.getMax(),
		     getMax() * c.getMin(),
		     getMin() * c.getMax(),
		     getMin() * c.getMin()) << std::endl;
    std::cout << "value = " << new_value << "; plus = " << new_plus
	      << "; minus = " << new_minus << std::endl;
  }
  */

  if ( isnan(new_value) ) {
    new_value = T();
    //std::cout << "new_value is nan in " << *this << " *= " << c << std::endl;
  }
  if ( isnan(new_plus) ) {
    new_plus = T();
    //std::cout << "new_plus is nan in " << *this << " *= " << c << std::endl;
  }
  if ( isnan(new_minus) ) {
    new_minus = T();
    //std::cout << "new_minus is nan in " << *this << " *= " << c << std::endl;
  }

  // If we are working with angles we might get very strange results
  // A hack is to just add the errors if they are outside the bounds
  if ( new_minus > T() ) {
    //std::cout << "new_minus = " << new_minus << " > T() in " << *this << " *= " << c << std::endl;
    new_minus = getMinus() + c.getMinus();
    if ( new_minus > T() ) {
      //std::cout << "new_minus = " << new_minus << " still > T() in " << *this << " *= " << c << std::endl;
      new_minus = getMinus();
    }
  }
  if ( new_plus < T() ) {
    //std::cout << "new_plus = " << new_plus << " < T() in " << *this << " *= " << c << std::endl;
    new_plus = getPlus() + c.getPlus();
    if ( new_plus < T() ) {
      //std::cout << "new_plus = " << new_plus << " still < T() in " << *this << " *= " << c << std::endl;
      new_plus = getPlus();
    }
  }

  value = new_value;
  plus = new_plus;
  minus = new_minus;

  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in op*=(" << c << ") = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
  return *this;
}

template <typename T>
WithError<T>& WithError<T>::operator/= (const T& c)
{
  value /= c;
  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in op/=(" << c << ") = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
  return *this;
}

template <typename T>
WithError<T>& WithError<T>::operator/= (const WithError<T>& c)
{
  T new_value = value / c.value;
  T new_plus = max(getMax() / c.getMax(),
		   getMax() / c.getMin(),
		   getMin() / c.getMax(),
		   getMin() / c.getMin()) - new_value;
  T new_minus = min(getMax() / c.getMax(),
		    getMax() / c.getMin(),
		    getMin() / c.getMax(),
		    getMin() / c.getMin()) - new_value;

  /*
  if ( isnan(new_value) || isnan(new_plus) || isnan(new_minus) ) {
    std::cout << *this << " /= " << c << " gave:" << std::endl;
    std::cout << "getMax() / c.getMax() = " << getMax() / c.getMax() << std::endl;
    std::cout << "getMax() / c.getMin() = " << getMax() / c.getMin() << std::endl;
    std::cout << "getMin() / c.getMax() = " << getMin() / c.getMax() << std::endl;
    std::cout << "getMin() / c.getMin() = " << getMin() / c.getMin() << std::endl;
    std::cout << "max(...) = "
	      << max(getMax() / c.getMax(),
		     getMax() / c.getMin(),
		     getMin() / c.getMax(),
		     getMin() / c.getMin()) << std::endl;
    std::cout << "min(...) = "
	      << min(getMax() / c.getMax(),
		     getMax() / c.getMin(),
		     getMin() / c.getMax(),
		     getMin() / c.getMin()) << std::endl;
    std::cout << "value = " << new_value << "; plus = " << new_plus
	      << "; minus = " << new_minus << std::endl;
  }
  */

  if ( isnan(new_value) ) {
    new_value = T();
    //std::cout << "new_value is nan in " << *this << " /= " << c << std::endl;
  }
  if ( isnan(new_plus) ) {
    new_plus = T();
    //std::cout << "new_plus is nan in " << *this << " /= " << c << std::endl;
  }
  if ( isnan(new_minus) ) {
    new_minus = T();
    //std::cout << "new_minus is nan in " << *this << " /= " << c << std::endl;
  }

  // If we are working with angles we might get very strange results
  // A hack is to just add the errors if they are outside the bounds
  if ( new_minus > T() ) {
    //std::cout << "new_minus = " << new_minus << " > T() in " << *this << " /= " << c << std::endl;
    new_minus = getMinus() + c.getMinus();
    if ( new_minus > T() ) {
      //std::cout << "new_minus = " << new_minus << " still > T() in " << *this << " /= " << c << std::endl;
      new_minus = getMinus();
    }
  }
  if ( new_plus < T() ) {
    //std::cout << "new_plus = " << new_plus << " < T() in " << *this << " /= " << c << std::endl;
    new_plus = getPlus() + c.getPlus();
    if ( new_plus < T() ) {
      //std::cout << "new_plus = " << new_plus << " still < T() in " << *this << " /= " << c << std::endl;
      new_plus = getPlus();
    }
  }

  value = new_value;
  plus = new_plus;
  minus = new_minus;

  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in op/=(" << c << ") = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
  return *this;
}

template <typename T>
WithError<T> WithError<T>::operator- () const
{
  // should plus and minus be switched here? No!
  return WithError(-value, plus, minus, T());
}

template <typename T>
WithError<T> WithError<T>::operator+ (const T& c) const
{
  return WithError<T>(value + c, plus, minus, T());
}

template <typename T>
WithError<T> WithError<T>::operator+ (const WithError<T>& c) const
{
  return WithError<T>(value + c.value,
		      plus + c.plus,
		      minus + c.minus, T());
}

template <typename T>
WithError<T> WithError<T>::operator- (const T& c) const
{
  return WithError<T>(value - c, plus, minus, T());
}

template <typename T>
WithError<T> WithError<T>::operator- (const WithError<T>& c) const
{
  return WithError<T>(value - c.value,
		      plus + c.plus,
		      minus + c.minus, T());
}

template <typename T>
WithError<T> WithError<T>::operator* (const T& c) const
{
  return WithError<T>(value * c, plus, minus, T());
}

template <typename T>
WithError<T> WithError<T>::operator* (const WithError<T>& c) const
{
  WithError<T> res(*this);
  res *= c;
  return res;
}

template <typename T>
WithError<T> WithError<T>::operator/ (const T& c) const
{
  return WithError<T>(value / c, plus, minus, T());
}

template <typename T>
WithError<T> WithError<T>::operator/ (const WithError<T>& c) const
{
  WithError<T> res(*this);
  res /= c;
  return res;
}

template <typename T>
bool WithError<T>::operator! () const
{
  return !value;
}

template <typename T>
bool WithError<T>::operator== (const T& op) const
{
  return (value == op);
}

template <typename T>
bool WithError<T>::operator== (const WithError<T>& op) const
{
  return ((value == op.value) &&
	  (plus == op.plus) &&
	  (minus == op.minus));
}

template <typename T>
bool WithError<T>::operator!= (const T& op) const
{
  return !(*value == op);
}

template <typename T>
bool WithError<T>::operator!= (const WithError<T>& op) const
{
  return !(*this == op);
}

template <typename T>
bool WithError<T>::operator< (const T& op) const
{
  return value < op;
}

template <typename T>
bool WithError<T>::operator< (const WithError<T>& op) const
{
  return value < op.value;
}

template <typename T>
bool WithError<T>::operator<= (const T& op) const {
  return value <= op;
}

template <typename T>
bool WithError<T>::operator<= (const WithError<T>& op) const {
  return value <= op.value;
}

template <typename T>
bool WithError<T>::operator> (const T& op) const {
  return value > op;
}

template <typename T>
bool WithError<T>::operator> (const WithError<T>& op) const {
  return value > op.value;
}

template <typename T>
bool WithError<T>::operator>= (const T& op) const {
  return value >= op;
}

template <typename T>
bool WithError<T>::operator>= (const WithError<T>& op) const {
  return value >= op.value;
}

template<typename T>
WithError<T> operator- (const T& op1, const WithError<T>& op2) {
  return (-op2) + op1;
}

template<typename T>
WithError<T> operator+ (const T& op1, const WithError<T>& op2) {
  return op2 + op1;
}

template<typename T>
WithError<T> operator* (const T& op1, const WithError<T>& op2) {
  return op2 * op1;
}

template<typename T>
WithError<T> operator/ (const T& op1, const WithError<T>& op2) {
  return WithError<T>(op1) / op2;
}

template<typename T>
bool operator==(const T& op1, const WithError<T>& op2) {
  return op2 == op1;
}

template<typename T>
bool operator!=(const T& op1, const WithError<T>& op2) {
  return op2 != op1;
}

template<typename T>
bool operator<(const T& op1, const WithError<T>& op2) {
  return op2 > op1;
}

template<typename T>
bool operator<=(const T& op1, const WithError<T>& op2) {
  return op2 >= op1;
}

template<typename T>
bool operator>(const T& op1, const WithError<T>& op2) {
  return op2 < op1;
}

template<typename T>
bool operator>=(const T& op1, const WithError<T>& op2) {
  return op2 <= op1;
}

template <typename T>
WithError<T>& WithError<T>::improve (const WithError<T>& op)
{
  // This will most likely not work for angles
  if (((getMax() >= op.getMin()) && (getMin() <= op.getMax())) ||
      ((getMin() <= op.getMax()) && (getMax() >= op.getMin()))) {
    T newMax = min(getMax(), op.getMax());
    T newMin = max(getMin(), op.getMin());
    value = (newMax + newMin) / 2.0;
    plus = newMax - value;
    minus = newMin - value;
  } else {
    RS_WARNING( "Trying to improve error with non-overlapping error intervals ["
		<< getMin() << ", " << getValue() << ", " << getMax() << "] and ["
		<< op.getMin() << ", " << op.getValue() << ", " << op.getMax() << "]" );
  }
  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in improve(" << op << ") = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
  return *this;
}

template <typename T>
WithError<T>& WithError<T>::disimprove (const WithError<T>& op)
{
  // This will most likely not work for angles
  T newMax = max(getMax(), op.getMax());
  T newMin = min(getMin(), op.getMin());
  value = (newMax + newMin) / 2.0;
  plus = newMax - value;
  minus = newMin - value;
  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in disimprove(" << op << ") = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
  return *this;
}

template <typename T>
T WithError<T>::getValue() const {
  return value;
}

template <typename T>
T WithError<T>::getMax() const {
  return value + plus;
}

template <typename T>
T WithError<T>::getMin() const {
  return value + minus;
}

template <typename T>
T WithError<T>::getPlus() const {
  return plus;
}

template <typename T>
T WithError<T>::getMinus() const {
  return minus;
}

template<typename T>
void WithError<T>::quantizeError(const T def_value, const Float64 quant_step)
{
  minus = invQMin(def_value, quant_step) - def_value;
  plus = invQMax(def_value, quant_step) - def_value;
  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in quantizeError(" << def_value
	      << ", " << quant_step << ") = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
}

template<typename T>
void WithError<T>::distanceError(const T def_value, const Float64 quant_step)
{
  minus = distanceMin(def_value, quant_step) - def_value;
  plus = distanceMax(def_value, quant_step) - def_value;

  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in distanceError(" << def_value
	      << ", " << quant_step << ") = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
}

template<typename T>
void WithError<T>::distchangeError(const T def_value,
				   const Float64 quant_step1,
				   const T distance,
				   const Float64 quant_step2)
{
  T dmin = distanceMin(distance, quant_step1);
  T dmax = distanceMax(distance, quant_step1);

  T dout = def_value/distance;
  T dminch = invQMin(dout, quant_step2);
  T dmaxch = invQMax(dout, quant_step2);

  plus = max(dmin*dmaxch, dmax*dmaxch) - def_value;
  minus = min(dmin*dminch, dmax*dminch) - def_value;
  
  if ( isnan(plus) ) {
    plus = T();
    /*
    std::cout << "plus is nan in distchangeError(" << def_value
	      << ", " << quant_step1 << ", " << distance
	      << ", " << quant_step2<< ")" << std::endl;
    */
  }
  if ( isnan(minus) ) {
    minus = T();
    /*
    std::cout << "minus is nan in distchangeError(" << def_value
	      << ", " << quant_step1 << ", " << distance
	      << ", " << quant_step2 << ")" << std::endl;
    */
  }

  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "dmin = " << dmin << "; dmax = " << dmax
	      << "; dout = " << dout << "; dmaxch = " << dmaxch
	      << "; dminch = " << dminch << std::endl;
    std::cout << "WithError wrong in distchangeError(" << def_value
	      << ", " << quant_step1 << ", " << distance
	      << ", " << quant_step2 << ") = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
}

template<typename T>
void WithError<T>::normalError(const T) {
  /** Fixme! */
  minus = T();
  plus = T();
  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in normalError() = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
}

template<typename T>
void WithError<T>::noError() {
  minus = T();
  plus = T();
  if ( !(getMinus() <= T() && getPlus() >= T()) ) {
    std::cout << "WithError wrong in noError() = " << *this << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( getMinus() <= T() );
  assert( getPlus() >= T() );
#endif
}

template<typename T>
T WithError<T>::min(T t1, T t2, T t3, T t4) const {
  T tmp = t1;
  if (tmp > t2) tmp = t2;
  if (tmp > t3) tmp = t3;
  if (tmp > t4) tmp = t4;
  return tmp;
}

template<typename T>
T WithError<T>::max(T t1, T t2, T t3, T t4) const {
  T tmp = t1;
  if (tmp < t2) tmp = t2;
  if (tmp < t3) tmp = t3;
  if (tmp < t4) tmp = t4;
  return tmp;
}

template<typename T>
T WithError<T>::min(T t1, T t2) const {
  if (t1 < t2) return t1;
  return t2;
}

template<typename T>
T WithError<T>::max(T t1, T t2) const {
  if (t1 < t2) return t2;
  return t1;
}

template<typename T>
std::ostream& operator<<(std::ostream& os, const WithError<T>& obj) {
#ifdef ERROR_HEAD
  if (obj.getPlus() != T() || obj.getMinus() != T())
    return os << ERROR_HEAD << "(" << toString(obj.getValue()) << " ("
	      << toString(obj.getPlus()) << ", "
	      << toString(obj.getMinus()) << "))";
  return os << ERROR_HEAD << "(" << toString(obj.getValue()) << ")";
#else
  return os << "(" << toString(obj.getValue()) << ")";
#endif
}

template<typename T>
const std::string toString(const WithError<T>& obj)
{
#ifdef ERROR_HEAD
  if (obj.getPlus() != T() || obj.getMinus() != T())
    return ERROR_HEAD + "(" + toString(obj.getValue()) + " (" +
      toString(obj.getPlus()) + ", " + toString(obj.getMinus()) + ")";
  return ERROR_HEAD + "(" + toString(obj.getValue()) + ")";
#else
  return "(" + toString(obj.getValue()) + ")";
#endif
}

template<typename T>
WithError<T> sqrt(const WithError<T>& op)
{
  // This will fail if value is negative!
  T value = sqrt(op.getValue());
  T plus = sqrt(op.getPlus());
  T minus = -sqrt(-op.getMinus());
  
  if ( isnan(value) ) {
    value = T();
    //std::cout << "value is nan in sqrt(" << op << ")" << std::endl;
  }
  if ( isnan(plus) ) {
    plus = T();
    //std::cout << "plus is nan in sqrt(" << op << ")" << std::endl;
  }
  if ( isnan(minus) ) {
    minus = T();
    //std::cout << "minus is nan in sqrt(" << op << ")" << std::endl;
  }
  
  if ( !((plus >= T()) && (minus <= T())) ) {
    std::cout << "sqrt(" << op << ") gave value = " << value
	      << "; plus = " << plus << "; minus = " << minus
	      << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( (plus >= T()) && (minus <= T()) );
#endif
  return WithError<T>(value, plus, minus, T());
}

template<typename T>
WithError<T> sqr(const WithError<T>& op)
{
  T value = sqr(op.getValue());
  T plus = sqr(plus);
  T minus = -sqr(minus);
  
  if ( isnan(value) ) {
    value = T();
    //std::cout << "value is nan in sqr(" << op << ")" << std::endl;
  }
  if ( isnan(plus) ) {
    plus = T();
    //std::cout << "plus is nan in sqr(" << op << ")" << std::endl;
  }
  if ( isnan(minus) ) {
    minus = T();
    //std::cout << "minus is nan in sqr(" << op << ")" << std::endl;
  }
  
  if ( !((plus >= T()) && (minus <= T())) ) {
    std::cout << "sqr(" << op << ") gave value = " << value
	      << "; plus = " << plus << "; minus = " << minus
	      << std::endl;
  }
#ifdef INVARIANT_CHECK
  assert( (plus >= T()) && (minus <= T()) );
#endif
  return WithError<T>(value, plus, minus, T());
}


RS_END_NAMESPACE

#endif // !WITHERROR_H
