/* -*- Mode: C++ -*-
 * $Id: WithUnknown.h,v 1.11 2002/10/01 16:08:13 s02davro Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: WithUnknown.h
 * Unitname: Library
 * $Revision: 1.11 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: s02davro $ $Date: 2002/10/01 16:08:13 $
 *
 *
 * HISTORY
 *
 * $Log: WithUnknown.h,v $
 * Revision 1.11  2002/10/01 16:08:13  s02davro
 * Small WithUnknown bug now /= do /= and not +=
 *
 * Revision 1.10  2002/08/31 14:56:49  s02davro
 * Corrected some small bugs
 *
 * Revision 1.9  2002/08/02 22:22:31  s02davro
 * Added some documentation
 *
 * Revision 1.8  2002/08/02 22:14:35  s02davro
 * Changed some things
 *
 * Revision 1.7  2001/08/29 21:08:29  frehe
 * Replaces class T with typename T where T can be a built in type.
 *
 * Revision 1.6  2000/10/11 18:40:09  frehe
 * Tried to fix a segmentation fault in WithUnknown, but without success.
 *
 * Revision 1.5  2000/10/08 17:33:14  frehe
 * Remade the sqrt template to a single function and also added a test to see
 * that the argument is positive.
 *
 * Revision 1.4  2000/09/03 20:05:52  frehe
 * Made minor changes.
 *
 * Revision 1.3  2000/09/03 14:23:20  frehe
 * Made minor changes to improve the documentation.
 *
 * Revision 1.2  2000/09/01 03:23:55  frehe
 * Made minor changes.
 *
 * Revision 1.1  2000/09/01 01:24:43  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 *
 */

/**
 * @file WithUnknown.h
 * Include file for the template WithUnknown.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.11 $
 * $Author: s02davro $
 * $Date: 2002/10/01 16:08:13 $
 */


#ifndef WITHUNKNOWN_H
#define WITHUNKNOWN_H


#include <string>
#include <iostream>

#include "math_utils.h"
#include "UnknownValueException.h"
#include "rs_throw.h"


RS_BEGIN_NAMESPACE


/**
 * @class WithUnknown WithUnknown.h
 * Template class for representing basic types that can be unknown.
 * The template argument only defines what type should be annotated
 * with a possibility to be unknown.
 * The only demands on the template argument T is that the arithmetic
 * functions are defined on it.
 *
 *
 * @par Unit:
 *      Library
 * $Revision: 1.11 $
 * $Author: s02davro $
 * $Date: 2002/10/01 16:08:13 $
 */
template <typename T> class WithUnknown {
public:
  /** The default constructor will set the value to unknown. */
  WithUnknown();

  /** Construct an object using a known value. */
  WithUnknown(const T&);

  /** Construct an object using a (unknown) value. */
  WithUnknown(const WithUnknown<T>&);
  
  /** @name Assignment opearators. */
  //@{
  WithUnknown<T>& operator= (const WithUnknown<T>& op);
  WithUnknown<T>& operator= (const T& op);
  //@}

  /** @name Arithmetic operators. */
  //@{
  WithUnknown<T>& operator++ ();
  WithUnknown<T>& operator-- ();
  WithUnknown<T>& operator+= (const WithUnknown<T>& op);
  WithUnknown<T>& operator+= (const T& op);
  WithUnknown<T>& operator-= (const WithUnknown<T>& op);
  WithUnknown<T>& operator-= (const T& op);
  WithUnknown<T>& operator*= (const WithUnknown<T>& op);
  WithUnknown<T>& operator*= (const T& op);
  WithUnknown<T>& operator/= (const WithUnknown<T>& op);
  WithUnknown<T>& operator/= (const T& op);
  WithUnknown<T> operator- () const;
  WithUnknown<T> operator+ (const WithUnknown<T>& op) const;
  WithUnknown<T> operator+ (const T& op) const;
  WithUnknown<T> operator- (const WithUnknown<T>& op) const;
  WithUnknown<T> operator- (const T& op) const;
  WithUnknown<T> operator* (const WithUnknown<T>& op) const;
  WithUnknown<T> operator* (const T& op) const;
  WithUnknown<T> operator/ (const WithUnknown<T>& op) const;
  WithUnknown<T> operator/ (const T& op) const;
  //@}

  /** @name Comparision operators on WithUnknown<T>
   *  !obj == true iff obj is unknown or !obj.value
   *  unknown is the largest number so:
   *  obj1 > obj2 iff obj1 is unknown and obj2 is known or
   *  obj1.value > obj2.value */
  //@{
  bool operator!  () const;
  bool operator== (const WithUnknown<T>& op) const;
  bool operator== (const T& op) const;
  bool operator!= (const WithUnknown<T>& op) const;
  bool operator!= (const T& op) const;
  bool operator<  (const WithUnknown<T>& op) const;
  bool operator<  (const T& op) const;
  bool operator<= (const WithUnknown<T>& op) const;
  bool operator<= (const T& op) const;
  bool operator>  (const WithUnknown<T>& op) const;
  bool operator>  (const T& op) const;
  bool operator>= (const WithUnknown<T>& op) const;
  bool operator>= (const T& op) const;
  //@}

  /** Make the value unknown. */
  WithUnknown<T>& makeUnknown();

  /** Check if the value is known. */
  bool isKnown() const;

  /** Check if the value is unknown. */
  bool isUnknown() const;

  /** Get the value. 
   *  @throws UnknownValueException if the value is unknown. */
  T getValue() const throw(UnknownValueException);
  
  /** Get the type string for the type that is unknown. */
  const std::string getTypeString() const;

  /** A string describing the base type. 
   *  It is used when outputting an unknown value of this type. */
  static const std::string typeString;

protected:
  /** The representation of the value. */
  T value;

  /** Is the value known or unknown. */
  bool unknown;
};

template <typename T> 
WithUnknown<T>::WithUnknown()
  : value(),
    unknown(true) {
}

template <typename T> 
WithUnknown<T>::WithUnknown(const T& val)
  : value(val),
    unknown(false) {
}

template <typename T> 
WithUnknown<T>::WithUnknown(const WithUnknown<T>& op)
  : value(op.value),
    unknown(op.unknown) {
}

template <typename T> 
WithUnknown<T>& WithUnknown<T>::makeUnknown() {
  unknown = true;
  return *this;
}

template <typename T> 
bool WithUnknown<T>::isUnknown() const {
  return unknown;
}

template <typename T> 
bool WithUnknown<T>::isKnown() const {
  return !unknown;
}

template <typename T> 
T WithUnknown<T>::getValue() const throw(UnknownValueException) {
  if ( unknown ) {
    RS_THROW(UnknownValueException,
	     "Tried to take the value of an unknown " 
	     + WithUnknown<T>::typeString);
    return T();
  } else
    return value;
}

template <typename T>
const std::string WithUnknown<T>::getTypeString() const {
  return typeString;
}

template <typename T>
WithUnknown<T>& WithUnknown<T>::operator= (const T& val) { 
  value = val;
  unknown = false;
  return *this;
}


template <typename T>
WithUnknown<T>& WithUnknown<T>::operator= (const WithUnknown<T>& op) {
  if ( &op != this ) {
    if (op.isKnown())
      value = op.getValue();
    else
      value = T();
    unknown = op.isUnknown();
  }
  return *this;
}


template <typename T>
WithUnknown<T>& WithUnknown<T>::operator++ () {
  if ( !unknown )
    ++value;
  return *this;
}


template <typename T>
WithUnknown<T>& WithUnknown<T>::operator-- () {
  if ( !unknown )
    --value;
  return *this;
}

template <typename T>
WithUnknown<T>& WithUnknown<T>::operator+= (const T& op){
  if ( !unknown )
    value += op;
  return *this;
}

template <typename T>
WithUnknown<T>& WithUnknown<T>::operator+= (const WithUnknown<T>& op) {
  if ( !unknown &&
       !op.unknown )
    value += op.value;
  return *this;
}

template <typename T>
WithUnknown<T>& WithUnknown<T>::operator-= (const T& op){
  if ( !unknown )
    value -= op;
  return *this;
}

template <typename T>
WithUnknown<T>& WithUnknown<T>::operator-= (const WithUnknown<T>& op) {
  if ( !unknown &&
       !op.unknown )
    value -= op.value;
  return *this;
}

template <typename T>
WithUnknown<T>& WithUnknown<T>::operator*= (const T& op) {
  if ( !unknown )
    value *= op;
  return *this;
}

template <typename T>
WithUnknown<T>& WithUnknown<T>::operator*= (const WithUnknown<T>& op) {
  if ( !unknown &&
       !op.unknown )
    value *= op.value;
  return *this;
}

template <typename T>
WithUnknown<T>& WithUnknown<T>::operator/= (const T& op){
  if ( !unknown )
    value /= op;  
  return *this;
}

template <typename T>
WithUnknown<T>& WithUnknown<T>::operator/= (const WithUnknown<T>& op) {
  if ( !unknown &&
       !op.unknown )
    value /= op.value;
  return *this;
}

template<typename T>
WithUnknown<T> WithUnknown<T>::operator- () const {
  if ( unknown )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(-value);
}

template<typename T>
WithUnknown<T> WithUnknown<T>::operator+ (const T& op) const {
  if ( unknown )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(value + op);
}

template<typename T>
WithUnknown<T> WithUnknown<T>::operator+ (const WithUnknown<T>& op) const {
  if ( unknown ||
       op.unknown )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(value + op.value);
}

template<typename T>
WithUnknown<T> WithUnknown<T>::operator- (const T& op) const {
  if ( unknown )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(value - op);
}

template<typename T>
WithUnknown<T> WithUnknown<T>::operator- (const WithUnknown<T>& op) const {
  if ( unknown ||
       op.unknown )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(value - op.value);
}

template<typename T>
WithUnknown<T> WithUnknown<T>::operator* (const T& op) const {
  if ( unknown )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(value * op);
}

template<typename T>
WithUnknown<T> WithUnknown<T>::operator* (const WithUnknown<T>& op) const {
  if ( unknown ||
       op.unknown )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(value * op.value);
}

template<typename T>
WithUnknown<T> WithUnknown<T>::operator/ (const T& op) const {
  if ( unknown )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(value / op);
}

template<typename T>
WithUnknown<T> WithUnknown<T>::operator/ (const WithUnknown<T>& op) const {
  if ( unknown ||
       op.unknown )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(value / op.value);
}


template<typename T>
bool WithUnknown<T>::operator!() const
{
  if ( unknown )
    return true;
  else
    return ( !value );
}

template<typename T>
bool WithUnknown<T>::operator==(const T& op) const {
  if ( !unknown )
    return (value == op);
  return false;
}

template<typename T>
bool WithUnknown<T>::operator==(const WithUnknown<T>& op) const {
  if ( unknown && op.unknown )
    return true;
  else if ( !unknown && !op.unknown )
    return ( value == op.value );
  else
    return false;
}

template<typename T>
bool WithUnknown<T>::operator!=(const T& op) const {
  return !( *this == op );
}

template<typename T>
bool WithUnknown<T>::operator!=(const WithUnknown<T>& op) const {
  return !( *this == op );
}

template<typename T>
bool WithUnknown<T>::operator<(const T& op) const {
  if ( !unknown )
    return ( value < op );
  return false;
}

template<typename T>
bool WithUnknown<T>::operator<(const WithUnknown<T>& op) const {
  if ( !unknown && !op.unknown )
    return ( value < op.value );
  else if ( !unknown && op.unknown )
    return true;
  else
    return false;
}

template<typename T>
bool WithUnknown<T>::operator<=(const T& op) const {
  return !( *this > op );
}

template<typename T>
bool WithUnknown<T>::operator<=(const WithUnknown<T>& op) const {
  return !( *this > op );
}

template<typename T>
bool WithUnknown<T>::operator>(const T& op) const {
  if ( !unknown )
    return ( value > op );
  return false;
}

template<typename T>
bool WithUnknown<T>::operator>(const WithUnknown<T>& op) const {
  if ( !unknown && !op.unknown )
    return ( value > op.value );
  else if ( unknown && !op.unknown )
    return true;
  else
    return false;
}

template<typename T>
bool WithUnknown<T>::operator>=(const T& op) const {
  return !( *this < op );
}

template<typename T>
bool WithUnknown<T>::operator>=(const WithUnknown<T>& op) const {
  return !( *this < op );
}

template<typename T>
WithUnknown<T> operator- (const T& op1, const WithUnknown<T>& op2) {
  return -(op2) + op1;
}

template<typename T>
WithUnknown<T> operator+ (const T& op1, const WithUnknown<T>& op2) {
  return op2 + op1;
}

template<typename T>
WithUnknown<T> operator* (const T& op1, const WithUnknown<T>& op2) {
  return op2 * op1;
}

template<typename T>
WithUnknown<T> operator/ (const T& op1, const WithUnknown<T>& op2) {
  return WithUnknown<T>(op1) / op2;
}

template<typename T>
bool operator==(const T& op1, const WithUnknown<T>& op2) {
  return op2 == op1;
}

template<typename T>
bool operator!=(const T& op1, const WithUnknown<T>& op2) {
  return op2 != op1;
}

template<typename T>
bool operator<(const T& op1, const WithUnknown<T>& op2) {
  return op2 >= op1;
}

template<typename T>
bool operator<=(const T& op1, const WithUnknown<T>& op2) {
  return op2 > op1;
}

template<typename T>
bool operator>(const T& op1, const WithUnknown<T>& op2) {
  return op2 <= op1;
}

template<typename T>
bool operator>=(const T& op1, const WithUnknown<T>& op2) {
  return op2 < op1;
}

template<typename T>
std::ostream& operator<<(std::ostream& os, const WithUnknown<T>& v) {
  if ( v.isUnknown() )
    return os << "UNKNOWN(" << v.getTypeString() << ")";
  else
    return os << v.getValue();
}

template<typename T>
const std::string toString(const WithUnknown<T>& obj) {
  if ( obj.isUnknown() )
    return "UNKNOWN(" + obj.getTypeString() + ")";
  else
    return toString(obj.getValue());
}

template<typename T>
WithUnknown<T> min(const WithUnknown<T>& op1, const WithUnknown<T>& op2) {
  if ( op1.isUnknown() ||
       op2.isUnknown() )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(min(op1.getValue(), op2.getValue()));
}

template<typename T>
WithUnknown<T> max(const WithUnknown<T>& op1, const WithUnknown<T>& op2) {
  if ( op1.isUnknown() ||
       op2.isUnknown() )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(max(op1.getValue(), op2.getValue()));
}

template<typename T>
WithUnknown<T> sqr(const WithUnknown<T>& op) {
  if ( op.isUnknown() )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(sqr(op.getValue()));
}

template<typename T>
WithUnknown<T> sqrt(const WithUnknown<T>& op) {
  if ( op.isUnknown() )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(sqrt(op.getValue()));
}

RS_END_NAMESPACE


#endif // !WITHUNKNOWN_H
